/*
 * Copyright (c) 2004 Ying Ge, Communication Research Center Canada.
 *
 * Copyright (c) 2002, 2003 Maoyu Wang, Communication Research Center Canada.
 *
 * By Ying Ge:
 * 1. Change the OLSR packet format and message processing procedure based 
 *    on the OLSR RFC.
 * 2. Add support of multiple interfaces to OLSR, including MID message 
 *    creating and processing procedure as specified in the OLSR RFC.
 * 3. Add QoS Support to OLSR
 *
 * By Maoyu Wang:
 * 1. Ported OLSR from IPv4 to IPv6.
 * 2. Added the Host and Network Association (HNA) functionality into OLSR.
 * 3. Added the default gateway functionality into OLSR by extending the HNA 
 *    message usage. The default gateway functionality supported the mobility
 *	  by cooperating with Mobile IPv6 for a mobile node as well as supported 
 * 	  Internet access for MANET nodes. 
 *
 * DISTRIBUTED WITH NO WARRANTY, EXPRESS OR IMPLIED.
 * See the GNU Library General Public License (file COPYING in the distribution)
 * for conditions of use and redistribution
 */
/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est-à-dire qu'il peut
 * être utilisé, copié, modifié, distribué à l'unique
 * condition que ce texte soit conservé afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant participé à l'élaboration de ce logiciel ne peut
 * être utilisé sans son accord préalable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilité d'aucune sorte.
 * Certaines parties de ce logiciel sont dérivées de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 */

/*
 * Copyright (c) 1983, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * From: @(#)af.c	5.11 (Berkeley) 2/28/91
 * From: @(#)af.c	8.1 (Berkeley) 6/5/93
 * From: @(#)af.c	1.5 (dholland Exp)  1996/11/25
 */

char af_rcsid[] = 
  "$Id: af.c,v 1.2 2000/12/06 10:36:11 prima Exp $";

#include "defs.h"

/*
 * Address family support routines
 */

void inet_canon(struct sockaddr *);
int inet_checkhost(struct sockaddr *);
char *inet_format(struct sockaddr *, char *, size_t);
void inet_hash(struct sockaddr *, struct afhash *);
int inet_netmatch(struct sockaddr *, struct sockaddr *);
int inet_portcheck(struct sockaddr *);
int inet_portmatch(struct sockaddr *);
void inet_output(int, int, struct sockaddr *, int);

void inet6_canon(struct sockaddr *);
int inet6_checkhost(struct sockaddr *);
char *inet6_format(struct sockaddr *, char *, size_t);
void inet6_hash(struct sockaddr *, struct afhash *);
int inet6_netmatch(struct sockaddr *, struct sockaddr *);
int inet6_portcheck(struct sockaddr *);
int inet6_portmatch(struct sockaddr *);
void inet6_output(int, int, struct sockaddr *, int);

#define NIL	{ 0 }
#define	INET \
	{ inet_hash,		inet_netmatch,  	inet_output, \
	  inet_portmatch,	inet_portcheck,  	inet_checkhost, \
	  inet_rtflags,		inet_sendroute, 	inet_canon, \
	  inet_format\
	}
	
#define	INET6 \
	{ inet6_hash,		inet6_netmatch,  	inet6_output, \
	  inet6_portmatch,	inet6_portcheck,  	inet6_checkhost, \
	  inet6_rtflags,	inet6_sendroute, 	inet6_canon, \
	  inet6_format\
	}

struct afswitch afswitch[AF_MAX+1] = 
{
	NIL,		/* 0- unused */
	NIL,		/* 1- Unix domain, unused */
	INET,		/* Internet */
	/*CRC v4->v6 added*/
	INET6,          /*Internet IPv6*/
};

int af_max = sizeof(afswitch) / sizeof(afswitch[0]);

struct sockaddr_in inet_default = 
{
	AF_INET, INADDR_ANY 
};

//CRC v4->v6 added
struct sockaddr_in6 inet6_default =
{
	AF_INET6
};

void inet_hash(struct sockaddr *sa, struct afhash *hp)
{
	struct sockaddr_in *sin=(struct sockaddr_in *)sa;
	int n;

	n = inet_netof_subnet(sin->sin_addr);
	if (n)
	    while ((n & 0xff) == 0)
		n >>= 8;
	hp->afh_nethash = n;
	hp->afh_hosthash = ntohl(sin->sin_addr.s_addr);
	hp->afh_hosthash &= 0x7fffffff;
}

int inet_netmatch(struct sockaddr *sa1, struct sockaddr *sa2)
{
	struct sockaddr_in *sin1=(struct sockaddr_in *)sa1;
	struct sockaddr_in *sin2=(struct sockaddr_in *)sa2;
	return (inet_netof_subnet(sin1->sin_addr) ==
	    inet_netof_subnet(sin2->sin_addr));
}


/*
 * Verify the message is from the right port.
 */

int inet_portmatch(struct sockaddr *sa)
{
	struct sockaddr_in *sin=(struct sockaddr_in *)sa;
	return (sin->sin_port == sp->s_port);
}

/*
 * Verify the message is from a "trusted" port.
 */
int inet_portcheck(struct sockaddr *sa)
{
	struct sockaddr_in *sin=(struct sockaddr_in *)sa;
	return (ntohs(sin->sin_port) <= IPPORT_RESERVED);
}

/*
 * Internet output routine. 
 */
void inet_output(int s, int flags, struct sockaddr *sa, int size)
{
	struct sockaddr_in *sin=(struct sockaddr_in *)sa;
	struct sockaddr_in dst;

	dst = *sin;
	sin = &dst;
	if (sin->sin_port == 0)
		sin->sin_port = sp->s_port;
	if (sendto(s, packet, size, flags,
	    (struct sockaddr *)sin, sizeof (*sin)) < 0)
		perror("sendto");

        outputsize = 0;        
}


/*
 * Return 1 if the address is believed
 * for an Internet host -- THIS IS A KLUDGE.
 */

int inet_checkhost(struct sockaddr *sa)
{
	struct sockaddr_in *sin=(struct sockaddr_in *)sa;
	int i = ntohl(sin->sin_addr.s_addr);

	if (IN_EXPERIMENTAL(i) || sin->sin_port != 0)
		return (0);
	if (i != 0 && (i & 0xff000000) == 0)
		return (0);
	for (i = 0; i < sizeof(sin->sin_zero)/sizeof(sin->sin_zero[0]); i++)
		if (sin->sin_zero[i])
			return (0);
	return (1);
}

void inet_canon(struct sockaddr *sa)
{
	struct sockaddr_in *sin=(struct sockaddr_in *)sa;
	sin->sin_port = 0;
        bzero(sin->sin_zero, sizeof(sin->sin_zero));
}

char *inet_format(struct sockaddr *sa, char *buf, size_t sz)
{
	struct sockaddr_in *sin=(struct sockaddr_in *)sa;
	strncpy(buf, inet_ntoa(sin->sin_addr), sz);
	buf[sz - 1] = '\0';
	return (buf);
}

//CRC v4->v6 add
void inet6_hash(struct sockaddr *sa, struct afhash *hp)
{
	return;	
}

int inet6_netmatch(struct sockaddr *sa1, struct sockaddr *sa2)
{
	return (0);
}

/*
 * Verify the message is from the right port.
 */

int inet6_portmatch(struct sockaddr *sa)
{
	struct sockaddr_in6 *sin=(struct sockaddr_in6 *)sa;
	return (sin->sin6_port == sp->s_port);
}

/*
 * Verify the message is from a "trusted" port.
 */
int inet6_portcheck(struct sockaddr *sa)
{
	struct sockaddr_in6 *sin=(struct sockaddr_in6 *)sa;
	return (ntohs(sin->sin6_port) <= IPPORT_RESERVED);
}

/*
 * Internet output routine. 
 */
void inet6_output(int s, int flags, struct sockaddr *sa, int size)
{
	struct sockaddr_in6 *sin = (struct sockaddr_in6 *)sa;
	struct sockaddr_in6 dst;

    	memcpy(&dst,sin,sizeof(struct sockaddr_in6));
	sin = &dst;
	
	if (sin->sin6_port == 0)
		sin->sin6_port = htons(OLSRPORT);
   	if (sendto(s, packet, size, 0,
	    (struct sockaddr *)sin, sizeof (*sin)) < 0)
		perror("sendto");
		        
        outputsize = 0;        
}

/*
 * Return 1 if the address is believed
 * for an Internet host -- THIS IS A KLUDGE.
 */

int inet6_checkhost(struct sockaddr *sa)
{
	return (1); /*always valid*/
}

void inet6_canon(struct sockaddr *sa)
{
	
	struct sockaddr_in6 *sin=(struct sockaddr_in6 *)sa;
	sin->sin6_port = 0;
        //bzero(sin->sin_zero, sizeof(sin->sin_zero));
}

char *inet6_format(struct sockaddr *sa, char* buf, size_t sz)
{
	struct sockaddr_in6 *sin=(struct sockaddr_in6 *)sa;
	char str[INET6_ADDRSTRLEN];	
	
	buf = (char*)inet_ntop(AF_INET6,&(sin->sin6_addr),str,INET6_ADDRSTRLEN);
	buf[sz - 1] = '\0';	
	return (buf);
}
