/*
 * Department of Systems and Computer Engineering
 * Carleton University, CANADA
 * Copyright (c) 2004 Liang Qin, Department of Systems and Computer Engineering,
 * Carleton University
 * 1. Porting CRC's OLSR with QoS support to XORP 1.0, this release enable OLSR
 *    adds and deletes route entries to RIB in XORP, instead of to Linux kernel. 
 *
 * Copyright (c) 2004 Ying Ge, Communication Research Center Canada.
 *
 * Copyright (c) 2002, 2003 Maoyu Wang, Communication Research Center Canada.
 *
 * By Ying Ge:
 * 1. Change the OLSR packet format and message processing procedure based 
 *    on the OLSR RFC.
 * 2. Add support of multiple interfaces to OLSR, including MID message 
 *    creating and processing procedure as specified in the OLSR RFC.
 * 3. Add QoS Support to OLSR
 *
 * By Maoyu Wang:
 * 1. Ported OLSR from IPv4 to IPv6.
 * 2. Added the Host and Network Association (HNA) functionality into OLSR.
 * 3. Added the default gateway functionality into OLSR by extending the HNA 
 *    message usage. The default gateway functionality supported the mobility
 *	  by cooperating with Mobile IPv6 for a mobile node as well as supported 
 * 	  Internet access for MANET nodes. 
 *
 * DISTRIBUTED WITH NO WARRANTY, EXPRESS OR IMPLIED.
 * See the GNU Library General Public License (file COPYING in the distribution)
 * for conditions of use and redistribution
 */
/* Code Modified by Joe Macker and Ron Lee July (NRL) 2001 
 * the following was added
 * (1) hysteresis code for hellos
 *   - added pending state and filter parameters
 * (2) parameterization of other setting
 * - hello intervals,
 * (3) added jitter functionality
 * (4) still don't like the polling routing used as an event handler
 * (5) changed gateway code so -g becomes enter a gateway address for stub
 * (6) added low delay TOS socket settings - oct 2001
 * (7) added some extra debug output for map plots in nam use debug level 2 or greater
 * (8) improved jitter variance statistics - Macker - nov 2001
 */
 
/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est-à-dire qu'il peut
 * être utilisé, copié, modifié, distribué à l'unique
 * condition que ce texte soit conservé afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant participé à l'élaboration de ce logiciel ne peut
 * être utilisé sans son accord préalable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilité d'aucune sorte.
 * Certaines parties de ce logiciel sont dérivées de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 */

/*
 * Copyright (c) 1983, 1988, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
char copyright[] =
  "@(#) Copyright (c) 2000, 2001\n"
  "  INRIA-HIPERCOM.  All rights reserved.\n";

/*
 * From: @(#)main.c	5.23 (Berkeley) 7/1/91
 * From: @(#)main.c	8.1 (Berkeley) 6/5/93
 * From: @(#)main.c	1.7 (dholland Exp)  1997/05/19
 */

char main_rcsid[] = 
  "$Id: main.c,v 1.2 2000/12/06 10:36:11 prima Exp $";

/*
 * Routing Table Management Daemon
 */

// modified by L.Qin, merged with XORP and changed it to C++ style
// these two header files are in c style

#include <sys/ioctl.h>
#include <sys/file.h>
#include <net/if.h>
#include <errno.h>
#include <signal.h>
#include <syslog.h>
#include <sys/utsname.h>
#include <netinet/ip.h>
#include <string>
#include <vector>
#include <sstream>

// added by L.Qin
#include "olsr_module.h"
#include "libxorp/eventloop.hh"
#include "libxorp/exceptions.hh"

#include "libxipc/finder_server.hh"
#include "libxipc/xrl_std_router.hh"
#include "libxipc/xrl_pf_sudp.hh"

#include "xrl_target6.hh"
#include "xrl_rib_notifier.h"

extern "C" {
	#include "defs.h"
	#include "version.h"
}

static const char* g_tgt_name = "olsr";
//*****

int	num_of_interface=0; //added by Y.Ge
struct sockaddr_in6 defgwIP;	/* NRL-NRL: Address used as Default Gateway */
struct olsr_ip_addr default_gateway;	/* NRL-NRL: sin_addr.s_addr member of variable defgwIP */
struct gateway	saved_gateway;
short hello_start=1;
int rt_reset=0,refresh=0;
//struct sockaddr_in6 addr; //commented by Y.Ge
struct sockaddr_in6 addr[MAXIFACENUM];       /*addedd by Y.Ge, multiple interfaces. Address used as Originator Address */
struct olsr_ip_addr addr_ip;  /*home address*/ //Y.Ge: use this as main address
//commented by L.Qin, already defined in defs.h
//olsr_u16_t addr_ip_prefix;	  /*home address prefix*/
struct olsr_ip_addr addr_coa;  /*care of address*/
//struct olsr_ip_addr addr_manet; /*address propogated in hello message*/ //commented by Y.Ge
struct olsr_ip_addr addr_manet[MAXIFACENUM]; //added by Y.Ge , my interface address
int	olsr_if_index[MAXIFACENUM]; //added by Y.Ge, keep the "int_ifindex" of "struct interface" of all OLSR interfaces
struct olsr_ip_addr my_main_addr; /* added by Y.Ge, main address of this node */
struct olsr_ip_addr multiaddr;	/* multicast addr using in hello and tc msg*/ 
//struct sockaddr_in6 addrsock6;	/* ipv6 address of daemon's socket */ //commented by  Y.Ge
struct sockaddr_in6	addrsock6[MAXIFACENUM]; //added by Y.Ge
//int s6;	//commented by Y.Ge
int s6[MAXIFACENUM]; //added by Y.Ge
//int sd; //commented by Y.Ge
int sd[MAXIFACENUM]; //added by Y.Ge
char packet[MAXPACKETSIZE+1];
int  outputsize = 0;         /* current size of the output buffer (packet) */
struct servent *sp;
sigset_t sigSet, osigset, psigset;
struct interf_name *interf_names = NULL; /* list of interface type on which olsr packets will be send */
int	supplier = 1;		/* process should supply updates */
int option_i = 0;
int filter = 0;
int	gateway = 0;		/* 1 if we are a gateway to parts beyond */
int	debug = 0;
int debug_level = 0;	/* NRL added debug levels .. TBD make this better if we keep using it*/
int	bufspace = 127*1024;	/* max. input buffer size to request */
struct	olsr *msg = (struct olsr *)packet;
//short hello_seqnum = 1;     /* Hello packet counter which is use as unique identifier to HELLO PACKET */ /*commented by  Y.Ge */
//short tc_seqnum = 1;        /* TC packet counter which is use as unique identifier to TC PACKET */ /*commented by Y.Ge */
//short	  hna_seqnum = 1;	/* HNA packet counter which is use as unique identifier to HNA PACKET */ /*commented by Y.Ge */
				   
short	  pkt_seqnum[MAXIFACENUM];	/* added by Y.Ge, olsr packet counter */

/**************** added by Y.Ge *******************/
struct olsr_msg_seq	 msg_seqnum; //the unique message sequence number for each type of message

/* store the message that should be forwarded */
struct tc_message	tcpacket_relay;
struct mid_message	midpacket_relay;
struct hna_message	hnapacket_relay;
struct qos_tc_message	qostcpacket_relay; //qos OLSR	

/* store the unknown message that should be relayed */
char unknown_packet[MAXPACKETSIZE+1];

/* store the message sequence number update when it has to be chunked */
int hello_chunk[MAXIFACENUM];
int tc_chunk[MAXIFACENUM];
int mid_chunk[MAXIFACENUM];
int hna_chunk[MAXIFACENUM];

/* use hysteresis or not ? */
int use_hyst = 0; //DEFAULT value

/* for debug */
FILE	*y_file;

/* qos OLSR */
olsr_u32_t	my_busy_time = 0;
int       	use_qos = 0;
double   	bw_update_threshold = 0.2;
int		default_busy_time = 100;
/************** end of revision **********************/	

short	my_willingness = WILL_DEFAULT; /* added by Y.Ge, set willingness */		
								 
int	role = MANETNODE;
				   
/*float hello_int = 0.25, hello_jitter = 0.05 , hello_int_wjit = 0.25, tc_int = 1.0, tc_jitter = 0.25, polling_int = 0.025, min_tc_int = 0.5;
int neighbor_timeout_mult = 8;
int topology_timeout_mult = 5;
u_int8_t tos = 16;
float topology_hold_time = 5.0 ;
//topology_hold_time = tc_int*topology_timeout_mult;
float neighbor_hold_time = 2.0;
//neighbor_hold_time = hello_int*neighbor_timeout_mult;*/  //commented by  Y.Ge

/*************** added by Y.Ge ************/
/*** redefine the above value based on olsr_v11 ***/
float	hello_int = HELLO_INTERVAL, tc_int = TC_INTERVAL, min_tc_int = MIN_TC_INTERVAL;
float	mid_int = MID_INTERVAL;
float	hello_jitter = 0.05, hello_int_wjit = 0.25, tc_jitter = 0.25, mid_jitter = 0.25, hna_jitter = 0.25, polling_int = 0.025; //the jitter Should be changed later
float	neighbor_hold_time = NEIGHB_HOLD_TIME, topology_hold_time = TOP_HOLD_TIME, mid_hold_time = MID_HOLD_TIME;
float	duplicate_hold_time = DUPLICATE_HOLD_TIME;
int  	neighbor_timeout_mult = 3, topology_timeout_mult = 3;
u_int8_t	tos = 16;
/**************** end of revision **********/

int tos_bits = 0;
int precedence = 0;
int  getsocket(int, int, struct sockaddr *);
void process(int, int, int); //Y.Ge: add another int to record the interface index
int setmulticast(int, struct interface*);
void init();

// ***** added by L.Qin
// this eventloop will be used for whole OLSR module
EventLoop e;
XrlStdRouter tgt_router(e, g_tgt_name /*, finder.addr(), finder.port()*/); 
string hostRole="gw";
string wirelessCard="wlan0";
string mAddr="ff05::11";
int helloInter=10;
int tcInter=10;
int minitcInter=10;
int neiborMulti=10;
int topoMulti=10;
int tosValue=10;
int TCjitter=10;
int HELLOjitter=10;
int willing=2;
int debugLEVEL=2;
int numberofInterfaces=1;
int hystEnabled=10;
int qosEnabled=10;

//*****

/******************** the following is commented by Y.Ge ********************/
/* int main(int argc, char *argv[])*/

/***************** rewritten by Y.Ge *********************/
int main(int argc, char *argv[])
{
	int n, nfd, tflags = 0;
	struct timeval *tvp, waittime;
	struct itimerval itval;
	fd_set ibits;
//	sigset_t sigset, osigset; //commented by L. Qin, already defined before
	struct tc_message tcpack;
	struct hello_message hellopack;
	struct interface *ifp;
	int iface_index=0; 
	int main_addr_set = 0;
	int index; 
	int i;
	struct interf_name *nt;
	
	/********** added by Y.Ge *********/
	/* init message sequence number */
	msg_seqnum.hello_seqnum = 1;
	msg_seqnum.tc_seqnum = 1;
	msg_seqnum.mid_seqnum = 1;
	msg_seqnum.hna_seqnum = 1;
	
	/* init packet sequence number */
	for (i=0; i<MAXIFACENUM; i++)
	{
		pkt_seqnum[i] = 1;
	}
	
	midpacket_relay.mid_interface_address=NULL;
	tcpacket_relay.multipoint_relay_selector_address = NULL;
	hnapacket_relay.hna_net_addrs = NULL;
	qostcpacket_relay.multipoint_relay_selector_busy_time = NULL;
	
	/* init the chunk count array */
	reset_chunk_count(hello_chunk);
	reset_chunk_count(tc_chunk);
	reset_chunk_count(mid_chunk);
	reset_chunk_count(hna_chunk);
	
	/* open my own trace file */
	if ((y_file=fopen("olsr_debug","w"))==NULL)
	{
		printf("Error with creating ying's File.\n");
		exit(0);
	}
	/********* end of revision ********/

	init();
	
	openlog("crcolsr6d", LOG_PID | LOG_ODELAY, LOG_DAEMON);
	setlogmask(LOG_UPTO(LOG_WARNING));

	sp = (struct servent *) malloc( sizeof(*sp));
	if (sp == 0) 
	{
		printf("olsrd: out of memory\n");
		exit(1);
	}
	sp->s_name = "olsr";
	sp->s_aliases = NULL;
	sp->s_port = htons(OLSRPORT);
	sp->s_proto = "udp";
	
 	/*
	 * Process olsrd options.
 	 */

	 // ***** modefied by L.Qin
	// when merge OLSR into XORP, we cannot provide parameters in 
	// command line, now, fix some in the program, and may make 
	// change to be flexible in the future

// added by L.Qin, start the OLSR target module  *****
	xlog_init(argv[0], 0);
    	xlog_set_verbose(XLOG_VERBOSE_HIGH);
    	xlog_add_default_output();
    	xlog_start();
    
    	XrlOlsrTarget tgt_target(&tgt_router);
    	printf("*-*-*- start running olsr target\n");

   	bool timeout = false;
    	XorpTimer t = e.set_flag_after_ms(2000, &timeout);
    	while (timeout == false && tgt_router.ready() == false)
		e.run();

        //wait for while
 	bool tout=false;
	t = e.set_flag_after_ms(10000, &tout);
	while (!tout){
		e.run();
	}

	 //read number of interfaces value from user's setting:
	 tgt_target.olsr_3_0_get_interface_number(numberofInterfaces);
         printf("When from Target, the number of interfaces becomes %d\n", numberofInterfaces);

		/*
	 * To specify the name of an interface to be used by olsrd.
	 * Is required if more than one interface have configuured.
	 */

//        string wirelessCard;
        int len;
        tgt_target.olsr_3_0_get_interface_name(wirelessCard);
	if(numberofInterfaces==1){
		//if the node only has one wireless interface
        	len=wirelessCard.length();
        	char *winterface=new char[len+1];

        	wirelessCard.copy(winterface,len,0);
        	winterface[len]=0;

        	printf("*********___\n");
        	printf(" The interface name is %s %d\n",wirelessCard.c_str(),len);
	//        const char *winterface=wirelessCard.data();
        	printf (" The interface to the ifchoice %s\n",winterface);

		option_i=1;
		ifchoice(winterface);
	 }
         else{
		// if we have more than one interface, we have specify in the 
		//configure file in a format like "eth1 eth2 eth3"
		//so that we can get all the interface name here

		string buf; //have a buffer string
		stringstream ss(wirelessCard); //insert the string into a stream
		vector<string> tokens; //creat vector to hold the interface names
		while (ss >> buf){
			tokens.push_back(buf);
		}

		option_i=1;
	//assign each interface
		for(int ifNO=0;ifNO<numberofInterfaces;ifNO++){
			len=tokens[ifNO].length();
			char *winterface=new char[len+1];
        		tokens[ifNO].copy(winterface,len,0);
        		winterface[len]=0;
//	        	printf(" The interface name is %s %d\n",wirelessCard.c_str(),len);
		        printf (" The interface to the ifchoice %s\n",winterface);
			ifchoice(winterface);
		}
	}

	/*
	 * To specify the host role in the networt, it can be
	 * gateway, mobilenode or manetnode
	 */
         tgt_target.olsr_3_0_get_host_role(hostRole);
	 if(hostRole=="gw"){
		role=GATEWAY;
		hello_start=1;
	 }
	 if(hostRole=="mb"){
		role=MOBILENODE;
	 }
	 if(hostRole=="mn"){
		role=MANETNODE;
		hello_start=1;
	 }
	 printf(" The host role is %d\n",role);

	/*
	 * To specify the multicast address to be used by olsrd.
	 * Is required 
	 */
         tgt_target.olsr_3_0_get_multicast_addr(mAddr);
         len=mAddr.length();
         char *multicastAdd=new char[len+1];
         mAddr.copy(multicastAdd,len,0);
	 multicastAdd[len]=0;
	 printf(" The multicast address is %s\n",multicastAdd);
         
         if(inet_pton(AF_INET6, multicastAdd, &multiaddr)<0){
		printf("Error: No multicast address!");
	        exit(1);
         }

         //read hello interval from user's setting:
	 tgt_target.olsr_3_0_get_hello_interval(helloInter);
         printf("When from target,hello interval becomes %d\n", helloInter);
	 hello_int=helloInter;

         //read TC interval from user's setting:
	 tgt_target.olsr_3_0_get_tc_interval(tcInter);
         printf("When from target,it becomes %d\n", tcInter);
	 tc_int=tcInter;

         //read minimum TC interval from user's setting:
	 tgt_target.olsr_3_0_get_mini_tc_interval(minitcInter);
         printf("When from Target, mini tc interval becomes %d\n", minitcInter);
	 min_tc_int=minitcInter;
	
	 //read neighbor timeout mult from user's setting:
	 tgt_target.olsr_3_0_get_neighbor_timeout_mult(neiborMulti);
         printf("When from Target, neighbor timeout multi becomes %d\n", neiborMulti);
         neighbor_timeout_mult=neiborMulti;

	 //read topo timeout mult from user's setting:
	 tgt_target.olsr_3_0_get_topo_timeout_mult(topoMulti);
         printf("When from Target, topo timeout multi becomes %d\n", topoMulti);
         topology_timeout_mult=topoMulti;

	 //read tos value from user's setting:
	 tgt_target.olsr_3_0_get_tos(tosValue);
         printf("When from Target, tos becomes %d\n", tosValue);
         tos=tosValue;

	 //read TC jitter value from user's setting:
	 tgt_target.olsr_3_0_get_tc_jitter(TCjitter);
         printf("When from Target, TC jitter becomes %d\n", TCjitter);
	// convert it to float
         tc_jitter=TCjitter*0.01;

	 //read Hello jitter value from user's setting:
	 tgt_target.olsr_3_0_get_hello_jitter(HELLOjitter);
         printf("When from Target, Hello jitter becomes %d\n", HELLOjitter);
	// convert it to float
         hello_jitter=HELLOjitter*0.01;

	 //read node willingness value from user's setting:
	 tgt_target.olsr_3_0_get_willingness(willing);
         printf("When from Target, node willingness becomes %d\n", willing);
	 my_willingness=willing;

	 //read debug level value from user's setting:
	 tgt_target.olsr_3_0_get_debug_level(debugLEVEL);
         printf("When from Target, debug level becomes %d\n", debugLEVEL);
	 debug_level=debugLEVEL;


	 //read hysteresis value from user's setting, actually it's boolean
         //then we treated value other than 0 as enabled:
	 tgt_target.olsr_3_0_get_hyst_enabled(hystEnabled);
         printf("When from Target, hyst becomes %d\n", hystEnabled);
	 if(hystEnabled!=0){
		use_hyst = 1;
	 }
	 printf("do we use hyst? %d\n",use_hyst);

	 //read QoS value from user's setting, actually it's boolean
         //then we treated value other than 0 as enabled:
	 tgt_target.olsr_3_0_get_qos_enabled(qosEnabled);
         printf("When from Target, QoS becomes %d\n", qosEnabled);
	 if(qosEnabled!=0){
		use_qos = 1;
	 }
	// *****
	
	/* update values for hold_time after input values are read */

	neighbor_hold_time = hello_int * neighbor_timeout_mult;
	topology_hold_time = tc_int * topology_timeout_mult;
	mid_hold_time = topology_hold_time; //added by Y.Ge
	mid_int = tc_int; //added by Y.Ge
	hello_int_wjit = hello_int;

	/* printout settings */
	printf("CRCOLSR: CRC IPv6 HNA Version %s \n\
		OLSR6d Parameters are: \n\
		hello interval = %f \n\
		hello jitter = %f \n\
		tc interval = %f \n\
		tc jitter = %f \n\
		polling interval = %f \n\
		neighbor_hold_time = %f \n\
		topology_hold_time = %f \n\
		neighbor_timeout_mult = %d \n\
		topology_timeout_mult = %d \n\
		tos setting = %d \n\
		", VERSION, hello_int, hello_jitter, \
		tc_int, tc_jitter, polling_int, \
		neighbor_hold_time, topology_hold_time, \
		neighbor_timeout_mult, topology_timeout_mult, tos);

	/*itval.it_interval.tv_sec = TIMER_RATE;
	itval.it_value.tv_sec = TIMER_RATE; */

 	itval.it_interval.tv_sec = (int) polling_int;
	itval.it_value.tv_sec = (int) polling_int;	
	itval.it_interval.tv_usec = (int) ((polling_int-itval.it_interval.tv_sec) * 1000000);
	itval.it_value.tv_usec = (int) ((polling_int-itval.it_interval.tv_sec) * 1000000);
	printf("	Polling interval in secs is %d\n", itval.it_interval.tv_sec);
	printf("	Polling interval in usecs is %d\n", itval.it_interval.tv_usec);	
	printf("        Node's WILLINGNESS is: %d\n", my_willingness);
	
	if(role==MANETNODE)
		printf("	A MANET NODE \n");
	if(role==MOBILENODE)
		printf("	A MOBILE NODE \n");
	if(role==GATEWAY)
		printf("	A GATEWAY NODE \n");
		
	for (nt = interf_names; nt != NULL; nt = nt->next)
	{
		printf ("----- interface name: %s\n", nt->name);
	}
	
	/* CRC v4->v6 chg */
	ifinit();


	(void) gettimeofday(&now,(struct timezone *)NULL);

	/* Choice of the Originator address. In this version ifnet contain one interface */
	/* v4->v6, we use the site address as a interface's ipv6 address*/
	//v4->v6 add
	
	for (index = 0; index <MAXIFACENUM; index++)
	{
		memset(&addr[index], 0, sizeof (addr));
	}		
	
	for (ifp = ifnet; ifp; ifp = ifp->int_next) 
	{
		printf ("...now look at inteface: %s\n",ifp->int_name);
		printf ("   this interface address is: %s\n", convert_address_to_string(&ifp->int6_site_addr));
		
		memset(&addrsock6[ifp->my_index], 0, sizeof (struct sockaddr_in6));
		addrsock6[ifp->my_index].sin6_family = AF_INET6;
		addrsock6[ifp->my_index].sin6_port = sp->s_port;
		
		//memcpy(&(addrsock6[ifp->my_index].sin6_addr.s6_addr), &(ifp->int6_site_addr), sizeof(struct olsr_ip_addr));
		
		//printf (" now looking at index: %d\n", ifp->my_index);
		//printf (" port is: %d\n", addrsock6[ifp->my_index].sin6_port);
		
		
		if((s6[ifp->my_index] = getsocket(AF_INET6, SOCK_DGRAM, (struct sockaddr *)&addrsock6[ifp->my_index]))<0)
		{
			printf("getsocket error\n");
			exit(1);
		}
		
		/* bind to interface. WHY??? */
		if((setsockopt(s6[ifp->my_index], SOL_SOCKET, SO_BINDTODEVICE, ifp->int_name, strlen(ifp->int_name)+1)) < 0)
		{
	  		perror("setsockopt(SO_BINDTODEVICE) error");
	  		exit(1);
		}

		/* CRC v4->v6 add 
		 Enable the multicast for all interface and set the mulitcast group
		 For the time being it is ok because we have only one interface. if we
		 have multiple interface, can the setsockopt still be used?*/
		
		if(setmulticast(s6[ifp->my_index], ifp)<0)
		{
			printf("Error: set multicast address. \n");
			exit(1);
		}

		// set tos from MDP code 
	 	// v4->v6 I am not sure if those socket options are support by ipv6
		precedence = IPTOS_PREC(tos);
		if (setsockopt(s6[ifp->my_index], SOL_SOCKET, SO_PRIORITY, (char*)&precedence, sizeof(precedence)) < 0)           
		{     
			perror("UdpSocket: setsockopt(SO_PRIORITY) error");
		}                          
		tos_bits = IPTOS_TOS(tos);
		if (setsockopt(s6[ifp->my_index], SOL_IP, IP_TOS, (char*)&tos_bits, sizeof(tos_bits)) < 0)    
		{            
			perror("UdpSocket: setsockopt(IP_TOS) error");
		}
		// end of tos 
			
		printf("	Interface name:%s\n",ifp->int_name);
		printf("	Interface num:%d\n",ifp->int_ifindex);
		printf("	Interface index:%d\n", ifp->my_index);
		printf("	s6 is: %d\n", s6[ifp->my_index]);
		
		if (ifp->int_flags & IFF_PASSIVE )
		{
			printf (" continue!\n");
			continue;
		}
		
		/* choose the main address */
		if (!main_addr_set) 
		{
			memcpy(&addr_ip,&(ifp->int6_site_addr),sizeof(struct olsr_ip_addr));
			addr_ip_prefix = ifp->int6_site_prefix;
			memcpy(&my_main_addr, &addr_ip, sizeof(struct olsr_ip_addr)); //added by Y.Ge
			main_addr_set = 1; //added by Y.Ge
		}
		
		memcpy(&addr_manet[iface_index], &(ifp->int6_site_addr), sizeof(struct olsr_ip_addr)); //added by Y.Ge	

		//CRC hna added
		if((sd[ifp->my_index]=socket(AF_INET6,SOCK_RAW,IPPROTO_ICMP))<0)
		{		
			perror("socket");
			syslog(LOG_ERR, "socket: %m");
			return (-1);
		}
		if(role==MOBILENODE)
		{
			if(ioctl(sd[ifp->my_index],SIOCSHNA,(void*)ifp->int_name)<0)
			{
				perror("ioctl");
				syslog(LOG_ERR, "ioctl: %m");
				return (-1);
			}	    
			//close(sd);
		}

		addr[iface_index].sin6_family = AF_INET6;
		addr[iface_index].sin6_port = htons(OLSRPORT);
		olsr_if_index[iface_index] = ifp->int_ifindex; //the int_ifindex of a certain interface;
		memcpy(&(addr[iface_index].sin6_addr.s6_addr), &addr_manet[iface_index],sizeof(struct olsr_ip_addr));
		iface_index++;
		num_of_interface++;
	}

	printf ("........ the main address used in this node is: %s.............\n", convert_address_to_string(&my_main_addr));
	printf ("......... number of interface is: %d........................\n", num_of_interface);
	
	/* Initialisation of differents tables to be used.*/

	olsr_init_tables();
	memset(&saved_gateway,0,sizeof(struct gateway));
 
	if(role==MANETNODE)
	{
		//delete_subnet_default_entry(&addr_ip,addr_ip_prefix); //commented by Y.Ge
		
		/********************* added by Y.Ge ****************/
		for (ifp = ifnet; ifp; ifp = ifp->int_next)
		{
			delete_subnet_default_entry(&addr_manet[ifp->my_index],ifp->int6_site_prefix,ifp->my_index); 
		}
		/******************** end of revision *************/
	}
    
	if (filter)    
		load_topology_filter();
     
	/* 
	 * Build and send empty HELLO.
	 */
        
	for (ifp = ifnet; ifp; ifp = ifp->int_next) 
	{
		//added by L.Qin, whether the hello message is built for eth1
		printf("in the main, the interface index is: %d\n",ifp->my_index);
		
		hellopack.neighbors = (struct hello_neighbor *)NULL;
		olsr_build_hello_packet(&hellopack, ifp->my_index);
		hello_build((struct sockaddr *)NULL, &hellopack, (struct tc_message *)NULL, (struct mid_message *)NULL, (struct hna_message *)NULL, 0, ifp);
		tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index);
	}
        
	//toall(hello_build, &hellopack, (struct tc_message *)NULL,(struct	hna_message *)NULL,(struct interface *)NULL);
	//toall(hna_build, (struct hello_message *)NULL,(struct tc_message *)NULL,&hnainfo,(struct interface *)NULL);
	//toall(sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL);
			
	signal(SIGALRM, timer);
	signal(SIGINT, olsr_delete_all_kernel_routes);  
	signal(SIGUSR1, sigtrace);
	signal(SIGUSR2, sigtrace);

	srandom(getpid());
	if (setitimer(ITIMER_REAL, &itval, (struct itimerval *)NULL) < 0)
		syslog(LOG_ERR, "setitimer: %m\n");

	nfd=0;
	for (i=0; i<num_of_interface; i++)
	{
		if((s6[i] + 1) > nfd)
			nfd = s6[i] + 1;
	}
	
// added by L.Qin, start the OLSR target module  *****
//    xlog_init(argv[0], 0);
//    xlog_set_verbose(XLOG_VERBOSE_HIGH);
//    xlog_add_default_output();
//    xlog_start();
    
//    XrlOlsrTarget tgt_target(&tgt_router);
//    printf("*-*-*- start running olsr target\n");
    
//    bool timeout = false;
//    XorpTimer t = e.set_flag_after_ms(2000, &timeout);
//    while (timeout == false && tgt_router.ready() == false)
//	e.run();

//create a routing table in the RIB
    XrlRibNotifier xn;
    xn.startup();
    printf("-*-*-* create a RIB table\n");
    
    if (timeout) {
	fprintf(stderr, "Timed out connecting to finder\n");
	exit(-1);
    }
//*****

	for (;;)
	{
		//added by L.Qin *****
		
  		e.run();
		//*****
		FD_ZERO(&ibits);
		
		for(i = 0;i < num_of_interface; i++)
		{
	  		FD_SET(s6[i], &ibits);	
		}
		
		tvp = (struct timeval *)NULL;

		n = select(nfd, &ibits, 0, 0, tvp);
		
		if (n <= 0) 
		{
			if (n < 0) 
			{
				if (errno == EINTR)
				{
					continue;
				}
				syslog(LOG_ERR, "select: %m");
			}
		}

		(void) gettimeofday(&now, (struct timezone *)NULL);
		
		sigemptyset(&sigSet);
		sigaddset(&sigSet, SIGALRM);
		sigprocmask(SIG_BLOCK, &sigSet, &osigset);

		/*********** added by Y.Ge **********/
		for (i=0; i<num_of_interface; i++)
		{
			if (FD_ISSET (s6[i], &ibits))
			{
				if (debug_level > 5)
				{
					printf ("&&&&&&&&&&&&&&&&&& packet received from iface:%d &&&&&&&&&&&&&&&&&&\n", i);
					fprintf (y_file,"&&&&&&&&&&&&&&&&&& packet received from iface:%d &&&&&&&&&&&&&&&&&&\n", i);
					//exit (1);
				}
			//	printf ("&&&&&&&&&&&&&&&&&& packet received from iface:%d &&&&&&&&&&&&&&&&&&\n", i);
				process(s6[i], sd[i], i);
			}
		}
		/********* end of revision *********/

		// handle ICMP redirects 
		sigprocmask(SIG_SETMASK, &osigset, NULL);
	} /* for(;;) */	
	
	//added by L.Qin
	xlog_stop();
        xlog_exit();

} /* main */
/*************************** end of revision *******************/

void process(int fd, int qd, int index) //"index" is added by Y.Ge, which interface the packet is recieved?
{
	struct sockaddr_in6 from;
	size_t fromlen = sizeof(struct sockaddr_in6);
	int cc, rtflags=0;
	union 
	{
		char	buf[MAXPACKETSIZE+1];
		struct	olsr olsr;
	} inbuf;

	char str[46];
	struct sockaddr_in6* sinfrom; 
	struct olsr *olsrptr = (struct olsr*)&(inbuf.olsr);
	struct olsr_ip_addr	inaddrany = in6addr_any;

	for (;;) 
	{
		fromlen = sizeof (from);
		cc = recvfrom(fd, &inbuf, sizeof (inbuf), 0, (struct sockaddr*)&from, &fromlen);

		if (cc <= 0) {
			if (cc < 0 && errno != EWOULDBLOCK)
				perror("recvfrom");
				break;
		}
		/*v4->v6 del if (fromlen != sizeof (struct sockaddr_in))*/
		if (fromlen != sizeof (struct sockaddr_in6))
			break;

        sinfrom = (struct sockaddr_in6*)&from;
	//added by L.Qin
//	printf("**************main: the index %d\n",index);
        olsr_input((struct sockaddr*)&from, &inbuf.olsr, cc, index); //"index" is added by Y.Ge
        
        //CRC added
        if(role==MOBILENODE){
        
        	if(debug_level>6)
        	{
        		printf("before ioctl(sd,SIOCSRTDEL) in main.c\n");
        		printf("saved_gateway: %s\n",inet_ntop(AF_INET6,&(saved_gateway.gw_addr),str,sizeof(str)));
        	}
        	
        	//if((rtflags=ioctl(sd,SIOCSRTDEL,0))<0){
		if((rtflags=ioctl(qd,SIOCSRTDEL,0))<0){
				perror("ioctl");
				syslog(LOG_ERR, "ioctl: %m");
			}
			
			if(rtflags==1){
				//if(ioctl(sd,SIOCSFLAGSET,0)<0){
				if(ioctl(qd,SIOCSFLAGSET,0)<0){
				perror("ioctl");
				syslog(LOG_ERR, "ioctl: %m");
				}
				olsr_release_tables(1);
				olsr_init_tables();
				
				if(0==memcmp(&inaddrany,&saved_gateway.gw_addr,sizeof(struct olsr_ip_addr)))
				{
					rt_reset=1;
					refresh=1;
				}
				else
				{
					rt_reset=0;
					refresh=0;
     			}
				
				if(debug_level>6)
				{
					printf("********routingtable*********\n");
 					olsr_print_routing_table(routingtable);
 					printf("********mirror_table*********\n");
  					olsr_print_routing_table(mirror_table);
 				}
			}
		}	 
	}
}


int getsocket(int domain, int type, struct sockaddr *sa)
{
	/*v4->v6 del struct sockaddr_in *sin=(struct sockaddr_in *)sa;*/
	//CRC v4->v6 add
	struct sockaddr_in6 *sin = (struct sockaddr_in6 *)sa;
	int sock, on = 1, longueur;
	
	/* memset(sin, 0 sizeof (struct sockaddr_in6));
	sin->sin6_family = AF_INET6;
	sin->sin6_port = sp->s_port;*/

	printf ("++++++++++++++++ in getsocket +++++++++++++\n");
	printf ("address is:  %s\n", convert_address_to_string (&sin->sin6_addr));

	if ((sock = socket(domain, type, 0)) < 0) {
		perror("socket");
		syslog(LOG_ERR, "socket: %m");
		return (-1);
	}

/* CRC v4->v6 del
#ifdef SO_BROADCAST
	if (setsockopt(sock, SOL_SOCKET, SO_BROADCAST, &on, sizeof (on)) < 0)
	  {
		syslog(LOG_ERR, "setsockopt SO_BROADCAST: %m");
		close(sock);
		return (-1);
	  }
#endif

#ifdef SO_RCVBUF
	for (on = bufspace; ; on -= 1024) {
		if (setsockopt(sock, SOL_SOCKET, SO_RCVBUF,
		    &on, sizeof (on)) == 0)
			break;
		if (on <= 8*1024) {
			syslog(LOG_ERR, "setsockopt SO_RCVBUF: %m");
			break;
		}
	}
	if (traceactions)
		fprintf(ftrace, "recv buf %d\n", on);
#endif*/

	printf ("address is: %s\n", convert_address_to_string (&sin->sin6_addr));
	
	//MAOYU 
	/********** reuse the in6addr_any ***********/
	if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof (on)) < 0)
	  {
		syslog(LOG_ERR, "setsockopt SO_REUSEADDR: %m");
		close(sock);
		return (-1);
	  }	
	
	
	if (bind(sock, (struct sockaddr *)sin, sizeof (*sin)) < 0) 
	{
		perror("bind");
		syslog(LOG_ERR, "bind: %m");
		close(sock);
		return (-1);
	}
	
	if (fcntl(sock, F_SETFL, O_NONBLOCK) == -1)
		syslog(LOG_ERR, "fcntl O_NONBLOCK: %m\n");

	
	return (sock);
}

//CRC v4->v6 add
int setmulticast(int sock,struct interface* ifp)
{
    struct ipv6_mreq mreq6;
    int if_index;

    memcpy(&mreq6.ipv6mr_multiaddr,&ifp->int6_multi_addr, sizeof(struct in6_addr));
    mreq6.ipv6mr_interface = ifp->int_ifindex;
    if_index=ifp->int_ifindex;
 
    if(setsockopt(sock, IPPROTO_IPV6, IPV6_ADD_MEMBERSHIP, &mreq6, sizeof(mreq6))<0)
    return (-1);
    
    if (setsockopt(sock, IPPROTO_IPV6, IPV6_JOIN_GROUP, &mreq6, sizeof(mreq6))<0)
    return -1;
    
    if(setsockopt(sock, IPPROTO_IPV6, IPV6_MULTICAST_IF, &if_index, sizeof(if_index))<0)
    return (-1);

    return (0);  
}

//CRC v4->v6 add
void init()
{
	int i;
	
	//memset(&addr,0,sizeof(struct sockaddr_in6)); //commented by Y.Ge
	memset(&addr_ip,0,sizeof(struct olsr_ip_addr));
	memset(&addr_coa,0,sizeof(struct olsr_ip_addr));
	//memset(&addr_manet,0,sizeof(struct olsr_ip_addr)); //commented by Y.Ge
	for (i=0; i<MAXIFACENUM; i++)
	{
		memset(&addr_manet[i], 0, sizeof(struct olsr_ip_addr));
		//memset(&addrsock6[i], 0, sizeof(struct sockaddr_in6));
		memset(&addr[i],0,sizeof(struct sockaddr_in6));
	} //added by Y.Ge
	memset (&my_main_addr, 0, sizeof (struct olsr_ip_addr)); //added by Y.Ge
	memset(&multiaddr,0,sizeof(struct olsr_ip_addr));
	memset(&addrsock6,0,sizeof(struct sockaddr_in6)); //comment by Y.Ge
	memset(packet,0,MAXPACKETSIZE+1);
	memset(&saved_gateway,0,sizeof(struct gateway));
}

