/*
 * Copyright (c) 2004 Ying Ge, Communication Research Center Canada.
 *
 * Copyright (c) 2002, 2003 Maoyu Wang, Communication Research Center Canada.
 *
 * By Ying Ge:
 * 1. Change the OLSR packet format and message processing procedure based 
 *    on the OLSR RFC.
 * 2. Add support of multiple interfaces to OLSR, including MID message 
 *    creating and processing procedure as specified in the OLSR RFC.
 * 3. Add QoS Support to OLSR
 *
 * By Maoyu Wang:
 * 1. Ported OLSR from IPv4 to IPv6.
 * 2. Added the Host and Network Association (HNA) functionality into OLSR.
 * 3. Added the default gateway functionality into OLSR by extending the HNA 
 *    message usage. The default gateway functionality supported the mobility
 *	  by cooperating with Mobile IPv6 for a mobile node as well as supported 
 * 	  Internet access for MANET nodes. 
 *
 * DISTRIBUTED WITH NO WARRANTY, EXPRESS OR IMPLIED.
 * See the GNU Library General Public License (file COPYING in the distribution)
 * for conditions of use and redistribution
 */
/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est-à-dire qu'il peut
 * être utilisé, copié, modifié, distribué à l'unique
 * condition que ce texte soit conservé afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant participé à l'élaboration de ce logiciel ne peut
 * être utilisé sans son accord préalable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilité d'aucune sorte.
 * Certaines parties de ce logiciel sont dérivées de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 *
 *	$Id: defs.h,v 1.2 2000/12/06 10:36:11 prima Exp $
 */

#include <sys/param.h>
#include <sys/socket.h>
#include <time.h>
#include <sys/time.h>
#include <net/route.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>

#include <netdb.h>
#include <signal.h>
#include <syslog.h>
#include <stdio.h>
#include <stdlib.h>
#include <search.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#include "types.h"
#include "packet.h"
#include "trace.h"
#include "interface.h"
#include "two_hop_neighbor_table.h"
#include "neighbor_table.h"
#include "mpr_selector_table.h"
#include "duplicate_table.h"
#include "topology_table.h"
#include "routing_table.h"
#include "af.h"
#include "olsr.h"
#include "top_statique.h"
#include "kernel_routes.h"
#include "ins_rem_q.h"
#include "main_addr_neigh_table.h"
#include "interface_association_table.h"
//CRC HNA add
//#include "hna_duplicate_table.h" //commented by Y.Ge
#include "hna_table.h"

#define equal(a1, a2) \
	(memcmp((a1), (a2), sizeof (struct sockaddr_in6)) == 0)

#define UP             1
#define DOWN           0

#define TRUE           1

//#define NOT_DETECTED   0  //commented by Y.Ge
#define NOT_DETECTED	16 //added by Y.Ge

#define OLSRPORT       698
#define INT_TIMEOUT		3
#define OLSRTOOLSPORT  IPPORT_RESERVED - 1

#define MAXIFACENUM	5 //added by Y.Ge, number of interface in this node

//CRC HNA add
/*---------------------------------------------------------------------------------*/
#define GATEWAY		    1	/* the host acts as a gateway*/
#define MANETNODE	    2	/* the host acts as a manet node*/
#define MOBILENODE	    3	/* the host acts as a mobile node*/
#define SIOCSHNA  0x8982	/* send interface device to kernel for radvd*/ 
#define SIOCSRTDEL 0x8983 /* to get the routing table clear message from kernel*/
#define SIOCSFLAGSET 0x8984	/* to erease the routing table clear message to kernel*/
/*---------------------------------------------------------------------------------*/


extern struct timeval  hold_time_neighbor ;
extern struct timeval  hold_time_topology;
extern struct timeval  hold_time_duplicate_tc;

/**************** added by Y.Ge *************/
extern	struct timeval  hold_time_expired;
extern	struct timeval  hold_time_duplicate;
/**************** end of revision ***********/

//CRC HNA add
/*---------------------------------------------------------------------------------*/
extern struct timeval  hold_time_duplicate_hna;
extern struct timeval  hold_time_hna;
/*---------------------------------------------------------------------------------*/

extern sigset_t psigset;  /* set of blocked signals which have been raised while blocked */

extern olsr_8_t changes_topology;
extern olsr_8_t changes_neighborhood;

//CRC HNA add
/*---------------------------------------------------------------------------------*/
extern olsr_8_t	changes_net;
/*---------------------------------------------------------------------------------*/

extern struct neighbor_table             neighbortable;  

extern struct neighbor2_hash             neighbor2table[HASHSIZE];

extern struct duplicatehash              duplicatetable[HASHSIZE];

extern struct mpr_selector_table         mprstable;

extern struct rthash                     routingtable[HASHSIZE];
extern struct rthash                     mirror_table[HASHSIZE];

extern struct topology_destination_hash  topologytable[HASHSIZE];
extern struct topology_last_hash         topologylasttable[HASHSIZE];

/*************** added by Y.Ge **************/
extern struct main_addr_neighhash             mainaddrneightable[HASHSIZE];

extern struct main_addr_hash	nodemainaddresstable[HASHSIZE];
extern struct iface_addr_hash	nodeifaceaddresstable[HASHSIZE];
/*************** end of revision ************/

//CRC HNA add
/*---------------------------------------------------------------------------------*/
extern struct hna_duplicatehash          hna_duplicatetable[HASHSIZE];
extern struct hna_host_hash              hnahosttable[HASHSIZE];
extern struct hna_subnet_hash	         hnasubnettable[HASHSIZE];
extern struct hna_message				 hnainfo;
extern struct gateway					 saved_gateway;
/*---------------------------------------------------------------------------------*/


#if defined(__FreeBSD__)
extern olsr_16_t    kernel_raw_sock;    /*used to communicate with the kernel*/
#define SEQ 9999
#endif

extern FILE *f_r_id;
extern FILE *hys_id;
extern FILE *y_file; //added by Y.Ge

/***** added by Y.Ge *****/
/* qos OLSR */
extern olsr_u32_t	my_busy_time;
extern int       	use_qos;
extern double    	bw_update_threshold;
extern int       	default_busy_time;
/***** end of revision *****/

//CRC v4->v6 add
/*--------------------------------------------------------------------------------*/
//extern struct sockaddr_in6 addr;	/* address of daemon's socket */ //commented by Y.Ge
extern struct sockaddr_in6 addr[MAXIFACENUM]; //added by Y.Ge
extern struct olsr_ip_addr addr_ip;	/* IP address of "in_addr" */
olsr_u16_t	addr_ip_prefix;			/* Prefix of "addr_ip" */
extern struct olsr_ip_addr addr_coa; /* care of address, it's a location address*/
//extern struct olsr_ip_addr addr_manet; /* the address propagated in the manet*/ //commented by Y.Ge
extern struct olsr_ip_addr addr_manet[MAXIFACENUM]; //added by Y.Ge
extern struct olsr_ip_addr my_main_addr; /* added by Y.Ge */
extern struct olsr_ip_addr multiaddr; /* multicast addr using in hello and tc msg*/
extern olsr_u8_t hw_addr[8];	/* hardware address of the olsr interface*/
extern struct olsr_ip_addr default_gateway;
//extern struct sockaddr_in6 addrsock6;	/* ipv6 address of daemon's socket */ //commeented by Y.Ge
extern struct sockaddr_in6 addrsock6[MAXIFACENUM]; //added by Y.Ge
//extern int s6;			/* source and sink of all data */ //commented by Y.Ge
extern int s6[MAXIFACENUM];	//added by Y.Ge
extern int olsr_if_index[MAXIFACENUM]; //added by Y.Ge, keep the "int_ifindex" of "struct interface" of all OLSR interfaces
/*-------------------------------------------------------------------------------*/
extern int supplier;		/* process should supply updates */
extern char	packet[MAXPACKETSIZE+1];
extern int      outputsize;
extern struct	olsr *msg;
/*extern short      hello_seqnum;*/  /* counter use to identifie HELLO PACKET */ /*commented by Y.Ge */
/*extern short      tc_seqnum;*/     /* counter use to identifie TC PACKET */ /*commented by Y.Ge */
/*extern short      hna_seqnum;*/     /* counter use to identifie HNA PACKET */ /*commented by Y.Ge */
extern char	unknown_packet[MAXPACKETSIZE+1]; /* added by Y.Ge, remember the unknown packet that should be forwarded */
extern short	  pkt_seqnum[MAXIFACENUM];    /* added by Y.Ge, counter use to identifie OLSR PACKET*/
extern struct olsr_msg_seq	msg_seqnum;	/*  added by Y.Ge, counter use to identifie OLSR MESSAGE */
extern struct tc_message	tcpacket_relay;  /* added by Y.Ge, remember the packet should be relayed out */
extern struct mid_message	midpacket_relay; /* added by Y.Ge, remember the packet should be relayed out */
extern struct hna_message	hnapacket_relay; /* added by Y.Ge, remember the packet should be relayed out */
extern struct qos_tc_message	qostcpacket_relay; /* added by Y.Ge, remember the packet should be relayed out */
extern int 			hello_chunk[MAXIFACENUM]; //added by Y.Ge
extern int 			tc_chunk[MAXIFACENUM]; //added by Y.Ge
extern int 			mid_chunk[MAXIFACENUM]; //added by Y.Ge
extern int 			hna_chunk[MAXIFACENUM]; //added by Y.Ge
extern short	my_willingness; /* added by Y.Ge, my willingness */
extern int	tc_to_relay;	/* added by Y.Ge, if node should relay a received tc */
extern int	qostc_to_relay;	/* added by Y.Ge, if node should relay a received qos tc */
extern int	mid_to_relay;	/* added by Y.Ge, if node should relay a received mid */
extern int 	hna_to_relay; 	/*added by Y.Ge, if the node should relay the hna */
extern int 	tc_relay_size;	/* added by Y.Ge. the size of the TC that should be relayed */
extern int 	mid_relay_size; /* added by Y.Ge. the size of the MID that should be relayed */
extern int 	hna_relay_size; /* added by Y.Ge. the size of the HNA that should be relayed */
extern int 	qostc_relay_size; /* added by Y.Ge. the size of the QoS TC that should be relayed */
extern float	tc_relay_jitter;  /* added by Y.Ge */
extern float	mid_relay_jitter; /* added by Y.Ge */ 
extern float 	hna_relay_jitter;	//added by Y.Ge
extern float	tc_relay_interval; /* added by Y.Ge */
extern float	mid_relay_interval; /* added by Y.Ge */
extern float 	hna_relay_interval;	//added by Y.Ge
extern int	use_hyst; /* added by Y.Ge, use hysteresis or not */
extern struct	servent *sp;
extern struct timeval now;		/* current idea of time */
extern struct tm *nowtm;		/* current idea of time (in tm) */
extern int changes;
extern int debug, debug_level;
extern float hello_int, hello_int_wjit, hello_jitter, tc_int, tc_jitter, mid_int, mid_jitter, hna_jitter, polling_int, min_tc_int;
extern float topology_hold_time, neighbor_hold_time, duplicate_hold_time, mid_hold_time;

//CRC HNA add
/*--------------------------------------------------------------------------------*/
//extern int sd; /*Raw Socket for ioctl()*/ //commented by Y.Ge
extern int sd[MAXIFACENUM]; //added by Y.Ge
extern int role;	/*define this host as a gateway,or a mobile node, or a manet node*/
extern int print_time_debug; //added by Y.Ge, for debugging "timeout" functions
extern short hello_start;
extern int rt_reset, refresh;
extern int num_of_interface; //added by Y.Ge
/*--------------------------------------------------------------------------------*/

u_long inet_netof_subnet(struct in_addr);
int inet_rtflags(struct sockaddr *);
int inet_sendroute(struct rt_entry *, struct sockaddr *);

//CRC v4->v6 add
/*--------------------------------------------------------------------------------*/
int inet6_rtflags(struct sockaddr *sa);
int inet6_sendroute(struct rt_entry *rt, struct sockaddr *sa);
/*--------------------------------------------------------------------------------*/

void olsr_input(struct sockaddr *, struct olsr *, int, int); //Y.Ge: add another int to record the index of the interface

void toall(void (*)(struct sockaddr *, struct hello_message *, struct tc_message *, struct mid_message *, struct hna_message *, int,  struct interface *), 
			struct hello_message *, struct tc_message *, struct mid_message *, struct hna_message *, struct interface *);

void tointerface (void (*)(struct sockaddr *, struct hello_message *, struct tc_message *, struct mid_message *, struct hna_message *, int,  struct interface *), 
			struct hello_message *, struct tc_message *, struct mid_message *, struct hna_message *, struct interface *, int); //added by Y.Ge

void sndmsg(struct sockaddr *, struct hello_message *, struct tc_message *, struct mid_message *,
	struct hna_message *, int, struct interface *);

void supply(struct sockaddr *, struct hello_message *, struct rt_message *, int, int); //last "int" is added by Y.Ge to define interface

void hello_build(struct sockaddr *, struct hello_message *, 
                 struct tc_message *, struct mid_message *, struct hna_message *, int, struct interface *);

void tc_build(struct sockaddr *, struct hello_message *, 
              struct tc_message *, struct mid_message *, struct hna_message *, int, struct interface *, int); //last "int" added by Y.Ge, to tell whether the packet is the relayed one

void hna_build(struct sockaddr *, struct hello_message *, 
              struct tc_message *, struct mid_message *, struct hna_message *, int, struct interface *, int);  //last "int" added by Y.Ge, to tell whether the packet is the relayed one

void mid_build(struct sockaddr *, struct hello_message *,
		struct tc_message *, struct mid_message *, struct hna_message *,int, struct interface *, int);  //last "int" added by Y.Ge, to tell whether the packet is the relayed one


void rt_build(struct sockaddr *, struct rt_message *, int, 
	      struct interface *);

void timer(int);

void olsr_delete_all_kernel_routes(int signal);

void olsr_del_all_kernel_routes();

int delete_subnet_default_entry(struct olsr_ip_addr* addrptr, olsr_u16_t addr_prefix, int iface_index); //"iface_index" added by Y.Ge

int add_subnet_default_entry(struct olsr_ip_addr* addrptr, olsr_u16_t addr_prefix, int iface_index); //"iface_index" added by Y.Ge

void ifinit(void);

/********* added by Y.Ge **********/
/* deal with packet sequence number of 
 * the chunked packet */
void reset_chunk_count(int *array); 

int check_chunk_count (int *array); 

/* qos OLSR */
void qoshello_build(struct sockaddr *, int, struct interface *);

void qostc_build(struct sockaddr *, struct qos_tc_message *, struct tc_message *, int, struct interface *, int);
/********* end of revision ********/
