/*
 * Copyright (c) 2004 Ying Ge, Communication Research Center Canada.
 *
 * Copyright (c) 2002, 2003 Maoyu Wang, Communication Research Center Canada.
 *
 * By Ying Ge:
 * 1. Change the OLSR packet format and message processing procedure based 
 *    on the OLSR RFC.
 * 2. Add support of multiple interfaces to OLSR, including MID message 
 *    creating and processing procedure as specified in the OLSR RFC.
 * 3. Add QoS Support to OLSR
 *
 * By Maoyu Wang:
 * 1. Ported OLSR from IPv4 to IPv6.
 * 2. Added the Host and Network Association (HNA) functionality into OLSR.
 * 3. Added the default gateway functionality into OLSR by extending the HNA 
 *    message usage. The default gateway functionality supported the mobility
 *	  by cooperating with Mobile IPv6 for a mobile node as well as supported 
 * 	  Internet access for MANET nodes. 
 *
 * DISTRIBUTED WITH NO WARRANTY, EXPRESS OR IMPLIED.
 * See the GNU Library General Public License (file COPYING in the distribution)
 * for conditions of use and redistribution
 */
/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est-à-dire qu'il peut
 * être utilisé, copié, modifié, distribué à l'unique
 * condition que ce texte soit conservé afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant participé à l'élaboration de ce logiciel ne peut
 * être utilisé sans son accord préalable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilité d'aucune sorte.
 * Certaines parties de ce logiciel sont dérivées de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 */

/*
 * Copyright (c) 1983, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * From: @(#)if.c	5.6 (Berkeley) 6/1/90
 * From: @(#)if.c	8.1 (Berkeley) 6/5/93
 * From: @(#)if.c	1.4 (dholland Exp)  1996/11/25
 */

char if_rcsid[] = 
  "$Id: if.c,v 1.2 2000/12/06 10:36:11 prima Exp $";

/*
 * Routing Table Management Daemon
 */

#include "defs.h"

/*
 * Find the interface with address addr.
 */

struct interface *if_ifwithaddr(struct sockaddr *addr)
{
	struct interface *ifp;

#define	same(a1, a2) \
	(memcmp((a1)->sa_data, (a2)->sa_data, 14) == 0) 
 
       for (ifp = ifnet; ifp; ifp = ifp->int_next) {

       if (ifp->int_addr.sa_family != addr->sa_family)
			continue;

       if (same(&ifp->int_addr, addr))
			break;

       if ((ifp->int_flags & IFF_BROADCAST) &&
		    same(&ifp->int_broadaddr, addr))
			break;
       	}
	return (ifp);
}

//CRC v4->v6 add
struct interface *if_ifwithaddr6(struct sockaddr *addr)
{
	struct sockaddr_in6 * sin = (struct sockaddr_in6*) addr;
	struct interface *ifp=NULL;
	char str[46];
	
	if(debug_level>6)
		printf("if_ifwithaddr6: %s\n", inet_ntop(AF_INET6,&sin->sin6_addr,str,46));
	
	for (ifp = ifnet; ifp; ifp = ifp->int_next) 
	{
		if (!memcmp(&ifp->int6_site_addr,&sin->sin6_addr,sizeof(struct olsr_ip_addr)))
			break;
 		if (!memcmp(&ifp->int6_link_addr,&sin->sin6_addr,sizeof(struct olsr_ip_addr)))
			break;
		if (!memcmp(&ifp->int6_multi_addr,&sin->sin6_addr,sizeof(struct olsr_ip_addr)))
			break;
		if (!memcmp(&ifp->int6_hw_addr[0],&sin->sin6_addr.s6_addr[8],8))
	   		break;  
	   /*future: add the handle for global addr*/
	}
       return (ifp);
}


/*
 * Find an interface from which the specified address
 * should have come from.  Used for figuring out which
 * interface a packet came in on -- for tracing.
 */
struct interface *if_iflookup(struct sockaddr *addr)
{
	struct interface *ifp, *maybe;
	int af = addr->sa_family;
	int (*netmatch)(struct sockaddr *, struct sockaddr *);

	if (af >= af_max)
		return (0);
	maybe = 0;
	netmatch = afswitch[af].af_netmatch;
	for (ifp = ifnet; ifp; ifp = ifp->int_next) {
		if (ifp->int_addr.sa_family != af)
			continue;
		if (same(&ifp->int_addr, addr))
			break;
		if ((ifp->int_flags & IFF_BROADCAST) &&
		    same(&ifp->int_broadaddr, addr))
			break;
		if (maybe == 0 && (*netmatch)(addr, &ifp->int_addr))
			maybe = ifp;
	}
	if (ifp == 0)
		ifp = maybe;
	return (ifp);
}

//RC v4->v6
struct interface *if_iflookup6(struct sockaddr *addr)
{
	struct interface *ifp, *maybe;
	struct sockaddr_in6 * sin = (struct sockaddr_in6*)addr;
	int af = sin->sin6_family;

	if (af >= af_max)
		return (0);
	maybe = 0;
	for (ifp = ifnet; ifp; ifp = ifp->int_next) {
		if (!memcmp(&ifp->int6_link_addr, &(sin->sin6_addr), sizeof(struct olsr_ip_addr)))
			break;
		if (!memcmp(&ifp->int6_site_addr, &(sin->sin6_addr), sizeof(struct olsr_ip_addr)))
			break;
		if (!memcmp(&ifp->int6_multi_addr, &(sin->sin6_addr), sizeof(struct olsr_ip_addr)))
			break;
		/* future: add function to handle global address and netmatch part, such as
		 *if (maybe == 0 && (*netmatch)(addr, &ifp->int_addr6))
			maybe = ifp;*/
	}
	if (ifp == 0)
		ifp = maybe;
	return (ifp);
}
