/*
 * Copyright (c) 2004 Ying Ge, Communication Research Center Canada.
 *
 * Copyright (c) 2002, 2003 Maoyu Wang, Communication Research Center Canada.
 *
 * By Ying Ge:
 * 1. Change the OLSR packet format and message processing procedure based 
 *    on the OLSR RFC.
 * 2. Add support of multiple interfaces to OLSR, including MID message 
 *    creating and processing procedure as specified in the OLSR RFC.
 * 3. Add QoS Support to OLSR
 *
 * By Maoyu Wang:
 * 1. Ported OLSR from IPv4 to IPv6.
 * 2. Added the Host and Network Association (HNA) functionality into OLSR.
 * 3. Added the default gateway functionality into OLSR by extending the HNA 
 *    message usage. The default gateway functionality supported the mobility
 *	  by cooperating with Mobile IPv6 for a mobile node as well as supported 
 * 	  Internet access for MANET nodes. 
 *
 * DISTRIBUTED WITH NO WARRANTY, EXPRESS OR IMPLIED.
 * See the GNU Library General Public License (file COPYING in the distribution)
 * for conditions of use and redistribution
 */
/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est-à-dire qu'il peut
 * être utilisé, copié, modifié, distribué à l'unique
 * condition que ce texte soit conservé afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant participé à l'élaboration de ce logiciel ne peut
 * être utilisé sans son accord préalable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilité d'aucune sorte.
 * Certaines parties de ce logiciel sont dérivées de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 *
 */

char input_rcsid[] = 
  "$Id: input.c,v 1.2 2000/12/06 10:36:11 prima Exp $";


/*
 * Routing Table Management Daemon
 */

#include "defs.h"

extern int filter;
struct timeval now;		/* current idea of time */
struct tm snowtm;
struct tm *nowtm=&snowtm;	/* current idea of time (in tm) */
int changes = 0;        /* is set if changes occur in MPRS set */ 
int tc_to_relay = 0;	/*added by Y.Ge, if the node should relay the tc */
int qostc_to_relay = 0;	/*added by Y.Ge, if the node should relay the qos tc */
int mid_to_relay = 0;	/*added by Y.Ge, if the node should relay the mid */
int hna_to_relay = 0; 	/*added by Y.Ge, if the node should relay the hna */
int tc_relay_size = 0;	/* added by Y.Ge. the size of the TC that should be relayed */
int mid_relay_size = 0; /* added by Y.Ge. the size of the MID that should be relayed */
int hna_relay_size = 0; /* added by Y.Ge. the size of the HNA that should be relayed */
int qostc_relay_size = 0; /* added by Y.Ge, the size of QoS HNA that should be relayed */

void tools_input(struct sockaddr *from, struct afswitch *afp, struct olsr *olsr, int size, int iface_index); //"iface_index" is added by Y.Ge 

/*
 * Build HELLO Message according to hello_message struct from olsrmsg struct.
 */

void hello_chgestruct(struct hello_message *hmsg, struct olsrmsg *m, struct sockaddr *from, int index)
{
 	struct hellomsg *h;
	struct hellinfo *hinfo, *hinf;
	//v4->v6 chg
	struct olsr_ip_addr *haddr, *hadr;
	struct hello_neighbor *nb;
	olsr_u8_t link; //added by Y.Ge
        
	hmsg->neighbors = (struct hello_neighbor *)NULL;

	if (!m)
		return;

	if (m->olsr_msgtype != HELLO_PACKET && m->olsr_msgtype != UPDATE_NEIGHBOR)
		return;

	h = m->olsr_hello;
	hinfo = h->hell_info;

	/*CRC v4->v6 del
	hmsg->source_addr = ((struct sockaddr_in *)from)->sin_addr.s_addr;*/
	//CRC v4->v6 add
	/*memcpy(&hmsg->source_addr,&(((struct sockaddr_in6 *)from)->sin6_addr),sizeof(struct olsr_ip_addr));*/
	//CRC HNA add
	//memcpy(&hmsg->source_addr,&(m->olsr_source),sizeof(struct olsr_ip_addr)); //commented by Y.Ge
	memcpy (&hmsg->source_addr,&(((struct sockaddr_in6 *)from)->sin6_addr), sizeof(struct olsr_ip_addr)); //added by Y.Ge, set the interface addr
	memcpy (&hmsg->originator_addr,&(m->olsr_source), sizeof(struct olsr_ip_addr)); //added by Y.Ge, set the originator main addr 
	memcpy (&hmsg->my_addr, &addr_manet[index], sizeof(struct olsr_ip_addr)); //added by Y.Ge
	/*---------------------------------------------------------------------*/	
	hmsg->v_time = m->olsr_vtime; //added by Y.Ge
	hmsg->packet_seq_number = ntohs(m->olsr_msgseq); //added by Y.Ge

	//hmsg->packet_seq_number = ntohs(h->hell_seqnum);  //commented by Y.Ge
	//hmsg->mpr_seq_number = ntohs(h->hell_mprseqnum);  //commented by Y.Ge

	hmsg->willingness = h->hell_willingness; //added by Y.Ge
	hmsg->rcvd_h_time = h->hell_htime;	//added by Y.Ge

	/********************** added by Y.Ge ******************/
	/*** debug ***/
	if (debug_level >3)
	{
		printf ("--------THE RECEIVED MESSAGE IS A HELLO with following information!\n");
		printf ("        sender's interface address: %s\n", convert_address_to_string (&hmsg->source_addr));
		printf ("        sender's originator address: %s\n", convert_address_to_string (&hmsg->originator_addr));
		printf ("        received from my interface: %s\n", convert_address_to_string (&hmsg->my_addr));
		printf ("        sender's willingness is: %d\n", hmsg->willingness);
		printf ("        sender has the following neighbor information:\n");
		
		fprintf (y_file,"--------THE RECEIVED MESSAGE IS A HELLO with following information!\n");
		fprintf (y_file,"        sender's interface address: %s\n", convert_address_to_string (&hmsg->source_addr));
		fprintf (y_file,"        sender's originator address: %s\n", convert_address_to_string (&hmsg->originator_addr));
		fprintf (y_file,"        received from my interface: %s\n", convert_address_to_string (&hmsg->my_addr));
		fprintf (y_file,"        sender's willingness is: %d\n", hmsg->willingness);
		fprintf (y_file,"        sender has the following neighbor information:\n");
	}
	/********************* end of revision *****************/
	
	/* the following "for" is commented and re-writed by Y.Ge **/	
	/*for (hinf = hinfo; (char *)hinf < (char *)m + ntohs(m->olsr_nextmsg);  hinf = (struct hellinfo *)((char *)hinf + ntohs(hinf->next_linktype)))*/	
	for (hinf = hinfo; (char *)hinf < (char *)m + ntohs(m->olsr_msgsize);  hinf = (struct hellinfo *)((char *)hinf + ntohs(hinf->link_msg_size)))
	{
		haddr = hinf->neigh_addr;

		/* the following "for" is commented and re-writed by Y.Ge **/
		//for (hadr = haddr; (char *)hadr < (char *)hinf + ntohs(hinf->next_linktype); hadr++)
		for (hadr = haddr; (char *)hadr < (char *)hinf + ntohs(hinf->link_msg_size); hadr++)
		{
			nb = (struct hello_neighbor *)malloc(sizeof (struct hello_neighbor));
			if (nb == 0)
			{
				printf("olsrd: out of memery \n");
				break;
			}
			/*CRC v4->v6 del nb->address = *hadr;*/
			//CRC v4->v6 add
			memcpy(&(nb->address),hadr,sizeof(struct olsr_ip_addr));		
			//nb->status = hinf->link_type;  //commented by Y.Ge
			/**************** added by  Y.Ge ************/
			link = hinf->link_code;
			
			if (debug_level > 3)
			{
				printf ("--link code: %d\n", link);
				fprintf (y_file,"--link code: %d\n", link);
			}
			
			if (link>15)  //link code must less than 15
			{
				printf ("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\n");
				printf ("XX ATTENTION: the link code in the received HELLO is invalid! XX\n");
				printf ("XXXXXXXXXXXXXXXXXXX     link code is: %d    XXXXXXXXXXXXXXXXXXXX\n", link);
				printf ("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\n");
				exit (1);
			}
			nb->neigh_status = (int)link/4;
			nb->link_status = link - (nb->neigh_status * 4);
			
			if (nb->link_status == SYM_LINK && nb->neigh_status == NOT_NEIGH)
			{
				printf ("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\n");
				printf ("XX ATTENTION: the status in the received HELLO is invalid! XX\n");
				printf ("XXXXXXXXXXXXX hello neighbor: %s XXXXXXXXXXX\n", convert_address_to_string(&nb->address));
				printf ("XXXXXXXXXXX neighbor_status: %d, link_status: %d XXXXXXXXXX\n", nb->neigh_status, nb->link_status);
				printf ("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\n");
				exit (1);
			}
			/**************** end of revision **********/
			nb->next = hmsg->neighbors;
			hmsg->neighbors = nb;

			/************************** added by Y.Ge *********************/
			/*** debug ***/
			if (debug_level >3)
			{
				printf ("   hello neighboor: %s,  neighbor_status: %d, link_status: %d\n", convert_address_to_string(&nb->address), nb->neigh_status, nb->link_status);
				//added by L.Qin
//				printf ("   input.c the neighbor_hysavg is %f\n",nb->neighbor_hysavg);	
				fprintf (y_file,"   hello neighboor: %s,  neighbor_status: %d, link_status: %d\n", convert_address_to_string(&nb->address), nb->neigh_status, nb->link_status);
			}
			/************************** end of revision  ******************/
		} 
	    
	}
}
/******************** added by Y.Ge ***********/
/*
 * Build QoS HELLO Message according to qos_hello_message struct from olsrmsg struct.
 */

void qoshello_chgestruct(struct qos_hello_message *qoshmsg, struct olsrmsg *m, struct sockaddr *from, int index)
{
 	struct qoshellomsg *qosh;
	olsr_u32_t	*in_busy, *out_busy;
	int        	count = 0;
        
	if (!m)
		return;

	if (m->olsr_msgtype != QOSHELLO_PACKET && m->olsr_msgtype != UPDATE_NEIGHBOR)
		return;

	qosh = m->olsr_qos_hello;
	in_busy = qosh->hell_busy;

	memcpy (&qoshmsg->source_addr,&(((struct sockaddr_in6 *)from)->sin6_addr), sizeof(struct olsr_ip_addr)); //set the interface addr
	memcpy (&qoshmsg->originator_addr,&(m->olsr_source), sizeof(struct olsr_ip_addr)); //set the originator main addr 
	memcpy (&qoshmsg->my_addr, &addr_manet[index], sizeof(struct olsr_ip_addr)); //added by Y.Ge
	qoshmsg->packet_seq_number = ntohs(m->olsr_msgseq);
	
	/*** debug ***/
	if (debug_level >3)
	{
		printf ("--------THE RECEIVED MESSAGE IS A QoS HELLO with following information!\n");
		printf ("        sender's interface address: %s\n", convert_address_to_string (&qoshmsg->source_addr));
		printf ("        sender's originator address: %s\n", convert_address_to_string (&qoshmsg->originator_addr));
		printf ("        received from my interface: %s\n", convert_address_to_string (&qoshmsg->my_addr));
		
		fprintf (y_file,"--------THE RECEIVED MESSAGE IS A QoS HELLO with following information!\n");
		fprintf (y_file,"        sender's interface address: %s\n", convert_address_to_string (&qoshmsg->source_addr));
		fprintf (y_file,"        sender's originator address: %s\n", convert_address_to_string (&qoshmsg->originator_addr));
		fprintf (y_file,"        received from my interface: %s\n", convert_address_to_string (&qoshmsg->my_addr));
	}
	
	for(out_busy = in_busy; (char *)out_busy < (char *)m + ntohs(m->olsr_msgsize); out_busy ++)
	{
		qoshmsg->rcvd_busy_time = *out_busy;
		count +=1;
		
		if (debug_level > 3)
		{
			printf ("received busy time: %d\n", qoshmsg->rcvd_busy_time);
			fprintf (y_file,"received busy time: %d\n", qoshmsg->rcvd_busy_time);
			
		}
	}
	
	if (count > 1)
	{
		printf ("XXXXXXXXXXX error in reqding the QoS message ! XXXXXXXX\n");
		exit (1);
	}	
}

/*
 * Build QoS TC Message according to qos_tc_message struct from olsrmsg struct.
 */

void qostc_chgestruct(struct qos_tc_message *qostmsg, struct olsrmsg *m, struct sockaddr *from, int index)
{

	struct qostcmsg *qostc;
	olsr_u32_t *mprsbusy, *mbusy;
	struct tc_mpr_busy_time *mprs_busy_time;

	qostmsg->multipoint_relay_selector_busy_time = (struct tc_mpr_busy_time *)NULL;

	if (!m)
		return;

	if (m->olsr_msgtype != QOSTC_PACKET)
		return;

	qostc = m->olsr_qos_tc;
	mprsbusy = qostc->tc_mpr_busy;
	
	memcpy(&(qostmsg->source_addr),&(((struct sockaddr_in6 *)from)->sin6_addr),sizeof(struct olsr_ip_addr)); 
	memcpy(&(qostmsg->originator),&(m->olsr_source),sizeof(struct olsr_ip_addr));
	memcpy (&qostmsg->my_addr, &addr_manet[index], sizeof(struct olsr_ip_addr));
	
	qostmsg->packet_seq_number = ntohs(m->olsr_msgseq);
	qostmsg->hop_count = m->olsr_hopcount;
	qostmsg->time_to_live = m->olsr_ttl;
	qostmsg->my_busy_time = qostc->qostc_originator_busy; 
	qostmsg->mssn =  ntohs(qostc->qostc_mprseqnum);
	
	if (debug_level >3)
	{
		printf ("--------THE RECEIVED MESSAGE IS A QoS TC with following information!\n");
		printf ("        sender's interface address: %s\n", convert_address_to_string (&qostmsg->source_addr));
		printf ("        sender's originator address: %s\n", convert_address_to_string (&qostmsg->originator));
		printf ("        received from my interface: %s\n", convert_address_to_string (&qostmsg->my_addr));
		printf ("	 originator's busy time: %d\n", qostmsg->my_busy_time);
		printf ("        sender has the following MPR selector busy time information:\n");
		
		fprintf (y_file,"--------THE RECEIVED MESSAGE IS A QoS TC with following information!\n");
		fprintf (y_file,"        sender's interface address: %s\n", convert_address_to_string (&qostmsg->source_addr));
		fprintf (y_file,"        sender's originator address: %s\n", convert_address_to_string (&qostmsg->originator));
		fprintf (y_file,"        received from my interface: %s\n", convert_address_to_string (&qostmsg->my_addr));
		fprintf (y_file,"	 originator's busy time: %d\n", qostmsg->my_busy_time);
		fprintf (y_file,"        sender has the following MPR selector busy time information:\n");
	}

	for (mbusy = mprsbusy; (char *)mbusy < (char *)m + ntohs(m->olsr_msgsize);  mbusy++)
	{
		mprs_busy_time = (struct tc_mpr_busy_time *)malloc(sizeof(struct tc_mpr_busy_time));
		if (mprs_busy_time == 0)
		{
			printf("olsrd: out of memery \n");
			break;
		}
		
		memcpy(&(mprs_busy_time->busy_time),mbusy,sizeof(olsr_u32_t));
		mprs_busy_time->next = qostmsg->multipoint_relay_selector_busy_time;
		qostmsg->multipoint_relay_selector_busy_time = mprs_busy_time;
		
		/*** debug ***/
		if (debug_level >3)
		{
			printf ("+++ mpr selector busy time: %d\n", mprs_busy_time->busy_time);
			
			fprintf (y_file, "+++ mpr selector busy time: %d\n", mprs_busy_time->busy_time);
		}
	}
}

/******************** end of revision **********/

/*
 * Build TC Message according to tc_message struct from olsrmsg struct.
 */

void tc_chgestruct(struct tc_message *tmsg, struct olsrmsg *m, struct sockaddr *from, int index)
{

	struct tcmsg *tc;
	//CRC v4->v6 chg
	struct olsr_ip_addr *mprsaddr, *maddr;
	struct tc_mpr_addr *mprs;

	tmsg->multipoint_relay_selector_address = (struct tc_mpr_addr *)NULL;

	if (!m)
		return;

	if (m->olsr_msgtype != TC_PACKET)
		return;

	tc = m->olsr_tc;
	mprsaddr = tc->tc_mprsaddr;

	/*CRC v4->v6 del
	tmsg->source_addr = ((struct sockaddr_in *)from)->sin_addr.s_addr;
	tmsg->originator = tc->tc_origaddr; */
	
	
	//CRC v4->v6 add
	/*memcpy(&(tmsg->source_addr),&(((struct sockaddr_in6 *)from)->sin6_addr),sizeof(struct olsr_ip_addr));*/
	//CRC HNA add
	/*-------------------------------------------------------------------*/
	/*memcpy(&(tmsg->source_addr),&(m->olsr_source),sizeof(struct olsr_ip_addr));*/ //commented by Y.Ge
	/*-------------------------------------------------------------------*/
	/*memcpy(&(tmsg->originator),&(tc->tc_origaddr),sizeof(struct olsr_ip_addr));*/ //commented by Y.Ge
	
	/********************** added by Y.Ge ********************/
	memcpy(&(tmsg->source_addr),&(((struct sockaddr_in6 *)from)->sin6_addr),sizeof(struct olsr_ip_addr)); 
	memcpy(&(tmsg->originator),&(m->olsr_source),sizeof(struct olsr_ip_addr));
	memcpy (&tmsg->my_addr, &addr_manet[index], sizeof(struct olsr_ip_addr));
	/********************** end of revision ******************/
	
	//tmsg->packet_seq_number = ntohs(tc->tc_seqnum); //commented by Y.Ge
	tmsg->packet_seq_number = ntohs(m->olsr_msgseq); //added by Y.Ge
	//tmsg->hop_count =  tc->tc_hopcnt; //commented by Y.Ge
	tmsg->hop_count = m->olsr_hopcount; //added by Y.Ge
	tmsg->time_to_live = m->olsr_ttl; //added by Y.Ge
	tmsg->mssn =  ntohs(tc->tc_mprseqnum);
	
	if (debug_level >3)
	{
		printf ("--------THE RECEIVED MESSAGE IS A TC with following information!\n");
		printf ("        sender's interface address: %s\n", convert_address_to_string (&tmsg->source_addr));
		printf ("        sender's originator address: %s\n", convert_address_to_string (&tmsg->originator));
		printf ("        received from my interface: %s\n", convert_address_to_string (&tmsg->my_addr));
		printf ("        sender has the following MPR selector information:\n");
		
		fprintf (y_file,"--------THE RECEIVED MESSAGE IS A TC with following information!\n");
		fprintf (y_file,"        sender's interface address: %s\n", convert_address_to_string (&tmsg->source_addr));
		fprintf (y_file,"        sender's originator address: %s\n", convert_address_to_string (&tmsg->originator));
		fprintf (y_file,"        received from my interface: %s\n", convert_address_to_string (&tmsg->my_addr));
		fprintf (y_file,"        sender has the following MPR selector information:\n");
	}

	/* the following "for" is commented and re-writed by Y.Ge **/
	/*for (maddr = mprsaddr; (char *)maddr < (char *)m + ntohs(m->olsr_nextmsg);  maddr++)*/
	for (maddr = mprsaddr; (char *)maddr < (char *)m + ntohs(m->olsr_msgsize);  maddr++)
	{
		mprs = (struct tc_mpr_addr *)malloc(sizeof(struct tc_mpr_addr));
		if (mprs == 0)
		{
			printf("olsrd: out of memery \n");
			break;
		}
		/*CRC v4->v6 del mprs->address = *maddr;*/
		//CRC v4->v6 add
		memcpy(&(mprs->address),maddr,sizeof(struct olsr_ip_addr));
		mprs->next = tmsg->multipoint_relay_selector_address;
		tmsg->multipoint_relay_selector_address = mprs;
		
		/*** debug ***/
		if (debug_level >3)
		{
			printf ("+++ mpr selector: %s\n", convert_address_to_string(&mprs->address));
			
			fprintf (y_file, "+++ mpr selector: %s\n", convert_address_to_string(&mprs->address));
		}
	}
}

/************************* added by Y.Ge *****************/
/*
 * Build MID Message according to tc_message struct from olsrmsg struct.
 */

void mid_chgestruct(struct mid_message *mmsg, struct olsrmsg *m, struct sockaddr *from, int index)
{

	struct midmsg *mid;
	//CRC v4->v6 chg
	struct olsr_ip_addr *midaddr, *maddr;
	struct mid_iface_addr *midinterfaces, *midinterfaces_debug;

	mmsg->mid_interface_address = (struct mid_iface_addr *)NULL;

	if (!m)
	    return;

	if (m->olsr_msgtype != MID_PACKET)
	    return;

	mid = m->olsr_mid;
	midaddr = mid->mid_ifaceaddr;

	memcpy(&(mmsg->source_addr),&(((struct sockaddr_in6 *)from)->sin6_addr),sizeof(struct olsr_ip_addr));
	memcpy(&(mmsg->originator_addr),&(m->olsr_source),sizeof(struct olsr_ip_addr)); 
	memcpy (&mmsg->my_addr, &addr_manet[index], sizeof(struct olsr_ip_addr));
	mmsg->packet_seq_number = ntohs(m->olsr_msgseq);
	mmsg->hop_count =  m->olsr_hopcount;
	mmsg->time_to_live = m->olsr_ttl;
	mmsg->v_time = m->olsr_vtime;

	/*** debug ***/
	if (debug_level >3)
	{
		printf ("--------THE RECEIVED MESSAGE IS A MID with following information!\n");
		printf ("        sender's interface address: %s\n", convert_address_to_string (&mmsg->source_addr));
		printf ("        sender's originator address: %s\n", convert_address_to_string (&mmsg->originator_addr));
		printf ("        received from my interface: %s\n", convert_address_to_string (&mmsg->my_addr));
		printf ("        sender has the following mid information:\n");
		
		fprintf (y_file,"--------THE RECEIVED MESSAGE IS A MID with following information!\n");
		fprintf (y_file,"        sender's interface address: %s\n", convert_address_to_string (&mmsg->source_addr));
		fprintf (y_file,"        sender's originator address: %s\n", convert_address_to_string (&mmsg->originator_addr));
		fprintf (y_file,"        received from my interface: %s\n", convert_address_to_string (&mmsg->my_addr));
		fprintf (y_file,"        sender has the following mid information:\n");		
	}
	
	for (maddr = midaddr; (char *)maddr < (char *)m + ntohs(m->olsr_msgsize);  maddr++)
	{
		/*printf (" +++ interface address: %s\n", convert_address_to_string(maddr));*/
		midinterfaces = (struct mid_iface_addr *)malloc(sizeof(struct mid_iface_addr));
		if (midinterfaces == 0)
		{
			printf("olsrd: out of memery \n");
			break;
		}
		memcpy(&(midinterfaces->interface_address),maddr,sizeof(struct olsr_ip_addr));
		midinterfaces->next = mmsg->mid_interface_address;
		mmsg->mid_interface_address = midinterfaces;
		
		/*printf ("... after adding this address, the message is:\n");
		midinterfaces_debug = mmsg->mid_interface_address;
		while(midinterfaces_debug!=NULL)
		{
			printf ("interface address: %s\n", convert_address_to_string(&midinterfaces_debug->interface_address));
			midinterfaces_debug = midinterfaces_debug->next;
		}
		printf ("... this adding round finished\n");*/

		/*** debug ***/
		if (debug_level >3)
		{
			printf ("mid interface address: %s\n", convert_address_to_string(&midinterfaces->interface_address));
			
			fprintf (y_file,"mid interface address: %s\n", convert_address_to_string(&midinterfaces->interface_address));
		}
	}
	
	/*printf ("+++++++++++++++++++++++ after MID chagestruct, the message is:\n");
	midinterfaces_debug = mmsg->mid_interface_address;
	while(midinterfaces_debug!=NULL)
	{
		printf ("interface address: %s\n", convert_address_to_string(&midinterfaces_debug->interface_address));
		midinterfaces_debug = midinterfaces_debug->next;
	}
	printf ("++++++++++++++++++++ MID changestruct finished ++++++++++++++++\n");*/
	
}
/************************* end of revision ***************/

//CRC HNA add
/* Build HNA Message according to hna_message struct from olsrmsg struct.*/
void hna_chgestruct(struct hna_message *hmsg, struct olsrmsg *m, struct sockaddr *from, int index) //"index" is added by Y.Ge
{
	struct hnamsg *hptr;
	struct hnaaddr *hna_addr, *haddr;
	struct hna_net_addr *netaddrptr;

	hmsg->hna_net_addrs =(struct hna_net_addr *)NULL;

	if (!m)
		return;

	if (m->olsr_msgtype != HNA_PACKET)
		return;

	hptr = m->olsr_hna;
	hna_addr = hptr->hna_netaddr;

	/* CRC v4->v6 add: memcpy(&(tmsg->source_addr),&(((struct sockaddr_in6
	 *)from)->sin6_addr),sizeof(struct olsr_ip_addr));*/
	//CRC HNA add
	/*------------------------------------------------------------------------*/
	//memcpy(&(hmsg->source_addr),&(m->olsr_source),sizeof(struct olsr_ip_addr)); //commented by Y.Ge
	/*------------------------------------------------------------------------*/
	/*memcpy(&(hmsg->originator),&(hptr->hna_origaddr),sizeof(struct olsr_ip_addr)); 
	hmsg->packet_seq_number = ntohs(hptr->hna_seqnum);
	hmsg->hop_count = hptr->hna_hopcnt;
	hmsg->msg_seq_number =  ntohs(hptr->hna_msgseqnum);*/ //commented by Y.Ge
	
	/************** added by Y.Ge ************/
	memcpy(&(hmsg->source_addr),&(((struct sockaddr_in6 *)from)->sin6_addr),sizeof(struct olsr_ip_addr));
	memcpy(&(hmsg->originator),&(m->olsr_source),sizeof(struct olsr_ip_addr));  
	memcpy (&hmsg->my_addr, &addr_manet[index], sizeof(struct olsr_ip_addr));
	hmsg->packet_seq_number = ntohs(m->olsr_msgseq);
	hmsg->hop_count =  m->olsr_hopcount;
	hmsg->time_to_live = m->olsr_ttl;
	hmsg->v_time = m->olsr_vtime;
	
	/*** debug ***/
	if (debug_level >3)
	{
		printf ("--------THE RECEIVED MESSAGE IS A HNA with following information!\n");
		printf ("        sender's interface address: %s\n", convert_address_to_string (&hmsg->source_addr));
		printf ("        sender's originator address: %s\n", convert_address_to_string (&hmsg->originator));
		printf ("        received from my interface: %s\n", convert_address_to_string (&hmsg->my_addr));
		printf ("        sender has the following HNA information:\n");
		
		fprintf (y_file,"--------THE RECEIVED MESSAGE IS A HNA with following information!\n");
		fprintf (y_file,"        sender's interface address: %s\n", convert_address_to_string (&hmsg->source_addr));
		fprintf (y_file,"        sender's originator address: %s\n", convert_address_to_string (&hmsg->originator));
		fprintf (y_file,"        received from my interface: %s\n", convert_address_to_string (&hmsg->my_addr));
		fprintf (y_file,"        sender has the following HNA information:\n");		
	}
	/************** end of revision **********/

	/* the following "for" is commented and re-writed by Y.Ge **/
	/*for (haddr = hna_addr; (char *)haddr < (char *)m + ntohs(m->olsr_nextmsg);  haddr++)*/
	for (haddr = hna_addr; (char *)haddr < (char *)m + ntohs(m->olsr_msgsize);  haddr++)
	{
		netaddrptr = (struct hna_net_addr *)malloc(sizeof(struct hna_net_addr));
		if (netaddrptr == 0)
		{
			printf("olsrd: out of memery \n");
			break;
		}
		memcpy(&(netaddrptr->net_addr),&(haddr->addr),sizeof(struct olsr_ip_addr));
		netaddrptr->prefix=ntohs(haddr->prefix);
		netaddrptr->next = hmsg->hna_net_addrs;
		hmsg->hna_net_addrs = netaddrptr;
		
		/*** debug ***/
		if (debug_level >3)
		{
			printf ("hna address: %s, prefix: %d\n", convert_address_to_string(&netaddrptr->net_addr),netaddrptr->prefix);
			
			fprintf (y_file,"hna address: %s, prefix: %d\n", convert_address_to_string(&netaddrptr->net_addr),netaddrptr->prefix);
		}
	}
}

/*
 * Process a newly received packet.
 */

void olsr_input(struct sockaddr *from, struct olsr *olsr, int size, int index) //"index" is added by Y.Ge, which interface the packet is recieved?
{
	struct olsrmsg *m = olsr->olsr_msg;
	struct hello_message hellopacket;
	struct tc_message tcpacket_in, *tcpacket_out;
	struct rt_message *routepacket = NULL;
	//CRC HNA add
	struct hna_message hnapacket_in, *hnapacket_out; 
	struct mid_message midpacket_in, *midpacket_out; //added by Y.Ge
	struct qos_hello_message qoshellopacket;	//added by Y.Ge
	struct qos_tc_message	qostcpacket_in, *qostcpacket_out;//added by Y.Ge
	int   	qos_tc_rcvd_size = 0;	//added by Y.Ge
	int	qos_tc_rcvd = 0;	//added by Y.Ge
	int	tc_rcvd = 0;	//added by Y.Ge
	int iface_index; //added by Y.Ge
	int found = 0;  //added by Y.Ge
	
	/********** added by Y.Ge *********/
	/* deal with unknown packet */
	int relay_unknown_pk = 0;
	int i;
	int unknown_pk_rcvd = 0;
	int unknown_msg_size = 0;
	/********* end of revision ********/
    
	/*CRC v4->v6 del
	olsr_ip_addr src_addr = ((struct sockaddr_in *)from)->sin_addr.s_addr;*/
	struct olsr_ip_addr src_addr; 
	struct interface *ifp;
	int minsize;      /* minimum size of olsr packet */
	int count;
	struct afswitch *afp;
	char buf1[256];
	
	//CRC v4->v6 add
	memcpy(&src_addr,&(((struct sockaddr_in6 *)from)->sin6_addr),sizeof(struct olsr_ip_addr));
	tcpacket_in.multipoint_relay_selector_address = (struct tc_mpr_addr *)NULL;
	hellopacket.neighbors = (struct hello_neighbor *)NULL;
	midpacket_in.mid_interface_address = (struct mid_iface_addr *)NULL; //added by Y.Ge
    
	//CRC HNA add
	hnapacket_in.hna_net_addrs=(struct hna_net_addr*)NULL;
	hnapacket_out=(struct hna_message*)NULL;
 
	//afp = &afswitch[from->sa_family];
	
	//v4->v6
	/*if (from->sa_family >= af_max || afp->af_hash == (void (*)(struct sockaddr *, struct afhash *))0)
	{
		syslog(LOG_INFO, "\"from\" address in unsupported address family (%d), msg %d\n",from->sa_family, m->olsr_msgtype);
		return;
	}*/

	/* if (ntohs(olsr->olsr_packlen) != size) 
	{
		fprintf(stderr, "Size error detected in received packet.\n");
		syslog(LOG_ERR, " packet length error in  packet received from %s!", *afswitch[from->sa_family].af_format)(from, buf1, sizeof(buf1)));
		return;
	}*/

	/*
	 * Process tools packets.
	 */

	//tools_input(from, afp, olsr, size);

	/* verify message came from a privileged port */
	/*if ((*afp->af_portmatch)(from) == 0)
		return;*/

	/*(*afp->af_canon)(from);*/
       
	/* are we talking to ourselves? */
	ifp = if_ifwithaddr6(from);
       
	if (ifp) 
	{
		if (ifp->int_flags & IFF_PASSIVE) 
		{
			fprintf(stderr, " Data from passive interface\n");
			syslog(LOG_ERR, "bogus input (from passive interface, %s)", (*afswitch[from->sa_family].af_format)(from, buf1, sizeof(buf1)));
		}
		return;
	}
      
	if (filter)
	{
  	  	/********** commented by Y.Ge *************/
		/*if (!is_neighbor(&addr, (struct sockaddr_in6 *)from))
		{
			return;
		}*/
		/********** end of revision**************/

		/********** added by Y.Ge *************/
	
		for (iface_index=0; iface_index<num_of_interface; iface_index++)
		{
			if (is_neighbor(&addr[iface_index], (struct sockaddr_in6 *)from))
			{
				found = 1;
				break;
			}
		}
		if (!found)
		{
			printf (" MESSAGE RECEIVED FROM A NON-NEIGHBOR\n");
			return;
		}
		/********** end of revision**************/
	}

	     
	ifp = 0;
	
	/***************** added by Y.Ge ****************/
	/***** debug *****/
	if (debug_level >3)
	{
		printf ("\n\n\n************time (%02d:%02d:%02d.%06lu) message received from interface %d!************\n", nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec, index);
		printf ("packet length: %d\n", ntohs(olsr->olsr_packlen));
		printf ("packet sequence: %d\n", ntohs(olsr->olsr_seqnum));
		
		fprintf (y_file,"\n\n\n************time (%02d:%02d:%02d.%06lu) message received from interface %d!************\n", nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec, index);
		fprintf (y_file,"packet length: %d\n", ntohs(olsr->olsr_packlen));
		fprintf (y_file,"packet sequence: %d\n", ntohs(olsr->olsr_seqnum));
		
	}
	/**************** end of revision ***************/

	//TRACE_INPUT(ifp, from, (char *)olsr, size);

	/*minsize = (int)sizeof(m->olsr_msgtype) + (int)sizeof(m->olsr_broad) + (int)sizeof(m->olsr_nextmsg);*/   /* commented by Y.Ge */
	
	minsize = 4*(int)sizeof(olsr_u8_t) + (int)sizeof(olsr_u16_t) + (int)sizeof (struct olsr_ip_addr);  //added by Y.Ge

	count = size - ((char *)m - (char *)olsr);
	if (count < minsize)
	{
		printf ("XXXXXXXXXXXX PACKET to small! count is: %d, minsize is: %d XXXXXXXXXXXXXX\n", count, minsize);
		exit(1);
		//return;
	}
      
	/* the following "for" is commented and re-writed by Y.Ge **/ 
	/*for ( ; count > 0; m = (struct olsrmsg *)((char *)m + ntohs(m->olsr_nextmsg)))*/
	for ( ; count > 0; m = (struct olsrmsg *)((char *)m + ntohs(m->olsr_msgsize)))
	{
		if (count < minsize)
		{
			printf ("XXXXXXXX MESSAGE too small! count is: %d, minsize is: %d XXXXXXXXXX\n", count, minsize);
			printf ("break here!\n");
			exit (1);
			//break;
	  	}
		//count -= ntohs(m->olsr_nextmsg); //commented  by Y.Ge
		count -= ntohs(m->olsr_msgsize); //added by Y.Ge
	   
		/************* added by Y.Ge ***************/
		/*** debug ***/
		if (debug_level >3)
		{
			printf ("\n\n\n.......................... the following is the olsr_msg info ...................\n");
			printf ("      message type: %d\n", m->olsr_msgtype);
			printf ("      vtime: %d\n", m->olsr_vtime);
			printf ("      message size: %d\n", ntohs(m->olsr_msgsize));
			printf ("      message source: %s\n", convert_address_to_string (&m->olsr_source));
			printf ("      ttl: %d\n", m->olsr_ttl);
			printf ("      hop count: %d\n", m->olsr_hopcount);
			printf ("      message sequence number: %d\n", ntohs(m->olsr_msgseq));
			printf ("\n\n");
			
			fprintf (y_file, "\n\n\n.......................... the following is the olsr_msg info ...................\n");
			fprintf (y_file, "      message type: %d\n", m->olsr_msgtype);
			fprintf (y_file, "      vtime: %d\n", m->olsr_vtime);
			fprintf (y_file, "      message size: %d\n", ntohs(m->olsr_msgsize));
			fprintf (y_file, "      message source: %s\n", convert_address_to_string (&m->olsr_source));
			fprintf (y_file, "      ttl: %d\n", m->olsr_ttl);
			fprintf (y_file, "      hop count: %d\n", m->olsr_hopcount);
			fprintf (y_file, "      message sequence number: %d\n", ntohs(m->olsr_msgseq));
			fprintf (y_file, "\n\n");
		}
		/************* end of revision *************/

		/************** added by Y.GE *********/
		/* if I understand qos OLSR */
		if (use_qos)
		{
			switch (m->olsr_msgtype)
			{
				case QOSHELLO_PACKET :
					qoshello_chgestruct(&qoshellopacket, m, from, index);
					
					olsr_process_received_qos_hello(index, &qoshellopacket);
					break;
				case QOSTC_PACKET :
					qostc_chgestruct(&qostcpacket_in, m, from, index);
					qos_tc_rcvd = 1;
					qos_tc_rcvd_size = ntohs(m->olsr_msgsize); //remember the size of the qos TC
					break;		
			}
		}
		/************* end of revision *******/
		
		switch (m->olsr_msgtype) 
		{
			case HELLO_PACKET:
				hello_chgestruct(&hellopacket, m, from, index); //"index" added by Y.Ge
					//added by L.Qin	
//					struct neighbor_entry *aNeighbor;
//					aNeighbor=(struct neighbor_entry *)malloc(sizeof(struct neighbor_entry)); 
//			printf("input.c The hyst value of the neighbor in %d is //%f\n",index,aNeighbor->neighbor_hysavg); 
					//***** 

				olsr_process_received_hello(&hellopacket, index); //"index" added by Y.Ge
				break;
			case TC_PACKET:
				tc_rcvd = 1;
				tc_chgestruct(&tcpacket_in, m, from, index); //"index" added by Y.Ge
								
				/******** added by Y.Ge *********/
				/* qos OLSR */
				if (qos_tc_rcvd) //I understand QoS OLSR
					olsr_process_received_qos_tc(index, &tcpacket_in, &tcpacket_out, &qostcpacket_in, &qostcpacket_out);
				else
					olsr_process_received_tc(index, &tcpacket_in, &tcpacket_out); //"index" added by Y.Ge
							
				if (tcpacket_out)
				{
					/*toall(tc_build, (struct hello_message *)NULL, tcpacket_out, (struct mid_message *)NULL, (struct hna_message *)NULL, (struct interface *)NULL);*/ //commented by Y.Ge
					
					/************** added by Y.Ge **************/
					/***** check if there is already another TC packet to be relayed *****/
					if (tc_to_relay ==1)
					{
						/* send out that TC packet even the jitter time has not expired */
						for (ifp = ifnet; ifp; ifp = ifp->int_next)
						{
							if (debug_level > 3)
							{
								printf ("XXXXXXXXXXXXXXXXXXXX send out the OLD RELAY TC! XXXXXXXXXXXXXXXXXXXXXXX\n"); 
								printf ("XXXXXXXXXXXXXXXXXXXX time  (%02d:%02d:%02d.%06lu) TC sent from interface %d!XXXXXXXXXXXXXXXXXX\n", nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec, ifp->my_index);
							
								fprintf (y_file,"XXXXXXXXXXXXXXXXXXXX send out the OLD RELAY TC! XXXXXXXXXXXXXXXXXXXXXXX\n"); 
								fprintf (y_file,"XXXXXXXXXXXXXXXXXXXX time  (%02d:%02d:%02d.%06lu) TC sent from interface %d!XXXXXXXXXXXXXXXXXX\n", nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec, ifp->my_index);
							}
							
							/* added by Y.Ge */
							/* check if there is a qos tc to be relayed */
							if (qostc_to_relay ==1)
								qostc_build((struct sockaddr *)NULL, &qostcpacket_relay, &tcpacket_relay, 0, ifp, 1);
							else
								tc_build((struct sockaddr *)NULL,(struct hello_message *)NULL, &tcpacket_relay, (struct mid_message *)NULL, (struct hna_message *)NULL,0,ifp, 1);
							tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index);
						}
						/* free memory */
						olsr_destroy_tc_message (&tcpacket_relay);
						olsr_destroy_qos_tc_message (&qostcpacket_relay);
						tcpacket_relay.multipoint_relay_selector_address = NULL;
						qostcpacket_relay.multipoint_relay_selector_busy_time = NULL;
						
						tc_to_relay = 0;
						qostc_to_relay = 0;
						tc_relay_interval = 0;
						tc_relay_size = 0;
						qostc_relay_size = 0;	
					}
					
					/* added by Y.Ge */
					/* qos OLSR */
					if (use_qos)
					{
						if(qos_tc_rcvd)
						{
							olsr_copy_qos_tc_packet(qostcpacket_out, &qostcpacket_relay);	
							qostc_to_relay = 1;
							
						}
					}
					olsr_copy_tc_packet(tcpacket_out, &tcpacket_relay);
					
					/*printf ("********** copy tc packet finished ! **********\n");
					printf ("after copy, the out tc is:\n");
					printf ("packet seq: %d, mssn: %d\n", tcpacket_relay.packet_seq_number, tcpacket_relay.mssn);
					
					fprintf (y_file,"********** copy tc packet finished ! **********\n");
					fprintf (y_file,"after copy, the out tc is:\n");
					fprintf (y_file,"packet seq: %d, mssn: %d\n", tcpacket_relay.packet_seq_number, tcpacket_relay.mssn);*/
					
					/***** set the timer to relay the tc *****/
					/* relay jitter is 1/4 of HELLO_INTERVAL */
					tc_relay_jitter = (float) random()/(((double)RAND_MAX +1)/((double)HELLO_INTERVAL/4));
					
					if(debug_level > 3)
					{
						printf (" ----- tc relay jetter is: %f\n", tc_relay_jitter);
						fprintf (y_file," ----- tc relay jetter is: %f\n", tc_relay_jitter);
					}
					
					tc_to_relay = 1;
					tc_relay_size =  qos_tc_rcvd_size + ntohs(m->olsr_msgsize); //remember the size of the TC
										
					/*** free the memery ***/
					olsr_destroy_tc_message (tcpacket_out);
					
					if (qos_tc_rcvd)
						olsr_destroy_qos_tc_message (qostcpacket_out);
					/*************** end of rrevision **********/
				}
				break;

			/******************* added by Y.Ge *************/
			case MID_PACKET:
				mid_chgestruct (&midpacket_in, m, from, index);
				olsr_process_received_mid (index, &midpacket_in, &midpacket_out);
				if (midpacket_out)
				{
					/***** check if there is already another MID packet to be relayed *****/
					if (mid_to_relay == 1)
					{
						/* send out that MID packet even the jitter time has not expired */
						for (ifp = ifnet; ifp; ifp = ifp->int_next)
						{
							if (debug_level > 3)
							{
								printf ("XXXXXXXXXXXXXXXXXXXX send out the OLD RELAY MID! XXXXXXXXXXXXXXXXXXXXXXX\n"); 
								printf ("XXXXXXXXXXXXXXXXXXXX time  (%02d:%02d:%02d.%06lu) MID sent from interface %d!XXXXXXXXXXXXXXXXXX\n", nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec, ifp->my_index);
							
								fprintf (y_file,"XXXXXXXXXXXXXXXXXXXX send out the OLD RELAY MID! XXXXXXXXXXXXXXXXXXXXXXX\n"); 
								fprintf (y_file,"XXXXXXXXXXXXXXXXXXXX time  (%02d:%02d:%02d.%06lu) MID sent from interface %d!XXXXXXXXXXXXXXXXXX\n", nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec, ifp->my_index);
							}
							
							mid_build((struct sockaddr *)NULL,(struct hello_message *)NULL, (struct tc_message *)NULL, &midpacket_relay, (struct hna_message *)NULL, 0, ifp, 1);
							tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index);
						}
						/* free memory */
						olsr_destroy_mid_message(&midpacket_relay);
						midpacket_relay.mid_interface_address=NULL;
						
						mid_to_relay = 0;
						mid_relay_interval = 0;
						mid_relay_size = 0;
					}
					/***** set the timer to relay the mid *****/
					olsr_copy_mid_packet(midpacket_out, &midpacket_relay);
					
					/***** set the timer to relay the tc *****/
					/* relay jitter is 1/4 of HELLO_INTERVAL */
					mid_relay_jitter = (float) random()/(((double)RAND_MAX +1)/((double)HELLO_INTERVAL/4));
					
					if (debug_level > 3)
					{
						printf (" ----- mid relay jetter is: %f\n", mid_relay_jitter);
						fprintf (y_file," ----- mid relay jetter is: %f\n", mid_relay_jitter);
					}
					
					mid_to_relay = 1;
					mid_relay_size =  ntohs(m->olsr_msgsize); //remember the size of the MID
					
					/*** free the memery ***/
					olsr_destroy_mid_message(midpacket_out);
				}	
				break;
			/****************** end of revision ************/
	   
			//CRC HNA add
			case HNA_PACKET:
				hna_chgestruct(&hnapacket_in, m, from, index); //"index" added by Y.Ge
				olsr_process_received_hna(index, &hnapacket_in, &hnapacket_out); //"index" added by Y.Ge
				if (hnapacket_out)
				{
					//toall(hna_build, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL, hnapacket_out, (struct interface *)NULL); //commented by Y.Ge
					/************** added by Y.Ge **************/
					/***** check if there is already another TC packet to be relayed *****/
					if (hna_to_relay ==1)
					{
						/* send out that TC packet even the jitter time has not expired */
						for (ifp = ifnet; ifp; ifp = ifp->int_next)
						{
							if (debug_level > 3)
							{
								printf ("XXXXXXXXXXXXXXXXXXXX send out the OLD RELAY HNA! XXXXXXXXXXXXXXXXXXXXXXX\n"); 
								printf ("XXXXXXXXXXXXXXXXXXXX time  (%02d:%02d:%02d.%06lu) HNA sent from interface %d!XXXXXXXXXXXXXXXXXX\n", nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec, ifp->my_index);
							
								fprintf (y_file,"XXXXXXXXXXXXXXXXXXXX send out the OLD RELAY HNA! XXXXXXXXXXXXXXXXXXXXXXX\n"); 
								fprintf (y_file,"XXXXXXXXXXXXXXXXXXXX time  (%02d:%02d:%02d.%06lu) HNA sent from interface %d!XXXXXXXXXXXXXXXXXX\n", nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec, ifp->my_index);
							}
							
							hna_build((struct sockaddr *)NULL,(struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,&hnapacket_relay , 0, ifp, 1);
							tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index);
						}
						/* free memory */
						olsr_destroy_hna_message (&hnapacket_relay);
						hnapacket_relay.hna_net_addrs = NULL;
						
						hna_to_relay = 0;
						hna_relay_interval = 0;
						hna_relay_size = 0;	
					}
					olsr_copy_hna_packet(hnapacket_out, &hnapacket_relay);
					
					/***** set the timer to relay the hna *****/
					/* relay jitter is 1/4 of HELLO_INTERVAL */
					hna_relay_jitter = (float) random()/(((double)RAND_MAX +1)/((double)HELLO_INTERVAL/4));
					
					if (debug_level > 3)
					{
						printf (" ----- hna relay jetter is: %f\n", hna_relay_jitter);
						fprintf (y_file," ----- hna relay jetter is: %f\n", hna_relay_jitter);
					}
					
					hna_to_relay = 1;
					hna_relay_size =  ntohs(m->olsr_msgsize); //remember the size of the HNA
										
					/*** free the memery ***/
					olsr_destroy_hna_message (hnapacket_out);
					/*************** end of rrevision **********/
					
				}	
				break;

			default:
				/************** added by Y.Ge *************/
				
				/* make sure I won't process QoS TC twice */
				if (m->olsr_ttl > 1 && !qos_tc_rcvd)
				{
					/* 
				 	* even if the node doesn't understand what is the received packet,
					* it still check if the message should be forwarded
				 	*/
					if (debug_level > 3)
					{
						printf ("******** UNKNOWN message! I may need to forward this message! ********\n");
						fprintf (y_file,"******** UNKNOWN message I may need to forward this message! ********\n");
					}
					
					relay_unknown_pk = olsr_forward_unknown_packets (m->olsr_msgtype,&(((struct sockaddr_in6 *)from)->sin6_addr),&m->olsr_source,ntohs(m->olsr_msgseq),index); 
				
					if (relay_unknown_pk)
					{
						unknown_pk_rcvd = 1;
					
						if (debug_level > 3)
						{
							printf ("+++++++ prepare to forward this message ! +++++++\n");
							fprintf (y_file,"+++++++ prepare to forward this message ! +++++++\n");
						}
						
						m->olsr_ttl -= 1;
						m->olsr_hopcount +=1;
						
						memcpy(((struct olsrmsg *)unknown_packet + unknown_msg_size), m, (ntohs(m->olsr_msgsize)));
						
						unknown_msg_size += ntohs(m->olsr_msgsize);
					}
				}
				/************* end of revision **********/ 
 				break;

		} /* switch */
	} /* for olsr_msg */ 
	
	/*********** added by Y.Ge ********/
	fprintf (y_file, "unknown_pk_rcvd: %d, tc_to_relay: %d, qostc_to_relay: %d\n",unknown_pk_rcvd, tc_to_relay, qostc_to_relay); 
	/* 
	 * if I should forward an unknown packet && I have received tc
	 * && it is possible that I have received a qos tc
	 * should forward the unknown packet together with tc
	 */
	if(unknown_pk_rcvd == 1)
	{
		/* 
		 * if I need to forward the received tc, don't waite until the tcrelay_jitter expires
		 * send it out immediately, as I may have received a QoS TC without understanding it
		 */	
		if (tc_to_relay ==1 && qostc_to_relay ==0)
		{
			/* send out that TC packet even the jitter time has not expired */
			for (ifp = ifnet; ifp; ifp = ifp->int_next)
			{
				if (debug_level > 3)
				{
					printf ("XXXXXXXXXXXXXXXXXXXX send out the RELAY TC together with unknown packet! XXXXXXXXXXXXXXXXXXXXXXX\n"); 
					printf ("XXXXXXXXXXXXXXXXXXXX time  (%02d:%02d:%02d.%06lu) TC sent from interface %d!XXXXXXXXXXXXXXXXXX\n", nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec, ifp->my_index);
							
					fprintf (y_file,"XXXXXXXXXXXXXXXXXXXX send out the RELAY TC together with unknown packet! XXXXXXXXXXXXXXXXXXXXXXX\n"); 
					fprintf (y_file,"XXXXXXXXXXXXXXXXXXXX time  (%02d:%02d:%02d.%06lu) TC sent from interface %d!XXXXXXXXXXXXXXXXXX\n", nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec, ifp->my_index);
				}
				
				/* copy the unknown packet to the output queue */
				/* the first 4 bytes is the common packet header */
				memcpy(((char *)packet +4), (char *)unknown_packet, unknown_msg_size);
				
				/* set the outputsize, don't neet to set the packet head now, as "tc_build" will deal with it */
				outputsize = unknown_msg_size + 4;
				
				tc_build((struct sockaddr *)NULL,(struct hello_message *)NULL, &tcpacket_relay, (struct mid_message *)NULL, (struct hna_message *)NULL,0,ifp, 1);
				
				tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index);
			}
			/* free memory */
			olsr_destroy_tc_message (&tcpacket_relay);
			tcpacket_relay.multipoint_relay_selector_address = NULL;
						
			tc_to_relay = 0;
			tc_relay_interval = 0;
			tc_relay_size = 0;	
		}
		else
		{
			if (debug_level > 3)
			{
				printf ("XXXXXXXXXXXXXXXXXXXX send out the unknown packet alone! XXXXXXXXXXXXXXXXXXXXXXX\n"); 
				printf ("XXXXXXXXXXXXXXXXXXXX time  (%02d:%02d:%02d.%06lu) unknown packet sent from interface %d!XXXXXXXXXXXXXXXXXX\n", nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec, ifp->my_index);
							
				fprintf (y_file,"XXXXXXXXXXXXXXXXXXXX send out the unknown packet alone! XXXXXXXXXXXXXXXXXXXXXXX\n"); 
				fprintf (y_file,"XXXXXXXXXXXXXXXXXXXX time  (%02d:%02d:%02d.%06lu) unknown packet sent from interface %d!XXXXXXXXXXXXXXXXXX\n", nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec, ifp->my_index);
			}
			
			for (ifp = ifnet; ifp; ifp = ifp->int_next)
			{
				/* copy the unknown packet to output queue */
				memcpy(((char *)packet +4), (char *)unknown_packet, unknown_msg_size);
				
				outputsize = 4 + unknown_msg_size;
				
				msg->olsr_packlen = htons (4 + unknown_msg_size);
				msg->olsr_seqnum = htons(pkt_seqnum[ifp->my_index] ++);
				
				tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index);
			}
		}
		
	}
	/* check if the qos stuff is correct */
	if (qos_tc_rcvd == 1 && tc_rcvd == 0)
	{
		printf("XXXXXXXXXX qos tc received but tc is not received! XXXXXXXXXX\n");
		exit(1);
	}
	
	/* Y.Ge: process changes at one time after all the messages are received */
	if (use_qos)
		olsr_process_changes_with_qos();
	else
		olsr_process_changes();
	
	if(debug_level > 2)
	{
		printf ("\n\n************ after process the received messages, the routing table:\n");
		if (use_qos)
			qos_olsr_print_routing_table(routingtable);
		else
			olsr_print_routing_table(routingtable);
			
		fprintf (y_file, "\n\n************ after process the received messages, the routing table:\n");
		if (use_qos)
			qos_olsr_print_trace_routing_table(routingtable);
		else
			olsr_print_trace_routing_table(routingtable);
	}
	/*********** end of revision ******/

	sigpending(&psigset);
    
	/* If SIGALRM is pending, wait for the end of   
	 * the received packet process to send the stored
	 * msg(s) with next broadcast msg(s).
	 */
	//added by L.Qin
	//test the status
//	printf("The SIGALARM is pending? %d %d\n",sigismember(&psigset, SIGALRM),outputsize);
      	//*****
	if ( outputsize && !sigismember(&psigset, SIGALRM) ){
		printf("Will you?\n");
		toall(sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL,(struct mid_message *)NULL, (struct hna_message *)NULL, (struct interface *)NULL );
	}

}


/*
* Process tools packets.
*/
void tools_input(struct sockaddr *from, struct afswitch *afp,struct olsr *olsr, int size, int iface_index) //"iface_index" added by Y.Ge
{
	struct olsrmsg *m = olsr->olsr_msg;
	struct hello_message hellopacket;
	struct rt_message *routepacket = NULL;
	struct interface *ifp = 0;

	switch (m->olsr_msgtype) 
	{
		case HELLO_REQUEST:
			olsr_build_hello_packet(&hellopacket, iface_index); //"iface_index" added by Y.Ge
			supply(from, &hellopacket, (struct rt_message *)NULL, m->olsr_msgtype, iface_index); //"iface_index" added by Y.Ge
			return;

		case ROUTE_REQUEST:
			olsr_build_routing_packet(&routepacket);
			supply(from, (struct hello_message *)NULL, routepacket,m->olsr_msgtype, iface_index);//"iface_index" added by Y.Ge
			return;

		case UPDATE_NEIGHBOR:
			hello_chgestruct(&hellopacket, m, from, iface_index);
			update_static_neighborhood(&hellopacket);
			return;

		case OLSRD_TRACEON:
		case OLSRD_TRACEOFF:
			/* verify message came from a privileged port */
			if ((*afp->af_portcheck)(from) == 0)
				return;
 			((char *)olsr)[size] = '\0';
			if (m->olsr_msgtype == OLSRD_TRACEON)
			{
				traceon(m->olsr_tracefile);
				bumploglevel();
				bumploglevel();
				bumploglevel();
			}
			else
				traceoff();
			return;
	}
}
