/*
 * Copyright (c) 2004 Ying Ge, Communication Research Center Canada.
 *
 * Copyright (c) 2002, 2003 Maoyu Wang, Communication Research Center Canada.
 *
 * By Ying Ge:
 * 1. Change the OLSR packet format and message processing procedure based 
 *    on the OLSR RFC.
 * 2. Add support of multiple interfaces to OLSR, including MID message 
 *    creating and processing procedure as specified in the OLSR RFC.
 * 3. Add QoS Support to OLSR
 *
 * By Maoyu Wang:
 * 1. Ported OLSR from IPv4 to IPv6.
 * 2. Added the Host and Network Association (HNA) functionality into OLSR.
 * 3. Added the default gateway functionality into OLSR by extending the HNA 
 *    message usage. The default gateway functionality supported the mobility
 *	  by cooperating with Mobile IPv6 for a mobile node as well as supported 
 * 	  Internet access for MANET nodes. 
 *
 * DISTRIBUTED WITH NO WARRANTY, EXPRESS OR IMPLIED.
 * See the GNU Library General Public License (file COPYING in the distribution)
 * for conditions of use and redistribution
 */
/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est-à-dire qu'il peut
 * être utilisé, copié, modifié, distribué à l'unique
 * condition que ce texte soit conservé afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant participé à l'élaboration de ce logiciel ne peut
 * être utilisé sans son accord préalable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilité d'aucune sorte.
 * Certaines parties de ce logiciel sont dérivées de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 */

/*
 * Copyright (c) 1983, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	From: @(#)interface.h	5.6 (Berkeley) 6/1/90
 *	From: @(#)interface.h	8.1 (Berkeley) 6/5/93
 *	From: @(#)interface.h   1.7 (dholland Exp) 1997/04/05 
 *	$Id: interface.h,v 1.2 2000/12/06 10:36:11 prima Exp $
 */


/*
 * Routing table management daemon.
 */


/*
 * An ``interface'' is similar to an ifnet structure,
 * except it doesn't contain q'ing and point to point info.
 * Only radio interfaces (broadcast interfaces) are maintained
 */

struct ipv6_addr_list {
	struct in6_addr addr;
	int prefix_len;
	int scope;
	struct ipv6_addr_list* next;
};
	
//CRC v4->v6 modify
struct interface {
	struct	interface *int_next;
	struct	sockaddr int_addr;		// ipv4 address on this host 
	struct	sockaddr int_broadaddr;         // broadcast address 
	//START INTERNET SPECIFIC 
	u_long	int_net;			// network # 
	u_long	int_netmask;			// net mask for addr 
	u_long	int_subnet;			// subnet # 
	u_long	int_subnetmask;			// subnet mask for addr 
	//END INTERNET SPECIFIC 
	struct	ifdebug int_input, int_output;	// packet tracing stuff 
	int	int_ipackets;			// input packets received 
	int	int_opackets;			// output packets sent 
	struct  in6_addr int6_link_addr;	// ipv6 link address, future use 
	olsr_u16_t	int6_link_prefix;
	struct  in6_addr int6_site_addr;     // ipv6 site address
	olsr_u16_t	int6_site_prefix;
	struct  ipv6_addr_list *int6_global_addr;// ipv6 global addresses,future use
	struct	in6_addr int6_multi_addr;     // ipv6 multicast addresses
	olsr_u8_t int6_hw_addr[8];			  //hardware address for this interface
	int	int_metric;			// init's routing entry 
	int	int_flags;			// see below 
	char	*int_name;			// from kernel interface structure 
	unsigned int int_ifindex;		// ipv6 interface index 
	int    	my_index; 			//added by Y.Ge, index in the interface array
	u_short	int_transitions;		// times gone up-down 
};


struct interf_name {
  char *name;
  struct interf_name *next;
};


#if defined(__GLIBC__) && (__GLIBC__ >= 2)
#include <net/if.h>
#else /* not glibc */

/*
 * 0x1 to 0x10 are reused from the kernel's ifnet definitions,
 * the others agree with the RTS_ flags defined elsewhere.
 */

#define	IFF_UP		0x1		/* interface is up */
#define	IFF_BROADCAST	0x2		/* broadcast address valid */
#define	IFF_DEBUG	0x4		/* turn on debugging */
#define	IFF_LOOPBACK	0x8		/* software loopback net */
#define	IFF_POINTOPOINT	0x10		/* interface is point-to-point link */
//CRC v4->v6 add
#define IFF_MULTICAST   0x1000		/* support multicast*/

#endif /* glibc/not glibc */
#define	IFF_SUBNET	0x100000	/* interface on subnetted network */
#define	IFF_PASSIVE	0x200000	/* can't tell if up/down */
#define	IFF_INTERFACE	0x400000	/* hardware interface */

//CRC v4->v6 add
#define IPV6_ADDR_GLOBAL	0x0000U
#define IPV6_ADDR_LOOPBACK	0x0010U
#define IPV6_ADDR_LINKLOCAL	0x0020U
#define IPV6_ADDR_SITELOCAL	0x0040U

extern struct interface *ifnet;

//CRC v4->v6 modify
struct  interface *if_ifwithaddr(struct sockaddr *addr);
struct	interface *if_ifwithaddr6(struct sockaddr *addr);
struct	interface *if_iflookup(struct sockaddr *addr);
struct	interface *if_iflookup6(struct sockaddr *addr);
void if_get_ipv6info (struct interface* ifptr);
void if_get_multiaddr(struct interface *ifptr);
void ifchoice(char *);  
