/*
 * Copyright (c) 2004 Ying Ge, Communication Research Center Canada.
 *
 * Copyright (c) 2002, 2003 Maoyu Wang, Communication Research Center Canada.
 *
 * By Ying Ge:
 * 1. Change the OLSR packet format and message processing procedure based 
 *    on the OLSR RFC.
 * 2. Add support of multiple interfaces to OLSR, including MID message 
 *    creating and processing procedure as specified in the OLSR RFC.
 * 3. Add QoS Support to OLSR
 *
 * By Maoyu Wang:
 * 1. Ported OLSR from IPv4 to IPv6.
 * 2. Added the Host and Network Association (HNA) functionality into OLSR.
 * 3. Added the default gateway functionality into OLSR by extending the HNA 
 *    message usage. The default gateway functionality supported the mobility
 *	  by cooperating with Mobile IPv6 for a mobile node as well as supported 
 * 	  Internet access for MANET nodes. 
 *
 * DISTRIBUTED WITH NO WARRANTY, EXPRESS OR IMPLIED.
 * See the GNU Library General Public License (file COPYING in the distribution)
 * for conditions of use and redistribution
 */
/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est--dire qu'il peut
 * tre utilis, copi, modifi, distribu  l'unique
 * condition que ce texte soit conserv afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant particip  l'laboration de ce logiciel ne peut
 * tre utilis sans son accord pralable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilit d'aucune sorte.
 * Certaines parties de ce logiciel sont drives de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 */
 
 /**************** the file is added by Y.Ge **************/

/*
 * Routing table management daemon.
 */
 
 char interface_association_table_rcsid[] = 
  "$Id: interface_association_table.c,v 1.2 2000/12/06 10:36:11 prima Exp $";

#include "defs.h"

struct main_addr_hash	nodemainaddresstable[HASHSIZE];
struct iface_addr_hash	nodeifaceaddresstable[HASHSIZE];

/*------------------------------------------------------------------------*/
void
olsr_delete_mainaddress_link (struct main_addr_entry  *my_main_address_entry, struct olsr_ip_addr *my_interface_address)
{
	struct interface_addr_list	*node_iface_list;
	struct interface_addr_list	*node_iface_list_tmp;
	
	node_iface_list = my_main_address_entry->list_of_interfaceaddress;
	my_main_address_entry->list_of_interfaceaddress = NULL;
	
	while (node_iface_list != NULL)
	{
		if (0==memcmp (&node_iface_list->interface_addresses->node_interface_address, my_interface_address, sizeof(struct olsr_ip_addr)))
		{
			node_iface_list_tmp = node_iface_list;
			node_iface_list = node_iface_list->next;
			free (node_iface_list_tmp);
		}
		else
		{
			node_iface_list_tmp = node_iface_list;
			node_iface_list = node_iface_list->next;
			node_iface_list_tmp = my_main_address_entry->list_of_interfaceaddress;
			my_main_address_entry->list_of_interfaceaddress = node_iface_list_tmp;
		}
	}
}
/*------------------------------------------------------------------------*/
void
olsr_delete_node_interfaceaddress_table (struct iface_addr_entry *ifaceaddress_entry)
{
	struct main_address_information	*link_to_my_main_address;
	struct main_addr_entry        	*my_main_address_entry;
	
	link_to_my_main_address = ifaceaddress_entry->node_mainaddress;
	my_main_address_entry = link_to_my_main_address->mainaddress;
	
	my_main_address_entry->interface_address_pointer--;
	
	olsr_delete_mainaddress_link (my_main_address_entry,&ifaceaddress_entry->node_interface_address);
	
	if (my_main_address_entry->interface_address_pointer<1)
	{
		/* delete the main address from the interface address table */
		olsr_remque((struct olsr_qelem *)my_main_address_entry);
		free((void *)my_main_address_entry);
	}
	
	free (link_to_my_main_address);
	
	olsr_remque((struct olsr_qelem *)ifaceaddress_entry);
	free ((void *)ifaceaddress_entry);
}
	
/*------------------------------------------------------------------------*/
void
olsr_delete_interfaceaddress_link (struct iface_addr_entry  *my_iface_addr_entry)
{
	struct main_address_information	*my_main_addr_link;
	
	my_main_addr_link = my_iface_addr_entry->node_mainaddress;
	
	my_iface_addr_entry->node_mainaddress = NULL;
	
	free (my_main_addr_link);	
}
/*------------------------------------------------------------------------*/
void
olsr_delete_node_mainaddress_table (struct main_addr_entry *my_main_address_entry)
{
	struct interface_addr_list	*list_of_my_interface_addresses;
	struct iface_addr_entry   	*my_iface_addr_entry;
	
	list_of_my_interface_addresses = my_main_address_entry->list_of_interfaceaddress;
	
	while (list_of_my_interface_addresses != NULL)
	{
		my_iface_addr_entry = list_of_my_interface_addresses->interface_addresses;
		olsr_delete_interfaceaddress_link (my_iface_addr_entry);
		my_main_address_entry->list_of_interfaceaddress = list_of_my_interface_addresses->next;
		free (list_of_my_interface_addresses);
		list_of_my_interface_addresses = my_main_address_entry->list_of_interfaceaddress;
	}
	
	olsr_remque((struct olsr_qelem *)my_main_address_entry);
	free ((void *)my_main_address_entry);		
}
/*------------------------------------------------------------------------*/
void
olsr_release_mid_table()
{
	olsr_u8_t             	index;
	
	struct main_addr_entry		*main;
	struct main_addr_entry		*main_tmp;
	struct main_addr_hash		*main_hash;
	
	struct iface_addr_entry		*interface_assoc;
	struct iface_addr_entry		*interface_assoc_tmp;
	struct iface_addr_hash		*interface_assoc_hash;
	
	for (index=0; index<HASHSIZE; index++)
	{
		interface_assoc_hash = &nodeifaceaddresstable[index];
		interface_assoc = interface_assoc_hash->interfaceaddress_forw;
		while (interface_assoc != (struct iface_addr_entry *)interface_assoc_hash);
		{
			interface_assoc_tmp = interface_assoc;
			interface_assoc = interface_assoc->iface_addr_forw;
			olsr_delete_node_interfaceaddress_table(interface_assoc_tmp);
		}
	}
	
	for (index=0; index<HASHSIZE; index++)
	{
		main_hash = &nodemainaddresstable[index];
		main = main_hash->mainaddress_forw;
		while (main != (struct main_addr_entry *)main_hash)
		{
			main_tmp = main;
			main = main->main_addr_forw;
			olsr_delete_node_mainaddress_table (main_tmp);
		}
	}
}
/*------------------------------------------------------------------------*/
struct main_addr_entry
*olsr_lookup_main_address_table (struct olsr_ip_addr *rcvd_main_addr)
{
	struct main_addr_entry	*main_address;
	struct main_addr_hash	*main_address_hash;
	olsr_u32_t            	hash;

	//printf ("****************** in lookup main address table *****************\n");	
	olsr_hashing(rcvd_main_addr, &hash);
	main_address_hash = &nodemainaddresstable[hash & HASHMASK];
	
	//printf ("input address:%s, hash: %d\n", convert_address_to_string (rcvd_main_addr), hash);
	
	for (main_address=main_address_hash->mainaddress_forw; main_address!=(struct main_addr_entry *)main_address_hash;main_address=main_address->main_addr_forw)
	{
		//printf ("..... now looking at %s, hash:%d\n", convert_address_to_string(&main_address->node_main_address), main_address->mainaddress_hash);
		if (main_address->mainaddress_hash != hash)
		{
			continue;
		}
			
		if (0==memcmp(&main_address->node_main_address, rcvd_main_addr, sizeof(struct olsr_ip_addr)))
		{
			return main_address;
		}
	}
	
	return NULL;
}
/*------------------------------------------------------------------------*/
struct iface_addr_entry
*olsr_lookup_interface_address_table (struct olsr_ip_addr *rcvd_interface_addr)
{
	struct iface_addr_entry 	*interface_address;
	struct iface_addr_hash  	*interface_address_hash;
	olsr_u32_t                 	hash;
	
	//printf ("****************** in lookup interface address table *****************\n");
	olsr_print_main_address_table();
	olsr_print_interface_address_table();
		
	olsr_hashing(rcvd_interface_addr, &hash);
	interface_address_hash = &nodeifaceaddresstable[hash & HASHMASK];
	
	//printf ("input address:%s, hash: %d\n", convert_address_to_string (rcvd_interface_addr), hash);
	
	for (interface_address=interface_address_hash->interfaceaddress_forw; interface_address!=(struct iface_addr_entry *)interface_address_hash; interface_address=interface_address->iface_addr_forw)
	{
		//printf ("..... now looking at %s, hash:%d\n", convert_address_to_string(&interface_address->node_interface_address), interface_address->interfaceaddress_hash);
		if (interface_address->interfaceaddress_hash != hash)
		{
			continue;
		}
			
		if (0==memcmp(&interface_address->node_interface_address, rcvd_interface_addr, sizeof(struct olsr_ip_addr)))
		{
			return interface_address;
		}
	}
	
	return NULL; 
}
/*------------------------------------------------------------------------*/
struct interface_addr_list
*olsr_in_list_of_interface_address(struct main_addr_entry *my_main_address_entry, struct olsr_ip_addr *rcvd_interface_addr)
{
	struct interface_addr_list	*list_of_my_interface_address;
	
	list_of_my_interface_address = my_main_address_entry->list_of_interfaceaddress;
	
	while (list_of_my_interface_address != NULL)
	{
		if (0==(memcmp(&list_of_my_interface_address->interface_addresses->node_interface_address, rcvd_interface_addr, sizeof(struct olsr_ip_addr))))
		{
			return list_of_my_interface_address;
		}
		
		list_of_my_interface_address = list_of_my_interface_address->next;
	}
	
	return NULL;
}
/*------------------------------------------------------------------------*/
void
olsr_insert_interface_address_table(struct iface_addr_entry *interface_assoc_entry, struct mid_iface_addr *message_interface_address)
{
	olsr_u32_t      	hash;
	struct iface_addr_hash	*interface_address_hash;
	
	olsr_hashing(&message_interface_address->interface_address, &hash);
	
	interface_address_hash = &nodeifaceaddresstable[hash & HASHMASK];
	
	memcpy(&interface_assoc_entry->node_interface_address, &message_interface_address->interface_address, sizeof (struct olsr_ip_addr));
	interface_assoc_entry->node_mainaddress = NULL;
	interface_assoc_entry->interfaceaddress_hash = hash;
	
	olsr_insque((struct olsr_qelem *)interface_assoc_entry, (struct olsr_qelem *)interface_address_hash);
}
/*------------------------------------------------------------------------*/
void
olsr_insert_main_address_table(struct main_addr_entry *my_main_address_entry, struct mid_message *message)
{
	olsr_u32_t      	hash;
	struct main_addr_hash	*my_main_address_hash;
	
	olsr_hashing(&message->originator_addr, &hash);
	my_main_address_hash = &nodemainaddresstable[hash & HASHMASK];
	
	memcpy(&my_main_address_entry->node_main_address, &message->originator_addr, sizeof (struct olsr_ip_addr));
	my_main_address_entry->list_of_interfaceaddress = NULL;
	my_main_address_entry->mainaddress_hash = hash;
	my_main_address_entry->interface_address_pointer = 0;
	
	olsr_insque((struct olsr_qelem *) my_main_address_entry, (struct olsr_qelem *)my_main_address_hash);	
}
/*------------------------------------------------------------------------*/
void
olsr_update_main_address_table (struct main_addr_entry *my_main_address_entry, struct mid_message *message)
{
	struct mid_iface_addr      	*message_interface_address;
	struct interface_addr_list	*list_of_my_interface_address;
	struct iface_addr_entry    	*interface_assoc_entry;
	struct main_address_information	*my_main_address;
	struct timeval          	hold_time_tmp;
	float                   	rcvd_vtime;
	
	//printf ("\n\n+++++++++++++++++ in olsr_update_main_address_table +++++++++++++++\n");
	
	rcvd_vtime = olsr_compute_time_from_vtime(message->v_time);
	olsr_init_timer(rcvd_vtime*1000,&hold_time_tmp);  //the v_time value in MID

	message_interface_address = message->mid_interface_address;
	
	while (message_interface_address!=NULL)
	{
		if (debug_level > 3)
		{
			printf (" \n.....now look at message interface address:%s\n", convert_address_to_string(&message_interface_address->interface_address));
			fprintf (y_file," \n.....now look at message interface address:%s\n", convert_address_to_string(&message_interface_address->interface_address));
		}
		
		if (NULL == (list_of_my_interface_address = olsr_in_list_of_interface_address(my_main_address_entry, &message_interface_address->interface_address)))
		{
			if (debug_level > 3)
			{
				printf ("      the list of interface is not in the interface_address_list!\n");
				fprintf (y_file,"      the list of interface is not in the interface_address_list!\n");
			}
			
			changes_topology=UP;
			
			interface_assoc_entry = (struct iface_addr_entry *)malloc(sizeof (struct iface_addr_entry ));
			olsr_insert_interface_address_table(interface_assoc_entry, message_interface_address);
			
			/* link the main address entry and interface address entry together */
			list_of_my_interface_address=(struct interface_addr_list *)malloc(sizeof (struct interface_addr_list));
			my_main_address = (struct main_address_information *)malloc(sizeof (struct main_address_information));
			
			list_of_my_interface_address->interface_addresses = interface_assoc_entry;
			list_of_my_interface_address->next = my_main_address_entry->list_of_interfaceaddress;
			my_main_address_entry->list_of_interfaceaddress = list_of_my_interface_address;
			
			my_main_address->mainaddress = my_main_address_entry;
			interface_assoc_entry->node_mainaddress = my_main_address;
			
			my_main_address_entry->interface_address_pointer ++;
			
			timeradd(&now, &hold_time_tmp, &interface_assoc_entry->interface_address_timer);			
		}
		else
		{
			interface_assoc_entry = list_of_my_interface_address->interface_addresses;
			timeradd(&now, &hold_time_tmp, &interface_assoc_entry->interface_address_timer);
		}
		
		message_interface_address=message_interface_address->next;		
	}	
}
/*------------------------------------------------------------------------*/
void
olsr_time_out_interface_address_table()
{
	olsr_u8_t       	index;
	struct iface_addr_entry	*my_interface_address;
	struct iface_addr_entry	*my_interface_address_tmp;
	struct iface_addr_hash	*interface_address_hash;
	int	i;
	int	changed = 0, interface_timed_out = 0;
	
	for (index=0; index<HASHSIZE; index++)
	{
		interface_address_hash = &nodeifaceaddresstable[index];
		my_interface_address = interface_address_hash->interfaceaddress_forw;
		while (my_interface_address != (struct iface_addr_entry	*)interface_address_hash)
		{
			if(olsr_timed_out (&my_interface_address->interface_address_timer))
			{
				if (debug_level > 1)
				{
					printf ("\n\n@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ interface timed out! @@@@@@@@@@@@@@@@@@@@@@@@@@@@\n");
					printf (" the timed out interface is: %s\n", convert_address_to_string(&my_interface_address->node_interface_address));
				
					fprintf (y_file,"\n\n@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ interface timed out! @@@@@@@@@@@@@@@@@@@@@@@@@@@@\n");
					fprintf (y_file," the timed out interface is: %s\n", convert_address_to_string(&my_interface_address->node_interface_address));
				}
				
				interface_timed_out = 1;
				my_interface_address_tmp = my_interface_address;
				my_interface_address = my_interface_address->iface_addr_forw;
				olsr_delete_node_interfaceaddress_table(my_interface_address_tmp);
				changes_topology = UP;
			}
			else
			{
				my_interface_address = my_interface_address->iface_addr_forw;
			}	
		}
	} 
	
	if (debug_level > 1 && interface_timed_out)
	{
		printf (" AFTER TIME OUT, THE TABLES ARE :\n");
		olsr_print_main_address_table();
		olsr_print_interface_address_table();
		
		fprintf (y_file," AFTER TIME OUT, THE TABLES ARE :\n");
		olsr_print_trace_main_address_table();
		olsr_print_trace_interface_address_table();
		
		interface_timed_out = 0;
	}
}
/*------------------------------------------------------------------------*/
void
olsr_print_main_address_table()
{
	struct main_addr_entry  	*main_address;
	struct main_addr_hash   	*main_address_hash;
	olsr_u32_t              	index;
	struct interface_addr_list	*list_of_my_interface_address;
	
	
	printf ("\n\nNODE MAIN ADDRESS TABLE:%02d:%02d:%02d.%06lu\n",nowtm->tm_hour, nowtm->tm_min, nowtm->tm_sec, now.tv_usec);
	
	for (index=0; index<HASHSIZE; index++)
	{
		main_address_hash = &nodemainaddresstable[index];
		
		for (main_address=main_address_hash->mainaddress_forw; main_address!=(struct main_addr_entry *)main_address_hash;main_address=main_address->main_addr_forw)
		{
			printf ("main address: %s, number of interface: %d [", convert_address_to_string(&main_address->node_main_address), main_address->interface_address_pointer);
			
			list_of_my_interface_address = main_address->list_of_interfaceaddress;
			
			while (list_of_my_interface_address!=NULL)
			{
				printf ("%s, ", convert_address_to_string(&list_of_my_interface_address->interface_addresses->node_interface_address));
				list_of_my_interface_address=list_of_my_interface_address->next;
			}
			printf ("]\n");
		}
	}
}
/*------------------------------------------------------------------------*/
void
olsr_print_interface_address_table()
{
	olsr_u8_t              	index;
	struct iface_addr_entry	*my_interface_address;
	struct iface_addr_hash 	*interface_address_hash;
	struct main_addr_entry	*my_main_address;
	
	printf ("\n\nNODE INTERFACE ADDRESS TABLE:%02d:%02d:%02d.%06lu\n",nowtm->tm_hour, nowtm->tm_min, nowtm->tm_sec, now.tv_usec);
	
	for (index=0; index<HASHSIZE; index++)
	{
		interface_address_hash = &nodeifaceaddresstable[index];
		my_interface_address = interface_address_hash->interfaceaddress_forw;
		for (my_interface_address=interface_address_hash->interfaceaddress_forw; my_interface_address!=(struct iface_addr_entry *)interface_address_hash; my_interface_address=my_interface_address->iface_addr_forw)
		{
			printf ("%s ", convert_address_to_string (&my_interface_address->node_interface_address));
			my_main_address = my_interface_address->node_mainaddress->mainaddress;
			printf ("-- %s (interface pointer: %d)\n", convert_address_to_string(&my_main_address->node_main_address),my_main_address->interface_address_pointer);
			printf ("----------> this interface's timer: %d, %d\n", my_interface_address->interface_address_timer.tv_sec,my_interface_address->interface_address_timer.tv_usec);
		}
		
	} 
	
}
/*------------------------------------------------------------------------*/
void
olsr_print_trace_main_address_table()
{
	struct main_addr_entry  	*main_address;
	struct main_addr_hash   	*main_address_hash;
	olsr_u32_t              	index;
	struct interface_addr_list	*list_of_my_interface_address;
	
	
	fprintf (y_file,"\n\nNODE MAIN ADDRESS TABLE:%02d:%02d:%02d.%06lu\n",nowtm->tm_hour, nowtm->tm_min, nowtm->tm_sec, now.tv_usec);
	
	for (index=0; index<HASHSIZE; index++)
	{
		main_address_hash = &nodemainaddresstable[index];
		
		for (main_address=main_address_hash->mainaddress_forw; main_address!=(struct main_addr_entry *)main_address_hash;main_address=main_address->main_addr_forw)
		{
			fprintf (y_file, "main address: %s, number of interface: %d [", convert_address_to_string(&main_address->node_main_address), main_address->interface_address_pointer);
			
			list_of_my_interface_address = main_address->list_of_interfaceaddress;
			
			while (list_of_my_interface_address!=NULL)
			{
				fprintf (y_file,"%s, ", convert_address_to_string(&list_of_my_interface_address->interface_addresses->node_interface_address));
				list_of_my_interface_address=list_of_my_interface_address->next;
			}
			fprintf (y_file,"]\n");
		}
	}
}
/*------------------------------------------------------------------------*/
void
olsr_print_trace_interface_address_table()
{
	olsr_u8_t              	index;
	struct iface_addr_entry	*my_interface_address;
	struct iface_addr_hash 	*interface_address_hash;
	struct main_addr_entry	*my_main_address;
	
	fprintf (y_file,"\n\nNODE INTERFACE ADDRESS TABLE:%02d:%02d:%02d.%06lu\n",nowtm->tm_hour, nowtm->tm_min, nowtm->tm_sec, now.tv_usec);
	
	for (index=0; index<HASHSIZE; index++)
	{
		interface_address_hash = &nodeifaceaddresstable[index];
		my_interface_address = interface_address_hash->interfaceaddress_forw;
		for (my_interface_address=interface_address_hash->interfaceaddress_forw; my_interface_address!=(struct iface_addr_entry *)interface_address_hash; my_interface_address=my_interface_address->iface_addr_forw)
		{
			fprintf (y_file,"%s ", convert_address_to_string (&my_interface_address->node_interface_address));
			my_main_address = my_interface_address->node_mainaddress->mainaddress;
			fprintf (y_file,"-- %s (interface pointer: %d)\n", convert_address_to_string(&my_main_address->node_main_address),my_main_address->interface_address_pointer);
			fprintf (y_file,"----------> this interface's timer: %d, %d\n", my_interface_address->interface_address_timer.tv_sec,my_interface_address->interface_address_timer.tv_usec);
		}
		
	} 
	
}
		
	
	
	
	
	
