/*
 * Copyright (c) 2004 Ying Ge, Communication Research Center Canada.
 *
 * Copyright (c) 2002, 2003 Maoyu Wang, Communication Research Center Canada.
 *
 * By Ying Ge:
 * 1. Change the OLSR packet format and message processing procedure based 
 *    on the OLSR RFC.
 * 2. Add support of multiple interfaces to OLSR, including MID message 
 *    creating and processing procedure as specified in the OLSR RFC.
 * 3. Add QoS Support to OLSR
 *
 * By Maoyu Wang:
 * 1. Ported OLSR from IPv4 to IPv6.
 * 2. Added the Host and Network Association (HNA) functionality into OLSR.
 * 3. Added the default gateway functionality into OLSR by extending the HNA 
 *    message usage. The default gateway functionality supported the mobility
 *	  by cooperating with Mobile IPv6 for a mobile node as well as supported 
 * 	  Internet access for MANET nodes. 
 *
 * DISTRIBUTED WITH NO WARRANTY, EXPRESS OR IMPLIED.
 * See the GNU Library General Public License (file COPYING in the distribution)
 * for conditions of use and redistribution
 */
/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est-à-dire qu'il peut
 * être utilisé, copié, modifié, distribué à l'unique
 * condition que ce texte soit conservé afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant participé à l'élaboration de ce logiciel ne peut
 * être utilisé sans son accord préalable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilité d'aucune sorte.
 * Certaines parties de ce logiciel sont dérivées de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 * 
 *	$Id: topology_table.h,v 1.2 2000/12/06 10:36:12 prima Exp $
 */

/*
 * Routing table management daemon.
 *
 */
/**************** the file is created by Y.Ge ****************/

/*
 * interface_associattion table (mid table) structure.
 */


struct main_address_information{
	struct main_addr_entry *mainaddress;       
};

struct iface_addr_infos{
	olsr_u32_t         		interfaceaddress_hash;
	struct olsr_ip_addr		node_interface_address;
	struct timeval             	interface_address_timer;
	struct main_address_information	*node_mainaddress;
};

struct iface_addr_entry{
	struct	iface_addr_entry	*iface_addr_forw;
	struct	iface_addr_entry	*iface_addr_back; 
	struct	iface_addr_infos	iface_addr_infor;
};

#define interfaceaddress_hash		iface_addr_infor.interfaceaddress_hash

#define node_interface_address		iface_addr_infor.node_interface_address

#define interface_address_timer		iface_addr_infor.interface_address_timer

#define node_mainaddress      		iface_addr_infor.node_mainaddress

struct interface_addr_list{
	struct interface_addr_list  	*next;
	struct iface_addr_entry 	*interface_addresses;
};

struct main_address_infos {
	olsr_u32_t                      mainaddress_hash;
	struct olsr_ip_addr             node_main_address;
	olsr_u16_t                	interface_address_pointer;        
	struct interface_addr_list	*list_of_interfaceaddress;
};


struct main_addr_entry {
	struct	main_addr_entry 	*main_addr_forw;
	struct	main_addr_entry 	*main_addr_back;
	struct  main_address_infos	main_addr_infor;
};


#define	mainaddress_hash         	main_addr_infor.mainaddress_hash

#define	node_main_address        	main_addr_infor.node_main_address      

#define	interface_address_pointer      	main_addr_infor.interface_address_pointer      /* how many interface address this node has? */

#define list_of_interfaceaddress	main_addr_infor.list_of_interfaceaddress 

struct main_addr_hash {
	struct	main_addr_entry		*mainaddress_forw;
	struct	main_addr_entry		*mainaddress_back;
};



struct iface_addr_hash {
	struct	iface_addr_entry	*interfaceaddress_forw;
	struct	iface_addr_entry	*interfaceaddress_back;
};

void
olsr_delete_mainaddress_link (struct main_addr_entry  *my_main_address_entry, struct olsr_ip_addr *my_interface_address);

void
olsr_delete_node_interfaceaddress_table (struct iface_addr_entry *ifaceaddress_entry);

void
olsr_delete_node_mainaddrress_table (struct main_addr_entry *mainaddress_entry);

void
olsr_delete_interfaceaddress_link (struct iface_addr_entry  *my_iface_addr_entry);

void
olsr_delete_node_mainaddress_table (struct main_addr_entry *my_main_address_entry);

void
olsr_release_mid_table();

struct main_addr_entry
*olsr_lookup_main_address_table (struct olsr_ip_addr *rcvd_main_addr);

struct iface_addr_entry
*olsr_lookup_interface_address_table (struct olsr_ip_addr *rcvd_interface_addr);

struct interface_addr_list
*olsr_in_list_of_interface_address(struct main_addr_entry *my_main_address_entry, struct olsr_ip_addr *rcvd_interface_addr);

void
olsr_insert_interface_address_table(struct iface_addr_entry *interface_assoc_entry, struct mid_iface_addr *message_interface_address);

void
olsr_insert_main_address_table(struct main_addr_entry *my_main_address_entry, struct mid_message *message);

void
olsr_update_main_address_table (struct main_addr_entry *my_main_address_entry, struct mid_message *message);

void
olsr_time_out_interface_address_table();

void
olsr_print_main_address_table();

void
olsr_print_interface_address_table();

void
olsr_print_trace_main_address_table();

void
olsr_print_trace_interface_address_table();
