/*
 * Copyright (c) 2004 Ying Ge, Communication Research Center Canada.
 *
 * Copyright (c) 2002, 2003 Maoyu Wang, Communication Research Center Canada.
 *
 * By Ying Ge:
 * 1. Change the OLSR packet format and message processing procedure based 
 *    on the OLSR RFC.
 * 2. Add support of multiple interfaces to OLSR, including MID message 
 *    creating and processing procedure as specified in the OLSR RFC.
 * 3. Add QoS Support to OLSR
 *
 * By Maoyu Wang:
 * 1. Ported OLSR from IPv4 to IPv6.
 * 2. Added the Host and Network Association (HNA) functionality into OLSR.
 * 3. Added the default gateway functionality into OLSR by extending the HNA 
 *    message usage. The default gateway functionality supported the mobility
 *	  by cooperating with Mobile IPv6 for a mobile node as well as supported 
 * 	  Internet access for MANET nodes. 
 *
 * DISTRIBUTED WITH NO WARRANTY, EXPRESS OR IMPLIED.
 * See the GNU Library General Public License (file COPYING in the distribution)
 * for conditions of use and redistribution
 */
/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est-à-dire qu'il peut
 * être utilisé, copié, modifié, distribué à l'unique
 * condition que ce texte soit conservé afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant participé à l'élaboration de ce logiciel ne peut
 * être utilisé sans son accord préalable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilité d'aucune sorte.
 * Certaines parties de ce logiciel sont dérivées de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 *
 *	$Id: main_addr_neigh_table.h,v 1.2 2000/12/06 10:36:11 prima Exp $
 */

/*
 * Routing table management daemon.
 */

/**************** the file is created by Y.Ge ****************/

/*
 * Main addr neighbor table structure.
 */
struct iface_addr_list_entry {
	struct iface_addr_list_entry 	*iface_addr_next;
	struct neighbor_entry       	*neighbor_iface;
};

struct main_addr_neigh_info {
	olsr_u32_t          		main_addr_neigh_hash;
	olsr_u8_t                   	covered_in_hello;          	/* if the interface address of this main neigbor is covered in hello message */
	olsr_u8_t                    	covered_in_routing;         	/* if the main address is covered in olsr_fill_routing_table_with_neighbors() */
	struct olsr_ip_addr 		main_addr_neigh_addr;       	/* neighbor main addr */
	olsr_u8_t           		main_addr_neigh_willingness;	/* status of the neighbor willingness */
	olsr_u8_t           		main_addr_neigh_status;     	/* status of the main_addr_neigh */
	olsr_u16_t                  	interface_pointer;          	/* how many interface addrress this neighbor has ? */
	struct iface_addr_list_entry	*neigh_iface_list;           	/* link to the interface address of this neighbor */
};

struct main_addr_neigh_entry 
{
	struct	main_addr_neigh_entry	*main_addr_neigh_forw;
	struct	main_addr_neigh_entry	*main_addr_neigh_back;
	struct	main_addr_neigh_info 	main_addr_neigh_infos;
};


#define	main_addr_neigh_hash         	main_addr_neigh_infos.main_addr_neigh_hash
#define	covered_in_hello           	main_addr_neigh_infos.covered_in_hello
#define	covered_in_routing           	main_addr_neigh_infos.covered_in_routing
#define	main_addr_neigh_addr     	main_addr_neigh_infos.main_addr_neigh_addr
#define	main_addr_neigh_status    	main_addr_neigh_infos.main_addr_neigh_status
#define	main_addr_neigh_willingness	main_addr_neigh_infos.main_addr_neigh_willingness
#define	interface_pointer          	main_addr_neigh_infos.interface_pointer	
#define	neigh_iface_list           	main_addr_neigh_infos.neigh_iface_list

struct main_addr_neighhash {
	struct	main_addr_neigh_entry	*mainaddrneigh_forw;
	struct	main_addr_neigh_entry	*mainaddrneigh_back;
};

/* neighbor type */
#define	NOT_NEIGH  	0
#define	SYM_NEIGH  	1
#define	MPR_NEIGH  	2

/* willingness of the node forward packets in the network */
#define	WILL_NEVER	0
#define	WILL_LOW  	1
#define	WILL_DEFAULT	3
#define	WILL_HIGH   	6
#define	WILL_ALWAYS 	7

struct iface_addr_list_entry
*olsr_lookup_iface_neighbor(struct main_addr_neigh_entry *main_neighbor,struct olsr_ip_addr *iface_address);

void
olsr_delete_iface_neighbor_pointer(struct main_addr_neigh_entry *main_addr_neigh, struct olsr_ip_addr *address);

void
olsr_delete_main_addr_neigh_table(struct main_addr_neigh_entry *main_addr_neigh);

void
olsr_release_main_addr_neigh_table();

struct main_addr_neigh_entry
*olsr_lookup_main_addr_neigh_table(struct olsr_ip_addr *n_main_addr);

void
olsr_insert_main_addr_neigh_table(struct main_addr_neigh_entry *main_addr_neigh);

void
olsr_clear_main_addr_neigh_table_mpr_status();

void
olsr_clear_main_addr_neigh_table_covered();

void
olsr_clear_main_addr_neigh_table_covered_in_routing();

void
olsr_print_main_addr_neigh_table();

void
olsr_print_trace_main_addr_neigh_table(); //added by Y.Ge

void
olsr_update_main_addr_neigh_status(struct main_addr_neigh_entry *main_addr_neigh);

void
olsr_update_main_addr_neigh_table_status ();


