/*
 * Copyright (c) 2004 Ying Ge, Communication Research Center Canada.
 *
 * Copyright (c) 2002, 2003 Maoyu Wang, Communication Research Center Canada.
 *
 * By Ying Ge:
 * 1. Change the OLSR packet format and message processing procedure based 
 *    on the OLSR RFC.
 * 2. Add support of multiple interfaces to OLSR, including MID message 
 *    creating and processing procedure as specified in the OLSR RFC.
 * 3. Add QoS Support to OLSR
 *
 * By Maoyu Wang:
 * 1. Ported OLSR from IPv4 to IPv6.
 * 2. Added the Host and Network Association (HNA) functionality into OLSR.
 * 3. Added the default gateway functionality into OLSR by extending the HNA 
 *    message usage. The default gateway functionality supported the mobility
 *	  by cooperating with Mobile IPv6 for a mobile node as well as supported 
 * 	  Internet access for MANET nodes. 
 *
 * DISTRIBUTED WITH NO WARRANTY, EXPRESS OR IMPLIED.
 * See the GNU Library General Public License (file COPYING in the distribution)
 * for conditions of use and redistribution
 */
/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est--dire qu'il peut
 * tre utilis, copi, modifi, distribu  l'unique
 * condition que ce texte soit conserv afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant particip  l'laboration de ce logiciel ne peut
 * tre utilis sans son accord pralable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilit d'aucune sorte.
 * Certaines parties de ce logiciel sont drives de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 * 
 *	$Id: mpr_selector_table.h,v 1.2 2000/12/06 10:36:11 prima Exp $
 */


/*
 * Routing table management daemon.
 */


/*
 * MultiPoint Relay Selector table structure.
 */

struct mpr_selector_info {

        olsr_u32_t      mprselector_hash;
        struct olsr_ip_addr      mpr_selector_addr;
	olsr_u16_t       mpr_selector_seq;
	
	/* qos OLSR */
	olsr_u32_t    	mpr_selector_busy_time; //added by Y.Ge
	
	struct timeval	mpr_selector_timer;		        
};


struct mpr_selector_entry {
	struct	mpr_selector_entry *mpr_selector_forw;
	struct	mpr_selector_entry *mpr_selector_back;
        struct  mpr_selector_info mpr_selector_infos;
};


#define	mprselector_hash      	mpr_selector_infos.mprselector_hash

#define	mpr_selector_addr      	mpr_selector_infos.mpr_selector_addr    /* MultiPoint Relay selector address */
#define	mpr_selector_seq    	mpr_selector_infos.mpr_selector_seq     /* MPRS sequence number with
									   which mprs_addr has selected
									   this node as its MPR */
#define mpr_selector_busy_time	mpr_selector_infos.mpr_selector_busy_time /* Y.Ge: mpr selector's busy  time */
#define	mpr_selector_timer     	mpr_selector_infos.mpr_selector_timer   /* for invalidation */

struct mpr_selector_hash {
  struct	mpr_selector_entry *mpr_selector_forw;
  struct	mpr_selector_entry *mpr_selector_back;
};

struct mpr_selector_table {
  olsr_u16_t      mssn;
  struct	mpr_selector_hash *mpr_selector_hash;
};


void
olsr_delete_mpr_selector_table(struct mpr_selector_entry *mprs);

void
olsr_insert_mpr_selector_table(struct mpr_selector_entry *mprs);

struct mpr_selector_entry
*olsr_lookup_mpr_selector_table(struct olsr_ip_addr *dst);

void
olsr_update_mpr_selector_table(struct hello_message *message);

void
olsr_update_mpr_selector_busy_time (struct qos_hello_message *message); //added by Y.Ge, qos OLSR

void
olsr_release_mpr_seletector_table();

void
olsr_time_out_mpr_selector_table();

void
olsr_print_mpr_selector_table();

void
olsr_print_trace_mpr_selector_table(); //added by Y.Ge
