/*
 * Copyright (c) 2004 Ying Ge, Communication Research Center Canada.
 *
 * Copyright (c) 2002, 2003 Maoyu Wang, Communication Research Center Canada.
 *
 * By Ying Ge:
 * 1. Change the OLSR packet format and message processing procedure based 
 *    on the OLSR RFC.
 * 2. Add support of multiple interfaces to OLSR, including MID message 
 *    creating and processing procedure as specified in the OLSR RFC.
 * 3. Add QoS Support to OLSR
 *
 * By Maoyu Wang:
 * 1. Ported OLSR from IPv4 to IPv6.
 * 2. Added the Host and Network Association (HNA) functionality into OLSR.
 * 3. Added the default gateway functionality into OLSR by extending the HNA 
 *    message usage. The default gateway functionality supported the mobility
 *	  by cooperating with Mobile IPv6 for a mobile node as well as supported 
 * 	  Internet access for MANET nodes. 
 *
 * DISTRIBUTED WITH NO WARRANTY, EXPRESS OR IMPLIED.
 * See the GNU Library General Public License (file COPYING in the distribution)
 * for conditions of use and redistribution
 */
/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est-à-dire qu'il peut
 * être utilisé, copié, modifié, distribué à l'unique
 * condition que ce texte soit conservé afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant participé à l'élaboration de ce logiciel ne peut
 * être utilisé sans son accord préalable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilité d'aucune sorte.
 * Certaines parties de ce logiciel sont dérivées de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 */

/*
 * Routing table management daemon.
 */

char neighbor_table_rcsid[] = 
  "$Id: neighbor_table.c,v 1.2 2000/12/06 10:36:11 prima Exp $";

#include "defs.h"


struct neighbor_table		neighbortable; 		

void 
olsr_delete_neighbor_2_pointer(struct neighbor_entry *neighbor,struct olsr_ip_addr *address)
{

	struct neighbor_2_list_entry *neighbor_two_list;
	struct neighbor_2_list_entry *neighbor_two_list_tmp;
  
	neighbor_two_list=neighbor->neighbor_2_list;

	neighbor->neighbor_2_list=NULL;

	while(neighbor_two_list!=NULL)
	{
		if(0==memcmp(&neighbor_two_list->neighbor_2->neighbor_2_addr,address,sizeof(struct olsr_ip_addr)))
		{
			neighbor_two_list_tmp= neighbor_two_list;
			neighbor_two_list=neighbor_two_list->neighbor_2_next;
			free(neighbor_two_list_tmp);	  
		}
		else
		{
			neighbor_two_list_tmp= neighbor_two_list;
			neighbor_two_list=neighbor_two_list->neighbor_2_next;
			neighbor_two_list_tmp->neighbor_2_next= neighbor->neighbor_2_list;
			neighbor->neighbor_2_list=neighbor_two_list_tmp;
		}      
	}
}

/*------------------------------------------------------------------------*/
/***************** added by Y.Ge **************/
void 
olsr_delete_main_neighbor_pointer(struct neighbor_entry *neighbor)
{

	struct link_to_main_neighbor	*main_neighbor_link;
	struct link_to_main_neighbor	*main_neighbor_link_tmp;
  
	main_neighbor_link=neighbor->neighbor_main_info;

	neighbor->neighbor_main_info=NULL;

	free(main_neighbor_link);	  
}
/**************** end of revision *************/
/*------------------------------------------------------------------------*/
struct neighbor_2_list_entry
*olsr_lookup_my_neighbors(struct neighbor_entry *neighbor,struct olsr_ip_addr *neighbor_address)
{
  
	struct neighbor_2_list_entry *neighbor_two_list;
  
	neighbor_two_list=neighbor->neighbor_2_list;
  
	while(neighbor_two_list!=NULL)
	{
      		if(0==memcmp(&neighbor_two_list->neighbor_2->neighbor_2_addr,neighbor_address,sizeof(struct olsr_ip_addr)))
			return(neighbor_two_list);
      
		neighbor_two_list=neighbor_two_list->neighbor_2_next;
      
	}
	return(NULL);
}
/*------------------------------------------------------------------------*/
void
olsr_delete_neighbor_table(struct neighbor_entry *neighbor)
{
  /*
    Delete a neighbor entry results the deletion of its 2 hop neighbors list!!!
   */
  
	struct  neighbor_2_list_entry *two_hop_list;
	struct  neighbor_2_entry      *two_hop_entry;
	struct  link_to_main_neighbor	*main_address_neighbor_entry; //added by Y.Ge
	struct  main_addr_neigh_entry	*main_neigh; //added by Y.Ge

	two_hop_list = neighbor->neighbor_2_list;
	main_address_neighbor_entry = neighbor->neighbor_main_info; //added by Y.Ge

	while(two_hop_list!=NULL)
	{
		two_hop_entry=two_hop_list->neighbor_2;
		two_hop_entry->neighbor_2_pointer--;

		olsr_delete_neighbor_pointer(two_hop_entry,&neighbor->neighbor_addr);
      
		if(two_hop_entry->neighbor_2_pointer<1)
		/* means this two hop entry is no more pointed, not reachable */
		{
			olsr_remque((struct olsr_qelem *)two_hop_entry);
			free((void *)two_hop_entry);
		}
		neighbor->neighbor_2_list=two_hop_list->neighbor_2_next;     
		free(two_hop_list);
		two_hop_list = neighbor->neighbor_2_list;
	}
	/***************added by Y.Ge *******************/
	/*
	 * also delete the link to the neighbor_main_addr table, update the neigh_status in neighbor_main_addr_table
	 */
	main_neigh = main_address_neighbor_entry->main_neighbor;
	main_neigh->interface_pointer --;
	//printf ("currently, main addr neighbor_info is: addr(%s), link(%d)\n", convert_address_to_string(&main_neigh->main_addr_neigh_addr), main_neigh->interface_pointer);
	olsr_delete_iface_neighbor_pointer(main_neigh, &neighbor->neighbor_addr);
	/*printf ("after deleting the neighbor pointer, the neighbor tables are: \n");
	olsr_print_neighbor_table();
	olsr_print_main_addr_neigh_table();*/
	if (main_neigh->interface_pointer < 1)
	{
		//printf ("-------- should delete the main address neighbor !\n");
		/* delete the neighbor from main_addr_neigh_table */
		olsr_remque ((struct olsr_qelem *)main_neigh);
		free((void *)main_neigh);
	}
	else
	{
		olsr_update_main_addr_neigh_status(main_neigh);
	}
	free (main_address_neighbor_entry); 
	/*************** end of revision ****************/
 
	if(neighbor->neighbor_ifp!=NULL)    
		free(neighbor->neighbor_ifp);

	TRACE_ACTION("DELETE NEIGHBOR", &neighbor->neighbor_addr,NULL,0);
	olsr_remque((struct olsr_qelem *)neighbor);
  
	free((void *)neighbor);
}
/*------------------------------------------------------------------------*/
void
olsr_insert_neighbor_table(struct neighbor_entry *neighbor)
{

	olsr_u32_t             hash;
	struct neighborhash    *hash_neighbor;

	olsr_hashing(&neighbor->neighbor_addr,&hash);
	neighbor->neighbor_hash=hash;

	hash_neighbor=&neighbortable.neighborhash[hash & HASHMASK];

	TRACE_ACTION("ADD NEIGHBOR", &neighbor->neighbor_addr,NULL,0);
	olsr_insque((struct olsr_qelem *)neighbor, (struct olsr_qelem *)hash_neighbor);
}
/*------------------------------------------------------------------------*/
/*********** commented by Y.Ge *************/
/*struct neighbor_entry 
*olsr_lookup_neighbor_table(struct olsr_ip_addr *dst)
{
	struct neighbor_entry  *neighbor;
	struct neighborhash    *hash_neighbor;
	olsr_u32_t             hash;

	olsr_hashing(dst,&hash);
	hash_neighbor=&neighbortable.neighborhash[hash & HASHMASK];

	for(neighbor=hash_neighbor->neighbor_forw;neighbor!=(struct neighbor_entry *)hash_neighbor;neighbor=neighbor->neighbor_forw)
	{
		if(neighbor->neighbor_hash!=hash)
			continue;

		if (0==memcmp(&neighbor->neighbor_addr,dst,sizeof(struct olsr_ip_addr)))
			return(neighbor);    
	}
	return (NULL);  
}*/
/*------------------------------------------------------------------------*/
/************ added by Y.Ge ***********/
struct neighbor_entry 
*olsr_lookup_neighbor_table(struct olsr_ip_addr *dst, struct olsr_ip_addr *my_addr)
{
	struct neighbor_entry	*neighbor, *neighbor_tmp;
	struct neighborhash   	*hash_neighbor;
	olsr_u32_t           	hash;
	int                  	i, mistake = 0;;

	olsr_hashing(dst,&hash);
	hash_neighbor=&neighbortable.neighborhash[hash & HASHMASK];

	for(neighbor=hash_neighbor->neighbor_forw;neighbor!=(struct neighbor_entry *)hash_neighbor;neighbor=neighbor->neighbor_forw)
	{
		if(neighbor->neighbor_hash!=hash)
			continue;

		if (0==memcmp(&neighbor->neighbor_addr,dst,sizeof(struct olsr_ip_addr)))
		{
			if (0==memcmp(&neighbor->my_iface_addr, my_addr, sizeof (struct olsr_ip_addr)))
				return (neighbor);    
			else
			{
				printf ("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\n");
				printf ("XXXXXXXXXXXXXXXXXXXX ATTENTION: MAYBE there is ERROR in IFACE ADDR! XXXXXXXXXXXXXXXX\n");
				printf ("XXXXXXXX the input addr is: neighbor(%s)\n",convert_address_to_string(dst));
				printf ("          mine(%s)\n", convert_address_to_string(my_addr));
				printf ("XXXXXXXX the found addr is: neighbor(%s)\n", convert_address_to_string(&neighbor->neighbor_addr));
				printf ("          mine(%s)\n",  convert_address_to_string(&neighbor->my_iface_addr));
				printf ("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\n");
				return (neighbor);				
			}
		}
	}

	return (NULL);  
}
/*------------------------------------------------------------------------*/
void
olsr_release_neighbor_table()
{
	olsr_u8_t              index;
	struct neighbor_entry  *neighbor;
	struct neighbor_entry  *neighbor_tmp;
	struct neighborhash    *hash_neighbor;
	int                    i;

	for(index=0;index<HASHSIZE;index++)
	{
		hash_neighbor=&neighbortable.neighborhash[index];

		//CRC modify: bugs in previous code:
		//for(neighbor=hash_neighbor->neighbor_forw;has_neighbor!=(struct neighbor_entry *)hash_neighbor;neighbor=neighbor->neighbor_forw)
	
		neighbor=hash_neighbor->neighbor_forw;
		while(neighbor!=(struct neighbor_entry*)hash_neighbor)
		{
			neighbor_tmp=neighbor;
			neighbor=neighbor->neighbor_forw;
			olsr_delete_neighbor_table(neighbor_tmp);
		}
	}
}
/*------------------------------------------------------------------------*/
void
olsr_time_out_two_hop_neighbors(struct neighbor_entry  *neighbor)
{
	struct  neighbor_2_list_entry *two_hop_list;
	struct  neighbor_2_list_entry *two_hop_list_tmp;
	struct  neighbor_2_entry      *two_hop_entry;
	int	two_hop_neighbor_time_out = 0;
	
	two_hop_list = neighbor->neighbor_2_list;
	neighbor->neighbor_2_list=NULL;
	
	while(two_hop_list!=NULL)
	{
		//printf ("two_hop_list !=NULL!\n");
		if(olsr_timed_out(&two_hop_list->neighbor_2_timer))
		{
			two_hop_neighbor_time_out = 1;
			
			if (debug_level > 1)
			{
				printf ("\n\n-----------------time(%d, %d), two hop neighbor timed out!--------------\n",now.tv_sec, now.tv_usec);
				printf (" the two hop neighbor is:%s\n", convert_address_to_string(&two_hop_list->neighbor_2->neighbor_2_addr));
				printf (" tow hop timer is: %d     %d\n\n", two_hop_list->neighbor_2_timer.tv_sec, two_hop_list->neighbor_2_timer.tv_usec);
			
				fprintf (y_file,"\n\n-----------------time(%d, %d), two hop neighbor timed out!--------------\n",now.tv_sec, now.tv_usec);
				fprintf (y_file," the two hop neighbor is:%s\n", convert_address_to_string(&two_hop_list->neighbor_2->neighbor_2_addr));
				fprintf (y_file," tow hop timer is: %d     %d\n\n", two_hop_list->neighbor_2_timer.tv_sec, two_hop_list->neighbor_2_timer.tv_usec);
			
			}
			two_hop_entry=two_hop_list->neighbor_2;
			two_hop_entry->neighbor_2_pointer--;
			olsr_delete_neighbor_pointer(two_hop_entry,&neighbor->neighbor_addr);
      
			if(two_hop_entry->neighbor_2_pointer<1)
			/* means this two hop entry is no more pointed, not reachable */
			{	    
				olsr_remque((struct olsr_qelem *)two_hop_entry);	      
				free((void *)two_hop_entry);
			}
			two_hop_list_tmp=two_hop_list;
			two_hop_list=two_hop_list->neighbor_2_next;	  
			free(two_hop_list_tmp);

			/* This flag is set to UP to recalculate the MPR set and the routing table*/
			changes_neighborhood=UP;
			changes_topology=UP;
		}
      		else
		{
			two_hop_list_tmp=two_hop_list;
			two_hop_list=two_hop_list->neighbor_2_next;
			two_hop_list_tmp->neighbor_2_next=neighbor->neighbor_2_list;
			neighbor->neighbor_2_list=two_hop_list_tmp;
		}
	}
	
	if (two_hop_neighbor_time_out && debug_level > 1)
	{
		printf ("XXXXXXXXXX after time out two hop neighbor table, the two hop neighbor table is:XXXXXXXXXXXX\n");
		olsr_print_two_hop_neighbor_table();
			
		fprintf (y_file,"XXXXXXXXXX after time out two hop neighbor table, the two hop neighbor table is:XXXXXXXXX\n");
		olsr_print_trace_neighbor_table(); 	
	}
}
/*------------------------------------------------------------------------*/
/***************** commented by Y.Ge ***************/
/*void
olsr_time_out_neighborhood_tables()
{
	olsr_u8_t              index;
	struct neighbor_entry  *neighbor;
	struct neighbor_entry  *neighbor_tmp;
	struct neighborhash    *hash_neighbor;
  
	for(index=0;index<HASHSIZE;index++)
	{
		hash_neighbor=&neighbortable.neighborhash[index];
		neighbor=hash_neighbor->neighbor_forw;
		while(neighbor!=(struct neighbor_entry *)hash_neighbor)
		{ 
			if(debug_level > 5)
			{
				printf("timeout: %d   %d\n ",neighbor->neighbor_timer.tv_sec,neighbor->neighbor_timer.tv_usec);
				printf("now: %d   %d\n ",now.tv_sec,now.tv_usec);
			}

			if(olsr_timed_out(&neighbor->neighbor_timer))
			{
				if((neighbor->neighbor_status==NBS_ASYM) ||(neighbor->neighbor_status==NBS_LOST))
				{
					if(debug_level > 2)
						printf("deleting neighbor entry\n");

					neighbor_tmp=neighbor;
					neighbor=neighbor->neighbor_forw;

					olsr_delete_neighbor_table(neighbor_tmp);
					// This flag is set to UP to recalculate the MPR set and the routing table
					changes_neighborhood=UP;
					changes_topology=UP;
				}
				else
				{
					neighbor->neighbor_status=NBS_LOST;
					neighbor=neighbor->neighbor_forw;
				}
			}
			else
			{
				olsr_time_out_two_hop_neighbors(neighbor);
				neighbor=neighbor->neighbor_forw;
			}
		}
	}
}*/
/*------------------------------------------------------------------------*/
/****************** added by Y.Ge ******************/
/* this function not only define if the link should be timed out, but define other link_changes as well */
void
olsr_time_out_neighborhood_tables()
{
	olsr_u8_t              index;
	struct neighbor_entry  *neighbor;
	struct neighbor_entry  *neighbor_tmp;
	struct neighborhash    *hash_neighbor;
	int                    i;
	int                    changed = 0;
	struct timeval        	tmp_timer, tmp_timeval; //added by Y.Ge   

  	for(index=0;index<HASHSIZE;index++)
	{
		hash_neighbor=&neighbortable.neighborhash[index];
		neighbor=hash_neighbor->neighbor_forw;
		while(neighbor!=(struct neighbor_entry *)hash_neighbor)
		{ 
			/*---------------*/
			/*if(debug_level > 2 && print_time_debug ==1)
			{
				printf ("\n\n in timeout neighbortables!\n");
				printf ("...NOW, looking at neighbor %s\n", convert_address_to_string(&neighbor->neighbor_addr));
				printf("neighbor_timer: %d   %d\n ",neighbor->neighbor_timer.tv_sec,neighbor->neighbor_timer.tv_usec);
				printf ("neighbor_link_sym_timer: %d   %d\n", neighbor->neighbor_link_sym_timer.tv_sec, neighbor->neighbor_link_sym_timer.tv_usec);
				printf ("neighbor_link_asym_timer: %d   %d\n", neighbor->neighbor_link_asym_timer.tv_sec, neighbor->neighbor_link_asym_timer.tv_usec);
				printf("now: %d   %d\n ",now.tv_sec,now.tv_usec);
			}*/
			/*--------------*/

			/************ added by Y.Ge *************/
			/* 
			 * hysteresis 
			 */
			if(use_hyst)
			{
			 	if(olsr_timed_out(&neighbor->next_hello_timer))
				{
					if (debug_level > 1)
					{
						printf ("\n\n............neighbor %s: in time out next hello timer!...............\n",convert_address_to_string(&neighbor->neighbor_addr));
						printf("next_hello_timer: %d   %d\n ",neighbor->next_hello_timer.tv_sec,neighbor->next_hello_timer.tv_usec);
						printf ("neighbor_link_lost_timer: %d   %d\n", neighbor->neighbor_link_lost_timer.tv_sec, neighbor->neighbor_link_lost_timer.tv_usec);
						printf ("neighbor_timer: %d   %d\n", neighbor->neighbor_timer.tv_sec, neighbor->neighbor_timer.tv_usec);
						printf("now: %d   %d\n ",now.tv_sec,now.tv_usec);
						printf ("originally, neighbor status: %d, neighbor hyst: %f\n",neighbor->neighbor_status, neighbor->neighbor_hysavg);
						
						fprintf (y_file,"\n\n............neighbor %s: in time out next hello timer!...............\n",convert_address_to_string(&neighbor->neighbor_addr));
						fprintf(y_file,"next_hello_timer: %d   %d\n ",neighbor->next_hello_timer.tv_sec,neighbor->next_hello_timer.tv_usec);
						fprintf (y_file,"neighbor_link_lost_timer: %d   %d\n", neighbor->neighbor_link_lost_timer.tv_sec, neighbor->neighbor_link_lost_timer.tv_usec);
						fprintf (y_file,"neighbor_timer: %d   %d\n", neighbor->neighbor_timer.tv_sec, neighbor->neighbor_timer.tv_usec);
						fprintf(y_file,"now: %d   %d\n ",now.tv_sec,now.tv_usec);
						fprintf (y_file,"originally, neighbor status: %d, neighbor hyst: %f\n",neighbor->neighbor_status, neighbor->neighbor_hysavg);
					}
					neighbor->neighbor_hysavg = hysteresis (neighbor->neighbor_hysavg, 0);
					
					/* update next hello timer */
					olsr_init_timer(neighbor->h_time_value*1000, &tmp_timer);
					timeradd(&now, &tmp_timer, &neighbor->next_hello_timer);
					
					if (debug_level > 1)
					{
						printf("after recalculation, neighbor hyst is: %f\n", neighbor->neighbor_hysavg);
						fprintf(y_file, "after recalculation, neighbor hyst is: %f\n", neighbor->neighbor_hysavg);
					}
					
					if (neighbor->neighbor_hysavg < HYST_THRESHOLD_LOW)
					{
						neighbor->link_pending = 1;
				
						if (neighbor->neighbor_status == SYM_LINK)
						{
							changes_neighborhood=UP;
							changes_topology=UP;
							changed = 1;
						}
						
						/* set the link lost timer */
						timeradd(&now, &hold_time_neighbor, &tmp_timeval);
						if (compare_timer(&neighbor->neighbor_timer, &tmp_timeval)==0)
							olsr_assign_timer(&neighbor->neighbor_timer, &neighbor->neighbor_link_lost_timer);
						else
							olsr_assign_timer(&tmp_timeval, &neighbor->neighbor_link_lost_timer);
							
						/* set the neighbor staatus */
						neighbor->neighbor_status = LOST_LINK;
						
						if (debug_level >1)
						{
							printf ("XXXXXXXXXXXX hyst low! change neighbor status to LOST link! XXXXXXXXX\n");
							printf (" now the link status is: %d\n", neighbor->neighbor_status);
							
							fprintf (y_file,"XXXXXXXXXXXX hyst low! change neighbor status to LOST link! XXXXXXXXX\n");
							fprintf (y_file," now the link status is: %d\n", neighbor->neighbor_status);			
						}
					}
					else
					{
						if (neighbor->neighbor_hysavg < HYST_THRESHOLD_HIGH)
						{
							if(neighbor->neighbor_status == SYM_LINK)
							{
								changes_neighborhood=UP;
								changes_topology=UP;
								changed = 1;
							}
							
							/* decide link status */
							if (!olsr_timed_out(&neighbor->neighbor_link_lost_timer))
							{
								neighbor->neighbor_status = LOST_LINK;								
							}
							else
							{
								neighbor->neighbor_status = ASYM_LINK;
							}
							
							if (debug_level >1)
							{
								printf ("XXXXXXXXXXXX hyst below high! change neighbor status! XXXXXXXXX\n");
								printf (" now the link status is: %d\n", neighbor->neighbor_status);
							
								fprintf (y_file,"XXXXXXXXXXXX hyst below high! change neighbor status! XXXXXXXXX\n");
								fprintf (y_file," now the link status is: %d\n", neighbor->neighbor_status);			
							}								
						}
					}
					
					if (debug_level > 1)
					{
						printf ("\n\n***** after next hello time out, the timers are \n");
						printf("next_hello_timer: %d   %d\n ",neighbor->next_hello_timer.tv_sec,neighbor->next_hello_timer.tv_usec);
						printf ("neighbor_link_lost_timer: %d   %d\n", neighbor->neighbor_link_lost_timer.tv_sec, neighbor->neighbor_link_lost_timer.tv_usec);
						printf ("neighbor_timer: %d   %d\n", neighbor->neighbor_timer.tv_sec, neighbor->neighbor_timer.tv_usec);
						printf("now: %d   %d\n ",now.tv_sec,now.tv_usec);
						printf ("new, neighbor status: %d, neighbor hyst: %f\n",neighbor->neighbor_status, neighbor->neighbor_hysavg);
						
						fprintf (y_file,"\n\n***** after next hello time out, the timers are\n");
						fprintf(y_file,"next_hello_timer: %d   %d\n ",neighbor->next_hello_timer.tv_sec,neighbor->next_hello_timer.tv_usec);
						fprintf (y_file,"neighbor_link_lost_timer: %d   %d\n", neighbor->neighbor_link_lost_timer.tv_sec, neighbor->neighbor_link_lost_timer.tv_usec);
						fprintf (y_file,"neighbor_timer: %d   %d\n", neighbor->neighbor_timer.tv_sec, neighbor->neighbor_timer.tv_usec);
						fprintf(y_file,"now: %d   %d\n ",now.tv_sec,now.tv_usec);
						fprintf (y_file,"now, neighbor status: %d, neighbor hyst: %f\n",neighbor->neighbor_status, neighbor->neighbor_hysavg);
					}
				}
			}
			/************* end of revision **********/
			/*
			 * the L_time expires,
			 * remove the link
			 */
			/*printf ("neighbor_timer: %d   %d\n", neighbor->neighbor_timer.tv_sec, neighbor->neighbor_timer.tv_usec);
			fprintf (y_file,"neighbor_timer: %d   %d\n", neighbor->neighbor_timer.tv_sec, neighbor->neighbor_timer.tv_usec);*/
			 
			if(olsr_timed_out(&neighbor->neighbor_timer))
			{
				/*-----------------*/
				if(debug_level > 6)
				{
					printf("\n\ndeleting neighbor entry %s\n", convert_address_to_string(&neighbor->neighbor_addr));
				}
				/*-----------------*/

				neighbor_tmp=neighbor;
				neighbor=neighbor->neighbor_forw;
				if (debug_level > 1)
				{
					printf ("\n\n............neighbor %s: in time out, delete neighbor table!...............\n",convert_address_to_string(&neighbor_tmp->neighbor_addr));
					printf("neighbor_timer: %d   %d\n ",neighbor_tmp->neighbor_timer.tv_sec,neighbor_tmp->neighbor_timer.tv_usec);
					printf ("neighbor_link_sym_timer: %d   %d\n", neighbor_tmp->neighbor_link_sym_timer.tv_sec, neighbor_tmp->neighbor_link_sym_timer.tv_usec);
					printf ("neighbor_link_asym_timer: %d   %d\n", neighbor_tmp->neighbor_link_asym_timer.tv_sec, neighbor_tmp->neighbor_link_asym_timer.tv_usec);
					printf("now: %d   %d\n ",now.tv_sec,now.tv_usec);
					
					fprintf (y_file,"\n\n............neighbor %s: in time out, delete neighbor table!...............\n",convert_address_to_string(&neighbor_tmp->neighbor_addr));
					fprintf(y_file,"neighbor_timer: %d   %d\n ",neighbor_tmp->neighbor_timer.tv_sec,neighbor_tmp->neighbor_timer.tv_usec);
					fprintf (y_file,"neighbor_link_sym_timer: %d   %d\n", neighbor_tmp->neighbor_link_sym_timer.tv_sec, neighbor_tmp->neighbor_link_sym_timer.tv_usec);
					fprintf (y_file,"neighbor_link_asym_timer: %d   %d\n", neighbor_tmp->neighbor_link_asym_timer.tv_sec, neighbor_tmp->neighbor_link_asym_timer.tv_usec);
					fprintf(y_file,"now: %d   %d\n ",now.tv_sec,now.tv_usec);
				}
				
				olsr_delete_neighbor_table(neighbor_tmp);
				
				/* This flag is set to UP to recalculate the MPR set and the routing table*/				
				changed = 1;				
			}
			else
			{
				/* check if links is asym */
				if (olsr_timed_out (&neighbor->neighbor_link_asym_timer))
				{
					if (debug_level>2)
					{
						
					}

					if (neighbor->neighbor_status != LOST_LINK)
					{
						if (debug_level>1)
						{
							printf ("\n\n.................neighbor %s -- Asym_timer expires!................\n", convert_address_to_string(&neighbor->neighbor_addr));
							printf ("neighbor_link_asym_timer: %d   %d\n", neighbor->neighbor_link_asym_timer.tv_sec, neighbor->neighbor_link_asym_timer.tv_usec);
							printf ("neighbor_link_sym_timer: %d   %d\n", neighbor->neighbor_link_sym_timer.tv_sec, neighbor->neighbor_link_sym_timer.tv_usec);
							printf("now: %d   %d\n ",now.tv_sec,now.tv_usec);		
							printf ("CHANGE LINK STATUS TO LOST_LINK!\n");
							
							fprintf (y_file,"\n\n.................neighbor %s -- Asym_timer expires!................\n", convert_address_to_string(&neighbor->neighbor_addr));
							fprintf (y_file,"neighbor_link_asym_timer: %d   %d\n", neighbor->neighbor_link_asym_timer.tv_sec, neighbor->neighbor_link_asym_timer.tv_usec);
							fprintf (y_file,"neighbor_link_sym_timer: %d   %d\n", neighbor->neighbor_link_sym_timer.tv_sec, neighbor->neighbor_link_sym_timer.tv_usec);
							fprintf(y_file,"now: %d   %d\n ",now.tv_sec,now.tv_usec);		
							fprintf (y_file,"CHANGE LINK STATUS TO LOST_LINK!\n");
						}
						neighbor->neighbor_status = LOST_LINK;
						changed = 1;
					}
				}
				else
				{					
					/* check if link is still sym */
					if (olsr_timed_out (&neighbor->neighbor_link_sym_timer))
					{
						
						if (neighbor->neighbor_status == SYM_LINK)
						{
							changed = 1;
							if (debug_level>1)
							{
								printf ("\n\n.............neighbor %s, sym_timer expires!................\n", convert_address_to_string(&neighbor->neighbor_addr));
								printf ("neighbor_link_asym_timer: %d   %d\n", neighbor->neighbor_link_asym_timer.tv_sec, neighbor->neighbor_link_asym_timer.tv_usec);
								printf ("neighbor_link_sym_timer: %d   %d\n", neighbor->neighbor_link_sym_timer.tv_sec, neighbor->neighbor_link_sym_timer.tv_usec);
								printf("now: %d   %d\n ",now.tv_sec,now.tv_usec);
								printf ("CHANGE LINK STATUS TO ASYM_LINK!\n");	
								
								fprintf (y_file,"\n\n.............neighbor %s, sym_timer expires!................\n", convert_address_to_string(&neighbor->neighbor_addr));
								fprintf (y_file,"neighbor_link_asym_timer: %d   %d\n", neighbor->neighbor_link_asym_timer.tv_sec, neighbor->neighbor_link_asym_timer.tv_usec);
								fprintf (y_file,"neighbor_link_sym_timer: %d   %d\n", neighbor->neighbor_link_sym_timer.tv_sec, neighbor->neighbor_link_sym_timer.tv_usec);
								fprintf(y_file,"now: %d   %d\n ",now.tv_sec,now.tv_usec);
								fprintf (y_file,"CHANGE LINK STATUS TO ASYM_LINK!\n");		
							}

						}
						
						if (neighbor->neighbor_status != LOST_LINK)
							neighbor->neighbor_status = ASYM_LINK;
					}
				}
				
				olsr_time_out_two_hop_neighbors(neighbor);
			}

			if (changed)
			{
				changes_neighborhood = 1;
				changes_topology = 1;
			}
			neighbor=neighbor->neighbor_forw;
		}
	}
	
	//also should update main_addr_neighbor_status based on the link status
	if (changed)
	{
		olsr_update_main_addr_neigh_table_status ();
		
		if (debug_level > 1)
		{
			printf ("XXXXXXXXXX after time out neighbor table, the neighbor table is:XXXXXXXXXXXX\n");
			olsr_print_neighbor_table();
			olsr_print_main_addr_neigh_table();
			
			fprintf (y_file,"XXXXXXXXXX after time out neighbor table, the neighbor table is:XXXXXXXXX\n");
			olsr_print_trace_neighbor_table();
			olsr_print_trace_main_addr_neigh_table(); 
		}
	}
}
/*------------------------------------------------------------------------*/
/************** commented by Y.Ge ************/
/*void
olsr_print_neighbor_table()
{
	olsr_u8_t                    index;
	struct neighbor_entry        *neighbor;
	struct neighborhash          *hash_neighbor;
	struct neighbor_2_list_entry *list_2;
	char str[46];

	printf("neighbor list: %02d:%02d:%02d.%06lu\n",nowtm->tm_hour, nowtm->tm_min, nowtm->tm_sec, now.tv_usec);
	for(index=0;index<HASHSIZE;index++)
	{
		hash_neighbor=&neighbortable.neighborhash[index];
		for(neighbor=hash_neighbor->neighbor_forw;neighbor!=(struct neighbor_entry *)hash_neighbor;neighbor=neighbor->neighbor_forw)
		{
			//printf("%s:%d:%1.2f[2hlist:",convert_address_to_string(&neighbor->neighbor_addr),neighbor->neighbor_status,neighbor->neighbor_hysavg);
			printf("%s;%d;%1.2f[2hlist;",inet_ntop(AF_INET6,(&neighbor->neighbor_addr),str,46),neighbor->neighbor_status,neighbor->neighbor_hysavg);
			list_2=neighbor->neighbor_2_list;
			while(list_2!=NULL)
			{
				//printf("%s:",convert_address_to_string(&list_2->neighbor_2->neighbor_2_addr));
				printf("%s;",inet_ntop(AF_INET6,&list_2->neighbor_2->neighbor_2_addr,str,46));
				list_2=list_2->neighbor_2_next;
			}
			printf("]\n");
		}
	}
}*/
/*------------------------------------------------------------------------*/
/************** added by Y.Ge ************/
void
olsr_print_neighbor_table()
{
	olsr_u8_t                    index;
	struct neighbor_entry        *neighbor;
	struct neighborhash          *hash_neighbor;
	struct neighbor_2_list_entry *list_2;
	char str[46];
	int  i;

	printf("\n\nneighbor list: %02d:%02d:%02d.%06lu\n",nowtm->tm_hour, nowtm->tm_min, nowtm->tm_sec, now.tv_usec);
	for(index=0;index<HASHSIZE;index++)
	{
		hash_neighbor=&neighbortable.neighborhash[index];
		for(neighbor=hash_neighbor->neighbor_forw;neighbor!=(struct neighbor_entry *)hash_neighbor;neighbor=neighbor->neighbor_forw)
		{
			//printf("%s:%d:%1.2f[2hlist:",convert_address_to_string(&neighbor->neighbor_addr),neighbor->neighbor_status,neighbor->neighbor_hysavg);
			printf ("---from my iface address (%d): %s\n",neighbor->my_iface_index, inet_ntop(AF_INET6,(&neighbor->my_iface_addr),str,46)); 
			printf("   neighbor:%s;%d;%1.2f[2hlist;",inet_ntop(AF_INET6,(&neighbor->neighbor_addr),str,46),neighbor->neighbor_status,neighbor->neighbor_hysavg);
			list_2=neighbor->neighbor_2_list;
			while(list_2!=NULL)
			{
				//printf("%s:",convert_address_to_string(&list_2->neighbor_2->neighbor_2_addr));
				printf("%s;",inet_ntop(AF_INET6,&list_2->neighbor_2->neighbor_2_addr,str,46));
				list_2=list_2->neighbor_2_next;
			}
			printf("]\n");
			
			/*** hyst ***/
			if (use_hyst)
			{
				printf("***** link pending: %d\n", neighbor->link_pending);
			}
			
			/* timers */
			printf ("       neighbor timers:\n");
			printf ("       neighbor_timer: %d   %d\n ",neighbor->neighbor_timer.tv_sec,neighbor->neighbor_timer.tv_usec);
			printf ("       neighbor_link_sym_timer: %d   %d\n", neighbor->neighbor_link_sym_timer.tv_sec, neighbor->neighbor_link_sym_timer.tv_usec);
			printf ("       neighbor_link_asym_timer: %d   %d\n", neighbor->neighbor_link_asym_timer.tv_sec, neighbor->neighbor_link_asym_timer.tv_usec);
			if (use_hyst)
			{
				printf ("       neighbor_link_lost_timer: %d   %d\n", neighbor->neighbor_link_lost_timer.tv_sec, neighbor->neighbor_link_lost_timer.tv_usec);
				printf ("       neighbor_next_hello_timer: %d  %d\n", neighbor->next_hello_timer.tv_sec, neighbor->next_hello_timer.tv_usec);
			}
			printf("        now: %d   %d\n ",now.tv_sec,now.tv_usec);
			
			/* qos OLSR */
			if (use_qos)
			{
				printf ("++++++ neighbor's busy time: %d\n", neighbor->neighbor_busy_time);
			}
			
			printf ("   --> this neighbor's main address info-address:%s, willingness: %d, status: %d\n", 
						inet_ntop(AF_INET6, (&neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_addr),str,46), 
						neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_willingness, 
						neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_status);
		}
	}
}
/*------------------------------------------------------------------------*/
void
olsr_print_trace_neighbor_table()
{
	olsr_u8_t                    index;
	struct neighbor_entry        *neighbor;
	struct neighborhash          *hash_neighbor;
	struct neighbor_2_list_entry *list_2;
	char str[46];
	int  i;

	fprintf(y_file,"\n\nneighbor list: %02d:%02d:%02d.%06lu\n",nowtm->tm_hour, nowtm->tm_min, nowtm->tm_sec, now.tv_usec);
	for(index=0;index<HASHSIZE;index++)
	{
		hash_neighbor=&neighbortable.neighborhash[index];
		for(neighbor=hash_neighbor->neighbor_forw;neighbor!=(struct neighbor_entry *)hash_neighbor;neighbor=neighbor->neighbor_forw)
		{
			//printf("%s:%d:%1.2f[2hlist:",convert_address_to_string(&neighbor->neighbor_addr),neighbor->neighbor_status,neighbor->neighbor_hysavg);
			fprintf (y_file,"---from my iface address (%d): %s\n",neighbor->my_iface_index, inet_ntop(AF_INET6,(&neighbor->my_iface_addr),str,46)); 
			fprintf(y_file,"   neighbor:%s;%d;%1.2f[2hlist;",inet_ntop(AF_INET6,(&neighbor->neighbor_addr),str,46),neighbor->neighbor_status,neighbor->neighbor_hysavg);
			list_2=neighbor->neighbor_2_list;
			while(list_2!=NULL)
			{
				//printf("%s:",convert_address_to_string(&list_2->neighbor_2->neighbor_2_addr));
				fprintf(y_file,"%s;",inet_ntop(AF_INET6,&list_2->neighbor_2->neighbor_2_addr,str,46));
				list_2=list_2->neighbor_2_next;
			}
			fprintf(y_file,"]\n");
			
			/*** hyst ***/
			if (use_hyst)
			{
				fprintf(y_file,"***** link pending: %d\n", neighbor->link_pending);
			}
			
			/*** timer ***/
			fprintf (y_file,"       neighbor timers:\n");
			fprintf (y_file,"       neighbor_timer: %d   %d\n ",neighbor->neighbor_timer.tv_sec,neighbor->neighbor_timer.tv_usec);
			fprintf (y_file,"       neighbor_link_sym_timer: %d   %d\n", neighbor->neighbor_link_sym_timer.tv_sec, neighbor->neighbor_link_sym_timer.tv_usec);
			fprintf (y_file,"       neighbor_link_asym_timer: %d   %d\n", neighbor->neighbor_link_asym_timer.tv_sec, neighbor->neighbor_link_asym_timer.tv_usec);
			if (use_hyst)
			{
				fprintf (y_file,"       neighbor_link_lost_timer: %d   %d\n", neighbor->neighbor_link_lost_timer.tv_sec, neighbor->neighbor_link_lost_timer.tv_usec);
				fprintf (y_file,"       neighbor_next_hello_timer: %d  %d\n", neighbor->next_hello_timer.tv_sec, neighbor->next_hello_timer.tv_usec);
			}
			fprintf(y_file,"        now: %d   %d\n ",now.tv_sec,now.tv_usec);
			
			/* qos OLSR */
			if (use_qos)
			{
				fprintf (y_file,"++++++ neighbor's busy time: %d\n", neighbor->neighbor_busy_time);
			}
			
			fprintf (y_file,"   --> this neighbor's main address info-address:%s, willingness: %d, status: %d\n", 
						inet_ntop(AF_INET6, (&neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_addr),str,46), 
						neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_willingness, 
						neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_status);
		}
	}
}
/*------------------------------------------------------------------------*/
struct neighbor_entry
*olsr_find_sym_neighbor(struct main_addr_neigh_entry *main_neighbor)
{
	struct iface_addr_list_entry	*list_of_iface;
	
	list_of_iface = main_neighbor->neigh_iface_list;
	
	while(list_of_iface != NULL)
	{
		if (list_of_iface->neighbor_iface->neighbor_status == SYM_LINK)
			return (list_of_iface->neighbor_iface);	
		list_of_iface = list_of_iface->iface_addr_next;	
	}
	
	return NULL;
	
}
/*------------------------------------------------------------------------*/
struct neighbor_entry
*olsr_find_highest_bw_neighbor(struct main_addr_neigh_entry *main_neighbor)
{
	struct iface_addr_list_entry	*list_of_iface;
	struct neighbor_entry       	*highest_bw_neighbor = NULL;
	int    current_busy_time = -1;
	
	list_of_iface = main_neighbor->neigh_iface_list;
	
	while(list_of_iface != NULL)
	{
		if (list_of_iface->neighbor_iface->neighbor_status == SYM_LINK && list_of_iface->neighbor_iface->neighbor_busy_time < current_busy_time)
		{
			current_busy_time = list_of_iface->neighbor_iface->neighbor_busy_time;
			highest_bw_neighbor = list_of_iface->neighbor_iface;
		}
		list_of_iface = list_of_iface->iface_addr_next;	
	}
	
	return (highest_bw_neighbor);
}
/********************************** end of revision *****************/











