/*
 * Copyright (c) 2004 Ying Ge, Communication Research Center Canada.
 *
 * Copyright (c) 2002, 2003 Maoyu Wang, Communication Research Center Canada.
 *
 * By Ying Ge:
 * 1. Change the OLSR packet format and message processing procedure based 
 *    on the OLSR RFC.
 * 2. Add support of multiple interfaces to OLSR, including MID message 
 *    creating and processing procedure as specified in the OLSR RFC.
 * 3. Add QoS Support to OLSR
 *
 * By Maoyu Wang:
 * 1. Ported OLSR from IPv4 to IPv6.
 * 2. Added the Host and Network Association (HNA) functionality into OLSR.
 * 3. Added the default gateway functionality into OLSR by extending the HNA 
 *    message usage. The default gateway functionality supported the mobility
 *	  by cooperating with Mobile IPv6 for a mobile node as well as supported 
 * 	  Internet access for MANET nodes. 
 *
 * DISTRIBUTED WITH NO WARRANTY, EXPRESS OR IMPLIED.
 * See the GNU Library General Public License (file COPYING in the distribution)
 * for conditions of use and redistribution
 */
/* Code Modified by Joe Macker and Ron Lee July (NRL) 2001 
 * the following was added
 * (1) hysteresis code for hellos
 *   - added pending state and filter parameters
 * (2) parameterization of other setting
 * - hello intervals,
 * - hold timeouts based upon interval settings
 * (3) added jitter functionality
 * (4) still don't like the polling routing used as an event handler
 * but deal with it by adding -T option which allows polling time to 
 * be decreased to have more effective jitter processing
 * (5) added debug levels and some fixed some compiler warnings
 * (6) added functions for default gateway settings
 * (7) added some function for adding associated subnets and addresses
 * to routing table of OLSR node (see README)
 * ces tres bon... bon chance
 */
 
/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est-à-dire qu'il peut
 * être utilisé, copié, modifié, distribué à l'unique
 * condition que ce texte soit conservé afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant participé à l'élaboration de ce logiciel ne peut
 * être utilisé sans son accord préalable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilité d'aucune sorte.
 * Certaines parties de ce logiciel sont dérivées de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 */

/*
 * Routing table management daemon.
 */

char olsr_rcsid[] = 
  "$Id: olsr.c,v 1.2 2000/12/06 10:36:11 prima Exp $";


#include "defs.h"
#include "math.h"

/* --------------------------------------------------------------------
    OLSR protocol
   -------------------------------------------------------------------- */

/* --------------------------------------------------------------------
    Forward declarations
   -------------------------------------------------------------------- */


/* --------------------------------------------------------------------
    Global variables and arrays
   -------------------------------------------------------------------- */


/* --------------------------------------------------------------------
    Functions
   -------------------------------------------------------------------- */


/*-------------------------------------------------------------------------*/

struct timeval  hold_time_neighbor ;
struct timeval  hold_time_topology;
struct timeval  hold_time_duplicate_tc;
struct timeval  hold_time_expired; //added by Y.Ge
struct timeval  hold_time_duplicate; //added by Y.Ge

//CRC HNA add
/*-------------------------------------------------------------------------*/
struct timeval  hold_time_hna;
struct timeval	hold_time_duplicate_hna;
/*-------------------------------------------------------------------------*/

//CRC HNA add
/*-------------------------------------------------------------------------*/
olsr_8_t changes_net;
/*-------------------------------------------------------------------------*/

olsr_8_t changes_topology;
olsr_8_t changes_neighborhood;

//CRC HNA add
olsr_8_t changes_gateway; 

FILE *f_r_id;
/* hysteresis trackig now included in debug 1 level neighbor message */
/*FILE *hys_id;*/

/* CRC v4->v6 add
 * copy from /usr/src/linux/include/net/addrconf.h:
 * ipv6_addr_hash(struct in6_addr) 
 */
/*-------------------------------------------------------------------------*/
void
olsr_hashing(struct olsr_ip_addr *address,olsr_u32_t *h)
{
	olsr_u32_t hash;
	hash = address->s6_addr[2]^address->s6_addr32[3];
	hash ^=(hash>>16);
	hash ^=(hash>>8);
	*h=((hash^(hash>>4))&0xff);
}
/*-------------------------------------------------------------------------*/

struct neighbor_2_list_entry 
*olsr_find_2_hop_neighbors_with_1_link()
{
	olsr_u8_t                    index;
	struct neighbor_2_list_entry *two_hop_list_tmp=NULL;
	struct neighbor_2_list_entry *two_hop_list=NULL;
	struct neighbor_2_entry      *two_hop_neighbor=NULL;
	struct neighbor2_hash        *hash_2_neighbor;

	for(index=0;index<HASHSIZE;index++)
	{
		hash_2_neighbor=&neighbor2table[index];

		for(two_hop_neighbor=hash_2_neighbor->neighbor2_forw;two_hop_neighbor!=(struct neighbor_2_entry  *)hash_2_neighbor;two_hop_neighbor=two_hop_neighbor->neighbor_2_forw)
		{
			two_hop_neighbor->neighbor_2_state=0;
			if(two_hop_neighbor->neighbor_2_pointer==1)
			{
				//if(two_hop_neighbor->neighbor_2_nblist->neighbor->neighbor_status==NBS_SYM)//commented by Y.Ge
				if(two_hop_neighbor->neighbor_2_nblist->neighbor->neighbor_status==SYM_LINK)//added by Y.Ge
				{
					two_hop_list_tmp =(struct neighbor_2_list_entry *)malloc(sizeof(struct neighbor_2_list_entry));
					two_hop_list_tmp->neighbor_2=two_hop_neighbor;
					two_hop_list_tmp->neighbor_2_next=two_hop_list;
					two_hop_list= two_hop_list_tmp;
				}
			}
		}      
	}
  	return(two_hop_list_tmp);
}
  
/*---------------------------------------------------------------------------*/
/************** added by Y.Ge ***************/
/* qos OLSR */
struct neighbor_entry 
*olsr_find_maximum_covered_with_qos(int willingness) //"willingness" is added by Y.Ge, define which willingness of the neighbor we should look in this round
{
	olsr_u16_t                  maximum=0;
	olsr_u8_t                   index;
	struct neighborhash         *neighbor_hash_table;
	struct neighbor_entry       *a_neighbor;
	struct neighbor_entry       *mpr_candidate=NULL;
	olsr_u32_t           	    min_busy_time = -1; /* set the busy time as the maximum posssible value */
   
	neighbor_hash_table=neighbortable.neighborhash;
	for (index=0;index<HASHSIZE;index++)
	{
		for(a_neighbor=neighbor_hash_table[index].neighbor_forw;a_neighbor!=(struct neighbor_entry  *)&neighbor_hash_table[index];a_neighbor=a_neighbor->neighbor_forw)
		{
			/* if(maximum<a_neighbor->neighbor_2_nocov) */ //commented by Y.Ge
			
			/* added by Y.Ge
			 * choose the certain willingness neighbor
			 * as MPR
			 */
			if(a_neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_willingness == willingness)
			{
				if(maximum<a_neighbor->neighbor_2_nocov)
				{
					maximum=a_neighbor->neighbor_2_nocov;
					mpr_candidate=a_neighbor;
					min_busy_time = a_neighbor->neighbor_busy_time;
				}
				else
				{
					if (maximum == a_neighbor->neighbor_2_nocov)
					{
						if (min_busy_time > a_neighbor->neighbor_busy_time)
						{
							mpr_candidate = a_neighbor;
							min_busy_time = a_neighbor->neighbor_busy_time;
						}
					}
				}
			}
		}
	}
	return (mpr_candidate);
}

/*---------------------------------------------------------------------------*/
struct iface_addr_list_entry
*olsr_find_will_always_neighbor()
{
	olsr_u8_t               	index;
	struct iface_addr_list_entry	*will_always_neighbor_list = NULL;
	struct iface_addr_list_entry	*will_always_neighbor_list_tmp = NULL;
	struct neighbor_entry       	*neighbor = NULL;
	struct neighborhash          	*hash_neighbor;
	
	for(index=0;index<HASHSIZE;index++)
	{
		hash_neighbor=&neighbortable.neighborhash[index];
		for(neighbor=hash_neighbor->neighbor_forw;neighbor!=(struct neighbor_entry *)hash_neighbor;neighbor=neighbor->neighbor_forw)
		{

			if (neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_willingness == WILL_ALWAYS) 
			{
				if (neighbor->neighbor_status == SYM_LINK)
				{
					will_always_neighbor_list_tmp = (struct iface_addr_list_entry *)malloc(sizeof(struct iface_addr_list_entry));
					will_always_neighbor_list_tmp->neighbor_iface = neighbor;
					will_always_neighbor_list_tmp->iface_addr_next = will_always_neighbor_list;
					will_always_neighbor_list = will_always_neighbor_list_tmp;
				}
			}
		}
	}	
	return (will_always_neighbor_list_tmp); 
}
/************** end of revision *************/
/*---------------------------------------------------------------------------*/

int
olsr_chosen_mpr(struct neighbor_entry *one_hop_neighbor,olsr_u16_t *two_hop_covered_count)
{
  /*
    This function  processes the chosen  and updates the counters
    used in calculations  
  */
  
	struct neighbor_list_entry   *the_one_hop_list;
  	struct neighbor_2_list_entry *second_hop_entries; 
	olsr_u16_t                   count=0;
  
	count=*two_hop_covered_count;
	//one_hop_neighbor->neighbor_status=NBS_MPR; //commented  by Y.Ge
	one_hop_neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_status = MPR_NEIGH; //added by Y.Ge

	TRACE_ACTION("CHANGE NEIGHBOR STATUS : MPR ",&one_hop_neighbor->neighbor_addr,NULL,0);
	second_hop_entries=one_hop_neighbor->neighbor_2_list;
	while(second_hop_entries!=NULL)
	{
		if(!second_hop_entries->neighbor_2->neighbor_2_state)
		{
			/*
			  Now the neighbor is covered by this mpr
			 */
			second_hop_entries->neighbor_2->neighbor_2_state=NB2S_COVERED;  /* =1*/	  
			the_one_hop_list=second_hop_entries->neighbor_2->neighbor_2_nblist;
			while(the_one_hop_list!=NULL)
			{
				//if((the_one_hop_list->neighbor->neighbor_status!=NBS_ASYM)&&(the_one_hop_list->neighbor->neighbor_status!=NBS_LOST)&&(the_one_hop_list->neighbor->neighbor_status!=NBS_PENDING))//commented by Y.Ge
				if(the_one_hop_list->neighbor->neighbor_status == SYM_LINK)//added by Y.Ge
				{
					count++;
					the_one_hop_list->neighbor->neighbor_2_nocov--;
				}
				the_one_hop_list=the_one_hop_list->neighbor_next;
			}
		}
		second_hop_entries=second_hop_entries->neighbor_2_next;
	}
	*two_hop_covered_count=count;
	return (0);
}
/*--------------------------------------------------------------------------*/
void
olsr_calculate_neighbors(olsr_u16_t *two_hop_count)
{
  /*
    This function calculates the number of  second hop neighbors for each 
    neighbor
  */

	olsr_u8_t                    index;
	struct neighborhash          *neighbor_hash_table;
	struct neighbor_entry        *a_neighbor;
	olsr_u16_t                   count=0;
	struct neighbor_2_list_entry *twohop_neighbors;
  
	neighbor_hash_table=neighbortable.neighborhash;
	for (index=0;index<HASHSIZE;index++)
	{
		for(a_neighbor=neighbor_hash_table[index].neighbor_forw;a_neighbor!=(struct neighbor_entry  *)&neighbor_hash_table[index];a_neighbor=a_neighbor->neighbor_forw)
		{ 
			count=0;
			//if((a_neighbor->neighbor_status==NBS_ASYM)||(a_neighbor->neighbor_status==NBS_LOST)||(a_neighbor->neighbor_status==NBS_PENDING)) //commented by Y.Ge
			if((a_neighbor->neighbor_status==ASYM_LINK)||(a_neighbor->neighbor_status==LOST_LINK)||(a_neighbor->neighbor_status==UNSPEC_LINK)) //addeded by Y.Ge
			{	    
				a_neighbor->neighbor_2_nocov=count;
				continue;
			}
			
			/************ added by Y.Ge ***********/
			/*
			 * don't calculate the two hop neighbors that who's 1-hop-neighbor is WILL_NEVER
			 */
			if (a_neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_status == WILL_NEVER)
			{
				a_neighbor->neighbor_2_nocov=count;
				continue;
			}/*********** end of revision ********/
				
			/* 
			Turn the neighbor status on symetric to re calculate the MPR set
			*/
			/*if(a_neighbor->neighbor_status==NBS_MPR)
				a_neighbor->neighbor_status=NBS_SYM;*/ //comented by Y.Ge
			count=0;
			twohop_neighbors=a_neighbor->neighbor_2_list;
			while(twohop_neighbors!=NULL)
			{
				count++;
				twohop_neighbors=twohop_neighbors->neighbor_2_next;
			}
			a_neighbor->neighbor_2_nocov=count;
			*two_hop_count= *two_hop_count+count;
		} 
	}
}
/*--------------------------------------------------------------------------*/
struct neighbor_entry 
*olsr_find_maximum_covered(int willingness) //"willingness" is added by Y.Ge, define which willingness of the neighbor we should look in this round
{
	olsr_u16_t                  maximum=0;
	olsr_u8_t                   index;
	struct neighborhash         *neighbor_hash_table;
	struct neighbor_entry       *a_neighbor;
	struct neighbor_entry       *mpr_candidate=NULL;
   
	neighbor_hash_table=neighbortable.neighborhash;
	for (index=0;index<HASHSIZE;index++)
	{
		for(a_neighbor=neighbor_hash_table[index].neighbor_forw;a_neighbor!=(struct neighbor_entry  *)&neighbor_hash_table[index];a_neighbor=a_neighbor->neighbor_forw)
		{
			/* if(maximum<a_neighbor->neighbor_2_nocov) */ //commented by Y.Ge
			
			/* added by Y.Ge
			 * choose the certain willingness neighbor
			 * as MPR
			 */
			if((maximum<a_neighbor->neighbor_2_nocov) && (a_neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_willingness == willingness))
			{
				maximum=a_neighbor->neighbor_2_nocov;
				mpr_candidate=a_neighbor;
			}
		}
	}
	return (mpr_candidate);
}


/*----------------------------------------------------------------------------*/
void
olsr_calculate_mpr()
     
{
	/*This function calculates the mpr neighbor*/
	olsr_u16_t                   two_hop_covered_count=0;
	olsr_u16_t                   two_hop_count=0;  
	struct neighbor_2_list_entry *two_hop_list=NULL;
	struct neighbor_2_list_entry *tmp;
	struct neighbor_entry        *mprs;
	struct iface_addr_list_entry *will_always_neighbor=NULL; //added by Y.Ge 
	struct iface_addr_list_entry *will_always_tmp; //added by Y.Ge
	int	i, should_try_next=0; //added by Y.Ge

	/*************** added by Y.Ge *************/	
	/* first, re-set the status in the main_addr_neigh_table*/
	olsr_clear_main_addr_neigh_table_mpr_status();	
	/*************** end of revision ***********/

	olsr_calculate_neighbors(&two_hop_count);
	
	two_hop_list=olsr_find_2_hop_neighbors_with_1_link();
	
	/*************** added by Y.Ge *************/	
	/* select neighbor who's wilingness is WILL_ALWAYS as MPR */
	/* 
	 * the function is added after "olsr_find_2_hop_neighbors_with_1_link()"
	 * because the neighbor_2_covered status is initialized in
	 * olsr_find_2_hop_neighbors_with_1_link()
	 */
	will_always_neighbor = olsr_find_will_always_neighbor();
	
	if (debug_level >3)
	{
		printf ("\n\n-------------- Node is selecting MPRs! ----------\n");
		printf ("my two hop neighbor count is: %d\n", two_hop_count);
		printf ("I have following neighbor as my WILL_ALWAYS neighbor:\n");
		
		fprintf (y_file,"\n\n-------------- Node is selecting MPRs! ----------\n");
		fprintf (y_file,"my two hop neighbor count is: %d\n", two_hop_count);
		fprintf (y_file,"I have following neighbor as my WILL_ALWAYS neighbor:\n");
	}
		
	while(will_always_neighbor != NULL)
	{
		if (debug_level >3)
		{
			printf ("WILL_ALWAYS neighbor: %s\n", convert_address_to_string(&will_always_neighbor->neighbor_iface->neighbor_addr));
			
			fprintf (y_file,"WILL_ALWAYS neighbor: %s\n", convert_address_to_string(&will_always_neighbor->neighbor_iface->neighbor_addr));
		}
		olsr_chosen_mpr(will_always_neighbor->neighbor_iface, &two_hop_covered_count);
		will_always_tmp = will_always_neighbor;
		will_always_neighbor = will_always_neighbor->iface_addr_next;
		free (will_always_tmp);
	}	
	/*************** end of revission *************/	

	/*--------------------*/
	if(debug_level > 3)
	{
		printf(" AFTER choose WILL_ALWAYS neighbor as mpr, Two hop count: %d\n",two_hop_count);
		printf("\nnow looking at 2-hop-neighbors that only have one link. these nodes are:\n");
		
		fprintf(y_file," AFTER choose WILL_ALWAYS neighbor as mpr, Two hop count: %d\n",two_hop_count);
		fprintf(y_file,"\nnow looking at 2-hop-neighbors that only have one link. these nodes are:\n");
	}
	/*--------------------*/

	while(two_hop_list!=NULL)
	{
		if (debug_level > 3)
		{
			printf ("1-LINK-2-HOP-NEIGHBOR:%s\n",convert_address_to_string(&two_hop_list->neighbor_2->neighbor_2_nblist->neighbor->neighbor_addr));
			
			fprintf (y_file,"1-LINK-2-HOP-NEIGHBOR:%s\n",convert_address_to_string(&two_hop_list->neighbor_2->neighbor_2_nblist->neighbor->neighbor_addr));
		}
		olsr_chosen_mpr(two_hop_list->neighbor_2->neighbor_2_nblist->neighbor,&two_hop_covered_count); 
		tmp=two_hop_list;
		two_hop_list=two_hop_list->neighbor_2_next;;
		free(tmp);
	}
	/*---------------------*/
	if(debug_level > 3)
	{
		printf(" AFTER select 1-link-2-hop-neighbor, Two hops covered: %d\n",two_hop_covered_count);
		printf(" \nnow looking at remaining 2-hop-neighbors\n");
		
		fprintf(y_file," AFTER select 1-link-2-hop-neighbor, Two hops covered: %d\n",two_hop_covered_count);
		fprintf(y_file," \nnow looking at remaining 2-hop-neighbors\n");
	}
	/*---------------------*/

	/*while(two_hop_covered_count<two_hop_count)
	{
		mprs=olsr_find_maximum_covered();
		if (mprs!=NULL)
		{
			olsr_chosen_mpr(mprs,&two_hop_covered_count);
		}
	}*/ //commented by Y.Ge, see the following
	
	/************** added by Y.Ge *************/
	/*
	 * do based on the OLSR RFC - choose the one that has the highest willingness as MPR
	 */
	for (i=WILL_HIGH; i>WILL_NEVER; i--)
	{
		if (debug_level > 3)
		{
			printf ("...looking at willingness %d in this round\n", i);		
			fprintf (y_file,"...looking at willingness %d in this round\n", i);
		}
		
		should_try_next=0;
		while(two_hop_covered_count<two_hop_count && (!should_try_next))
		{
			/* qos OLSR */
			if (use_qos)
				mprs = olsr_find_maximum_covered_with_qos(i); //i is the willingness that we should look in this round
			else
				mprs = olsr_find_maximum_covered(i); //i is the willingness that we should look in this round
			if (mprs!=NULL)
			{
				if (debug_level > 3)
				{
					printf ("MPR:%s\n", convert_address_to_string(&mprs->neighbor_addr));
					fprintf (y_file,"MPR:%s\n", convert_address_to_string(&mprs->neighbor_addr));
				}
				olsr_chosen_mpr(mprs,&two_hop_covered_count);
			}
			else
			{
				should_try_next=1;
			}
		
		}
	}
	/*************** end of revision ***********/
	
	/*
	 increment the mpr sequence number
	*/
	neighbortable.neighbor_mpr_seq++;
}
/*-----------------------------------------------------------------------------*/
/************** added by Y.Ge ************/
/*** compute vtime ***/
olsr_u8_t
olsr_compute_vtime (int input_time)
{
	olsr_u8_t	vtime_value;
	int      	a, b;
   
	//printf ("/************** in olsr_compute_vtime *************/\n");
	//printf ("		    the input time is: %d\n", input_time);
	b = (int)((log((double)input_time * SCALING_FACTOR))/(log((double)2)));
	//printf ("b = %d\n", b);
	a = (int)(16*((double)input_time/((double)pow(2,b)/SCALING_FACTOR)-1));
	//printf ("a = %d\n", a);
   
	if (a==16)
	{
		a = 0;
		b += 1;
	}
   
	vtime_value = (olsr_u8_t) a*16 + b;
   
	//printf ("		   a = %d, b=%d\n", a,b);
	//printf ("               vtime_value is: %d\n", vtime_value);
	//printf ("\n");
	return  vtime_value; 
}	
/************* end of revision **********/
/*-----------------------------------------------------------------------------*/
float
olsr_compute_time_from_vtime (olsr_u8_t vtime_value)
{
	int	result_time;
	int	a,b;

	a = (int)vtime_value/16;
	b = vtime_value - a*16;

	result_time = ((1+(double)a/16)*pow(2,b))/SCALING_FACTOR;

	return result_time;
}
/*-----------------------------------------------------------------------------*/
int
olsr_timed_out(struct timeval *timer)
{
 	return( timercmp(timer,&now,<));
}
/*-----------------------------------------------------------------------------*/
void
olsr_init_timer(olsr_u32_t time_value,struct timeval *hold_timer)
{ 
	olsr_u16_t  time_value_sec=0;
	olsr_u16_t  time_value_msec=0;

	time_value_sec=time_value/1000;
	time_value_msec=time_value-(time_value_sec*1000);

	hold_timer->tv_sec=time_value_sec;
	hold_timer->tv_usec=time_value_msec*1000; 
  
}
/*-----------------------------------------------------------------------------*/
/************* added by Y.Ge ************/
void
olsr_assign_timer(const struct timeval *given_timer, struct timeval *change_timer)
{
	change_timer->tv_sec = given_timer->tv_sec;
	change_timer->tv_usec = given_timer->tv_usec;
}
/*-----------------------------------------------------------------------------*/
void
timerminus (struct timeval *change_timer)
{
	change_timer->tv_sec = -1;
}
/*-----------------------------------------------------------------------------*/
olsr_u8_t
compare_timer(const struct timeval *timer_a, const struct timeval *timer_b)
{
	if (timer_a->tv_sec > timer_b->tv_sec)
		return 1;
	else
	{
		if (timer_a->tv_sec == timer_b->tv_sec)
		{
			if (timer_a->tv_usec > timer_b->tv_usec)
				return 1;
		}
		return 0;
	}
	return 0;
}
/************* end of revision **********/
/*-------------------------------------------------------------------------*/
void
olsr_linking_this_2_entries(struct neighbor_entry *neighbor,struct neighbor_2_entry *two_hop_neighbor)
{
	struct neighbor_list_entry    *list_of_1_neighbors;
	struct neighbor_2_list_entry  *list_of_2_neighbors;

	list_of_1_neighbors=(struct neighbor_list_entry *)malloc(sizeof(struct neighbor_list_entry));
	list_of_2_neighbors=(struct neighbor_2_list_entry *)malloc(sizeof(struct neighbor_2_list_entry));

	list_of_1_neighbors->neighbor=neighbor;
	list_of_1_neighbors->neighbor_next=two_hop_neighbor->neighbor_2_nblist;
	two_hop_neighbor->neighbor_2_nblist=list_of_1_neighbors;

	list_of_2_neighbors->neighbor_2=two_hop_neighbor;
	timeradd(&now,&hold_time_neighbor,&list_of_2_neighbors->neighbor_2_timer);

	list_of_2_neighbors->neighbor_2_next=neighbor->neighbor_2_list;
	neighbor->neighbor_2_list=list_of_2_neighbors;
  
	/*increment the pointer counter*/
	two_hop_neighbor->neighbor_2_pointer++;
}
/*-------------------------------------------------------------------------*/
/***************** added by Y.Ge **************/
void
olsr_linking_neighbor_main_neighbor(struct neighbor_entry *neighbor,struct main_addr_neigh_entry *main_address_neighbor)
{
	struct iface_addr_list_entry	*list_of_iface_neighbors;
	struct link_to_main_neighbor	*main_neighbor_link;

	list_of_iface_neighbors=(struct iface_addr_list_entry *)malloc(sizeof(struct iface_addr_list_entry));
	main_neighbor_link=(struct link_to_main_neighbor *)malloc(sizeof(struct link_to_main_neighbor));

	list_of_iface_neighbors->neighbor_iface=neighbor;
	list_of_iface_neighbors->iface_addr_next=main_address_neighbor->neigh_iface_list;
	main_address_neighbor->neigh_iface_list=list_of_iface_neighbors;

	main_neighbor_link->main_neighbor=main_address_neighbor;
	neighbor->neighbor_main_info=main_neighbor_link;

	main_address_neighbor->interface_pointer++;	
}
/*-------------------------------------------------------------------------*/
int
olsr_lookup_my_own_address (struct olsr_ip_addr *rcvd_interface_addr)
{
	int i;
	
	for (i=0; i<num_of_interface; i++)
	{
		if (0==memcmp(rcvd_interface_addr,&addr_manet[i],sizeof(struct olsr_ip_addr)))
			return 1;
	}
	return 0;
}
/*************** end of revision **************/
/*-------------------------------------------------------------------------*/
/*handle the neighbor linked list in the HELLO message*/
void
olsr_process_message_neighbors(struct neighbor_entry *neighbor,struct hello_message *message, int iface_index) //"iface_index" added by Y.Ge
{
	olsr_u8_t                  	ok_insert=0;// boolean
	struct hello_neighbor      	*message_neighbors;
	struct neighbor_entry      	*in_neighbor;
	struct neighbor_2_list_entry	*two_hop_neighbor_yet;
	struct neighbor_2_entry     	*two_hop_neighbor;
	float                       	rcvd_vtime;  //added by Y.Ge
	struct timeval              	hold_time_tmp; //added by Y.Ge
	struct iface_addr_entry     	*message_neighbor_interface_address; //added by Y.Ge
	struct main_addr_entry      	*message_neighbor_main_address; //added by Y.Ge
	struct olsr_ip_addr             *rcvd_main_address; //added by Y.Ge
	struct main_addr_neigh_entry	*in_main_addr_neighbor; //added by Y.Ge
	struct neighbor_2_list_entry	*two_hop_list; //added by Y.Ge
	struct neighbor_2_list_entry	*two_hop_list_tmp; //added by Y.Ge
	struct neighbor_2_entry     	*two_hop_entry; //added by Y.Ge
	struct topology_last_entry  	*t_last; //added by Y.Ge
	struct destination_list     	*topo_dest; //added by Y.Ge
	struct destination_list     	*topo_dest_tmp; //added by Y.Ge
	struct topology_destination_entry	*dest_entry; //added by Y.Ge
	
	int count=0;
	rcvd_vtime = olsr_compute_time_from_vtime(message->v_time); //added by Y.Ge
	olsr_init_timer(rcvd_vtime*1000,&hold_time_tmp);  //added by Y.Ge , rcvd v_time
	
	if (debug_level > 3)
	{
		printf ("\n\n-------------- in process message neighbors ----------------\n");
		printf ("the HELLO sender's main status is: %d\n", neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_status);
	
		fprintf (y_file,"\n\n-------------- in process message neighbors ----------------\n");
		fprintf (y_file,"the HELLO sender's main status is: %d\n", neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_status);
	}
	
	message_neighbors=message->neighbors;
	while(message_neighbors!=NULL)
	{
		if (debug_level > 3)
		{
			printf ("\n\n...now looking at message neighbbor %s\n", convert_address_to_string(&message_neighbors->address));
			fprintf (y_file,"\n\n...now looking at message neighbbor %s\n", convert_address_to_string(&message_neighbors->address));
		}
		
		/* the one hop neighbor is not the host*/
		//CRC modify
		//if(0!=memcmp(&message_neighbors->address,&addr_ip,sizeof(struct olsr_ip_addr))) //commented by Y.Ge
			
		if (olsr_lookup_my_own_address (&message_neighbors->address)==0) //added by Y.Ge
		{
			if (debug_level > 3)
			{
				printf (" this message neighbor is not my self!\n"); 
				fprintf (y_file," this message neighbor is not my self!\n");
			} 
			
			/******************* added by Y.Ge ********************/
			/*find main address of the message neighbor from MID */
			message_neighbor_interface_address = olsr_lookup_interface_address_table (&message_neighbors->address);
			if (message_neighbor_interface_address == NULL)
			{
				if (debug_level > 3)
				{
					printf (" -- the message neighbor is NOT an interface address\n");
					fprintf (y_file," -- the message neighbor is NOT an interface address\n");
				}
				
				rcvd_main_address = &(message_neighbors->address);
			}
			else
			{
				if (debug_level > 3)
				{
					printf (" the neighbor have main address\n");
					fprintf (y_file," the neighbor have main address\n");
				}
				
				message_neighbor_main_address = message_neighbor_interface_address->node_mainaddress->mainaddress;
				rcvd_main_address = &(message_neighbor_main_address->node_main_address);
			}
			
			if (debug_level > 3)
			{	
				printf ("   the main address of this neighbor is: %s\n", convert_address_to_string(rcvd_main_address));
				fprintf (y_file,"   the main address of this neighbor is: %s\n", convert_address_to_string(rcvd_main_address));
			}
			/******************* end of revision *****************/
			
			/*the one hop neighbor is NBS_SYM or NBS_MPR status*/	   	    
			//if((message_neighbors->status==NBS_SYM)||(message_neighbors->status==NBS_MPR))//commented by Y.Ge
			if(((message_neighbors->neigh_status==SYM_NEIGH)||(message_neighbors->neigh_status==MPR_NEIGH))&& neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_status!=NOT_NEIGH)//added by Y.Ge			
			{
				if (debug_level > 3)
				{
					printf ("-- the message neighbor has a sym_link!\n");
					fprintf (y_file,"-- the message neighbor has a sym_link!\n");
				}
								
				//if((two_hop_neighbor_yet=olsr_lookup_my_neighbors(neighbor,&message_neighbors->address))!=NULL) //commented by Y.Ge
				if((two_hop_neighbor_yet=olsr_lookup_my_neighbors(neighbor,rcvd_main_address))!=NULL) //added by Y.Ge (use main address)
				{
					if (debug_level > 3)
					{
						printf ("-- the message neighbor is my two hop neighbor\n");
						fprintf (y_file,"-- the message neighbor is my two hop neighbor\n");
					}
					
					/*
					 Updating the holding time for this neighbor
					*/
					//timeradd(&now,&hold_time_neighbor,&two_hop_neighbor_yet->neighbor_2_timer); //commented by Y.Ge
					timeradd(&now, &hold_time_tmp, &two_hop_neighbor_yet->neighbor_2_timer); //added by Y.Ge

					/*-------------------*/
					if(debug_level > 3)
					{
						printf("%s=============>Update second hop neighbor time\n",convert_address_to_string(&message_neighbors->address));
						fprintf(y_file,"%s=============>Update second hop neighbor time\n",convert_address_to_string(&message_neighbors->address));
					}
					/*-------------------*/
				}
				else
				{					
					/*-------------------*/
					if(debug_level > 3)
					{
						printf ("-- the message neighbor is not my two hop neighbor!\n");
						printf("%s=============>MAYBE A new second hop neighbor\n",convert_address_to_string(&message_neighbors->address));
						
						fprintf (y_file,"-- the message neighbor is not my two hop neighbor!\n");
						fprintf(y_file,"%s=============>MAYBE A new second hop neighbor\n",convert_address_to_string(&message_neighbors->address));
					}
					/*-------------------*/

					/*
					 Looking if this node is already in the two hop neighbor table!!!
					*/
					/* Y.Ge: use the main address! */
					//if((two_hop_neighbor=olsr_lookup_two_hop_neighbor_table(&message_neighbors->address))==NULL) //commented by Y.Ge
					if((two_hop_neighbor=olsr_lookup_two_hop_neighbor_table(rcvd_main_address))==NULL) //added by Y.Ge
					{
						if (debug_level > 3)
						{
							printf ("-- this message neighbor is NOT in the two hop neighbor table!\n");
							fprintf (y_file,"-- this message neighbor is NOT in the two hop neighbor table!\n");
						}
						
						/*this node is not in the two hop neighbor table*/
						ok_insert=0;

						/*
						This node will be inserted in the two hop neighbor table even if it exist in the neighbor
						table with the status ASYM or LOST. In this way we can have connectivity with this neighbor
						*/
						
						/* Y.Ge: we should look at main address neighbor table AND also the neighbor(link)table */
								 
						/*if((in_neighbor=olsr_lookup_neighbor_table(&message_neighbors->address))!=NULL)
						{
							if((in_neighbor->neighbor_status==NBS_ASYM)||(in_neighbor->neighbor_status==NBS_LOST)||(in_neighbor->neighbor_status==NBS_PENDING)) //commented by Y.Ge
								ok_insert=1;			      
						}*/ //commented by Y.Ge
						/************************* added by Y.Ge *******************/
						in_neighbor = olsr_lookup_neighbor_table(&message_neighbors->address, &addr_manet[iface_index]);
						if (in_neighbor==NULL)
						{
							if (debug_level > 3)
							{
								printf (" --- the sender is NOT in my neighbor table!\n");
								fprintf (y_file," --- the sender is NOT in my neighbor table!\n");
							}
							
							if ((in_main_addr_neighbor=olsr_lookup_main_addr_neigh_table(rcvd_main_address))!=NULL)
							{
								if (debug_level > 3)
								{
									printf (" it is in the main_address_neighbor table!\n");
									printf ("_____ address: %s, status:%d\n", convert_address_to_string(&in_main_addr_neighbor->main_addr_neigh_addr), in_main_addr_neighbor->main_addr_neigh_status);
								
									fprintf (y_file," it is in the main_address_neighbor table!\n");
									fprintf (y_file,"_____ address: %s, status:%d\n", convert_address_to_string(&in_main_addr_neighbor->main_addr_neigh_addr), in_main_addr_neighbor->main_addr_neigh_status);
								}
								
								if(in_main_addr_neighbor->main_addr_neigh_status == NOT_NEIGH)
									ok_insert = 1;
							}					
							
							else
								ok_insert=1;
						}
						else
						{
							if (debug_level > 3)
							{
								printf (" --- the sender is in my neighbor table, its status is:%d!\n", in_neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_status);
								fprintf (y_file," --- the sender is in my neighbor table, its status is:%d!\n", in_neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_status);
							}
							
							if (in_neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_status == NOT_NEIGH)
								ok_insert=1;
						}
		  				/************************* end of revision *****************/
						
						if(ok_insert)
						{
							if (debug_level > 3)
							{
								printf("%s=============>IT IS A new second hop neighbor\n",convert_address_to_string(rcvd_main_address));
								fprintf(y_file,"%s=============>IT IS A new second hop neighbor\n",convert_address_to_string(rcvd_main_address));
							}
							
							changes_neighborhood=UP;
							changes_topology=UP;
							two_hop_neighbor=(struct neighbor_2_entry *)malloc(sizeof(struct neighbor_2_entry));
							two_hop_neighbor->neighbor_2_nblist=NULL;
							two_hop_neighbor->neighbor_2_pointer=0;
							//memcpy(&two_hop_neighbor->neighbor_2_addr,&message_neighbors->address,sizeof(struct olsr_ip_addr)); //commented by Y.Ge
							memcpy(&two_hop_neighbor->neighbor_2_addr,rcvd_main_address,sizeof(struct olsr_ip_addr)); //added by Y.Ge, use main address
							olsr_insert_two_hop_neighbor_table(two_hop_neighbor);
							olsr_linking_this_2_entries(neighbor,two_hop_neighbor);
						}
					}
					else
					{
						if (debug_level > 3)
						{
							printf ("-- this message neighbor is in the two hop neighbor table!\n");
							fprintf (y_file,"-- this message neighbor is in the two hop neighbor table!\n");
						}
						/*
						 linking to this two_hop_neighbor entry
						*/
						/*this node is already in the two hop neighbor table*/	
						olsr_linking_this_2_entries(neighbor,two_hop_neighbor);
						
						/******** added by Y.Ge ***********/
						/*
						 * the hello sender has one more 2-hop-neighbor,
						 * reselect MPR!
						 */
						changes_neighborhood=UP;
						changes_topology=UP;
						/************** end of revision ********/  
					}
				}
			}
			/******************* added by Y.Ge ****************/
			/* only the HELLO sender who is a SYM neighbor of me can have 2 hop neighbors */
			/* delete this neighbor's two hop neighbor if the neighbor status in the message is NOT_NEIGH */
			else
			{
				if (debug_level > 3)
				{
					printf (" ----- this message neighbor or its HELLO sender is not a SYM neighbor, check if it is in the two hop list!\n");
					fprintf (y_file," ----- this message neighbor or its HELLO sender is not a SYM neighbor, check if it is in the two hop list!\n");
				}
				
				two_hop_list = neighbor->neighbor_2_list;
				neighbor->neighbor_2_list = NULL;
				while (two_hop_list != NULL)
				{
					/*
					 * check both message neighbor's interface address and main address
					 * incase the MID has not been received
					 */
					if((0==memcmp(rcvd_main_address, &two_hop_list->neighbor_2->neighbor_2_addr, sizeof(struct olsr_ip_addr))) || (0 == memcmp(&message_neighbors->address, &two_hop_list->neighbor_2->neighbor_2_addr, sizeof(struct olsr_ip_addr)))) 
					{
						/* node exist in two_hop_neighbor list, delete it */
						if (debug_level > 3)
						{
							printf (" ----- it is in the two hop list from the HELLO sender! DELETE IT!!!\n");
						}
						two_hop_entry=two_hop_list->neighbor_2;
						two_hop_entry->neighbor_2_pointer --;
						olsr_delete_neighbor_pointer(two_hop_entry, &neighbor->neighbor_addr);
					
						if (two_hop_entry->neighbor_2_pointer<1)
						{
							olsr_remque((struct olsr_qelem *)two_hop_entry);
							free ((void *) two_hop_entry);
						}
						two_hop_list_tmp = two_hop_list;
						two_hop_list = two_hop_list->neighbor_2_next;
						free (two_hop_list_tmp);
						
						changes_neighborhood=UP;
						changes_topology=UP;
					}
					else
					{
						two_hop_list_tmp = two_hop_list;
						two_hop_list = two_hop_list->neighbor_2_next;
						two_hop_list_tmp->neighbor_2_next = neighbor->neighbor_2_list;
						neighbor->neighbor_2_list = two_hop_list_tmp;
					}
				}
			}
			/******************** end of revision ***************/					
		}
		else
		{
			if (debug_level > 3)
			{
				printf(" this neighbor is myself!\n");
				fprintf(y_file," this neighbor is myself!\n");
			}
		}    
     		message_neighbors=message_neighbors->next;
	}
}
/*------------------------------------------------------------------------*/
int
olsr_process_changes()
{
	/*------------------*/
	if(debug_level > 8)
	{
		if(changes_neighborhood)
		printf("CHANGES IN NEIGHBORHOOD\n");
		if(changes_topology)
		printf("CHANGES IN TOPOLOGY\n");
	}
	/*------------------*/
  
	if(changes_neighborhood)
	{
		/*
		Calculate new mprs and routing table
		*/
		olsr_calculate_mpr();
		olsr_calculate_routing_table(); 
		changes_neighborhood=DOWN;
		changes_topology=DOWN;
		//CRC HNA add
		/*-------------------------*/
		changes_net=DOWN;
		/*-------------------------*/
		if (debug_level ==1 || debug_level ==2)
		{
			printf ("\n\n.........time (%02d:%02d:%02d.%06lu), changes in neighborhood! .........\n", nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec);
			printf (" now, the routing table is\n");
			olsr_print_routing_table(routingtable);
			
			fprintf (y_file,"\n\n\n\n.........time (%02d:%02d:%02d.%06lu), changes in neighborhood! .........\n", nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec);
			fprintf (y_file," now, the routing table is\n");
			olsr_print_trace_routing_table(routingtable);
		}
		
		if (debug_level ==2)
		{
			printf ("\n the neighbor table is\n");
			olsr_print_neighbor_table();
			printf ("\n the two hop neighbor table is\n");
			olsr_print_two_hop_neighbor_table();
			printf ("\n the main address routing table is\n");
			olsr_print_main_addr_neigh_table();
			
			fprintf (y_file,"\n the neighbor table is\n");
			olsr_print_trace_neighbor_table();
			fprintf (y_file,"\n the two hop neighbor table is\n");
			olsr_print_trace_two_hop_neighbor_table();
			fprintf (y_file,"\n the main address routing table is\n");
			olsr_print_trace_main_addr_neigh_table();
			
		}
		return 0;
	}
	if(changes_topology)
	{
		/*
		calculate the routing table
		*/
		olsr_calculate_routing_table(); 
		changes_topology=DOWN;
		//CRC HNA add
		/*-------------------------*/
		changes_net=DOWN;
		/*-------------------------*/
		
		if (debug_level ==1 || debug_level ==2)
		{
			printf ("\n\n.........time (%02d:%02d:%02d.%06lu),changes in topology! .........\n", nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec);
			printf (" now, the routing table is\n");
			olsr_print_routing_table(routingtable);
			
			fprintf (y_file,"\n\n.........time (%02d:%02d:%02d.%06lu),changes in topology! .........\n", nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec);
			fprintf (y_file," now, the routing table is\n");
			olsr_print_trace_routing_table(routingtable);
		}
		
		if (debug_level ==2)
		{
			printf ("\n the topology table is\n");
			olsr_print_topology_table();
			printf ("\n the mid table is:\n");
			olsr_print_main_address_table();
			olsr_print_interface_address_table();
						
			fprintf (y_file,"\n the topology table is\n");
			olsr_print_trace_topology_table();
			fprintf (y_file,"\n the mid table is:\n");
			olsr_print_trace_main_address_table();
			olsr_print_trace_interface_address_table();
		}
		return 0;
	}
	//CRC HNA add
	/*------------------------------*/
	if(changes_net)
	{
		/*
		copy the routing table from mirror table 
		and modify it to reflect the hna changes
		*/
		//olsr_modify_routing_table_from_hna();
		olsr_calculate_routing_table();
		changes_net=DOWN; 
		
		if (debug_level ==1 || debug_level ==2)
		{
			printf (".........changes in network! .........\n");
			printf (" now, the routing table is\n");
			olsr_print_routing_table(routingtable);
			
			fprintf (y_file,".........changes in network! .........\n");
			fprintf (y_file," now, the routing table is\n");
			olsr_print_trace_routing_table(routingtable);
		}
		return 0;
	}
}
/*------------------------------------------------------------------------*/
/********************** added by Y.Ge *****************/
/* qos OLSR */
int
olsr_process_changes_with_qos()
{
	/*------------------*/
	if(debug_level > 8)
	{
		if(changes_neighborhood)
		printf("CHANGES IN NEIGHBORHOOD\n");
		if(changes_topology)
		printf("CHANGES IN TOPOLOGY\n");
	}
	/*------------------*/
  
	if(changes_neighborhood)
	{
		/*
		Calculate new mprs and routing table
		*/
		olsr_calculate_mpr();
		olsr_calculate_routing_table_with_qos(); 
		changes_neighborhood=DOWN;
		changes_topology=DOWN;
		//CRC HNA add
		/*-------------------------*/
		changes_net=DOWN;
		/*-------------------------*/
		if (debug_level ==1 || debug_level ==2)
		{
			printf ("\n\n.........time (%02d:%02d:%02d.%06lu), changes in neighborhood! .........\n", nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec);
			printf (" now, the routing table is\n");
			olsr_print_routing_table(routingtable);
			
			fprintf (y_file,"\n\n\n\n.........time (%02d:%02d:%02d.%06lu), changes in neighborhood! .........\n", nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec);
			fprintf (y_file," now, the routing table is\n");
			olsr_print_trace_routing_table(routingtable);
		}
		
		if (debug_level ==2)
		{
			printf ("\n the neighbor table is\n");
			olsr_print_neighbor_table();
			printf ("\n the two hop neighbor table is\n");
			olsr_print_two_hop_neighbor_table();
			printf ("\n the main address routing table is\n");
			olsr_print_main_addr_neigh_table();
			
			fprintf (y_file,"\n the neighbor table is\n");
			olsr_print_trace_neighbor_table();
			fprintf (y_file,"\n the two hop neighbor table is\n");
			olsr_print_trace_two_hop_neighbor_table();
			fprintf (y_file,"\n the main address routing table is\n");
			olsr_print_trace_main_addr_neigh_table();
			
		}
		return 0;
	}
	if(changes_topology)
	{
		/*
		calculate the routing table
		*/
		olsr_calculate_routing_table_with_qos(); 
		changes_topology=DOWN;
		//CRC HNA add
		/*-------------------------*/
		changes_net=DOWN;
		/*-------------------------*/
		
		if (debug_level ==1 || debug_level ==2)
		{
			printf ("\n\n.........time (%02d:%02d:%02d.%06lu),changes in topology! .........\n", nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec);
			printf (" now, the routing table is\n");
			olsr_print_routing_table(routingtable);
			
			fprintf (y_file,"\n\n.........time (%02d:%02d:%02d.%06lu),changes in topology! .........\n", nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec);
			fprintf (y_file," now, the routing table is\n");
			olsr_print_trace_routing_table(routingtable);
		}
		
		if (debug_level ==2)
		{
			printf ("\n the topology table is\n");
			olsr_print_topology_table();
			printf ("\n the mid table is:\n");
			olsr_print_main_address_table();
			olsr_print_interface_address_table();
						
			fprintf (y_file,"\n the topology table is\n");
			olsr_print_trace_topology_table();
			fprintf (y_file,"\n the mid table is:\n");
			olsr_print_trace_main_address_table();
			olsr_print_trace_interface_address_table();
		}
		return 0;
	}
	//CRC HNA add
	/*------------------------------*/
	if(changes_net)
	{
		/*
		copy the routing table from mirror table 
		and modify it to reflect the hna changes
		*/
		//olsr_modify_routing_table_from_hna();
		olsr_calculate_routing_table_with_qos();
		changes_net=DOWN; 
		
		if (debug_level ==1 || debug_level ==2)
		{
			printf (".........changes in network! .........\n");
			printf (" now, the routing table is\n");
			olsr_print_routing_table(routingtable);
			
			fprintf (y_file,".........changes in network! .........\n");
			fprintf (y_file," now, the routing table is\n");
			olsr_print_trace_routing_table(routingtable);
		}
		return 0;
	}
}
/*------------------------------------------------------------------------*/
int
olsr_process_received_hello(struct hello_message *message, int index)
{
	if (use_hyst)
		olsr_process_received_hello_with_hyst(message, index);
	else
		olsr_process_received_hello_no_hyst(message, index);	
}
/*------------------------------------------------------------------------*/
/*------ This function WITH hysteresis----*/
int
olsr_process_received_hello_with_hyst(struct hello_message *message, int iface_index) //"iface_index" added by Y.Ge
{
	olsr_u8_t                   	link=0;
	olsr_u8_t               	link_type=0, neigh_type = 0;
	struct neighbor_entry       	*neighbor;
	struct neighbor_2_entry      	*two_hop_neighbor;
	struct main_addr_neigh_entry	*main_addr_neighbor;  //added by Y.Ge
	struct iface_addr_list_entry	*neighbor_iface_link; //added by Y.Ge
	float                   	rcvd_vtime;  //added by Y.Ge 
	float                       	rcvd_htime;	//added by Y.Ge  
	olsr_u8_t               	change_hold_time = 0; //added by Y.Ge 
	struct timeval 	        	hold_time_tmp;	//added by Y.Ge
	struct timeval              	next_hello_time; //added by Y.Ge 	
	int                          	changed = 0;  //IF (changed), the main address neighbor status should be updated
	struct timeval            	tmp_timeval; //added by Y.Ge
	
	if(message==NULL)
		return 1;
    
	//link=olsr_lookup_link_status(message); //commented by Y.Ge
	
	if (debug_level > 4)
	{
		printf ("\n\nbefore process received HELLO WITH HYST, the tables are:\n");
		olsr_print_neighbor_table();
		olsr_print_main_addr_neigh_table(); 
		olsr_print_two_hop_neighbor_table();
		olsr_print_mpr_selector_table();
		olsr_print_topology_table();
		olsr_print_main_address_table();
		olsr_print_interface_address_table();
		if (use_qos)
			qos_olsr_print_routing_table(routingtable);
		else
			olsr_print_routing_table(routingtable);
		
		fprintf (y_file,"\n\nbefore process received HELLO WITH HYST, the tables are:\n");
		olsr_print_trace_neighbor_table();
		olsr_print_trace_main_addr_neigh_table(); 
		olsr_print_trace_two_hop_neighbor_table();
		olsr_print_trace_mpr_selector_table();
		olsr_print_trace_topology_table();
		olsr_print_trace_main_address_table();
		olsr_print_trace_interface_address_table();
		if (use_qos)
			qos_olsr_print_trace_routing_table(routingtable);
		else
			olsr_print_trace_routing_table(routingtable);
	}

	/********************** added by Y.Ge ********************/
	/*
	 * decompose link - link = neighbor type + link type 
	 */

	/********************** added by Y.Ge ******************/
	link_type = olsr_lookup_link_status(message, iface_index);
	neigh_type = olsr_lookup_neighbor_status(message, iface_index);
	rcvd_vtime = olsr_compute_time_from_vtime(message->v_time);
	rcvd_htime = olsr_compute_time_from_vtime(message->rcvd_h_time) * 1.25; //Y.Ge: consider Hello jitter

	/* a message that announce link type == link_sym && neighbor type == not_neigh is invalid */
	if (link_type == SYM_LINK && neigh_type == NOT_NEIGH)
	{
		printf ("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\n");
		printf ("XATTENTION: the link and neigghbor code in HELLO  does't match!X\n");
		printf ("XXXXXXX   link type is: %d, neighbor type is: %d  XXXXXXXXXXXXXX\n", link_type, neigh_type);
		printf ("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\n");
		olsr_destroy_hello_message(message);
		exit  (1);
	}
	olsr_init_timer(rcvd_vtime*1000,&hold_time_tmp);  //the v_time value in HELLO
	olsr_init_timer(rcvd_htime*1000, &next_hello_time); //the h_time value in HELLO
	/******************** end of revision *******************/

	//if(link==NBS_MPR) //commented by Y.Ge
	if (neigh_type == MPR_NEIGH) //added by Y.Ge
		olsr_update_mpr_selector_table(message);
  
	//if(NULL== (neighbor=olsr_lookup_neighbor_table(&message->source_addr))) //commented by Y.Ge
	if(NULL== (neighbor=olsr_lookup_neighbor_table(&message->source_addr, &addr_manet[iface_index]))) //added by Y.Ge
	{
		//The neighbor entry doesn't exist in neighbtable	
	
		/********************** added by Y.Ge ******************/
		/*** debug ***/
		if (debug_level >3)
		{
			printf ("\n+++ new neighbor detected ! +++\n"); //added by Y.Ge
			fprintf (y_file,"\n+++ new neighbor detected ! +++\n"); //added by Y.Ge
		}
		/********************* end of revision *****************/

		neighbor=(struct neighbor_entry *)malloc(sizeof(struct neighbor_entry));   

		memcpy(&neighbor->my_iface_addr, &message->my_addr, sizeof(struct olsr_ip_addr)); //added by  Y.Ge 
		memcpy(&neighbor->neighbor_addr, &message->source_addr, sizeof(struct olsr_ip_addr)); //added by Y.Ge
		neighbor->my_iface_index = iface_index; //added by Y.Ge, iface_index that connects to that neighbor
		//memcpy(&neighbor->neighbor_addr,&message->originator_addr,sizeof(struct olsr_ip_addr));	//commented by Y.Ge
		
		/*********** added by Y.Ge ***********/
		/* qos OLSR */
		if (use_qos)
		{
			neighbor->neighbor_hello_seq = message->packet_seq_number;
			neighbor->neighbor_busy_time = default_busy_time;
		}
		/********** end of revision *********/
		
		/* timers */				
		timerminus(&neighbor->neighbor_link_sym_timer); //added by Y.Ge, expire the L_SYM_TIME
		timeradd(&now, &hold_time_tmp, &neighbor->neighbor_link_asym_timer); //added by Y.Ge
		timeradd(&now, &hold_time_tmp, &neighbor->neighbor_timer); //added  by Y.Ge
		//timeradd(&now,&hold_time_neighbor,&neighbor->neighbor_timer); //commented by Y.Ge

		neighbor->neighbor_ifp=NULL;
		neighbor->neighbor_2_list=NULL;
		neighbor->neighbor_2_nocov=0;
		neighbor->neighbor_main_info=NULL;
		
		/* added by Y.Ge, dealing with hysteresis */
  		//added by L. Qin, 
		neighbor->neighbor_hysavg = HYST_SCALING;
		printf("in olsr.c the neighbor_hysavg is %f\n",neighbor->neighbor_hysavg);
		neighbor->link_pending = 0;
		neighbor->h_time_value = rcvd_htime;
		timerminus(&neighbor->neighbor_link_lost_timer);
		timeradd(&now, &next_hello_time, &neighbor->next_hello_timer);

		olsr_insert_neighbor_table(neighbor);		

		/* update the link status */
		//neighbor->neighbor_status=NBS_ASYM; //commented by Y.Ge 
		neighbor->neighbor_status = ASYM_LINK; //added by Y.Ge
		
		/* L_time = max (L_time, L_ASYM_time), added by Y.Ge */
		if (compare_timer(&neighbor->neighbor_timer, &neighbor->neighbor_link_asym_timer) ==0)
			olsr_assign_timer(&neighbor->neighbor_link_asym_timer, &neighbor->neighbor_timer); 
 		
		/******************** added by Y.Ge **************/
		/*
		 update main_addr_neighbor_table
		*/
		/* a new neighbor */
		if (NULL == (main_addr_neighbor = olsr_lookup_main_addr_neigh_table(&message->originator_addr)))
		{
			main_addr_neighbor = (struct main_addr_neigh_entry *)malloc(sizeof(struct main_addr_neigh_entry));
			memcpy(&main_addr_neighbor->main_addr_neigh_addr, &message->originator_addr, sizeof(struct olsr_ip_addr));
			main_addr_neighbor->main_addr_neigh_willingness = message->willingness;
			main_addr_neighbor->main_addr_neigh_status = NOT_NEIGH; 
			main_addr_neighbor->neigh_iface_list=NULL;
			main_addr_neighbor->interface_pointer=0;
			olsr_insert_main_addr_neigh_table(main_addr_neighbor);
			/* LINK neighbor to main_neighbor ENTRIES */
			olsr_linking_neighbor_main_neighbor(neighbor, main_addr_neighbor);
		}
		else
		{
			/* LINK neighbor to main_neighbor ENTRIES */
			olsr_linking_neighbor_main_neighbor(neighbor, main_addr_neighbor);
		}
		/******************* end of revision ***************/							  
	}
	else  //existing link
	{      
		/********************** added by Y.Ge ******************/
		/*** debug ***/
		if (debug_level >3)
		{
			printf ("\n+++ existing neighbor detected ! +++\n"); //added by Y.Ge
			printf ("     link type is: %d\n", link_type); //added by Y.Ge
			printf (" neighbor is: %s\n", convert_address_to_string (&neighbor->neighbor_addr));
			
			fprintf (y_file,"\n+++ existing neighbor detected ! +++\n"); //added by Y.Ge
			fprintf (y_file,"     link type is: %d\n", link_type); //added by Y.Ge
			fprintf (y_file," neighbor is: %s\n", convert_address_to_string (&neighbor->neighbor_addr));
		}
		
		/* update hysteresis */
		neighbor->neighbor_hysavg = hysteresis (neighbor->neighbor_hysavg, 1);
		neighbor->h_time_value = rcvd_htime;
		
		if (debug_level >3)
		{
			printf ("----- the hyst is: %f\n",neighbor->neighbor_hysavg); 
			fprintf (y_file,"----- the hyst is: %f\n",neighbor->neighbor_hysavg);
		} 
		
		/*********** added by Y.Ge ***********/
		/* qos OLSR */
		if (use_qos)
		{
			neighbor->neighbor_hello_seq = message->packet_seq_number;
		}
		/********** end of revision *********/
		
		/* update timer */
		timeradd(&now, &next_hello_time, &neighbor->next_hello_timer);
		timeradd(&now, &hold_time_tmp, &neighbor->neighbor_link_asym_timer);
		timeradd(&now, &hold_time_tmp, &neighbor->neighbor_timer);
		
		if (link_type != NOT_DETECTED && link_type != UNSPEC_LINK)
		{
			if (neighbor->neighbor_hysavg > HYST_THRESHOLD_HIGH)
			{
				neighbor->link_pending = 0;
				timerminus(&neighbor->neighbor_link_lost_timer);
			
				switch (neighbor->neighbor_status)
				{
					case LOST_LINK:
						timerminus(&neighbor->neighbor_link_sym_timer); //expire the L_SYM_TIME
						timeradd(&now, &hold_time_tmp, &neighbor->neighbor_timer);
						neighbor->neighbor_status = ASYM_LINK;
						break;
					case SYM_LINK:
					case ASYM_LINK:
						if (neighbor->neighbor_status != SYM_LINK)
						{
							changes_neighborhood=UP;
							changes_topology=UP;
							changed = 1;						
						}
						neighbor->neighbor_status = SYM_LINK;
						timeradd(&now, &hold_time_tmp, &neighbor->neighbor_link_sym_timer);
						olsr_assign_timer(&neighbor->neighbor_link_sym_timer, &neighbor->neighbor_timer);
						timeradd(&neighbor->neighbor_timer, &hold_time_neighbor, &neighbor->neighbor_timer);
						break;
					default:
						break;
				}		
			}
			else
			{
				/* HYST_THRESHOLD_LOW > L_link_qualitty */
				if (neighbor->neighbor_hysavg < HYST_THRESHOLD_LOW)
				{
					neighbor->link_pending = 1;
				
					if (neighbor->neighbor_status == SYM_LINK)
					{
						changes_neighborhood=UP;
						changes_topology=UP;
						changed = 1;
					}
					neighbor->neighbor_status = LOST_LINK;
				
					timeradd(&now, &hold_time_tmp, &neighbor->neighbor_timer);
					timeradd(&now, &hold_time_neighbor, &tmp_timeval);
					if (compare_timer(&neighbor->neighbor_timer, &tmp_timeval) ==0)
						olsr_assign_timer(&neighbor->neighbor_timer, &neighbor->neighbor_link_lost_timer); 
					else
						olsr_assign_timer(&tmp_timeval, &neighbor->neighbor_link_lost_timer); 	
				}
			
				/* HYST_THRESHOLD_LOW <=L_link_qualitty<=HYST_THRESHOLD_HIGH */
				else
				{
					timeradd(&now, &hold_time_tmp, &neighbor->neighbor_link_asym_timer); //update L_ASYM_time
					
					if (link_type == LOST_LINK)
					{
						if (neighbor->neighbor_status == SYM_LINK)
						{
							changes_neighborhood=UP;
							changes_topology=UP;
							changed = 1;
						}
						neighbor->neighbor_status = LOST_LINK;
						timerminus(&neighbor->neighbor_link_sym_timer); //expire the L_SYM_TIME
						timeradd(&now, &hold_time_neighbor, &neighbor->neighbor_link_lost_timer);
					}				
					else
					{
						if (link_type == SYM_LINK || link_type == ASYM_LINK)
						{
							if (olsr_timed_out(&neighbor->neighbor_link_lost_timer) && neighbor->link_pending ==0)
							{
								if (neighbor->neighbor_status != SYM_LINK)
								{
									changes_neighborhood=UP;
									changes_topology=UP;
									changed = 1;						
								}
								neighbor->neighbor_status = SYM_LINK;
								timeradd(&now, &hold_time_tmp, &neighbor->neighbor_link_sym_timer);
								olsr_assign_timer(&neighbor->neighbor_link_sym_timer, &neighbor->neighbor_timer);
								timeradd(&neighbor->neighbor_timer, &hold_time_neighbor, &neighbor->neighbor_timer);
							}
						}
					} 
			
					/* L_time = max (L_time, L_ASYM_time) */
					if (compare_timer(&neighbor->neighbor_timer, &neighbor->neighbor_link_asym_timer) ==0)
						olsr_assign_timer(&neighbor->neighbor_link_asym_timer, &neighbor->neighbor_timer); 

				}
			}
				
			/* again, check the main_addr_neighbor table */
			if (NULL == (main_addr_neighbor = olsr_lookup_main_addr_neigh_table(&message->originator_addr)))
			{
				printf ("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\n");
				printf ("XATTENTION: something wrong with the main_addr_neighbor_table!X\n");
				printf ("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\n");
				olsr_destroy_hello_message(message);
				exit  (1);
			}
			else
			{
				/* check if the main_addr_neighbor has a link to current neighbbor_iface */
				if (NULL == (neighbor_iface_link = olsr_lookup_iface_neighbor(main_addr_neighbor, &neighbor->neighbor_addr)))
				{
					/* something must be wrong again, as the link should exist */
					{
						printf ("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\n");
						printf ("XATTENTION: something wrong with the main_addr_neighbor_link!X\n");
						printf ("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\n");
						olsr_destroy_hello_message(message);
						exit  (1);
					}
				
				}
				
				/* check if the neighbor has changed its willingness */
				if (main_addr_neighbor->main_addr_neigh_willingness != message->willingness)
				{
					if (debug_level > 3)
					{
						printf ("*********** the neighbor's willingness has been changed, recalculate MPRs!\n");
						printf (" the original willingness is: %d, the new willingness is: %d\n",main_addr_neighbor->main_addr_neigh_willingness,message->willingness); 
						
						fprintf (y_file, "*********** the neighbor's willingness has been changed, recalculate MPRs!\n");
						fprintf (y_file, " the original willingness is: %d, the new willingness is: %d\n",main_addr_neighbor->main_addr_neigh_willingness,message->willingness); 
					}
					
					main_addr_neighbor->main_addr_neigh_willingness = message->willingness;
					
					changes_neighborhood = UP;
				}	

				/* the switch function is deleted completely */

				/*************** end of revision *****************/
			}
		}
					
	}
	
	// Deleting this neighbor from the two hop list IF exist
	//if((neighbor->neighbor_status==NBS_MPR)||(neighbor->neighbor_status==NBS_SYM)) //commented by Y.Ge
	if(neighbor->neighbor_status==SYM_LINK) //added by Y.Ge
	{
		if (debug_level > 3)
		{
			printf (" ********** the neighor status is SYM_LINK, check if it is in the two hop neighbor table!\n");
			fprintf (y_file," ********** the neighor status is SYM_LINK, check if it is in the two hop neighbor table!\n");
		}
		
		if((two_hop_neighbor=olsr_lookup_two_hop_neighbor_table(&neighbor->neighbor_addr))!=NULL)
		{
			if (debug_level > 3)
			{
				printf (" delete the hello sender from the 2-hop-neighbor table!\n");
				fprintf (y_file," delete the hello sender from the 2-hop-neighbor table!\n");
			}
			
			olsr_delete_two_hop_neighbor_table(two_hop_neighbor);
			// This flag is set to UP to recalculate the MPR set and the routing table
			changes_neighborhood=UP;
			changes_topology=UP;
			changed = 1;
		}
				
		/***************** added by Y.Ge ****************/
		/*
		 * Y.Ge: should check both neighbor interface address and main address 
		 * as at the time the HELLO is received, MID information may still not avaliable
		 */
		if((two_hop_neighbor=olsr_lookup_two_hop_neighbor_table(&message->originator_addr))!=NULL)
		{
			if (debug_level > 3)
			{
				printf (" delete the hello sender from the 2-hop-neighbor table!\n");
				fprintf (y_file," delete the hello sender from the 2-hop-neighbor table!\n");
			}
			
			olsr_delete_two_hop_neighbor_table(two_hop_neighbor);
			// This flag is set to UP to recalculate the MPR set and the routing table
			changes_neighborhood=UP;
			changes_topology=UP;
			changed = 1;
		}
		/***************** end of revision *************/
	}

	processing: ;
	// Processing the list of neighbors in the received message
	
	if (changed)
		olsr_update_main_addr_neigh_status(main_addr_neighbor); //added by Y.Ge, update the neighbor status based on the link status
	
	/* 
	 * Y.Ge : "IF" added by Y.Ge. 
	 * don't link any two hop neighbors to a neighbor whose willingness is WILL_NEVER
	 * this makes MPR selection and routing table calculation easier
	 */ 
	if (message->willingness != WILL_NEVER) 
		olsr_process_message_neighbors(neighbor,message, iface_index);

	olsr_destroy_hello_message(message);

	/* olsr_process_changes(); */ //commented by Y.Ge, DO it after all the received messages are processed

	if(debug_level > 2)
	{
		printf("\n\n*********** after process received hello, the tables are: \n");
		olsr_print_neighbor_table();
		olsr_print_main_addr_neigh_table(); 
		olsr_print_two_hop_neighbor_table();
		olsr_print_mpr_selector_table();
		
		fprintf(y_file,"\n\n*********** after process received hello, the tables are: \n");
		olsr_print_trace_neighbor_table();
		olsr_print_trace_main_addr_neigh_table(); 
		olsr_print_trace_two_hop_neighbor_table();
		olsr_print_trace_mpr_selector_table();
	}
}
/*------------------------------------------------------------------------*/

/*------ This function without hysteresis----*/
int
olsr_process_received_hello_no_hyst(struct hello_message *message, int iface_index) //"iface_index" added by Y.Ge
{
	olsr_u8_t                   	link=0;
	olsr_u8_t               	link_type=0, neigh_type = 0;
	struct neighbor_entry       	*neighbor;
	struct neighbor_2_entry      	*two_hop_neighbor;
	struct main_addr_neigh_entry	*main_addr_neighbor;  //added by Y.Ge
	struct iface_addr_list_entry	*neighbor_iface_link; //added by Y.Ge
	float                   	rcvd_vtime;  //added by Y.Ge  
	olsr_u8_t               	change_hold_time = 0; //added by Y.Ge 
	struct timeval 	        	hold_time_tmp; 	
	int                          	changed = 0;  //IF (changed), the main address neighbor status should be updated


//added by L.Qin
//	printf("             ******* received without hyst\n");
	
	if(message==NULL)
		return 1;
    
	//link=olsr_lookup_link_status(message); //commented by Y.Ge
	
	if (debug_level > 4)
	{
		printf ("\n\nbefore process received HELLO, the tables are:\n");
		olsr_print_neighbor_table();
		olsr_print_main_addr_neigh_table(); 
		olsr_print_two_hop_neighbor_table();
		olsr_print_mpr_selector_table();
		olsr_print_topology_table();
		olsr_print_main_address_table();
		olsr_print_interface_address_table();
		olsr_print_routing_table(routingtable);
		
		fprintf (y_file,"\n\nbefore process received HELLO, the tables are:\n");
		olsr_print_trace_neighbor_table();
		olsr_print_trace_main_addr_neigh_table(); 
		olsr_print_trace_two_hop_neighbor_table();
		olsr_print_trace_mpr_selector_table();
		olsr_print_trace_topology_table();
		olsr_print_trace_main_address_table();
		olsr_print_trace_interface_address_table();
		olsr_print_trace_routing_table(routingtable);
	}

	/********************** added by Y.Ge ********************/
	/*
	 * decompose link - link = neighbor type + link type 
	 */

	/********************** added by Y.Ge ******************/
	link_type = olsr_lookup_link_status(message, iface_index);
	neigh_type = olsr_lookup_neighbor_status(message, iface_index);
	rcvd_vtime = olsr_compute_time_from_vtime(message->v_time);
	
	/*printf (" the received v time is: %f\n", rcvd_vtime);
	fprintf (y_file," the received v time is: %f\n", rcvd_vtime);*/

	/* a message that announce link type == link_sym && neighbor type == not_neigh is invalid */
	if (link_type == SYM_LINK && neigh_type == NOT_NEIGH)
	{
		printf ("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\n");
		printf ("XATTENTION: the link and neigghbor code in HELLO  does't match!X\n");
		printf ("XXXXXXX   link type is: %d, neighbor type is: %d  XXXXXXXXXXXXXX\n", link_type, neigh_type);
		printf ("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\n");
		olsr_destroy_hello_message(message);
		exit  (1);
	}
	olsr_init_timer(rcvd_vtime*1000,&hold_time_tmp);  //the v_time value in HELLO
	/******************** end of revision *******************/

	//if(link==NBS_MPR) //commented by Y.Ge
	if (neigh_type == MPR_NEIGH) //added by Y.Ge
		olsr_update_mpr_selector_table(message);
  
	//if(NULL== (neighbor=olsr_lookup_neighbor_table(&message->source_addr))) //commented by Y.Ge
	if(NULL== (neighbor=olsr_lookup_neighbor_table(&message->source_addr, &addr_manet[iface_index]))) //added by Y.Ge
	{
		//The neighbor entry doesn't exist in neighbtable	
	
		/********************** added by Y.Ge ******************/
		/*** debug ***/
		if (debug_level >3)
		{
			printf ("\n+++ new neighbor detected ! +++\n"); //added by Y.Ge
			fprintf (y_file,"\n+++ new neighbor detected ! +++\n"); //added by Y.Ge
		}
		/********************* end of revision *****************/

		neighbor=(struct neighbor_entry *)malloc(sizeof(struct neighbor_entry));   

		memcpy(&neighbor->my_iface_addr, &message->my_addr, sizeof(struct olsr_ip_addr)); //added by  Y.Ge 
		memcpy(&neighbor->neighbor_addr, &message->source_addr, sizeof(struct olsr_ip_addr)); //added by Y.Ge
		neighbor->my_iface_index = iface_index; //added by Y.Ge, iface_index that connects to that neighbor
		//memcpy(&neighbor->neighbor_addr,&message->originator_addr,sizeof(struct olsr_ip_addr));	//commented by Y.Ge
		
		/*********** added by Y.Ge ***********/
		/* qos OLSR */
		if (use_qos)
		{
			neighbor->neighbor_hello_seq = message->packet_seq_number;
			neighbor->neighbor_busy_time = default_busy_time;
		}
		/********** end of revision *********/
		
		/* timers */				
		timerminus(&neighbor->neighbor_link_sym_timer); //added by Y.Ge, expire the L_SYM_TIME
		timeradd(&now, &hold_time_tmp, &neighbor->neighbor_link_asym_timer); //added by Y.Ge
		timeradd(&now, &hold_time_tmp, &neighbor->neighbor_timer); //added  by Y.Ge
		//timeradd(&now,&hold_time_neighbor,&neighbor->neighbor_timer); //commented by Y.Ge

		//added by L.Qin, check the hyst value *****
//		printf("The hyst vakue of the neighbor is %f\n",neighbor->neighbor_hysavg);
//		printf("then we set it to zero\n");
//		neighbor->neighbor_hysavg=0,
                // *****

		neighbor->neighbor_ifp=NULL;
		neighbor->neighbor_2_list=NULL;
		neighbor->neighbor_2_nocov=0;
		neighbor->neighbor_main_info=NULL;

		olsr_insert_neighbor_table(neighbor);
		
		//if((link==NBS_SYM)||(link==NBS_MPR)||(link==NBS_ASYM)) //commented by Y.Ge
		if (link_type == SYM_LINK || link_type == ASYM_LINK) //added by Y.Ge
		{
			//neighbor->neighbor_status=NBS_SYM;  //commeted by Y.Ge
			neighbor->neighbor_status = SYM_LINK; //added by Y.Ge
			
			/* change timer again */
			timeradd(&now, &hold_time_tmp, &neighbor->neighbor_link_sym_timer);
			olsr_assign_timer(&neighbor->neighbor_link_sym_timer, &neighbor->neighbor_timer);
			timeradd(&neighbor->neighbor_timer, &hold_time_neighbor, &neighbor->neighbor_timer);
			
			// This flag is set to UP to recalculate the MPR set and the routing table			
			changes_neighborhood=UP;
			changes_topology=UP;
			changed = 1;  //IF (changed), the main address neighbor status should be updated
			
			// remove this neighbor from the two hop table IF exist
			if((two_hop_neighbor=olsr_lookup_two_hop_neighbor_table(&neighbor->neighbor_addr))!=NULL)
				olsr_delete_two_hop_neighbor_table(two_hop_neighbor);
		}
		else
		{
			/* update the link status */
			//neighbor->neighbor_status=NBS_ASYM; //commented by Y.Ge 
			neighbor->neighbor_status = ASYM_LINK; //added by Y.Ge
		}
		
		/* L_time = max (L_time, L_ASYM_time), added by Y.Ge */
		if (compare_timer(&neighbor->neighbor_timer, &neighbor->neighbor_link_asym_timer) ==0)
			olsr_assign_timer(&neighbor->neighbor_link_asym_timer, &neighbor->neighbor_timer); 
 		
		/******************** added by Y.Ge **************/
		/*
		 update main_addr_neighbor_table
		*/
		/* a new neighbor */
		if (NULL == (main_addr_neighbor = olsr_lookup_main_addr_neigh_table(&message->originator_addr)))
		{
			main_addr_neighbor = (struct main_addr_neigh_entry *)malloc(sizeof(struct main_addr_neigh_entry));
			memcpy(&main_addr_neighbor->main_addr_neigh_addr, &message->originator_addr, sizeof(struct olsr_ip_addr));
			main_addr_neighbor->main_addr_neigh_willingness = message->willingness;
			main_addr_neighbor->main_addr_neigh_status = NOT_NEIGH; 
			main_addr_neighbor->neigh_iface_list=NULL;
			main_addr_neighbor->interface_pointer=0;
			olsr_insert_main_addr_neigh_table(main_addr_neighbor);
			/* LINK neighbor to main_neighbor ENTRIES */
			olsr_linking_neighbor_main_neighbor(neighbor, main_addr_neighbor);
		}
		else
		{
			/* LINK neighbor to main_neighbor ENTRIES */
			olsr_linking_neighbor_main_neighbor(neighbor, main_addr_neighbor);
		}
		/******************* end of revision ***************/							  
	}
	else  //existing link
	{      
		//timeradd(&now,&hold_time_neighbor,&neighbor->neighbor_timer); //commented by Y.Ge
		/********************** added by Y.Ge ******************/
		/*** debug ***/
		if (debug_level >3)
		{
			printf ("\n+++ existing neighbor detected ! +++\n"); //added by Y.Ge
			printf ("     link type is: %d\n", link_type); //added by Y.Ge
			printf (" neighbor is: %s\n", convert_address_to_string (&neighbor->neighbor_addr));
			
			fprintf (y_file,"\n+++ existing neighbor detected ! +++\n"); //added by Y.Ge
			fprintf (y_file,"     link type is: %d\n", link_type); //added by Y.Ge
			fprintf (y_file," neighbor is: %s\n", convert_address_to_string (&neighbor->neighbor_addr));
		}
		/********************* end of revision *****************/
				
		/*************** added by Y.Ge ****************/
		timeradd(&now, &hold_time_tmp, &neighbor->neighbor_link_asym_timer); //update L_ASYM_time
		
		/*********** added by Y.Ge ***********/
		/* qos OLSR */
		if (use_qos)
		{
			neighbor->neighbor_hello_seq = message->packet_seq_number;
		}
		/********** end of revision *********/
		
		if (link_type != NOT_DETECTED)
		{
					
			if (link_type == LOST_LINK)
			{
				if (neighbor->neighbor_status == SYM_LINK)
				{
					changes_neighborhood=UP;
					changes_topology=UP;
					changed = 1;
				}
				neighbor->neighbor_status = ASYM_LINK;
				timerminus(&neighbor->neighbor_link_sym_timer); //expire the L_SYM_TIME
			}
			else
			{
				if (link_type == SYM_LINK || link_type == ASYM_LINK)
				{
					if (neighbor->neighbor_status != SYM_LINK)
					{
						changes_neighborhood=UP;
						changes_topology=UP;
						changed = 1;						
					}
					neighbor->neighbor_status = SYM_LINK;
					timeradd(&now, &hold_time_tmp, &neighbor->neighbor_link_sym_timer);
					olsr_assign_timer(&neighbor->neighbor_link_sym_timer, &neighbor->neighbor_timer);
					timeradd(&neighbor->neighbor_timer, &hold_time_neighbor, &neighbor->neighbor_timer);
				}
			} 
			
			/* L_time = max (L_time, L_ASYM_time) */
			if (compare_timer(&neighbor->neighbor_timer, &neighbor->neighbor_link_asym_timer) ==0)
				olsr_assign_timer(&neighbor->neighbor_link_asym_timer, &neighbor->neighbor_timer); 

			/* again, check the main_addr_neighbor table */
			if (NULL == (main_addr_neighbor = olsr_lookup_main_addr_neigh_table(&message->originator_addr)))
			{
				printf ("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\n");
				printf ("XATTENTION: something wrong with the main_addr_neighbor_table!X\n");
				printf ("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\n");
				olsr_destroy_hello_message(message);
				exit  (1);
			}
			else
			{
				/* check if the main_addr_neighbor has a link to current neighbbor_iface */
				if (NULL == (neighbor_iface_link = olsr_lookup_iface_neighbor(main_addr_neighbor, &neighbor->neighbor_addr)))
				{
					/* something must be wrong again, as the link should exist */
					{
						printf ("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\n");
						printf ("XATTENTION: something wrong with the main_addr_neighbor_link!X\n");
						printf ("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\n");
						olsr_destroy_hello_message(message);
						exit  (1);
					}
				}
				
				/* check if the neighbor has changed its willingness */
				if (main_addr_neighbor->main_addr_neigh_willingness != message->willingness)
				{
					if (debug_level > 3)
					{
						printf ("*********** the neighbor's willingness has been changed, recalculate MPRs!\n");
						printf (" the original willingness is: %d, the new willingness is: %d\n",main_addr_neighbor->main_addr_neigh_willingness,message->willingness); 
						
						fprintf (y_file, "*********** the neighbor's willingness has been changed, recalculate MPRs!\n");
						fprintf (y_file, " the original willingness is: %d, the new willingness is: %d\n",main_addr_neighbor->main_addr_neigh_willingness,message->willingness); 
					}
					
					main_addr_neighbor->main_addr_neigh_willingness = message->willingness;
					
					changes_neighborhood = UP;
				}	
			}	

			/* the switch function is deleted completely */

			/*************** end of revision *****************/
		}
					
	}
	
	// Deleting this neighbor from the two hop list IF exist
	//if((neighbor->neighbor_status==NBS_MPR)||(neighbor->neighbor_status==NBS_SYM)) //commented by Y.Ge
	if(neighbor->neighbor_status==SYM_LINK) //added by Y.Ge
	{
		if (debug_level > 3)
		{
			printf (" ********** the neighor status is SYM_LINK, check if it is in the two hop neighbor table!\n");
			fprintf (y_file," ********** the neighor status is SYM_LINK, check if it is in the two hop neighbor table!\n");
		}
		
		if((two_hop_neighbor=olsr_lookup_two_hop_neighbor_table(&neighbor->neighbor_addr))!=NULL)
		{
			if (debug_level > 3)
			{
				printf (" delete the hello sender from the 2-hop-neighbor table!\n");
				fprintf (y_file," delete the hello sender from the 2-hop-neighbor table!\n");
			}
			
			olsr_delete_two_hop_neighbor_table(two_hop_neighbor);
			// This flag is set to UP to recalculate the MPR set and the routing table
			changes_neighborhood=UP;
			changes_topology=UP;
			changed = 1;
		}
				
		/***************** added by Y.Ge ****************/
		/*
		 * Y.Ge: should check both neighbor interface address and main address 
		 * as at the time the HELLO is received, MID information may still not avaliable
		 */
		if((two_hop_neighbor=olsr_lookup_two_hop_neighbor_table(&message->originator_addr))!=NULL)
		{
			if (debug_level > 3)
			{
				printf (" delete the hello sender from the 2-hop-neighbor table!\n");
				fprintf (y_file," delete the hello sender from the 2-hop-neighbor table!\n");
			}
			
			olsr_delete_two_hop_neighbor_table(two_hop_neighbor);
			// This flag is set to UP to recalculate the MPR set and the routing table
			changes_neighborhood=UP;
			changes_topology=UP;
			changed = 1;
		}
		/***************** end of revision *************/
	}

	processing: ;
	// Processing the list of neighbors in the received message
	
	if (changed)
		olsr_update_main_addr_neigh_status(main_addr_neighbor); //added by Y.Ge, update the neighbor status based on the link status
	
	/* 
	 * Y.Ge : "IF" added by Y.Ge. 
	 * don't link any two hop neighbors to a neighbor whose willingness is WILL_NEVER
	 * this makes MPR selection and routing table calculation easier
	 */ 
	if (message->willingness != WILL_NEVER) 
		olsr_process_message_neighbors(neighbor,message, iface_index);

	olsr_destroy_hello_message(message);

	/*olsr_process_changes();*/ //commented by Y.Ge, DO it after all the received messages are processed

	if(debug_level > 2)
	{
		printf("\n\n*********** after process received hello, the tables are: \n");
		olsr_print_neighbor_table();
		olsr_print_main_addr_neigh_table(); 
		olsr_print_two_hop_neighbor_table();
		olsr_print_mpr_selector_table();
		
		fprintf(y_file,"\n\n*********** after process received hello, the tables are: \n");
		olsr_print_trace_neighbor_table();
		olsr_print_trace_main_addr_neigh_table(); 
		olsr_print_trace_two_hop_neighbor_table();
		olsr_print_trace_mpr_selector_table();
	}
}
/*------------------------------------------------------------------------*/
/************ commented by Y.Ge ***************/
/* function with hysteresis is deleted completely */
/*int
olsr_process_received_hello(struct hello_message *message)*/
/*********** end of revission ****************/
/*------------------------------------------------------------------------*/
/************ added by Y.Ge ***************/
/* qos OLSR */
int
olsr_process_received_qos_hello (int index, struct qos_hello_message *message)
{
	struct neighbor_entry	*neighbor;
	
	if (debug_level > 3)
	{
		printf ("******************* in process_received_qos_hello *****************\n");
		fprintf (y_file,"******************* in process_received_qos_hello *****************\n");
	}
	
	if ((neighbor = olsr_lookup_neighbor_table(&message->source_addr, &message->my_addr))!=NULL)
	{
		if (neighbor->neighbor_hello_seq != message->packet_seq_number)
		{
			if (debug_level > 3)
			{
				printf ("********** ATTENTION! packet sequence number is different! HELLO sequence: %d, QoS HELLO sequence: %d\n", neighbor->neighbor_hello_seq, message->packet_seq_number);
				fprintf (y_file,"********** ATTENTION! packet sequence number is different! HELLO sequence: %d, QoS HELLO sequence: %d\n", neighbor->neighbor_hello_seq, message->packet_seq_number);  
			}
			
			if (wrap_around_compare(neighbor->neighbor_hello_seq,message->packet_seq_number) == -1)
			{
				printf ("XXXXX QoS HELLO received before HELLO! ERROR! XXXXX\n");
				fprintf (y_file,"XXXXX QoS HELLO received before HELLO! ERROR! XXXXX\n");
				
				exit (1);
			}
		}
		else
		{
			if (neighbor->neighbor_busy_time != message->rcvd_busy_time)
			{
				/* should update the new busy time */
				if (debug_level > 3)
				{
					printf ("************ neighbor busy time changed! old busy time:%d, new busy time:%d\n",neighbor->neighbor_busy_time, message->rcvd_busy_time); 
					fprintf (y_file, "************ neighbor busy time changed! old busy time:%d, new busy time:%d\n",neighbor->neighbor_busy_time, message->rcvd_busy_time);
				}
				
				changes_neighborhood = UP;
				neighbor->neighbor_busy_time = message->rcvd_busy_time;  
			}
			else
			{
				if (debug_level > 3)
				{
					printf ("************ neighbor busy time not changed!\n"); 
					fprintf (y_file, "************ neighbor busy time not changed! \n");
				}
			}
			
			/* should also check if it is the MPR selector. If yes, mpr selector table should also be changed */
			if (debug_level > 3)
			{
				printf ("..... update the MPR selector busy time!\n");
				fprintf (y_file,"..... update the MPR selector busy time!\n");
			}
			olsr_update_mpr_selector_busy_time(message);
		}
	}
	else
	{
		printf ("XXXXX QoS HELLO received while there is no established neighbor! ERROR! XXXXX\n");
		fprintf (y_file,"XXXXX QoS HELLO while there is no established neigghbor! ERROR! XXXXX\n");
				
		exit (1);
	}
	
	if (debug_level > 2)
	{
		printf("\n\n*********** after process received qos hello, the tables are: \n");
		olsr_print_neighbor_table();
		olsr_print_mpr_selector_table();
		
		fprintf(y_file,"\n\n*********** after process received qos hello, the tables are: \n");
		olsr_print_trace_neighbor_table();
		olsr_print_trace_mpr_selector_table();
	}
	
	return 0;	
}
/*------------------------------------------------------------------------*/
int
olsr_process_received_qos_tc(int index, struct tc_message *message,struct tc_message **message_out,
				struct qos_tc_message *qosmessage, struct qos_tc_message **qosmessage_out)
{
	struct	neighbor_entry     	*neighbor;
	struct	topology_last_entry	*t_last;
	struct	tc_mpr_addr       	*mpr;
	struct	tc_mpr_busy_time	*mprbusy;
	int	should_forward=0, not_processed_yet = 1; 
	struct	duplicate_entry     	*dup_message;	

	*message_out=NULL;
	*qosmessage_out=NULL;
    
	if(message==NULL)
		return 1;
	
	if(qosmessage == NULL)
		return 1;
		
	if (message->packet_seq_number != qosmessage->packet_seq_number)
	{
		printf("XXXXXXXXXXX ERROR in QoS TC packet sequence number! XXXXXXXXXX\n");
		printf("tc seq: %d, qos tc seq: %d\n", message->packet_seq_number, qosmessage->packet_seq_number);
		olsr_destroy_tc_message(message);
		olsr_destroy_qos_tc_message(qosmessage);
		exit(1);
	}
	
	if (message->mssn != qosmessage->mssn)
	{
		printf("XXXXXXXXXXX ERROR in QoS TC packet mssn number! XXXXXXXXXX\n");
		printf("tc mssn: %d, qos tc mssn: %d\n", message->mssn, qosmessage->mssn);
		olsr_destroy_tc_message(message);
		olsr_destroy_qos_tc_message(qosmessage);
		exit(1);
	}
	
	if (memcmp(&message->originator, &qosmessage->originator,sizeof(struct olsr_ip_addr))!=0)
	{
		printf("XXXXXXXXXXX ERROR in QoS TC packet originator address! XXXXXXXXXX\n");
		printf("tc originator: %s\n", convert_address_to_string(&message->originator));
		printf("qos tc mssn: %s\n", convert_address_to_string(&qosmessage->originator));
		olsr_destroy_tc_message(message);
		olsr_destroy_qos_tc_message(qosmessage);
		exit(1);
	}

	/*----------------------------*/
	if(debug_level > 3)
	{
		printf ("\n\n******************** in process received_qos_tc *********************\n");
		printf (" message mssn: %d\n",message->mssn);
		fprintf (y_file,"\n\n******************** in process received_qos_tc *********************\n");
		fprintf (y_file," message mssn: %d\n",message->mssn);
	}
	/*----------------------------*/
	
	/* if I'm the originator of the TC */
	if(0==memcmp(&message->originator,&my_main_addr,sizeof(struct olsr_ip_addr))) //added by Y.Ge
	{
		olsr_destroy_tc_message(message);
		olsr_destroy_qos_tc_message(qosmessage);
		
		if (debug_level > 3)
		{
			printf ("XXXXXXXXXXXXXXX I'm the originator of the QoS TC, DROP! XXXXXXXXXXXXX\n");
			fprintf (y_file,"XXXXXXXXXXXXXXX I'm the originator of the QoS TC, DROP! XXXXXXXXXXXXX\n");
		}
		
		return 1;
	}

	/*Check whether  the tc is received from a sym neighbor or not.
	Be careful we should not check the originator address*/
	if ((neighbor=olsr_lookup_neighbor_table(&message->source_addr, &addr_manet[index]))!=NULL) //added by Y.Ge
	{
		//the tc message from a neighbor
		//if((neighbor->neighbor_status==NBS_ASYM)||(neighbor->neighbor_status==NBS_LOST)||(neighbor->neighbor_status==NBS_PENDING)) //commented by Y.Ge
		if (neighbor->neighbor_status != SYM_LINK) 
		{
			olsr_destroy_tc_message(message);
			olsr_destroy_qos_tc_message(qosmessage);
			
			if (debug_level > 3)
			{
				printf ("XXXXXXXXXXXXXXX QoS TC sender is a asymmetric neighbor, DROP! XXXXXXXXXXXXX\n");
				fprintf (y_file,"XXXXXXXXXXXXXXX QoS TC sender is a asymmetric neighbor, DROP! XXXXXXXXXXXXX\n");
			}
			
			return 1;
		}
	}
	else
	{
		/*The TC message source address is not in the neighbor table !!!!!!*/
		olsr_destroy_tc_message(message);
		olsr_destroy_qos_tc_message(qosmessage);
		
		if (debug_level > 3)
		{
			printf ("XXXXXXXXXXXXXXX TC sender is NOT a neighbor, DROP! XXXXXXXXXXXXX\n");
			fprintf (y_file,"XXXXXXXXXXXXXXX TC sender is NOT a neighbor, DROP! XXXXXXXXXXXXX\n");
		}
		
		return 1;
	}
	
	/* if we should forward the qos tc */
	if (debug_level > 3)
	{
	 	printf ("\n\nmy mpr selector table is: \n");
		olsr_print_mpr_selector_table();
		
		fprintf (y_file,"\n\nmy mpr selector table is: \n");
		olsr_print_trace_mpr_selector_table();
	}
	if(NULL!=(olsr_lookup_mpr_selector_table(&neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_addr)))
	{
		should_forward = 1;
	}
	    
	/* till now, we can say that the tc message from a sym neighbor */
	if ((dup_message = olsr_lookup_duplicate_table(TC_PACKET, message->originator, message->packet_seq_number, index))!=NULL) //added by Y.Ge
	{
		if (debug_level > 3)
		{
			printf ("&&&&&&&&&&&&& THE MESSAGE IS IN THE DUPLICATE TABLE! &&&&&&&&&&&&&&&&\n");
			fprintf (y_file,"&&&&&&&&&&&&& THE MESSAGE IS IN THE DUPLICATE TABLE! &&&&&&&&&&&&&&&&\n");
		
			if (debug_level > 5)
			{
				olsr_print_duplicate_table();
				olsr_print_trace_duplicate_table();
			}
		}
		
		not_processed_yet = 0; //the message has been processed
		
		/* check if the receiving interface address is in the duplicate entry */
		if (olsr_lookup_interface_address_from_duplicate_entry(dup_message, &addr_manet[index])==1)
		{
			if (debug_level > 5)
			{
				printf (" my current interface is already in the duplicate table !\n");
				fprintf (y_file," my current interface is already in the duplicate table !\n");
				
				olsr_print_duplicate_table();
				olsr_print_trace_duplicate_table();
			}
		}
		else
		{
			olsr_insert_interface_address_into_duplicate_entry(dup_message, &addr_manet[index]);
			
			if (debug_level > 5)
			{
				printf (" ........after add my current address, the duplicate table is:\n");
				olsr_print_duplicate_table();
				
				fprintf (y_file," ........after add my current address, the duplicate table is:\n");
				olsr_print_trace_duplicate_table();
			}
		}
		
		/* if the duplicate message is only processed, but not forwarded , then do it */
		if (should_forward == 1)
		{
			if (dup_message->duplicate_retransmitted == 0) //	should forward this message
			{
				dup_message->duplicate_retransmitted =1;
			
				if (debug_level > 3)
				{
					printf ("\n***** I did not retransmitted the message before, DO it this time\n");
					fprintf (y_file,"\n***** I did not retransmitted the message before, DO it this time\n");
				}
			}
			else //need not forward this message
			{
				olsr_destroy_tc_message(message);
				olsr_destroy_qos_tc_message(qosmessage);
			
				if (debug_level > 3)
				{
					printf ("\n***** the message has been retransmitted. DROP it!\n");
					fprintf (y_file,"\n***** the message has been retransmitted. DROP it!\n");
				}
				return 1;
			}
		}
		else
		{
			olsr_destroy_tc_message(message);
			olsr_destroy_qos_tc_message(qosmessage);
					
			if (debug_level > 3)
			{
				printf ("\n***** the message need NOT to be retransmitted. DROP it!\n");
				fprintf (y_file,"\n***** the message need NOT to be retransmitted. DROP it!\n");
			}
			return 1;
		}
	}
	else
	{
		/*the tc message is not in the duplicate table, insert new entry in duplicate table*/
		olsr_insert_duplicate_table(TC_PACKET, message->originator, message->packet_seq_number, index, should_forward); //added by Y.Ge
		
		if (debug_level > 3)
		{
			printf ("\n++++++++++++++ new QoS TC packet, insert into duplicate table\n");
			fprintf (y_file,"\n++++++++++++++ new QoS TC packet, insert into duplicate table\n");
		}
	}
	
	if (not_processed_yet) //a message maybe processed, but not forwarded
	{
		/*check if this node receive a tc message from the same host*/
		t_last=olsr_lookup_last_topology_table(&message->originator);
  
		if(t_last!=NULL)
		{
			/*----------------------*/
			if(debug_level > 3)
			{
				printf("*******THE T_LAST EXISTS\n");
				fprintf(y_file,"*********THE T_LAST EXISTS\n");
			}
			/*----------------------*/

			if (wrap_around_compare(t_last->topology_seq, message->mssn) == 1) //consider the wrap-around
			{
				olsr_destroy_tc_message(message);
				olsr_destroy_qos_tc_message(qosmessage);
				
				if(debug_level > 3)
				{
					printf("QoS TC_message TOO OLD\n");
					fprintf(y_file,"QoS TC_message TOO OLD\n");
				}
				return 1;
			}

			/*-----------------------*/
			if(debug_level > 3)
			{
				printf("THE NEW MSSN is %d and the old one is %d\n",message->mssn,t_last->topology_seq);
				fprintf(y_file,"THE NEW MSSN is %d and the old one is %d\n",message->mssn,t_last->topology_seq);
			}
			/*-----------------------*/
      
			if (wrap_around_compare(t_last->topology_seq, message->mssn) == -1) //consider the wrap-around
			{
				olsr_delete_last_topolgy_table(t_last);
				/*
			 	 We must insert the new entries contained in the received message
				*/ 
				
				if (debug_level > 3)
				{
					printf (" ----- a TC with more receint MSSN!\n");
					fprintf (y_file," ----- a TC with more receint MSSN!\n");
				}
				
				/*
				 * if it is an empty TC, this means that the t_last is no longer valid
				 * delete it from the topology table
				 */
				 
				if (message->multipoint_relay_selector_address == NULL)
				{
					if (debug_level > 3)
					{
						printf (" &&& this is an empty TC! just delete it from the topology table\n");
						fprintf (y_file," &&& this is an empty TC! just delete it from the topology table\n");
					}
					
					changes_topology = UP;
				}
				else
				{		
					changes_topology=UP;
					t_last=(struct topology_last_entry *)malloc(sizeof(struct topology_last_entry));		
					olsr_insert_last_topology_table(t_last,message);
					t_last->topology_last_busy = qosmessage->my_busy_time; //t_last's busy time
					
					olsr_update_last_table_with_qos(t_last,message, qosmessage);		
				}
			}
			else
			{
				/*
				 This the case where t_last->topology_seq==message->mssn, so
			 	we must update the time out and create the new dest entries
				*/				
				if (debug_level > 3)
				{
					printf (" ----- a TC with same MSSN, update timer!\n");
					fprintf (y_file," ----- a TC with same MSSN, update timer!\n");
				}
				
				/* update t_last busy time */
				if (t_last->topology_last_busy != qosmessage->my_busy_time)
				{
					t_last->topology_last_busy = qosmessage->my_busy_time;
					changes_topology = UP;	
				}
				olsr_update_last_table_with_qos(t_last,message, qosmessage);
			}
		}
		else
		{
			/*----------------------*/
			if(debug_level > 3)
			{
				printf("*******THE T_LAST does NOT EXISTS, insert into topology table!\n");
				fprintf(y_file,"********THE T_LAST does NOT EXISTS, insert into topology table!\n");
			}
			/*----------------------*/
			changes_topology=UP; /*changes_topology will be set to DOWN  after recalculating the routing table*/
			t_last=(struct topology_last_entry *)malloc(sizeof(struct topology_last_entry));
			olsr_insert_last_topology_table(t_last,message);
			t_last->topology_last_busy = qosmessage->my_busy_time; //t_last's busy time
			olsr_update_last_table_with_qos(t_last,message, qosmessage);
		}
	}
	
	/*Check whether this node is a mpr of the previous one*/
	//if(NULL!=(olsr_lookup_mpr_selector_table(&message->source_addr))) //commented by Y.Ge
	if (should_forward) //added by Y.Ge
	{
		/*Yes we are an mpr of the previous hop */
		/*set the source_addr to the current one*/
		/*---------------------*/
		if(debug_level > 3)
		{
			printf(".................WE ARE MPR OF %s\n",convert_address_to_string(&message->source_addr));
			fprintf(y_file,".................WE ARE MPR OF %s\n",convert_address_to_string(&message->source_addr));
		}
		/*---------------------*/

		/*send this message */
      
		//IMPORTANT Don't forget to decrement the hop count in the message!!!!!!!
		/*--------------------*/
		if(debug_level > 3)
			printf("The hop_count is %d\n",message->hop_count);
		/*--------------------*/
      
		message->hop_count ++;    
		message->time_to_live --;
		
		qosmessage->hop_count ++;
		qosmessage->time_to_live --;
             
		if(message->time_to_live >0) 
		{
			if (debug_level > 3)
			{
				printf ("--------------should forward this message ! -----------------\n");
				fprintf (y_file,"--------------should forward this message ! -----------------\n");
			}
			
			*message_out=message;
			*qosmessage_out=qosmessage;
		}
		else
		{
			olsr_destroy_tc_message(message);
			olsr_destroy_qos_tc_message(qosmessage);
		}
	}
	else
	{
		olsr_destroy_tc_message(message); 
		olsr_destroy_qos_tc_message(qosmessage);
	}     

	/* olsr_process_changes(); */ //commented by Y.Ge, DO it after all the received messages are processed
	
	if (debug_level > 2)
	{
		printf ("\n\n**************** after process received TC, the tables are:\n");
		olsr_print_topology_table();
		
		fprintf (y_file,"\n\n**************** after process received TC, the tables are:\n");
		olsr_print_trace_topology_table();
	}
	return 0; 

}
/*------------------------------------------------------------------------*/

int
olsr_process_received_mid (int index, struct mid_message *message, struct mid_message **message_out)
{

	struct neighbor_entry	*neighbor;
	struct main_addr_entry	*rcvd_node_main_address_entry;
	struct mid_iface_addr	*rcvd_node_interface_address;
	struct duplicate_entry	*dup_message;
	
	int	should_forward=0, not_processed_yet=1;

	*message_out = NULL;

	if (message == NULL)
		return 1;

	/*----------------------------*/
	if (debug_level > 3)
	{
		rcvd_node_interface_address = message->mid_interface_address;
		printf ("\n\n******************* in process_received_mid *****************\n");
				
		/* printf ("... sending node: %s\n", convert_address_to_string (&message->source_addr));
		printf ("    main address: %s\n", convert_address_to_string (&message->originator_addr));
		printf ("    interface addresses: [");
		while (rcvd_node_interface_address !=NULL)
		{
			printf ("%s,  ", convert_address_to_string (&rcvd_node_interface_address->interface_address));
			rcvd_node_interface_address = rcvd_node_interface_address->next;
		}
		printf ("]\n");*/
		
		fprintf (y_file,"\n\n******************* in process_received_mid *****************\n");
	}
	/*----------------------------*/

	/*
	 * check if I'm the sender of the mid message (check the main address)
	 */
	if (0==memcmp (&message->originator_addr, &my_main_addr, sizeof (struct olsr_ip_addr)))
	{
		olsr_destroy_mid_message  (message);
		
		if (debug_level > 3)
		{
			printf ("XXXXXXXXXXXXXXX I'm the originator of the MID, DROP! XXXXXXXXXXXXX\n");
			fprintf (y_file,"XXXXXXXXXXXXXXX I'm the originator of the MID, DROP! XXXXXXXXXXXXX\n");
		}
		
		return 1;
	}

	/*
	 * Check whether the mid is received from a sym neighbor or not (check sender_interface)
	 */
	if ((neighbor = olsr_lookup_neighbor_table(&message->source_addr, &message->my_addr))!=NULL)
	{
		if(neighbor->neighbor_status != SYM_LINK)
		{
			olsr_destroy_mid_message (message);
			
			if (debug_level > 3)
			{
				printf ("XXXXXXXXXXXXXXX MID sender is a asymmetric neighbor, DROP! XXXXXXXXXXXXX\n");
				fprintf (y_file,"XXXXXXXXXXXXXXX MID sender is a asymmetric neighbor, DROP! XXXXXXXXXXXXX\n");
			}
			
			return 1;
		}
	}
	else
	{
		olsr_destroy_mid_message (message);
		
		if (debug_level > 3)
		{
			printf ("XXXXXXXXXXXXXXX MID sender is NOT neighbor, DROP! XXXXXXXXXXXXX\n");
			fprintf (y_file,"XXXXXXXXXXXXXXX MID sender is NOT neighbor, DROP! XXXXXXXXXXXXX\n");
		}
		
		return 1;
	}
	
	/*
	 * check if we should forward this packet (check the main address)
	 */
	if (debug_level > 3)
	{
		printf ("\n\nmy mpr selector table is: \n");
		olsr_print_mpr_selector_table();
		printf (" mpr selector table finished\n");
		
		fprintf (y_file,"\n\nmy mpr selector table is: \n");
		olsr_print_trace_mpr_selector_table();
		fprintf (y_file," mpr selector table finished\n");
	}
	if (NULL!=(olsr_lookup_mpr_selector_table(&neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_addr)))
	{
		should_forward = 1;
		if (debug_level > 3)
		{
			printf ("\n***** I'm the MPR of the MID sender!\n");
			fprintf (y_file,"\n***** I'm the MPR of the MID sender!\n");
		}
	}
	 
	/*
	 * check whether is in the duplicate table. if not, insert it. (Should add the wrap off to check if the message is old, change it later)
	 */	
	if((dup_message = olsr_lookup_duplicate_table(MID_PACKET, message->originator_addr, message->packet_seq_number, index))!=NULL)
	{
		if (debug_level > 3)
		{
			printf ("&&&&&&&&&&&&& THE MESSAGE IS IN THE DUPLICATE TABLE! &&&&&&&&&&&&&&&&\n");
			if (debug_level > 5)
				olsr_print_duplicate_table();
		
			fprintf (y_file,"&&&&&&&&&&&&& THE MESSAGE IS IN THE DUPLICATE TABLE! &&&&&&&&&&&&&&&&\n");
			if (debug_level > 5)
				olsr_print_trace_duplicate_table();
		}
		
		not_processed_yet = 0; //the message has been processed
		
		/* check if the receiving interface address is in the duplicate entry */
		if (olsr_lookup_interface_address_from_duplicate_entry(dup_message, &addr_manet[index])==1)
		{
			if (debug_level > 5)
			{
				printf (" my current interface is already in the duplicate table !\n");
				fprintf (y_file," my current interface is already in the duplicate table !\n");
			}
		}
		else
		{
			olsr_insert_interface_address_into_duplicate_entry(dup_message, &addr_manet[index]);
			
			if (debug_level > 5)
			{
				printf (" ........after add my current address, the duplicate table is:\n");			
				olsr_print_duplicate_table();
			
				fprintf (y_file," ........after add my current address, the duplicate table is:\n");			
				olsr_print_trace_duplicate_table();
			}
		}
		
		/* if the duplicate message is only processed, but not forwarded , then do it */
		if (should_forward == 1)
		{
			if (dup_message->duplicate_retransmitted == 0) //	should forward this message
			{
				dup_message->duplicate_retransmitted =1;
			
				if (debug_level > 3)
				{
					printf ("\n***** I did not retransmitted the message before, DO it this time\n");
					fprintf (y_file,"\n***** I did not retransmitted the message before, DO it this time\n");
				}
			}
			else //need not forward this message
			{
				if (debug_level > 3)
				{
					printf ("\n***** the message has been retransmitted. DROP it!\n");
					fprintf (y_file,"\n***** the message has been retransmitted. DROP it!\n");
				}
			
				olsr_destroy_mid_message(message);
				return 1;
			}
		}
		else
		{
			olsr_destroy_mid_message(message);
					
			if (debug_level > 3)
			{
				printf ("\n***** the message need NOT to be retransmitted. DROP it!\n");
				fprintf (y_file,"\n***** the message need NOT to be retransmitted. DROP it!\n");
			}
			return 1;
		}
	}
	else
	{
		olsr_insert_duplicate_table(MID_PACKET, message->originator_addr, message->packet_seq_number, index, should_forward);
		
		if (debug_level > 3)
		{
			printf ("\n++++++++++++++ new MID packet, insert into duplicate table\n");
			fprintf (y_file,"\n++++++++++++++ new MID packet, insert into duplicate table\n");
		}
	}
	
	/*
	 * update tables
	 */
	
	if (not_processed_yet)
	{
		/* check if the received main address is already in the node main address table */
		rcvd_node_main_address_entry = olsr_lookup_main_address_table (&message->originator_addr);
	
		if (rcvd_node_main_address_entry != NULL)
		{
			if (debug_level > 3)
			{
				printf ("xxxxxxxxxxxxxx the main address is in the main addrress table !xxxxxxxxxxxxxx\n");
				fprintf (y_file,"xxxxxxxxxxxxxx the main address is in the main addrress table !xxxxxxxxxxxxxx\n");
			}
			
			olsr_update_main_address_table(rcvd_node_main_address_entry, message);
		}
		else
		{
			if (debug_level > 3)
			{
				printf ("xxxxxxxxxxxxxx the main address is NOT in the main addrress table !xxxxxxxxxxxxxx\n");
				fprintf (y_file,"xxxxxxxxxxxxxx the main address is NOT in the main addrress table !xxxxxxxxxxxxxx\n");
			}
			
			changes_topology = UP;
			rcvd_node_main_address_entry = (struct main_addr_entry *)malloc(sizeof (struct main_addr_entry));
			olsr_insert_main_address_table (rcvd_node_main_address_entry, message);
			olsr_update_main_address_table (rcvd_node_main_address_entry, message);
		}
	}
	
	if (debug_level > 2)
	{
		printf ("\n\n================ AFTER PROCESS RECEIVED MID, THE MID TABLES ARE: ===========\n");
		olsr_print_main_address_table();
		olsr_print_interface_address_table();
	
		fprintf (y_file,"\n\n================ AFTER PROCESS RECEIVED MID, THE MID TABLES ARE: ===========\n");
		olsr_print_trace_main_address_table();
		olsr_print_trace_interface_address_table();
	}
	
	if (should_forward)
	{
		if (debug_level > 3)
		{
			printf ("--------------should forward this message ! -----------------\n");
			fprintf (y_file,"--------------should forward this message ! -----------------\n");
		}
		
		message->hop_count++;
		message->time_to_live--;
		
		if (message->time_to_live >0)
			*message_out=message;
		else
			olsr_destroy_mid_message(message); 
	}
	else
		olsr_destroy_mid_message(message);
		
	/* olsr_process_changes(); */ //commented by Y.Ge, DO it after all the received messages are processed
	return 0;
}
/*********** end of revission ****************/
/*------------------------------------------------------------------------*/
int
olsr_process_received_tc(int index, struct tc_message *message,struct tc_message **message_out) //"index" added by Y.Ge
{

  /* 
     1. If there exist some entry in the topology table whose T_last
     corresponds to the originator address of the TC message and
     whose T_seq is greater in value than the MSSN in the received
     message, then no further processing of this TC message is 
     performed and it is silently discarded (case: message
         received out of order).
	 
	 2. If there exist some entry in the topology table whose T_last
         corresponds to the originator address of the TC message and
         whose T_seq is lesser in value than the MSSN in the received
         message, then that topology entry is removed.
	 
	 3. For each of the MPR Selector address received in the TC
         message:
	 
         3.1 If there exist some entry in the topology table whose
	 T_dest corresponds to the MPR Selector address and the
	 T_last corresponds to the originator address of the TC
	 message, then the holding time T_time of that topology
	 entry is refreshed to TOP_HOLD_TIME.
	 
         3.2 Otherwise, a new topology entry is recorded in the
	 topology table whereas:
	 
	 - T_dest is set to the MPR Selector address,
	 - T_last is set to the originator address of the TC
	 message,
	 - T_seq is set to the value of MSSN received in the TC
	 message,
	 - T_time is set to the value of TOP_HOLD_TIME.
	 
	 
  */
  
	struct	neighbor_entry     	*neighbor;
	struct	topology_last_entry	*t_last;
	struct	tc_mpr_addr       	*mpr;
	int	should_forward=0, not_processed_yet = 1; //added by Y.Ge
	struct	duplicate_entry     	*dup_message; //added by Y.Ge	

	*message_out=NULL;
    
	if(message==NULL)
		return 1;

	/*----------------------------*/
	if(debug_level > 3)
	{
		mpr=message->multipoint_relay_selector_address;
		printf ("\n\n******************** in process received_tc *********************\n");
		
		/*printf ("sending node: %s\n", convert_address_to_string(&message->source_addr));
		printf ("originator: %s\n", convert_address_to_string(&message->originator));
		printf (" message mssn: %d\n",message->mssn);
		printf("mpr_selector_list:[");      
		while(mpr!=NULL)
		{
			printf("%s:",convert_address_to_string(&mpr->address));
			mpr=mpr->next;
		}
		printf("]\n");*/
		
		fprintf (y_file,"\n\n******************** in process received_tc *********************\n");
		fprintf (y_file," message mssn: %d\n",message->mssn);
	}
	/*----------------------------*/

	//CRC modified
	//if(0==memcmp(&message->originator,&addr_ip,sizeof(struct olsr_ip_addr)))
	//if(0==memcmp(&message->originator,&addr_manet,sizeof(struct olsr_ip_addr))) //commented by Y.Ge
	
	/* if I'm the originator of the TC */
	if(0==memcmp(&message->originator,&my_main_addr,sizeof(struct olsr_ip_addr))) //added by Y.Ge
	{
		olsr_destroy_tc_message(message);
		
		if (debug_level > 3)
		{
			printf ("XXXXXXXXXXXXXXX I'm the originator of the TC, DROP! XXXXXXXXXXXXX\n");
			fprintf (y_file,"XXXXXXXXXXXXXXX I'm the originator of the TC, DROP! XXXXXXXXXXXXX\n");
		}
		
		return 1;
	}

	/*Check whether  the tc is received from a sym neighbor or not.
	Be careful we should not check the originator address*/

	//if ((neighbor=olsr_lookup_neighbor_table(&message->source_addr))!=NULL) //commented by Y.Ge
	if ((neighbor=olsr_lookup_neighbor_table(&message->source_addr, &addr_manet[index]))!=NULL) //added by Y.Ge
	{
		//the tc message from a neighbor
		//if((neighbor->neighbor_status==NBS_ASYM)||(neighbor->neighbor_status==NBS_LOST)||(neighbor->neighbor_status==NBS_PENDING)) //commented by Y.Ge
		if (neighbor->neighbor_status != SYM_LINK) //added by Y.Ge
		{
			olsr_destroy_tc_message(message);
			
			if (debug_level > 3)
			{
				printf ("XXXXXXXXXXXXXXX TC sender is a asymmetric neighbor, DROP! XXXXXXXXXXXXX\n");
				fprintf (y_file,"XXXXXXXXXXXXXXX TC sender is a asymmetric neighbor, DROP! XXXXXXXXXXXXX\n");
			}
			
			return 1;
		}
	}
	else
	{
		/*The TC message source address is not in the neighbor table !!!!!!*/
		olsr_destroy_tc_message(message);
		
		if (debug_level > 3)
		{
			printf ("XXXXXXXXXXXXXXX TC sender is NOT a neighbor, DROP! XXXXXXXXXXXXX\n");
			fprintf (y_file,"XXXXXXXXXXXXXXX TC sender is NOT a neighbor, DROP! XXXXXXXXXXXXX\n");
		}
		
		return 1;
	}
	
	/************************* added by Y.Ge ************/
	/* if we should forward the tc */
	if (debug_level > 3)
	{
	 	printf ("\n\nmy mpr selector table is: \n");
		olsr_print_mpr_selector_table();
		
		fprintf (y_file,"\n\nmy mpr selector table is: \n");
		olsr_print_trace_mpr_selector_table();
	}
	if(NULL!=(olsr_lookup_mpr_selector_table(&neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_addr)))
	{
		should_forward = 1;
	}
	/********************** end of revision ************/
    
	/* till now, we can say that the tc message from a sym neighbor */
	//if(NULL!=olsr_lookup_duplicate_table(message))  //commented by Y.Ge
	if ((dup_message = olsr_lookup_duplicate_table(TC_PACKET, message->originator, message->packet_seq_number, index))!=NULL) //added by Y.Ge
	{
		/************************************ added by Y.Ge *******************/
		if (debug_level > 3)
		{
			printf ("&&&&&&&&&&&&& THE MESSAGE IS IN THE DUPLICATE TABLE! &&&&&&&&&&&&&&&&\n");
			fprintf (y_file,"&&&&&&&&&&&&& THE MESSAGE IS IN THE DUPLICATE TABLE! &&&&&&&&&&&&&&&&\n");
		
			if (debug_level > 5)
			{
				olsr_print_duplicate_table();
				olsr_print_trace_duplicate_table();
			}
		}
		
		not_processed_yet = 0; //the message has been processed
		
		/* check if the receiving interface address is in the duplicate entry */
		if (olsr_lookup_interface_address_from_duplicate_entry(dup_message, &addr_manet[index])==1)
		{
			if (debug_level > 5)
			{
				printf (" my current interface is already in the duplicate table !\n");
				fprintf (y_file," my current interface is already in the duplicate table !\n");
				
				olsr_print_duplicate_table();
				olsr_print_trace_duplicate_table();
			}
		}
		else
		{
			olsr_insert_interface_address_into_duplicate_entry(dup_message, &addr_manet[index]);
			
			if (debug_level > 5)
			{
				printf (" ........after add my current address, the duplicate table is:\n");
				olsr_print_duplicate_table();
				
				fprintf (y_file," ........after add my current address, the duplicate table is:\n");
				olsr_print_trace_duplicate_table();
			}
		}
		
		/* if the duplicate message is only processed, but not forwarded , then do it */
		if (should_forward == 1)
		{
			if (dup_message->duplicate_retransmitted == 0) //	should forward this message
			{
				dup_message->duplicate_retransmitted =1;
			
				if (debug_level > 3)
				{
					printf ("\n***** I did not retransmitted the message before, DO it this time\n");
					fprintf (y_file,"\n***** I did not retransmitted the message before, DO it this time\n");
				}
			}
			else //need not forward this message
			{
				olsr_destroy_tc_message(message);
			
				if (debug_level > 3)
				{
					printf ("\n***** the message has been retransmitted. DROP it!\n");
					fprintf (y_file,"\n***** the message has been retransmitted. DROP it!\n");
				}
				return 1;
			}
		}
		else
		{
			olsr_destroy_tc_message(message);
					
			if (debug_level > 3)
			{
				printf ("\n***** the message need NOT to be retransmitted. DROP it!\n");
				fprintf (y_file,"\n***** the message need NOT to be retransmitted. DROP it!\n");
			}
			return 1;
		}
		/*********************************** end of revision *****************/
	}
	else
	{
		/*the tc message is not in the duplicate table, insert new entry in duplicate table*/
		//olsr_insert_duplicate_table(message); //commented by Y.Ge
		olsr_insert_duplicate_table(TC_PACKET, message->originator, message->packet_seq_number, index, should_forward); //added by Y.Ge
		
		if (debug_level > 3)
		{
			printf ("\n++++++++++++++ new TC packet, insert into duplicate table\n");
			fprintf (y_file,"\n++++++++++++++ new TC packet, insert into duplicate table\n");
		}
	}
	
	if (not_processed_yet) //added by Y.Ge, a message maybe processed, but not forwarded
	{
		/*check if this node receive a tc message from the same host*/
		t_last=olsr_lookup_last_topology_table(&message->originator);
  
		if(t_last!=NULL)
		{
			/*----------------------*/
			if(debug_level > 3)
			{
				printf("*******THE T_LAST EXISTS\n");
				fprintf(y_file,"*********THE T_LAST EXISTS\n");
			}
			/*----------------------*/

			//if(t_last->topology_seq>message->mssn) //commented by Y.Ge
			if (wrap_around_compare(t_last->topology_seq, message->mssn) == 1) //added by Y.Ge, consider the wrap-around
			{
				olsr_destroy_tc_message(message);
				if(debug_level > 3)
				{
					printf("TC_message TOO OLD\n");
					fprintf(y_file,"TC_message TOO OLD\n");
				}
				return 1;
			}

			/*-----------------------*/
			if(debug_level > 3)
			{
				printf("THE NEW MSSN is %d and the old one is %d\n",message->mssn,t_last->topology_seq);
				fprintf(y_file,"THE NEW MSSN is %d and the old one is %d\n",message->mssn,t_last->topology_seq);
			}
			/*-----------------------*/
      
			//if(t_last->topology_seq<message->mssn) //commented by Y.Ge
			if (wrap_around_compare(t_last->topology_seq, message->mssn) == -1) //added by Y.Ge, consider the wrap-around
			{
				olsr_delete_last_topolgy_table(t_last);
				/*
			 	 We must insert the new entries contained in the received message
				*/ 
				
				if (debug_level > 3)
				{
					printf (" ----- a TC with more receint MSSN!\n");
					fprintf (y_file," ----- a TC with more receint MSSN!\n");
				}
				
				/************ added by Y.Ge *************/
				/*
				 * if it is an empty TC, this means that the t_last is no longer valid
				 * delete it from the topology table
				 */
				 
				if (message->multipoint_relay_selector_address == NULL)
				{
					if (debug_level > 3)
					{
						printf (" &&& this is an empty TC! just delete it from the topology table\n");
						fprintf (y_file," &&& this is an empty TC! just delete it from the topology table\n");
					}
					
					changes_topology = UP;
				}
				/*************** end of revision ***********/
				else
				{		
					changes_topology=UP;
					t_last=(struct topology_last_entry *)malloc(sizeof(struct topology_last_entry));
					olsr_insert_last_topology_table(t_last,message);
					
					/******** added by Y.Ge *******/
					/* qos OLSR, if we did not receive a qos tc, suppose all the busy time is 0 */ 
					if (use_qos)
						t_last->topology_last_busy = default_busy_time; //t_last's busy time
					/******** end of revision ****/
					
					olsr_update_last_table(t_last,message);
				}
			}
			else
			{
				/*
				 This the case where t_last->topology_seq==message->mssn, so
			 	we must update the time out and create the new dest entries
				*/				
				if (debug_level > 3)
				{
					printf (" ----- a TC with same MSSN, update timer!\n");
					fprintf (y_file," ----- a TC with same MSSN, update timer!\n");
				}
				
				olsr_update_last_table(t_last,message);
			}
		}
		else
		{
			/*----------------------*/
			if(debug_level > 3)
			{
				printf("*******THE T_LAST does NOT EXISTS, insert into topology table!\n");
				fprintf(y_file,"********THE T_LAST does NOT EXISTS, insert into topology table!\n");
			}
			/*----------------------*/
			changes_topology=UP; /*changes_topology will be set to DOWN  after recalculating the routing table*/
			t_last=(struct topology_last_entry *)malloc(sizeof(struct topology_last_entry));
			olsr_insert_last_topology_table(t_last,message);
			
			/******** added by Y.Ge *******/
			/* qos OLSR, if we did not receive a qos tc, suppose all the busy time is 0 */ 
			if (use_qos)
				t_last->topology_last_busy = default_busy_time; //t_last's busy time
			/******** end of revision ****/
			
			olsr_update_last_table(t_last,message);
		}
	}
	
	/*Check whether this node is a mpr of the previous one*/
	//if(NULL!=(olsr_lookup_mpr_selector_table(&message->source_addr))) //commented by Y.Ge
	if (should_forward) //added by Y.Ge
	{
		/*Yes we are an mpr of the previous hop */
		/*set the source_addr to the current one*/
		/*---------------------*/
		if(debug_level > 3)
		{
			printf(".................WE ARE MPR OF %s\n",convert_address_to_string(&message->source_addr));
			fprintf(y_file,".................WE ARE MPR OF %s\n",convert_address_to_string(&message->source_addr));
		}
		/*---------------------*/
      
		//CRC modified
		//memcpy(&message->source_addr,&addr_ip,sizeof(struct olsr_ip_addr));
		//memcpy(&message->source_addr,&addr_manet,sizeof(struct olsr_ip_addr)); //commented by Y.Ge, need not to DO so

		/*---------------------*/
		/*if(debug_level > 3)
		{
			printf("........::.THE ORIGINATOR OF THIS MPR %s\n",convert_address_to_string(&message->originator));
			printf("...NOW WE R  THE SOURCE OF THIS MSG %s\n",convert_address_to_string(&message->source_addr));
		}*/  //commented by Y.Ge
		/*---------------------*/
      
		/*send this message */
      
		//IMPORTANT Don't forget to decrement the hop count in the message!!!!!!!
		/*--------------------*/
		if(debug_level > 3)
			printf("The hop_count is %d\n",message->hop_count);
		/*--------------------*/
      
		//message->hop_count --;   //commented by Y.Ge
		message->hop_count ++;     //added by Y.Ge
		message->time_to_live --;  //added by Y.Ge
             
		//if(message->hop_count>0)  //commented by Y.Ge
		if(message->time_to_live >0) //added by Y.Ge
		{
			if (debug_level > 3)
			{
				printf ("--------------should forward this message ! -----------------\n");
				fprintf (y_file,"--------------should forward this message ! -----------------\n");
			}
			
			*message_out=message;
		}
		else
			olsr_destroy_tc_message(message);
	}
	else
		olsr_destroy_tc_message(message);      

	/* olsr_process_changes(); */ //commented by Y.Ge, DO it after all the received messages are processed
	
	if (debug_level > 2)
	{
		printf ("\n\n**************** after process received TC, the tables are:\n");
		olsr_print_topology_table();
		olsr_print_routing_table(routingtable);
		
		fprintf (y_file,"\n\n**************** after process received TC, the tables are:\n");
		olsr_print_trace_topology_table();
		olsr_print_trace_routing_table(routingtable);		
	}
	return 0; 
}


void
olsr_delete_time_out_entries()
{
	olsr_time_out_duplicate_table();
	olsr_time_out_mpr_selector_table();
	olsr_time_out_neighborhood_tables();
	olsr_time_out_topology_table();
	olsr_time_out_interface_address_table(); //added by Y.Ge
	//CRC HNA add
	/*olsr_time_out_hna_duplicate_table();*/ //commented by  Y.Ge
	
	olsr_time_out_hna_table(); 
	olsr_process_changes(); 
}

/*------------------------------------------------------------------------*/
void
olsr_release_tables(int number)
{
  /*
    neighbortable
    neighbor2table
    duplicatetable
    mprstable
    routingtable
    topologytable
    hna_duplicatetable
    hnahosttable
    hnasubnettable
    hnainfo
    mainaddrneightable
  */

	olsr_release_duplicate_table();
	olsr_release_mpr_seletector_table();
	olsr_release_neighbor_table();

	olsr_release_routing_table(mirror_table);
	olsr_release_routing_table(routingtable);

	olsr_release_topology_table();
	olsr_release_two_hop_neighbor_table();

	/********** added by Y.Ge **********/
	olsr_release_main_addr_neigh_table();
	olsr_release_mid_table();
	/********** end of revision ********/
  
	//CRC HNA add
	/*------------------------------------------------------------*/
	/*olsr_release_hna_duplicate_table();*/ //commented by Y.Ge
	olsr_release_hna_table();
	olsr_release_hna_configuration_list(&hnainfo);
	/*------------------------------------------------------------*/
}
/*------------------------------------------------------------------------*/

void
olsr_init_tables()
{

  /*   
    neighbortable
    neighbor2table
    duplicatetable
    mprstable
    routingtable
    topologytable
    topologylasttable
    hna_duplicatetable
    hnahosttable
    hnasubnettable
    hnainfo
    mainaddrneightable	
 */

	olsr_16_t index;

	/*-------------------*/
	olsr_init_trace_route_file();
	/*-------------------*/

	/*NRL----------------*/
	olsr_init_trace_hysteresis_file();
	/*NRL----------------*/

	/*olsr_init_timer(NEIGHB_HOLD_TIME*1000,&hold_time_neighbor);
	olsr_init_timer(TOP_HOLD_TIME*1000,&hold_time_topology);
	olsr_init_timer(DUPLICATE_HOLD_TIME*1000,&hold_time_duplicate_tc);*/

	olsr_init_timer((olsr_u32_t)(neighbor_hold_time*1000),&hold_time_neighbor);
	olsr_init_timer((olsr_u32_t)(topology_hold_time*1000),&hold_time_topology);
	olsr_init_timer((olsr_u32_t)((tc_int*3.0)*1000),&hold_time_duplicate_tc);
	olsr_init_timer((olsr_u32_t)((MINUS_TIME)*1000),&hold_time_expired);	//added by Y.Ge
	olsr_init_timer((olsr_u32_t)(duplicate_hold_time*1000),&hold_time_duplicate);	//added by Y.Ge
  
	//CRC HNA add
	/*-----------------------------------------------------------------------------*/
	olsr_init_timer(HNA_DUPLICATE_HOLD_TIME*1000, &hold_time_duplicate_hna);
	olsr_init_timer(HNA_HOLD_TIME*1000, &hold_time_hna);
	/*-----------------------------------------------------------------------------*/
  
	neighbortable.neighborhash =(struct neighborhash *) calloc (HASHSIZE,sizeof(struct neighborhash));
	mprstable.mpr_selector_hash=(struct mpr_selector_hash *) calloc(HASHSIZE,sizeof(struct mpr_selector_hash));
  
	for(index=0;index<HASHSIZE;index++)
	{
		neighbortable.neighborhash[index].neighbor_forw=(struct neighbor_entry *)&neighbortable.neighborhash[index];
		neighbortable.neighborhash[index].neighbor_back=(struct neighbor_entry *)&neighbortable.neighborhash[index];
		neighbortable.neighbor_mpr_seq=0;
      
		neighbor2table[index].neighbor2_forw=(struct neighbor_2_entry *)&neighbor2table[index];
		neighbor2table[index].neighbor2_back=(struct neighbor_2_entry *)&neighbor2table[index];

		/********************** added by Y.Ge ******************/
		mainaddrneightable[index].mainaddrneigh_forw = (struct main_addr_neigh_entry *)&mainaddrneightable[index];
		mainaddrneightable[index].mainaddrneigh_back = (struct main_addr_neigh_entry *)&mainaddrneightable[index];
		
		nodemainaddresstable[index].mainaddress_forw = (struct main_addr_entry *)&nodemainaddresstable[index];
		nodemainaddresstable[index].mainaddress_back = (struct main_addr_entry *)&nodemainaddresstable[index];
		
		nodeifaceaddresstable[index].interfaceaddress_forw = (struct iface_addr_entry *)&nodeifaceaddresstable[index];
		nodeifaceaddresstable[index].interfaceaddress_back = (struct iface_addr_entry *)&nodeifaceaddresstable[index];
		/********************** end of revission ***************/

		duplicatetable[index].duplicate_forw=(struct duplicate_entry *)&duplicatetable[index];
		duplicatetable[index].duplicate_back=(struct duplicate_entry *)&duplicatetable[index];

		mprstable.mpr_selector_hash[index].mpr_selector_forw=(struct mpr_selector_entry *)&mprstable.mpr_selector_hash[index];
		mprstable.mpr_selector_hash[index].mpr_selector_back=(struct mpr_selector_entry *)&mprstable.mpr_selector_hash[index];
		mprstable.mssn=0;

		routingtable[index].rt_forw=(struct rt_entry *)&routingtable[index];
		routingtable[index].rt_back=(struct rt_entry *)&routingtable[index];

		mirror_table[index].rt_forw=(struct rt_entry *)&mirror_table[index];
		mirror_table[index].rt_back=(struct rt_entry *)&mirror_table[index];

		topologytable[index].topology_destination_forw=(struct topology_destination_entry *)&topologytable[index];
		topologytable[index].topology_destination_back=(struct topology_destination_entry *)&topologytable[index];

		topologylasttable[index].topology_last_forw=(struct topology_last_entry *)&topologylasttable[index];
		topologylasttable[index].topology_last_back=(struct topology_last_entry *)&topologylasttable[index];
    
		//CRC HNA add
		/*----------------------------------------------------------------------------------------*/
		/*hna_duplicatetable[index].hna_duplicate_forw=(struct hna_duplicate_entry *)&hna_duplicatetable[index];
		hna_duplicatetable[index].hna_duplicate_back=(struct hna_duplicate_entry *)&hna_duplicatetable[index];*/ //commented by Y.Ge
      
		hnasubnettable[index].hna_subnet_forw=(struct hna_subnet_entry *)&hnasubnettable[index];
		hnasubnettable[index].hna_subnet_back=(struct hna_subnet_entry *)&hnasubnettable[index];

		hnahosttable[index].hna_host_forw=(struct hna_host_entry *)&hnahosttable[index];
		hnahosttable[index].hna_host_back=(struct hna_host_entry *)&hnahosttable[index];
		/*----------------------------------------------------------------------------------------------------*/
	}

	//CRC HNA add
	/*----------------------------*/
	olsr_init_hna_info();
	/*----------------------------*/
}
/*------------------------------------------------------------------------*/
/*********************** revised by Y.Ge  ********************/
int
olsr_build_hello_packet(struct hello_message *message, int iface_index)//"iface_index" is added by Y.Ge
{
	if (use_hyst)
		olsr_build_hello_packet_with_hyst(message, iface_index);
	else
		olsr_build_hello_packet_no_hyst(message, iface_index);
		
	return 0;	
}
/*------------------------------------------------------------------------*/
int
olsr_build_hello_packet_with_hyst(struct hello_message *message, int iface_index)//"iface_index" is added by Y.Ge
{
	struct hello_neighbor *message_neighbor;
	struct neighbor_entry *neighbor;
	struct neighborhash   *hash_neighbor;
	olsr_u16_t             index;
	struct main_addr_neigh_entry	*main_addr_neigh; //added by Y.Ge
	struct main_addr_neigh_entry	*main_addr_neigh_tmp; //added by  Y.Ge
	struct main_addr_neighhash 	*main_addr_neighhash; //added by Y.Ge

  
	message->neighbors=NULL;
	message->packet_seq_number=0;

	//message->mpr_seq_number=neighbortable.neighbor_mpr_seq; //commented by Y.Ge
  
	//CRC v4->v6 modify
	//memcpy(&message->source_addr,&addr_ip,sizeof(struct olsr_ip_addr));
	memcpy(&message->source_addr,&addr_manet[iface_index],sizeof(struct olsr_ip_addr));
	memcpy (&message->originator_addr, &my_main_addr, sizeof(struct olsr_ip_addr)); //added by Y.Ge
	message->willingness = my_willingness; //added by Y.Ge
	message->packet_seq_number = msg_seqnum.hello_seqnum; //added by Y.Ge
	
	olsr_clear_main_addr_neigh_table_covered(); //added by  Y.Ge, clear the "covered" status of main_addr_neigh_table

	/*-----------------*/
	if(debug_level > 3)
	{
		/************** added by Y.Ge ***************/
		if (debug_level > 3)
		{
			printf ("\n\n\n-------------------time (%02d:%02d:%02d.%06lu) node is sending HELLO (WITH HYST) from interface %d --------------------\n",nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec, iface_index);
			printf ("originator address: %s\n", convert_address_to_string(&message->originator_addr));
			printf ("source address: %s\n", convert_address_to_string(&message->source_addr));
			printf ("my willingness: %d\n", my_willingness);
			
			if (debug_level > 5)
			{
				printf ("the related tables are: \n");
				olsr_print_neighbor_table();
				olsr_print_main_addr_neigh_table();
			}
			printf ("\n\n ................neighbors included in the message ---\n");
		
			fprintf (y_file,"\n\n\n-------------------time (%02d:%02d:%02d.%06lu) node is sending HELLO (WITH HYST) from interface %d --------------------\n",nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec, iface_index);
			fprintf (y_file,"originator address: %s\n", convert_address_to_string(&message->originator_addr));
			fprintf (y_file,"source address: %s\n", convert_address_to_string(&message->source_addr));
			fprintf (y_file,"my willingness: %d\n", my_willingness);
			
			if (debug_level > 5)
			{
				fprintf (y_file,"the related tables are: \n");
				olsr_print_trace_neighbor_table();
				olsr_print_trace_main_addr_neigh_table();
			}
			fprintf (y_file,"\n\n ................neighbors included in the message ---\n");
		}
		/************** end of revision *************/
	}
	/*-----------------*/

 	for(index=0;index<HASHSIZE;index++)
	{       
		hash_neighbor=&neighbortable.neighborhash[index];
		for(neighbor=hash_neighbor->neighbor_forw;neighbor!=(struct neighbor_entry *)hash_neighbor;neighbor=neighbor->neighbor_forw)
		{	
			if (neighbor->my_iface_index == iface_index) //added by Y.Ge
			{
				if (!olsr_timed_out(&neighbor->neighbor_link_lost_timer))
				{
					if (neighbor->neighbor_status != LOST_LINK)
					{
						printf ("XXXXXXXXX error in define LOST LINK WITH HYST ! XXXXXXXXXXX\n");
						exit (1);
					}
					if (neighbor->link_pending == 0)
					{
						message_neighbor=(struct hello_neighbor *)malloc(sizeof(struct hello_neighbor));
						//message_neighbor->status=neighbor->neighbor_status; //commented by Y.Ge
						message_neighbor->link_status = neighbor->neighbor_status; //added by Y.Ge
						message_neighbor->neigh_status = neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_status; //added by Y.Ge
						//CRC v4->v6 modify
						memcpy(&message_neighbor->address,&neighbor->neighbor_addr,sizeof(struct olsr_ip_addr));
						neighbor->neighbor_main_info->main_neighbor->covered_in_hello = 1; //added by Y.Ge, this main address neighbor is covered in this HELLO

						/*--------------------*/
						if(debug_level > 3)
						{
							printf("%s           \n ",convert_address_to_string(&message_neighbor->address));
							//printf(" status  %d\n",message_neighbor->status); //commented by Y.Ge
							printf(" neighbor status  %d\n",message_neighbor->neigh_status); //added by Y.Ge
							printf(" link status  %d\n",message_neighbor->link_status); //added by Y.Ge
					
							fprintf(y_file,"%s           \n ",convert_address_to_string(&message_neighbor->address));
							fprintf(y_file," neighbor status  %d\n",message_neighbor->neigh_status); //added by Y.Ge
							fprintf(y_file," link status  %d\n",message_neighbor->link_status); //added by Y.Ge
						}
						/*--------------------*/
						message_neighbor->next=message->neighbors;
						message->neighbors=message_neighbor;					
					}
					else //SHOULD NOT advertise this neighbor
					{
						/*--------------------*/
						if(debug_level > 3)
						{
							printf("%s           \n ",convert_address_to_string(&neighbor->neighbor_addr));
							//printf(" status  %d\n",message_neighbor->status); //commented by Y.Ge
							printf(" neighbor status  %d\n",neighbor->neighbor_status); //added by Y.Ge
							printf ("XXXXXXXXX this link is not advertised !\n");
					
							fprintf(y_file,"%s           \n ",convert_address_to_string(&neighbor->neighbor_addr));
							fprintf(y_file," neighbor status  %d\n",neighbor->neighbor_status); //added by Y.Ge
							fprintf (y_file, "XXXXXXXXX this link is not advertised !\n");
						}
						/*--------------------*/
						
						/******* added by Y.Ge **********/
						/*pretend this main address neighbor has been included in the HELLO */
						neighbor->neighbor_main_info->main_neighbor->covered_in_hello = 1; 
						/******** end if revision *******/
						 	
					}
					
				} 
				else
				{   
					message_neighbor=(struct hello_neighbor *)malloc(sizeof(struct hello_neighbor));
					//message_neighbor->status=neighbor->neighbor_status; //commented by Y.Ge
					message_neighbor->link_status = neighbor->neighbor_status; //added by Y.Ge
					message_neighbor->neigh_status = neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_status; //added by Y.Ge
					//CRC v4->v6 modify
					memcpy(&message_neighbor->address,&neighbor->neighbor_addr,sizeof(struct olsr_ip_addr));
					neighbor->neighbor_main_info->main_neighbor->covered_in_hello = 1; //added by Y.Ge, this main address neighbor is covered in this HELLO

					/*--------------------*/
					if(debug_level > 3)
					{
						printf("%s           \n ",convert_address_to_string(&message_neighbor->address));
						//printf(" status  %d\n",message_neighbor->status); //commented by Y.Ge
						printf(" neighbor status  %d\n",message_neighbor->neigh_status); //added by Y.Ge
						printf(" link status  %d\n",message_neighbor->link_status); //added by Y.Ge
					
						fprintf(y_file,"%s           \n ",convert_address_to_string(&message_neighbor->address));
						fprintf(y_file," neighbor status  %d\n",message_neighbor->neigh_status); //added by Y.Ge
						fprintf(y_file," link status  %d\n",message_neighbor->link_status); //added by Y.Ge
					}
					/*--------------------*/
					message_neighbor->next=message->neighbors;
					message->neighbors=message_neighbor;
				}
			}	    
		}
	}

	/************ added by Y.Ge ************/
	/* check main address neighbor table, include main_address */	
	for(index=0;index<HASHSIZE;index++)
	{
		main_addr_neighhash=&mainaddrneightable[index];
		for(main_addr_neigh=main_addr_neighhash->mainaddrneigh_forw;main_addr_neigh!=(struct main_addr_neigh_entry *)main_addr_neighhash;main_addr_neigh=main_addr_neigh->main_addr_neigh_forw)
		{
			if (main_addr_neigh->covered_in_hello == 0) //this neighbor has not been covered
			{
				main_addr_neigh->covered_in_hello = 1;
				message_neighbor=(struct hello_neighbor *)malloc(sizeof(struct hello_neighbor));
				message_neighbor->link_status = UNSPEC_LINK;
				message_neighbor->neigh_status = main_addr_neigh->main_addr_neigh_status;
				//CRC v4->v6 modify
				memcpy(&message_neighbor->address,&main_addr_neigh->main_addr_neigh_addr,sizeof(struct olsr_ip_addr));

				/*--------------------*/
				if(debug_level > 3)
				{
					printf("%s           \n ",convert_address_to_string(&message_neighbor->address));
					printf(" neighbor status  %d\n",message_neighbor->neigh_status); 
					printf(" link status  %d\n",message_neighbor->link_status);
					
					fprintf(y_file,"%s           \n ",convert_address_to_string(&message_neighbor->address));
					fprintf(y_file," neighbor status  %d\n",message_neighbor->neigh_status); 
					fprintf(y_file," link status  %d\n",message_neighbor->link_status);
				}
				/*--------------------*/

				message_neighbor->next=message->neighbors;
				message->neighbors=message_neighbor;	    
			}
				
		}
	}
	/********************* end of revision **************/
	
	return 0;
}
/*------------------------------------------------------------------------*/

int
olsr_build_hello_packet_no_hyst(struct hello_message *message, int iface_index)//"iface_index" is added by Y.Ge
{
	struct hello_neighbor *message_neighbor;
	struct neighbor_entry *neighbor;
	struct neighborhash   *hash_neighbor;
	olsr_u16_t             index;
	struct main_addr_neigh_entry	*main_addr_neigh; //added by Y.Ge
	struct main_addr_neigh_entry	*main_addr_neigh_tmp; //added by  Y.Ge
	struct main_addr_neighhash 	*main_addr_neighhash; //added by Y.Ge
	int i;
  
	message->neighbors=NULL;
	message->packet_seq_number=0;

	//message->mpr_seq_number=neighbortable.neighbor_mpr_seq; //commented by Y.Ge
  
	//CRC v4->v6 modify
	//memcpy(&message->source_addr,&addr_ip,sizeof(struct olsr_ip_addr));
	memcpy(&message->source_addr,&addr_manet[iface_index],sizeof(struct olsr_ip_addr));
	memcpy (&message->originator_addr, &my_main_addr, sizeof(struct olsr_ip_addr)); //added by Y.Ge
	message->willingness = my_willingness; //added by Y.Ge
	message->packet_seq_number = msg_seqnum.hello_seqnum; //added by Y.Ge

	olsr_clear_main_addr_neigh_table_covered(); //added by  Y.Ge, clear the "covered" status of main_addr_neigh_table

	/*-----------------*/
	if(debug_level > 3)
	{
		if (debug_level > 3)
		{
			printf ("\n\n\n-------------------time (%02d:%02d:%02d.%06lu) node is sending HELLO from interface %d --------------------\n",nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec, iface_index);
			printf ("originator address: %s\n", convert_address_to_string(&message->originator_addr));
			printf ("source address: %s\n", convert_address_to_string(&message->source_addr));
			printf ("my willingness: %d\n", my_willingness);
			
			if (debug_level > 5)
			{
				printf ("the related tables are: \n");
				olsr_print_neighbor_table();
				olsr_print_main_addr_neigh_table();
			}
			
			printf ("\n\n ................neighbors included in the message ---\n");
		
			fprintf (y_file,"\n\n\n-------------------time (%02d:%02d:%02d.%06lu) node is sending HELLO from interface %d --------------------\n",nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec, iface_index);
			fprintf (y_file,"originator address: %s\n", convert_address_to_string(&message->originator_addr));
			fprintf (y_file,"source address: %s\n", convert_address_to_string(&message->source_addr));
			fprintf (y_file,"my willingness: %d\n", my_willingness);
			
			if (debug_level > 5)
			{
				fprintf (y_file,"the related tables are: \n");
				olsr_print_trace_neighbor_table();
				olsr_print_trace_main_addr_neigh_table();
			}
			
			fprintf (y_file,"\n\n ................neighbors included in the message ---\n");
		}
	}
	/*-----------------*/

 	for(index=0;index<HASHSIZE;index++)
	{       
		hash_neighbor=&neighbortable.neighborhash[index];
		for(neighbor=hash_neighbor->neighbor_forw;neighbor!=(struct neighbor_entry *)hash_neighbor;neighbor=neighbor->neighbor_forw)
		{	
			if (neighbor->my_iface_index == iface_index) //added by Y.Ge
			{
				message_neighbor=(struct hello_neighbor *)malloc(sizeof(struct hello_neighbor));
				//message_neighbor->status=neighbor->neighbor_status; //commented by Y.Ge
				message_neighbor->link_status = neighbor->neighbor_status; //added by Y.Ge
				message_neighbor->neigh_status = neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_status; //added by Y.Ge
				//CRC v4->v6 modify
				memcpy(&message_neighbor->address,&neighbor->neighbor_addr,sizeof(struct olsr_ip_addr));
				neighbor->neighbor_main_info->main_neighbor->covered_in_hello = 1; //added by Y.Ge, this main address neighbor is covered in this HELLO

				/*--------------------*/
				if(debug_level > 3)
				{
					printf("%s           \n ",convert_address_to_string(&message_neighbor->address));
					//printf(" status  %d\n",message_neighbor->status); //commented by Y.Ge
					printf(" neighbor status  %d\n",message_neighbor->neigh_status); //added by Y.Ge
					printf(" link status  %d\n",message_neighbor->link_status); //added by Y.Ge
					
					fprintf(y_file,"%s           \n ",convert_address_to_string(&message_neighbor->address));
					fprintf(y_file," neighbor status  %d\n",message_neighbor->neigh_status); //added by Y.Ge
					fprintf(y_file," link status  %d\n",message_neighbor->link_status); //added by Y.Ge
				}
				/*--------------------*/
				
				message_neighbor->next=message->neighbors;
				message->neighbors=message_neighbor;
			}	    
		}
	}

	/************ added by Y.Ge ************/
	/* check main address neighbor table, include main_address */	
	for(index=0;index<HASHSIZE;index++)
	{
		main_addr_neighhash=&mainaddrneightable[index];
		for(main_addr_neigh=main_addr_neighhash->mainaddrneigh_forw;main_addr_neigh!=(struct main_addr_neigh_entry *)main_addr_neighhash;main_addr_neigh=main_addr_neigh->main_addr_neigh_forw)
		{
			if (main_addr_neigh->covered_in_hello == 0) //this neighbor has not been covered
			{
				main_addr_neigh->covered_in_hello = 1;
				message_neighbor=(struct hello_neighbor *)malloc(sizeof(struct hello_neighbor));
				message_neighbor->link_status = UNSPEC_LINK;
				message_neighbor->neigh_status = main_addr_neigh->main_addr_neigh_status;
				//CRC v4->v6 modify
				memcpy(&message_neighbor->address,&main_addr_neigh->main_addr_neigh_addr,sizeof(struct olsr_ip_addr));

				/*--------------------*/
				if(debug_level > 3)
				{
					printf("%s           \n ",convert_address_to_string(&message_neighbor->address));
					printf(" neighbor status  %d\n",message_neighbor->neigh_status); 
					printf(" link status  %d\n",message_neighbor->link_status);
					
					fprintf(y_file,"%s           \n ",convert_address_to_string(&message_neighbor->address));
					fprintf(y_file," neighbor status  %d\n",message_neighbor->neigh_status); 
					fprintf(y_file," link status  %d\n",message_neighbor->link_status);
				}
				/*--------------------*/

				message_neighbor->next=message->neighbors;
				message->neighbors=message_neighbor;	    
			}
				
		}
	}
	/********************* end of revision **************/
	
	return 0;
}
/***************************  end of revision *******************/
/*------------------------------------------------------------------------*/
int
olsr_build_tc_packet(struct tc_message *message, int iface_index)
{
	struct tc_mpr_addr        *message_mpr;
	struct mpr_selector_entry *mprs;
	struct mpr_selector_hash  *mprs_hash;
	olsr_u16_t          index;
	int i;

	message->multipoint_relay_selector_address=NULL;
	
	/* message->packet_seq_number=0;	
	message->hop_count=HOPCNT_MAX; */ //commented by Y.Ge
	
	/************* added by Y.Ge ************/
	message->packet_seq_number = msg_seqnum.tc_seqnum; 
	message->hop_count = 0;
	message->time_to_live=HOPCNT_MAX;
	message->v_time = olsr_compute_vtime (topology_hold_time);
	/************* end of revision *********/
	
	message->mssn=mprstable.mssn;

	//CRC v4->v6 modify
	//memcpy(&message->originator,&addr_ip,sizeof(struct olsr_ip_addr));
	//memcpy(&message->source_addr,&addr_ip,sizeof(struct olsr_ip_addr));
	
	/*memcpy(&message->originator,&addr_manet,sizeof(struct olsr_ip_addr));
	memcpy(&message->source_addr,&addr_manet,sizeof(struct olsr_ip_addr));*/ //commented by Y.Ge
  
  	/************* added by Y.Ge *************/
	memcpy(&message->originator, &my_main_addr, sizeof (struct olsr_ip_addr));
	memcpy(&message->source_addr, &addr_manet[iface_index], sizeof (struct olsr_ip_addr));
	/*-----------------*/
	if(debug_level > 3)
	{
		printf ("\n\n\n-------------------time (%02d:%02d:%02d.%06lu) node is sending TC from interface %d--------------------\n",nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec,iface_index);
		printf ("originator address: %s\n", convert_address_to_string(&message->originator));
		printf ("source address: %s\n", convert_address_to_string(&message->source_addr));
		printf ("message seqnumber: %d\n", message->packet_seq_number);
		printf ("mpr selection sequence number: %d\n", message->mssn);
		
		if (debug_level > 5)
		{
			printf ("the related tables are: \n");
			olsr_print_mpr_selector_table();
		}
		
		printf ("................mpr selector addresses included in the message ---\n");
		
		fprintf (y_file,"\n\n\n-------------------time (%02d:%02d:%02d.%06lu) node is sending TC from interface %d--------------------\n",nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec,iface_index);
		fprintf (y_file,"originator address: %s\n", convert_address_to_string(&message->originator));
		fprintf (y_file,"source address: %s\n", convert_address_to_string(&message->source_addr));
		fprintf (y_file,"message seqnumber: %d\n", message->packet_seq_number);
		fprintf (y_file,"mpr selection sequence number: %d\n", message->mssn);
		
		if (debug_level > 5)
		{
			fprintf (y_file,"the related tables are: \n");
			olsr_print_trace_mpr_selector_table();
		}
		
		fprintf (y_file,"................mpr selector addresses included in the message ---\n");
	}
	/*-----------------*/
	
	/************* end of revision ************/
	
	for(index=0;index<HASHSIZE;index++)
	{
		mprs_hash=&mprstable.mpr_selector_hash[index];
		for(mprs=mprs_hash->mpr_selector_forw;mprs!=(struct mpr_selector_entry *)mprs_hash;mprs=mprs->mpr_selector_forw)
		{
			message_mpr=(struct tc_mpr_addr *)malloc(sizeof(struct tc_mpr_addr));
			//CRC v4->v6 modify 
			memcpy(&message_mpr->address,&mprs->mpr_selector_addr,sizeof(struct olsr_ip_addr));
			if (debug_level > 3)
			{
				printf ("MPR selector: %s\n", convert_address_to_string(&message_mpr->address));
				fprintf (y_file,"MPR selector: %s\n", convert_address_to_string(&message_mpr->address));
			}
			message_mpr->next=message->multipoint_relay_selector_address;
			message->multipoint_relay_selector_address=message_mpr;
		}      
	}
	return 0;
}
/*------------------------------------------------------------------------*/
/************* added by Y.Ge ***********/
/* qos OLSR */
int
olsr_build_qos_tc_packet(struct qos_tc_message *message, int iface_index)
{
	struct tc_mpr_busy_time        *message_busy_time;
	struct mpr_selector_entry *mprs;
	struct mpr_selector_hash  *mprs_hash;
	olsr_u16_t          index;
	int i;

	message->multipoint_relay_selector_busy_time=NULL;
	
	message->packet_seq_number = msg_seqnum.tc_seqnum; 
	message->hop_count = 0;
	message->time_to_live=HOPCNT_MAX;
	
	message->mssn=mprstable.mssn;

	memcpy(&message->originator, &my_main_addr, sizeof (struct olsr_ip_addr));
	memcpy(&message->source_addr, &addr_manet[iface_index], sizeof (struct olsr_ip_addr));
	
	message->my_busy_time = my_busy_time;
	
	/*-----------------*/
	if(debug_level > 3)
	{
		printf ("\n\n\n-------------------time (%02d:%02d:%02d.%06lu) node is sending QoS TC from interface %d--------------------\n",nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec,iface_index);
		printf ("originator address: %s\n", convert_address_to_string(&message->originator));
		printf ("source address: %s\n", convert_address_to_string(&message->source_addr));
		printf ("my busy time: %d\n", message->my_busy_time);
		printf ("message seqnumber: %d\n", message->packet_seq_number);
		printf ("mpr selection sequence number: %d\n", message->mssn);
		
		if (debug_level > 5)
		{
			printf ("the related tables are: \n");
			olsr_print_mpr_selector_table();
		}
		
		printf ("................mpr selector busy time included in the message ---\n");
		
		fprintf (y_file,"\n\n\n-------------------time (%02d:%02d:%02d.%06lu) node is sending QoS TC from interface %d--------------------\n",nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec,iface_index);
		fprintf (y_file,"originator address: %s\n", convert_address_to_string(&message->originator));
		fprintf (y_file,"source address: %s\n", convert_address_to_string(&message->source_addr));
		fprintf (y_file,"my busy time: %d\n", message->my_busy_time);
		fprintf (y_file,"message seqnumber: %d\n", message->packet_seq_number);
		fprintf (y_file,"mpr selection sequence number: %d\n", message->mssn);
		
		if (debug_level > 5)
		{
			fprintf (y_file,"the related tables are: \n");
			olsr_print_trace_mpr_selector_table();
		}
		
		fprintf (y_file,"................mpr selector busy time included in the message ---\n");
	}
	/*-----------------*/
	
	for(index=0;index<HASHSIZE;index++)
	{
		mprs_hash=&mprstable.mpr_selector_hash[index];
		for(mprs=mprs_hash->mpr_selector_forw;mprs!=(struct mpr_selector_entry *)mprs_hash;mprs=mprs->mpr_selector_forw)
		{
			message_busy_time=(struct tc_mpr_busy_time *)malloc(sizeof(struct tc_mpr_busy_time));
			memcpy(&message_busy_time->busy_time,&mprs->mpr_selector_busy_time,sizeof(olsr_u32_t));
			
			message_busy_time->next=message->multipoint_relay_selector_busy_time;
			message->multipoint_relay_selector_busy_time=message_busy_time;
			
			if (debug_level > 3)
			{
				printf ("MPR selector: %s, busy time: %d\n", convert_address_to_string(&mprs->mpr_selector_addr), message_busy_time->busy_time);
				fprintf (y_file,"MPR selector: %s, busy time: %d\n", convert_address_to_string(&mprs->mpr_selector_addr), message_busy_time->busy_time);
			}
		}      
	}
	return 0;
}

/*------------------------------------------------------------------------*/
int
olsr_build_mid_packet(struct mid_message *message, int iface_index)
{
	struct mid_iface_addr	*message_mid, *mid_for_print;
	olsr_u16_t          index;
	int i;
	
	message->mid_interface_address=NULL;
	
	message->hop_count=0;
	message->time_to_live = HOPCNT_MAX;
	message->v_time = olsr_compute_vtime (mid_hold_time);
	message->packet_seq_number = msg_seqnum.mid_seqnum;
	memcpy(&message->originator_addr,&my_main_addr,sizeof(struct olsr_ip_addr));
	memcpy(&message->source_addr,&addr_manet[iface_index],sizeof(struct olsr_ip_addr)); //Do we need it? Should change later
  
	/*-----------------*/
	if(debug_level > 3)
	{
		printf ("\n\n\n-------------------time (%02d:%02d:%02d.%06lu) node is sending MID from interface %d--------------------\n",nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec, iface_index);
		printf ("originator address: %s\n", convert_address_to_string(&message->originator_addr));
		printf ("source address: %s\n", convert_address_to_string(&message->source_addr));
		printf ("message seqnumber: %d\n", message->packet_seq_number);
		printf ("................iface addresses included in the message ---\n");
		
		fprintf (y_file,"\n\n\n-------------------time (%02d:%02d:%02d.%06lu) node is sending MID from interface %d--------------------\n",nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec, iface_index);
		fprintf (y_file,"originator address: %s\n", convert_address_to_string(&message->originator_addr));
		fprintf (y_file,"source address: %s\n", convert_address_to_string(&message->source_addr));
		fprintf (y_file,"message seqnumber: %d\n", message->packet_seq_number);
		fprintf (y_file,"................iface addresses included in the message ---\n");
	}
	/*-----------------*/
	for(index=0;index<num_of_interface;index++)
	{
		if (memcmp(&addr_manet[index], &my_main_addr, sizeof(struct olsr_ip_addr))!=0)
		{
			if (debug_level > 3)
			{
				printf ("interface address: %s\n", convert_address_to_string (&addr_manet[index]));
				fprintf (y_file,"interface address: %s\n", convert_address_to_string (&addr_manet[index]));
			}
			
			message_mid=(struct mid_iface_addr *)malloc(sizeof(struct mid_iface_addr));
			memcpy(&message_mid->interface_address,&addr_manet[index],sizeof(struct olsr_ip_addr));			
			message_mid->next=message->mid_interface_address;
			message->mid_interface_address=message_mid;
			/*printf ("... after adding, the message iface address is:\n");
			mid_for_print = message->mid_interface_address;
			while (mid_for_print!=NULL)
			{
				printf ("iface address: %s\n", convert_address_to_string (&mid_for_print->interface_address));
				mid_for_print = mid_for_print->next;
			} 
			printf ("... this round of adding finished!\n");*/
		}
	}

	if (debug_level > 3)
	{
		/*printf ("\n\n.......... after adding all interface address!\n");
		mid_for_print = message->mid_interface_address;
		while (mid_for_print!=NULL)
		{
			printf ("iface address: %s\n", convert_address_to_string (&mid_for_print->interface_address));
			mid_for_print = mid_for_print->next;
		} */ 
	}    
	
	return 0;
}
/************* end of revision *************/
/*------------------------------------------------------------------------*/

int
olsr_build_routing_packet(struct rt_message **message)
{

	struct rt_entry     *destination;
	struct rthash       *routing_hash;
	olsr_u8_t           index;
	struct rt_message   *routing_message=NULL;

	for(index=0;index<HASHSIZE;index++)
	{
		routing_hash=&routingtable[index];
		for(destination=routing_hash->rt_forw;destination!=(struct rt_entry *)routing_hash;destination=destination->rt_forw)
		{
			routing_message= (struct rt_message *) malloc(sizeof(struct rt_message));
			//CRC v4->v6 modify
			memcpy(&routing_message->destination,&destination->rt_dst,sizeof(struct olsr_ip_addr));

			if(debug_level > 6)
				printf("%s            \n",convert_address_to_string(&destination->rt_dst));

			//CRC v4->v6 modify
			memcpy(&routing_message->next_hop,&destination->rt_router,sizeof(struct olsr_ip_addr));

			if(debug_level > 6)
				printf("%s            \n",convert_address_to_string(&destination->rt_router));

			routing_message->hop_count=destination->rt_metric;
			routing_message->next=*message;
			*message=routing_message;
			if (debug_level > 6)
				fprintf(stdout, "Built routing message\n");
		}
	}
	return(0);
}

/*------------------------------------------------------------------------*/
void
olsr_print_tables()
{
	olsr_print_duplicate_table();
	olsr_print_mpr_selector_table();
	olsr_print_neighbor_table();
	olsr_print_main_addr_neigh_table(); //added by Y.Ge
	olsr_print_two_hop_neighbor_table();
	olsr_print_routing_table(routingtable);
	olsr_print_topology_table();
}
/*------------------------------------------------------------------------*/
char 
*convert_sockaddr_to_string(struct sockaddr *address_to_convert)
{
	/*CRC v4->v6 del
	struct sockaddr_in      *address;
	address=(struct sockaddr_in *)address_to_convert; 
	return(inet_ntoa(address->sin_addr));*/
             
	//CRC v4->v6 add
	char* 	strptr;
	static char	str[INET6_ADDRSTRLEN];
	struct sockaddr_in6	   *address;
 
	address=(struct sockaddr_in6 *)address_to_convert; 
	strptr=(char*)inet_ntop(AF_INET6, &address->sin6_addr, str, sizeof(str));
	return(strptr);
}
/*------------------------------------------------------------------------*/
char 
*convert_address_to_string(const struct olsr_ip_addr *address)
{
	/* CRC v4->v6 del
	struct in_addr in;
	in.s_addr=*address;
	return(inet_ntoa(in));*/

	//CRC v4->v6 add
	char* 	strptr;
	static char	str[INET6_ADDRSTRLEN];
	struct in6_addr in;
	memcpy(&in.s6_addr,address, sizeof(struct in6_addr));
	strptr=(char*)inet_ntop(AF_INET6,&in,str,sizeof(str));
	return(strptr);
}
/*--------------------------------------------------------------------*/
void
olsr_init_trace_route_file()
{
	if ((f_r_id=fopen("routes","w"))==NULL)
	{
		printf("erreur\n");
		exit(0);
	}

}

/*--------------------------------------------------------------------*/
/*--------------------------------------------------------------------*/
void
olsr_init_trace_hysteresis_file()
{
	/******************
	if ((hys_id=fopen("hysteresis","w"))==NULL)
	{
		printf("Error with creating Hysteresis File.\n");
		exit(0);
	}
	*****************/
}

/*--------------------------------------------------------------------*/
/******************** the following is re-written by Y.Ge ***************/
/*float
hysteresis(float hysavg, olsr_u16_t seq_old, olsr_u16_t seq_new)
{
	int i, diff, final;
	float a,b=0.0; 
	#define MAXSEQ  65535 // ala USHRT_MAX 
	if(seq_new < seq_old)
		diff = MAXSEQ - seq_old + seq_new + 1;
	else
		diff = seq_new - seq_old;

	if(diff > 100)
		final = 100;
	else
		final = diff;

	a = hysavg;
		for (i=1; i < final; i++)
	a = (a * 0.8) + (0 * 0.2);
	a = (a * 0.8) + (1.0 * 0.2);
	return a;
}*/ //commented by Y.Ge

/* added by Y.Ge */
float
hysteresis(float hysavg, int action) // Y.Ge: action == 0: timeout, actiion ==1: time add
{
	float new_hyst;
	
	if (action)
		new_hyst = (1- HYST_SCALING) * hysavg + HYST_SCALING;
	else
		new_hyst = (1- HYST_SCALING) * hysavg;
	
	return new_hyst;
}
/***************** end of revision ****************/
/*------------------------------------------------------------------*/

//CRC HNA add
/*-----------------------------------------------------------------------------------*/
struct hna_net_addr*
olsr_lookup_default_gateway(struct hna_message* message)
{
	struct olsr_ip_addr			inaddrany = in6addr_any;
	struct hna_net_addr			*net_addr_ptr;
   	
	net_addr_ptr=message->hna_net_addrs;
	while(net_addr_ptr!=NULL)
	{
		if(0 == memcmp(&inaddrany,&(net_addr_ptr->net_addr),sizeof(struct olsr_ip_addr)))
			return (net_addr_ptr);
		net_addr_ptr = net_addr_ptr->next;   	
	}
	return NULL;
}

void
olsr_config_care_of_address(struct hna_message* message)
{
	int i;
	for(i=0;i<8;i++)
		memcpy(&addr_coa.s6_addr[i+8],&hw_addr[i],sizeof(olsr_u8_t));
	memcpy(&(addr_coa.s6_addr32[0]),&((message->originator).s6_addr32[0]),sizeof((message->originator).s6_addr32[0]));
	memcpy(&(addr_coa.s6_addr32[1]),&((message->originator).s6_addr32[1]),sizeof((message->originator).s6_addr32[1]));
	memcpy(&addr_manet,&addr_coa,sizeof(struct olsr_ip_addr));
	return;
}

void
olsr_wait_route_table_clear()
{
	int rtflags=0;
	struct interface *ifp; //added by Y.Ge
	
	printf("before ioctl(sd,SIOCSRTDEL) in olsr.c\n");
	
	for (ifp = ifnet; ifp; ifp = ifp->int_next)  //"for" loop is added by Y.Ge
	{
		for(;;)
		{
			//if((rtflags=ioctl(sd,SIOCSRTDEL,0))<0) //commented by Y.Ge
			if((rtflags=ioctl(sd[ifp->my_index],SIOCSRTDEL,0))<0) //added by Y.Ge
			{
				perror("ioctl");
				syslog(LOG_ERR, "ioctl: %m");
			}
				
			if(rtflags==1)
			{
				printf("in for cycle and rtflags==1\n");
				break;
			}
		}
		if(rtflags==1)
		{
			//if(ioctl(sd,SIOCSFLAGSET,0)<0) //commented by Y.Ge
			if(ioctl(sd[ifp->my_index],SIOCSFLAGSET,0)<0) //added by Y.Ge
			{
				perror("ioctl");
				syslog(LOG_ERR, "ioctl: %m");
			}
		}
	}
}

void
olsr_send_radvd(struct olsr_ip_addr* radvdaddr)
{
	pid_t pid;
	char addrbuf[46];
  	
	//send radvd to kernel, should be modified later!!!
	inet_ntop(AF_INET6,radvdaddr,addrbuf,sizeof(addrbuf));

	if((pid = fork())<0)
		printf("There was an error with forking.\n");
  		
	if (pid == 0)
	{
		execlp("radvd","radvd","-net",addrbuf,NULL);
		exit(0);
	} 
	else
	{
		waitpid(pid, NULL, 0);
	} 
	return;
}

void
olsr_process_received_default_gateway(struct hna_message* message)
{
	struct olsr_ip_addr	*received_gw;
	struct olsr_ip_addr *saved_gw;
	struct olsr_ip_addr	inaddrany = in6addr_any;
	
	received_gw = &(message->originator);
	saved_gw = &(saved_gateway.gw_addr);
	
	if(0==memcmp(&inaddrany,saved_gw,sizeof(struct olsr_ip_addr)))
	{
		olsr_config_care_of_address(message);
		olsr_send_radvd(received_gw);
		if(refresh==1)
		{
			olsr_wait_route_table_clear();
			rt_reset=0;
			refresh=0;
		}
		else
		{
			olsr_del_all_kernel_routes();
		}	
		
		olsr_release_tables(1);
		olsr_init_tables();
	
		memcpy(&saved_gateway.gw_addr,&message->originator,sizeof(struct olsr_ip_addr));
		timeradd(&now,&hold_time_hna,&saved_gateway.gw_timer);
		return;	
	}
	else if(0!=memcmp(saved_gw,received_gw,sizeof(struct olsr_ip_addr)))
	{
		olsr_config_care_of_address(message);
		olsr_send_radvd(received_gw);
		olsr_wait_route_table_clear();
		olsr_release_tables(1);
		olsr_init_tables();
		rt_reset=0;
		refresh=0;

		memcpy(&saved_gateway.gw_addr,&message->originator,sizeof(struct olsr_ip_addr));
		timeradd(&now,&hold_time_hna,&saved_gateway.gw_timer);
		return;
	}
	else if(rt_reset==1)
	{
		olsr_config_care_of_address(message);
		olsr_send_radvd(received_gw);
		olsr_wait_route_table_clear();
		olsr_release_tables(1);
		olsr_init_tables();
		rt_reset=0;
		
		memcpy(&saved_gateway.gw_addr,&message->originator,sizeof(struct olsr_ip_addr));
		timeradd(&now,&hold_time_hna,&saved_gateway.gw_timer);
		return;	
	}
	
	olsr_send_radvd(received_gw);
	timeradd(&now,&hold_time_hna,&saved_gateway.gw_timer);
	return;
}

/*-------------------------------------------------------------------------*/

void
olsr_time_out_gw()
{
	if(olsr_timed_out(&saved_gateway.gw_timer))
		memset(&saved_gateway,0,sizeof(struct gateway));
}

void
mobile_start_hello()
{	
	struct hello_message hellopack;
	struct interface *ifp; //added by Y.Ge

	/****************** added by Y.Ge *****************/
	for (ifp = ifnet; ifp; ifp = ifp->int_next) 
	{
		hellopack.neighbors = (struct hello_neighbor *)NULL;
		olsr_build_hello_packet(&hellopack, ifp->my_index);
		hello_build((struct sockaddr *)NULL, &hellopack, (struct tc_message *)NULL, (struct mid_message *)NULL, (struct hna_message *)NULL, 0, (struct interface *)NULL);
		tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL, (struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index);
	}
	/***************** end of revision *************/
	
	/*hellopack.neighbors = (struct hello_neighbor *)NULL;
	olsr_build_hello_packet(&hellopack);
        
	toall(hello_build, &hellopack, (struct tc_message *)NULL,(struct	hna_message *)NULL,(struct interface *)NULL);
     toall(sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL);*/ //commented by Y.Ge
		
}
/*------------------------------------------------------------------------*/
int
olsr_process_received_hna(int index, struct hna_message *message,struct hna_message **message_out) //"index" added by Y.Ge
{

  /* 
  |---------------------------------------------------------------------------------------|
  | 	Check if there is a default gateway info in this HNA. If a default                |
  |		gateway exists, check if need to process handoff. if no, process this message     | 
  |		as a general HNA according to the following steps.                                |
  |---------------------------------------------------------------------------------------|
     1. If there exist some entry in the hnahosttable whose hna_host_addr
     	corresponds to the originator address of the hna message and
     	whose hna_host_seq is greater in value than the MSSN in the received
     	message, then no further processing of this HNA message is 
     	performed and it is silently discarded (case: message 
     	received out of order).
	 
	 2. If there exist some entry in the hna table whose hna_host_addr
         corresponds to the originator address of the HNA message and
         whose hna_host_seq is lesser in value than the MSSN in the received
         message, then that hna entry is removed.
	 
	 3. For each of the subnet address received in the HNA
         message:
	 
         3.1 If there exist some entry in the hnasubnettable whose
	 hna_subnet_addr corresponds to the subnet address and the
	 hna_host_addr corresponds to the originator address of the HNA
	 message, then the holding time hna_host_time of that topology
	 entry is refreshed to HNA_HOLD_TIME.
	 
         3.2 Otherwise, a new hna entry is recorded in the
	 hna table whereas:
	 
	 - hna_subnet_addr is set to the subnet address,
	 - hna_host_addr is set to the originator address of the HNA
	 message,
	 - hna_host_seq is set to the value of MSSN received in the HNA
	 message,
	 - hna_host_timer is set to the value of HNA_HOLD_TIME.	 
  */
	struct neighbor_entry	*neighbor;
	struct olsr_ip_addr   	inaddrany6 = in6addr_any;
	struct hna_net_addr   	*def_gw;
	struct hna_host_entry	*hna_host;
	
	struct duplicate_entry	*dup_message;		//added by Y.Ge
	int should_forward=0, not_processed_yet=1;	//added by Y.Ge
  
	*message_out=NULL;
    
	if(message==NULL)
		return 1;
	
	if (debug_level > 3)
	{	
		printf ("\n\n************************* in process received hna! *************\n");
		fprintf (y_file,"\n\n************************* in process received hna! *************\n");
	}

	//if(0==memcmp(&message->originator,&addr_ip,sizeof(struct olsr_ip_addr)))
	//if(0==memcmp(&message->originator,&addr_manet,sizeof(struct olsr_ip_addr))) //commented by Y.Ge
	if(0==memcmp(&message->originator,&my_main_addr,sizeof(struct olsr_ip_addr))) //added by Y.Ge
	{
		olsr_destroy_hna_message(message);
		
		if (debug_level > 3)
		{
			printf ("XXXXXXXXXXXXXXXXXXX I'm the originator of the HNA, DROP!XXXXXXXXXXXXXXXXXX\n");
			fprintf (y_file,"XXXXXXXXXXXXXXXXXXX I'm the originator of the HNA, DROP!XXXXXXXXXXXXXXXXXX\n");
		}
		
		return 1;
	}
  	
	if(role==MOBILENODE)
	{
		if(NULL!=(def_gw=olsr_lookup_default_gateway(message)))
		{
			olsr_process_received_default_gateway(message);
		}
	}
  	
	/*Check whether  the hna is received from a sym neighbor or not.
	Be careful we should not check the originator address*/

	//if ((neighbor=olsr_lookup_neighbor_table(&message->source_addr))!=NULL) //commented by Y.Ge
	if ((neighbor=olsr_lookup_neighbor_table(&message->source_addr, &message->my_addr))!=NULL) //added by Y.Ge
	{
		//the hna message from a neighbor
		//if((neighbor->neighbor_status==NBS_ASYM)||(neighbor->neighbor_status==NBS_LOST)||(neighbor->neighbor_status==NBS_PENDING)) //commented by Y.Ge
		if (neighbor->neighbor_status != SYM_LINK) //added by Y.Ge
		{
			olsr_destroy_hna_message(message);
			
			if (debug_level > 3)
			{
				printf ("XXXXXXXXXXXXXXXXXXX HNA sender is a asymmetric neighbor, DROP!XXXXXXXXXXXXXXXXXX\n");
				fprintf (y_file,"XXXXXXXXXXXXXXXXXXX HNA sender is a asymmetric neighbor, DROP!XXXXXXXXXXXXXXXXXX\n");
			}
			
			return 1;
		}
	}
	else
	{
		/*The HNA message source address is not in the neighbor table !*/
		olsr_destroy_hna_message(message);
		
		if (debug_level > 3)
		{
			printf ("XXXXXXXXXXXXXXXXXXX HNA sender is a NOT neighbor, DROP!XXXXXXXXXXXXXXXXXX\n");
			fprintf (y_file,"XXXXXXXXXXXXXXXXXXX HNA sender is a NOT neighbor, DROP!XXXXXXXXXXXXXXXXXX\n");
		}
		
		return 1;
	}
    
	/********************** commented by Y.Ge ******************/
	/*till now, we can say that the hna message from a sym neighbor*/
	/*if(NULL!=olsr_lookup_hna_duplicate_table(message)) 	
	{
		olsr_destroy_hna_message(message);
		return 1;
	}
	//the hna message is not in the hna duplicate table, insert new entry in hna duplicate table
	olsr_insert_hna_duplicate_table(message);*/ 
	/*********************************** end of revision *****************/
	
	/********************** added by Y.Ge ******************/
	/* if we should forward the hna */
	if (debug_level > 3)
	{
	 	printf ("\n\nmy mpr selector table is: \n");
		olsr_print_mpr_selector_table();
		
		fprintf (y_file,"\n\nmy mpr selector table is: \n");
		olsr_print_trace_mpr_selector_table();
	}
	if(NULL!=(olsr_lookup_mpr_selector_table(&neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_addr)))
	{
		should_forward = 1;
		
		if (debug_level > 3)
		{
			printf ("\n********** I'm the MPR of the HNA sender!\n");
			fprintf (y_file,"\n********** I'm the MPR of the HNA sender!\n");
		}
	}
	
	/*** check the duplicate table ***/
	if ((dup_message = olsr_lookup_duplicate_table(HNA_PACKET, message->originator, message->packet_seq_number, index))!=NULL)
	{
		if (debug_level > 3)
		{
			printf ("&&&&&&&&&&&&& THE MESSAGE IS IN THE DUPLICATE TABLE! &&&&&&&&&&&&&&&&\n");
			fprintf (y_file,"&&&&&&&&&&&&& THE MESSAGE IS IN THE DUPLICATE TABLE! &&&&&&&&&&&&&&&&\n");
		
			if (debug_level > 5)
			{
				olsr_print_duplicate_table();
				olsr_print_trace_duplicate_table();
			}
		}
		
		not_processed_yet = 0; //the message has been processed
		
		/* check if the receiving interface address is in the duplicate entry */
		if (olsr_lookup_interface_address_from_duplicate_entry(dup_message, &addr_manet[index])==1)
		{
			if (debug_level > 5)
			{
				printf (" my current interface is already in the duplicate table !\n");
				olsr_print_duplicate_table();
				
				fprintf (y_file," my current interface is already in the duplicate table !\n");
				olsr_print_trace_duplicate_table();
			}
		}
		else
		{
			olsr_insert_interface_address_into_duplicate_entry(dup_message, &addr_manet[index]);
			
			if (debug_level > 5)
			{
				printf (" ........after add my current address, the duplicate table is:\n");			
				olsr_print_duplicate_table();
				
				fprintf (y_file," ........after add my current address, the duplicate table is:\n");			
				olsr_print_trace_duplicate_table();
			}
		}
		
		/* if the duplicate message is only processed, but not forwarded , then do it */
		if (should_forward ==1)
		{
			if (dup_message->duplicate_retransmitted == 0) //	should forward this message
			{
				dup_message->duplicate_retransmitted =1;
			
				if (debug_level > 5)
				{
					printf ("\n***** I did not retransmitted the message before, DO it this time\n");
					fprintf (y_file,"\n***** I did not retransmitted the message before, DO it this time\n");
				}
			}
			else //need not forward this message
			{
				olsr_destroy_hna_message(message);
			
				if (debug_level > 5)
				{
					printf ("\n***** the message has been retransmitted. DROP it!\n");
					fprintf (y_file,"\n***** the message has been retransmitted. DROP it!\n");
				}
			
				return 1;
			}
		}
		else
		{
			olsr_destroy_hna_message(message);
					
			if (debug_level > 3)
			{
				printf ("\n***** the message need NOT to be retransmitted. DROP it!\n");
				fprintf (y_file,"\n***** the message need NOT to be retransmitted. DROP it!\n");
			}
			return 1;
		}
	}
	else
	{
		/*the hna message is not in the duplicate table, insert new entry in duplicate table*/
		olsr_insert_duplicate_table(HNA_PACKET, message->originator, message->packet_seq_number, index, should_forward); 
		
		if (debug_level > 3)
		{
			printf ("\n++++++++++++++ new HNA packet, insert into duplicate table\n");
			fprintf (y_file,"\n++++++++++++++ new HNA packet, insert into duplicate table\n");
		}
	}
	/*********************************** end of revision *****************/

	
	if (not_processed_yet) //"if" added by Y.Ge
	{
		/*check if this node receive a hna message from the same host*/
		hna_host=olsr_lookup_hna_host_table(&message->originator);
  
		if(hna_host!=NULL)
		{
			if (debug_level > 3)
			{
				printf ("XXXXXXXXXXX the HNA host is in HNA table! XXXXXXXXXX\n");
				fprintf (y_file, "XXXXXXXXXXX the HNA host is in HNA table! XXXXXXXXXX\n");
			}
			
			/************** the following is commented by Y.Ge *******/
			/* in RFC, the HNA message does not have a HNA sequence */
			/*if(hna_host->hna_host_seq>message->packet_seq_number) 
			{
				printf ("HNA message TOO old!\n");
				fprintf (y_file, "HNA message TOO old!\n");
				
				olsr_destroy_hna_message(message);
				return 1;
			}
   
			if(hna_host->hna_host_seq<message->packet_seq_number)
			{
				olsr_delete_hna_host_table(hna_host);
				//We must insert the new entries contained in the received message
				hna_host=(struct hna_host_entry *)malloc(sizeof(struct hna_host_entry));
				olsr_insert_hna_host_table(hna_host,message);
				olsr_update_hna_host_table(hna_host,message);
	  		}
			else
			{		
				// This the case where hna_host->hna_seq==message->msg_seq_number, so
				// we must update the time out and create the new dest entries
				olsr_update_hna_host_table(hna_host,message);
			}*/ //end of comment
			
			
			/************** added by Y.Ge ******************/
			olsr_update_hna_host_table (hna_host, message);
			/**********************end of revision *************/
			
		}
		else
		{
			if (debug_level > 3)
			{
				printf ("XXXXXXXXXXX the HNA host is NOT in HNA table! XXXXXXXXXX\n");
				fprintf (y_file, "XXXXXXXXXXX the HNA host is NOT in HNA table! XXXXXXXXXX\n");
			}
			
			hna_host=(struct hna_host_entry *)malloc(sizeof(struct hna_host_entry));
			olsr_insert_hna_host_table(hna_host,message);
			olsr_update_hna_host_table(hna_host,message);
		}
	}
  	
	/*Check whether this node is a mpr of the previous one*/
	//if(NULL!=(olsr_lookup_mpr_selector_table(&message->source_addr))) //commented by Y.Ge
	if (should_forward) //added by Y.Ge
	{
		/*Yes we are an mpr of the previous hop */
		/*set the source_addr to the current one*/
		//memcpy(&message->source_addr,&addr_manet,sizeof(struct olsr_ip_addr)); //commented by Y.Ge

		/*send this message */
		// Don't forget to decrement the hop count in the message!
		//message->hop_count --;  //commented by Y.Ge
		
		message->time_to_live --; //added by Y.Ge
		message->hop_count ++;	//added by Y.Ge

		//if(message->hop_count>0)  //commented by Y.Ge
		if (message->time_to_live > 0) //added by Y.Ge
		{
			if (debug_level > 3)
			{
				printf ("--------------should forward this message ! -----------------\n");
				fprintf (y_file,"--------------should forward this message ! -----------------\n");
			}
			*message_out=message;
		}
		else
			olsr_destroy_hna_message(message);
       
	}
	else
		olsr_destroy_hna_message(message);      

	/* olsr_process_changes(); */ //commented by Y.Ge, DO it after all the received messages are processed
	
	if (debug_level > 2)
	{
		printf ("\n\n**************** after process received HNA, the tables are:\n");
		olsr_print_hna_host_table();
		olsr_print_hna_subnet_table();
		olsr_print_routing_table(routingtable);
		
		fprintf (y_file,"\n\n**************** after process received HNA, the tables are:\n");
		olsr_print_trace_hna_host_table();
		olsr_print_trace_hna_subnet_table();
		olsr_print_trace_routing_table(routingtable);		
	}
  
	return 0; 
}
/*-----------------------------------------------------------------------------------*/
/******** added by Y.Ge  *********/
int wrap_around_compare (int s1, int s2)
{
	if (s1 == s2)
		return 0;
		
	if ( 
		((s1 > s2) && ((s1-s2)<=MAXVALUE/2))
		||
		((s1 < s2) && ((s1-s1)>MAXVALUE/2))
	    )
	{
		return 1;
	}
	else
	{
		return -1;
	}
		
	
}

/*-----------------------------------------------------------------------------------*/
int
olsr_forward_unknown_packets (int message_type, struct olsr_ip_addr *sender, struct olsr_ip_addr *originator, int message_seq, int index)
{
	struct neighbor_entry	*neighbor;
	int	should_forward = 0;
	struct 	duplicate_entry	*dup_message;
	
	if ((neighbor=olsr_lookup_neighbor_table(sender, &addr_manet[index]))!=NULL) 
	{
		//the unknown message from a neighbor
		if (neighbor->neighbor_status != SYM_LINK)
		{
			if (debug_level > 3)
			{
				printf ("XXXXXXXXXXXXXXXXXXX UNKNOWN message sender is a asymmetric neighbor, DO NOT forward!XXXXXXXXXXXXXXXXXX\n");
				fprintf (y_file,"XXXXXXXXXXXXXXXXXXX UNKNOWN sender is a asymmetric neighbor, DO NOT forward!XXXXXXXXXXXXXXXXXX\n");
			}
			
			return 0;
		}
	}
	else
	{
		/*The message source address is not in the neighbor table !*/
		
		if (debug_level > 3)
		{
			printf ("XXXXXXXXXXXXXXXXXXX UNKNOWN message sender is a NOT neighbor, DO NOT forward!XXXXXXXXXXXXXXXXXX\n");
			fprintf (y_file,"XXXXXXXXXXXXXXXXXXX UNKNOWN message sender is a NOT neighbor, DO NOT forward!XXXXXXXXXXXXXXXXXX\n");
		}
		
		return 0;
	}
    
	/* if we should forward the message */
	if (debug_level > 3)
	{
	 	printf ("\n\nmy mpr selector table is: \n");
		olsr_print_mpr_selector_table();
		
		fprintf (y_file,"\n\nmy mpr selector table is: \n");
		olsr_print_trace_mpr_selector_table();
	}
	if(NULL!=(olsr_lookup_mpr_selector_table(&neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_addr)))
	{
		should_forward = 1;
		
		if (debug_level > 3)
		{
			printf ("\n********** I'm the MPR of the UNKNOWN message sender!\n");
			fprintf (y_file,"\n********** I'm the MPR of the UNKNOWN message sender!\n");
		}
	}
	
	/*** check the duplicate table ***/
	if ((dup_message = olsr_lookup_duplicate_table(message_type, (*originator), message_seq, index))!=NULL)
	{
		if (debug_level > 3)
		{
			printf ("&&&&&&&&&&&&& THE UNKNOWN MESSAGE IS IN THE DUPLICATE TABLE! &&&&&&&&&&&&&&&&\n");
			fprintf (y_file,"&&&&&&&&&&&&& THE MESSAGE IS IN THE DUPLICATE TABLE! &&&&&&&&&&&&&&&&\n");
		
			if (debug_level > 5)
			{
				olsr_print_duplicate_table();
				olsr_print_trace_duplicate_table();
			}
		}
		
		/* check if the receiving interface address is in the duplicate entry */
		if (olsr_lookup_interface_address_from_duplicate_entry(dup_message, &addr_manet[index])==1)
		{
			if (debug_level > 5)
			{
				printf (" my current interface is already in the duplicate table !\n");
				olsr_print_duplicate_table();
				
				fprintf (y_file," my current interface is already in the duplicate table !\n");
				olsr_print_trace_duplicate_table();
			}
		}
		else
		{
			olsr_insert_interface_address_into_duplicate_entry(dup_message, &addr_manet[index]);
			
			if (debug_level > 5)
			{
				printf (" ........after add my current address, the duplicate table is:\n");			
				olsr_print_duplicate_table();
				
				fprintf (y_file," ........after add my current address, the duplicate table is:\n");			
				olsr_print_trace_duplicate_table();
			}
		}
		
		/* if the duplicate message is only processed, but not forwarded , then do it */
		if (should_forward ==1)
		{
			if (dup_message->duplicate_retransmitted == 0) //	should forward this message
			{
				dup_message->duplicate_retransmitted =1;
			
				if (debug_level > 5)
				{
					printf ("\n***** I did not retransmitted the message before, DO it this time\n");
					fprintf (y_file,"\n***** I did not retransmitted the message before, DO it this time\n");
				}
				
				return 1;
			}
			else //need not forward this message
			{
			
				if (debug_level > 5)
				{
					printf ("\n***** the UNKNOWN message has been retransmitted. DO NOT forward it!\n");
					fprintf (y_file,"\n***** the UNKNOWN message has been retransmitted. DO NOT forward it!\n");
				}
			
				return 0;
			}
		}
		else
		{					
			if (debug_level > 3)
			{
				printf ("\n***** the message need NOT to be retransmitted. DROP it!\n");
				fprintf (y_file,"\n***** the message need NOT to be retransmitted. DROP it!\n");
			}
			return 0;
		}
	}
	else
	{
		/*the hna message is not in the duplicate table, insert new entry in duplicate table*/
		olsr_insert_duplicate_table(message_type, (*originator), message_seq, index, should_forward); 
		
		if (debug_level > 3)
		{
			printf ("\n++++++++++++++ new UNKNOWN packet, insert into duplicate table\n");
			fprintf (y_file,"\n++++++++++++++ new UNKNOWN packet, insert into duplicate table\n");
		}
		
		return should_forward;
	}
	/*********************************** end of revision *****************/

}

