/*
 * Copyright (c) 2004 Ying Ge, Communication Research Center Canada.
 *
 * Copyright (c) 2002, 2003 Maoyu Wang, Communication Research Center Canada.
 *
 * By Ying Ge:
 * 1. Change the OLSR packet format and message processing procedure based 
 *    on the OLSR RFC.
 * 2. Add support of multiple interfaces to OLSR, including MID message 
 *    creating and processing procedure as specified in the OLSR RFC.
 * 3. Add QoS Support to OLSR
 *
 * By Maoyu Wang:
 * 1. Ported OLSR from IPv4 to IPv6.
 * 2. Added the Host and Network Association (HNA) functionality into OLSR.
 * 3. Added the default gateway functionality into OLSR by extending the HNA 
 *    message usage. The default gateway functionality supported the mobility
 *	  by cooperating with Mobile IPv6 for a mobile node as well as supported 
 * 	  Internet access for MANET nodes. 
 *
 * DISTRIBUTED WITH NO WARRANTY, EXPRESS OR IMPLIED.
 * See the GNU Library General Public License (file COPYING in the distribution)
 * for conditions of use and redistribution
 */
/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est-à-dire qu'il peut
 * être utilisé, copié, modifié, distribué à l'unique
 * condition que ce texte soit conservé afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant participé à l'élaboration de ce logiciel ne peut
 * être utilisé sans son accord préalable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilité d'aucune sorte.
 * Certaines parties de ce logiciel sont dérivées de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 *
 *	$Id: olsr.h,v 1.2 2000/12/06 10:36:11 prima Exp $
 */

/*
 * Routing Table Management Daemon
 */

#ifndef _PROTOCOLS_OLSRD_H
#define	_PROTOCOLS_OLSRD_H 1

#include <sys/socket.h>

/*CRC HNA add----------------------------------------------------*/
#include <sys/types.h>		/*fork() and wait() system call*/
#include <sys/wait.h>		/*fork() and wait() system call*/
/*--------------------------------------------------------------*/

/*
 * Optimized Link State Routing Protocol
 */

/*
 * Hello packet informations
 */

struct hellinfo {
	//olsr_u8_t        	link_type;    	/* link type */ /* commented by Y.Ge */
	olsr_u8_t         	link_code;   	/* added by Y.Ge */
	olsr_u8_t         	res[1];     	/* padding   */
	olsr_u16_t        	link_msg_size;	/* link message size, added by Y.Ge */		
	//olsr_u16_t       	next_linktype;	/* offset to the next link type */ /*commented by Y.Ge */
	struct olsr_ip_addr	neigh_addr[1];	/* neighbor address (variable length)*/
};

struct hellomsg {
	olsr_u16_t     	hell_resvd [1];   	/* reserved, added by Y.Ge */
	olsr_u8_t      	hell_htime;     	/* Htime, added by Y.Ge */
	olsr_u8_t      	hell_willingness; 	/* willingness to participate, added by Y.Ge */		
	/*olsr_u16_t   	hell_seqnum;*/   	/* msg sequence number  */ /* commented by Y.Ge */
	/*olsr_u16_t   	hell_mprseqnum;*/	/* mpr sequence number  */ /* commented by Y.Ge */
	struct hellinfo	hell_info[1];   	/* variable length... */
};
 
/*
 * Topology Control packet informations
 */

struct tcmsg {
        /*olsr_u16_t  tc_seqnum;*/        	/* msg sequence number  */ /* commented by Y.Ge */
        olsr_u16_t  tc_mprseqnum;         	/* mprs sequence number  */
        /*olsr_u8_t  tc_hopcnt;*/         	/* maximum number of hops */ /* commented by Y.Ge */
        /*olsr_u8_t tc_res[3]; */         	/* pad to 32-bit boundary */ /* commented by Y.Ge */
	olsr_u8_t   tc_res[2];          	/* padding, added by Y.Ge */
        /*struct olsr_ip_addr tc_origaddr;*/ 	/* originator address */ /* commented by Y.Ge */
        struct olsr_ip_addr tc_mprsaddr[1];	/* mprs address (variable length)*/
};

/*************************** added by Y.Ge **********************/
/*
 * Multiple Interface Declaration (MID)
 */

struct midmsg {
	struct olsr_ip_addr mid_ifaceaddr[1]; /*olsr interface address (variable length) */
};

/*
 * QoS information 
 */
 
 struct qoshellomsg {
 	olsr_u32_t	hell_busy[1];	/* busy time of the hello sender */
 };
 
 struct qostcmsg{
 	olsr_u16_t	qostc_mprseqnum;	/* mprs sequence number, must be the same as the related tcmsg  */
	olsr_u8_t	qostc_res[2];  		/* padding */
	olsr_u32_t	qostc_originator_busy;	/* busy time of the TC originator */
	olsr_u32_t	tc_mpr_busy[1];		/* mpr selector's busy time */		
 };
/************************** end of revision *******************/

/*
 * ROUTE_REQUEST response packet informations
 */

struct rtmsg {
        olsr_u8_t  rt_hopcnt;			/* number of hops to destination */
        olsr_u8_t rt_res[3];            	/* pad to 32-bit boundary   */
        struct olsr_ip_addr rt_destaddr;       	/* destination address */
        struct olsr_ip_addr rt_nextaddr;	/* next hop address */
};


//CRC add for HNA
/*-----------------------------------------------------------------------------*/
/*
 * Host and Network Association packet informations
 */
struct hnaaddr{
		struct olsr_ip_addr		addr;
		olsr_u16_t			prefix;
		olsr_u8_t			res[2];
};

struct hnamsg {
        /*olsr_u16_t	hna_seqnum;		// msg sequence number 
        olsr_u16_t	hna_msgseqnum;		// the sequence number of attached network
        olsr_u8_t	hna_hopcnt;		// maximum number of hops 
        olsr_u8_t	hna_res[3];		// pad to 32-bit boundary  
        struct olsr_ip_addr hna_origaddr;	// originator address */ //commented by Y.Ge
        struct hnaaddr hna_netaddr[1];		/* network address */
};
/*-------------------------------------------------------------------------------*/

/*
 * OLSR packet
 */

struct olsrmsg {
	olsr_u8_t  		olsr_msgtype;	/* msg type */
	olsr_u8_t  		olsr_vtime;	/* added by Y.Ge, the Vtime */

	/*olsr_u8_t		olsr_broad;*/	/* msg to broadcast? */ /* commentedd by  Y.Ge */
	/*olsr_u16_t		olsr_nextmsg;*//* next link msg */	/* commented by Y.Ge */

	olsr_u16_t		olsr_msgsize;	/* added by Y.Ge , Messagge Size */
	struct olsr_ip_addr	olsr_source;	/* originator address, new added by CRC - it shold be main address*/
	olsr_u8_t  		olsr_ttl;	/* added by Y.Ge, time to live */
	olsr_u8_t  		olsr_hopcount;	/* added by Y.Ge, hop count  */
	olsr_u16_t  		olsr_msgseq;	/* added by Y.Ge, message sequence number*/
	
	union {
		struct	hellomsg	un_hello[1];	/* variable length... */
		struct	tcmsg   	un_tc[1];  	/* variable length... */
		
		//added by Y.Ge, MID
		struct	midmsg	un_mid[1];	/* variable length... */
		
		//added by Y.Ge, qos OLSR
		struct qoshellomsg	un_qoshello[1];
		struct qostcmsg     	un_qostc[1];

		//CRC add for HNA
		struct	hnamsg  	un_hna[1];     /* variable length, only need 1*/
		struct	rtmsg   	un_rt[1];		/* response to ROUTE_REQUEST */
	    	char  	        	un_tracefile[1];/* trace file for trace fonctioonalities */
		} olsrun;
		
	#define	olsr_hello	olsrun.un_hello
	#define	olsr_tc		olsrun.un_tc

	/******** added by  Y.Ge ********/
	#define	olsr_mid  	olsrun.un_mid
	#define olsr_qos_hello	olsrun.un_qoshello
	#define olsr_qos_tc   	olsrun.un_qostc
	/********* end of revision *****/

	//CRC HNA add
	#define	olsr_hna		olsrun.un_hna
	#define	olsr_rt 		olsrun.un_rt
	#define	olsr_tracefile olsrun.un_tracefile
};

struct olsr {
	olsr_u16_t	olsr_packlen;		/* packet length */
	olsr_u16_t	olsr_seqnum;		/* added  by Y.Ge, the packet sequence number */
	/*olsr_u8_t	olsr_res[2];*/		/* pad to 32-bit boundary */ /* commented by Y.Ge */
    	struct olsrmsg	olsr_msg[1];
};
 
//CRC HNA add 
struct gateway{
	struct 			olsr_ip_addr gw_addr;
	struct timeval		gw_timer;
};


struct olsr_msg_seq
{
	short	hello_seqnum;
	short	tc_seqnum;
	short	mid_seqnum;
	short	hna_seqnum;
	short	rt_seqnum;
};

/*
 * Message types.
 */
 
/* the numbers are revised by Y.Ge to be consistant with OLSR_V11 */

#define	HELLO_PACKET   		1     /* neighborhood  info */
#define	TC_PACKET      		2     /* topology information */
#define	MID_PACKET     		3     /* multiple interface Declaration */
#define	HNA_PACKET     		4     /* host and netowrk info*/

#define	HELLO_REQUEST  		5     /* to obtain neighborhood  info */
#define	ROUTE_REQUEST  		6     /* to obtain routing table contents */
#define	ROUTE_RESPONSE 		7     /* routing table information */
#define	OLSRD_TRACEON  		8     /* to enable trace fonctionnalities */
#define	OLSRD_TRACEOFF 		9     /* to desable trace fonctionnalities */
#define	UPDATE_NEIGHBOR		10    /* to update the neighborhood  */

/* qos message type, added by Y.Ge */
#define QOSHELLO_PACKET		11    /* message to inform HELLO sender's busy time */
#define QOSTC_PACKET   		12    /* message to imform TC originator and MPR selector's busy time */


#define	OLSRTYPE_MAX   		13    /*revised  by Y.Ge, original value is 10 */

#ifdef OLSRTYPES
char *olsrtypes[OLSRTYPE_MAX] =
  { "#0", "HELLO MSG", "TC MSG", "MID MSG", "HNA MSG", "HELLO REQUEST", "ROUTE REQUEST",
    "ROUTE RESPONSE", "TRACE ON MSG", "TRACE OFF MSG", 
    "UPDATE NEIGHBORHOOD MSG", "QOSHELLO MSG", "QOSTC MSG"};		/* revised  by Y.Ge based on the definition above */
#endif

#define	MSG_NOBROADCAST	0	/* don't broadcast the msg */
#define	MSG_BROADCAST		1	/* msg to broadcast */

#define	HOPCNT_MAX		255	/* changed by Y.Ge. Original value: 16. maximum hops number */
#define	MAXPACKETSIZE		256	/* max broadcast size */

#define MAXVALUE        	65535 /* added by Y.Ge, the maximum possible value of sequence numbe -- 2^16-1 */

/*CRC HNA add-------------------------------------------------*/
#define	PREFIXLEN    		64	/*network prefix length*/
#define	GW_PREFIXLEN		0

/************* added by Y.Ge **************/
/* scaling factor to compute vtime */
#define SCALING_FACTOR		16

/* constants used in hysteresis */
#define HYST_SCALING		0.5
#define HYST_THRESHOLD_HIGH	0.8
#define HYST_THRESHOLD_LOW	0.3

/*********** end of revision *************/					
					

/*
 * Timer values used in managing the routing table.
 * Neighbor tables are broadcast every HELLO_INTERVAL seconds and topology 
 * tables every TC_INTERVAL.
 * If changes occur in the MPR Selector set, the next TC information is 
 * sent after TC_MIN_INTERVAL, time starting from the time the last TC 
 * message was sent.
 * Every update of a neighbor or topology entry forces an entry's timer 
 * to be reset.
 * After .._HOLD_TIME without updates,the entry is delete.
 */

#define	TIMER_RATE      	2              /* alarm clocks every 2 seconds */
#define	HELLO_INTERVAL  	2              /* time to supply neighbor database */
#define	REFRESH_INTERVAL	2              /* added by Y.Ge */
#define	TC_INTERVAL     	5              /* time to supply topology informations */
#define	MIN_TC_INTERVAL  	2              /* min. interval to broadcast topology changes */
#define	MAX_TC_INTERVAL  	MIN_TC_INTERVAL	/* max. time to delay topology changes */
#define	HNA_INTERVAL      	TC_INTERVAL    	/* time to supple host and network info, revised by Y.Ge according to olsr_v11, the original value is 2 */					      
#define	MID_INTERVAL		TC_INTERVAL    	/* added by Y.Ge */

/*#define NEIGHB_HOLD_TIME	20*/  			/* time to delete entry in neighbor table */

#define	NEIGHB_HOLD_TIME	(3 * REFRESH_INTERVAL)	/*time to delete entry in neighbor table, revised by Y.Ge based on olsr_v11, the original value is 6 */
#define	TOP_HOLD_TIME	    	(3 * TC_INTERVAL)      	/* time to delete entry in topology table, revised by Y.Ge based on olsr_v11, the original value is 16 */
#define	DUPLICATE_HOLD_TIME	30                   	/* revised by Y.Ge based on olsr_v11, the originaal value is (2 * TC_INTERVAL)*/ 						   
#define	MID_HOLD_TIME	    	(3 * MID_INTERVAL)    	/* added by Y.Ge */
#define	MINUS_TIME         	-1                    	/* added by Y.Ge */

//CRC HNA add
#define	HNA_HOLD_TIME	    	(3 * HNA_INTERVAL)    	/* time to delete hna entry in routing table, revised by Y.Ge based on olsr_v11, the original value is 20 */
#define	HNA_DUPLICATE_HOLD_TIME 10

#endif /* protocols/olsrd.h */

struct neighbor_2_list_entry 
*olsr_find_2_hop_neighbors_with_1_link();

struct iface_addr_list_entry
*olsr_find_will_always_neighbor(); //added by Y.Ge

int
olsr_chosen_mpr(struct neighbor_entry *one_hop_neighbor,olsr_u16_t *two_hop_covered_count);

void
olsr_calculate_neighbors(olsr_u16_t *two_hop_count);

struct neighbor_entry 
*olsr_find_maximum_covered(int willingness); //"willingness" is added by Y.Ge

void
olsr_calculate_mpr();

/************** added by Y.Ge ************/
/*** compute vtime ***/
olsr_u8_t
olsr_compute_vtime (int input_time);

float
olsr_compute_time_from_vtime (olsr_u8_t vtime_value);
/************* end of revision **********/

int
olsr_timed_out(struct timeval *timer);

void
olsr_init_timer(olsr_u32_t time_value,struct timeval *hold_timer);

/************* added by Y.Ge ************/
void
olsr_assign_timer(const struct timeval *given_timer, struct timeval *change_timer);

void
timerminus (struct timeval *change_timer);

olsr_u8_t
compare_timer(const struct timeval *timer_a, const struct timeval *timer_b);
/************* end of revision **********/

void
olsr_linking_this_2_entries(struct neighbor_entry *neighbor,struct neighbor_2_entry *two_hop_neighbor);

void
olsr_linking_neighbor_main_neighbor(struct neighbor_entry *neighbor,struct main_addr_neigh_entry *main_address_neighbor); //added by Y.Ge

int
olsr_lookup_my_own_address (struct olsr_ip_addr *rcvd_interface_addr); //added by Y.Ge. IF the address is one of my interface address, RETURN 1; otherwise, RETURN 0

void
olsr_process_message_neighbors(struct neighbor_entry *neighbor,struct hello_message *message, int iface_index); //"iface_index" is addedd by Y.Ge

int
olsr_process_received_hello_no_hyst(struct hello_message *message, int index); //added by Y.Ge

int
olsr_process_received_hello_with_hyst(struct hello_message *message, int index); //added by Y.Ge

int
olsr_process_received_hello(struct hello_message *message, int index); //"index" added by Y.Ge, the interface index the HELLO is received

/* qos OLSR */
int
olsr_process_received_qos_hello (int index, struct qos_hello_message *message); //added by Y.Ge

int
olsr_process_received_qos_tc(int index, struct tc_message *message,struct tc_message **message_out,
				struct qos_tc_message *qosmessage, struct qos_tc_message **qosmessage_out); //added by Y.Ge

int
olsr_process_received_tc(int index, struct tc_message *message,struct tc_message **message_out); //"index" added by Y.Ge, the interface index the HELLO is received

int
olsr_process_received_mid (int index, struct mid_message *message, struct mid_message **message_out); //added by Y.Ge

//CRC HNA add
/*--------------------------------------------------------------------------*/
struct hna_net_addr*
olsr_lookup_default_gateway(struct hna_message* message);

void
olsr_config_care_of_address(struct hna_message* message);

void
olsr_wait_route_table_clear();

void
olsr_send_radvd(struct olsr_ip_addr* radvdaddr);

void
olsr_process_received_default_gateway(struct hna_message* message);

void
olsr_time_out_gw();

int
olsr_process_received_hna(int index, struct hna_message *message,struct hna_message **message_out);//"index" added by Y.Ge, the interface index the HNA is received
/*---------------------------------------------------------------------------------*/

void
olsr_delete_time_out_entries();

void
olsr_release_tables(int number);

void
olsr_init_tables();

int
olsr_build_hello_packet(struct hello_message *message, int iface_index); //"iface_index" is added by Y.Ge

int
olsr_build_tc_packet(struct tc_message *message, int iface_index); //"iface_index" is added by Y.Ge

int
olsr_build_qos_tc_packet(struct qos_tc_message *message, int iface_index); //added by Y.Ge, qos OLSR

int
olsr_build_mid_packet (struct mid_message *message, int iface_index); //added by Y.Ge

int
olsr_build_routing_packet(struct rt_message **message);

void
olsr_print_tables();

char 
*convert_sockaddr_to_string(struct sockaddr *address_to_convert);

void
olsr_hashing(struct olsr_ip_addr *address,olsr_u32_t *h);

char 
*convert_address_to_string(const struct olsr_ip_addr *address);

void
olsr_init_trace_route_file();

void
olsr_init_trace_hysteresis_file();

/* float hysteresis (float, short unsigned int, short unsigned int); */ //commented by Y.Ge

float hysteresis (float, int); //added by Y.Ge

int wrap_around_compare (int, int);// added by Y.Ge, IF s1 > s2, RETURN 1; IF s1 < s2 RETURN -1; IF s1 = s2, RETURN 0

int olsr_forward_unknown_packets (int message_type, struct olsr_ip_addr *sender, struct olsr_ip_addr *originator, int message_seq, int index); //added by Y.Ge
//CRC add for HNA
/*-------------------------------------------------------------------------------*/
void
olsr_time_out_hna();

void
mobile_start_hello();

void
execute_radvd();
/*-------------------------------------------------------------------------------*/
