/*
 * Copyright (c) 2004 Ying Ge, Communication Research Center Canada.
 *
 * Copyright (c) 2002, 2003 Maoyu Wang, Communication Research Center Canada.
 *
 * By Ying Ge:
 * 1. Change the OLSR packet format and message processing procedure based 
 *    on the OLSR RFC.
 * 2. Add support of multiple interfaces to OLSR, including MID message 
 *    creating and processing procedure as specified in the OLSR RFC.
 * 3. Add QoS Support to OLSR
 *
 * By Maoyu Wang:
 * 1. Ported OLSR from IPv4 to IPv6.
 * 2. Added the Host and Network Association (HNA) functionality into OLSR.
 * 3. Added the default gateway functionality into OLSR by extending the HNA 
 *    message usage. The default gateway functionality supported the mobility
 *	  by cooperating with Mobile IPv6 for a mobile node as well as supported 
 * 	  Internet access for MANET nodes. 
 *
 * DISTRIBUTED WITH NO WARRANTY, EXPRESS OR IMPLIED.
 * See the GNU Library General Public License (file COPYING in the distribution)
 * for conditions of use and redistribution
 */
/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est-à-dire qu'il peut
 * être utilisé, copié, modifié, distribué à l'unique
 * condition que ce texte soit conservé afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant participé à l'élaboration de ce logiciel ne peut
 * être utilisé sans son accord préalable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilité d'aucune sorte.
 * Certaines parties de ce logiciel sont dérivées de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 */

/*
 * Copyright (c) 1983, 1988, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * From: @(#)output.c	5.15 (Berkeley) 2/28/91
 * From: @(#)output.c	8.1 (Berkeley) 6/5/93
 * From: @(#)output.c	1.6 (dholland Exp)  1996/11/25
 */

char output_rcsid[] = 
  "$Id: output.c,v 1.2 2000/12/06 10:36:12 prima Exp $";



 

/*
 * Routing Table Management Daemon
 */


#include "defs.h"

/*
 * added by Y.Ge
 * calculate size of common olsr message head
 */
 
int comm_head_size = sizeof (msg->olsr_packlen) + sizeof (msg->olsr_seqnum);

/*
 * Apply the function "f" to all non-passive
 * interfaces.Use broadcast address.
 */

void toall(void (*f)(struct sockaddr *, struct hello_message *, struct tc_message *,struct mid_message *, struct hna_message *,int, struct interface *), struct hello_message *hmsg, struct tc_message *tmsg, struct mid_message *mmsg, struct hna_message *nmsg,struct interface *skipif)
{
	struct interface *ifp;
	/* CRC v4->v6 del
	struct sockaddr *dst;*/
	
	//CRC v4->v6 add	
	struct sockaddr *dstptr;
	struct sockaddr_in6 dst;
	
	int flags;

	/*CRC v4->v6 del
	for (ifp = ifnet; ifp; ifp = ifp->int_next) {
		if (ifp->int_flags & IFF_PASSIVE || ifp == skipif)
			continue;
		dst = ifp->int_flags & IFF_BROADCAST ? &ifp->int_broadaddr :
		      &ifp->int_addr;
		flags = ifp->int_flags & IFF_INTERFACE ? MSG_DONTROUTE : 0;

		(*f)(dst, hmsg, tmsg, flags, ifp);
	}*/
	//		printf ("...now look at inteface: %s\n",ifp->int_name);
	//	printf ("   this interface address is: %s\n", convert_address_to_string(&ifp->int6_site_addr));
	//CRC v4->v6 add
	for (ifp = ifnet; ifp; ifp = ifp->int_next) 
	{
		printf ("...now look at inteface: %s\n",ifp->int_name);
		printf ("   this interface address is: %s\n", convert_address_to_string(&ifp->int6_site_addr));
		
		if (ifp->int_flags & IFF_PASSIVE || ifp == skipif)
		{
			continue;
		}
					
		flags=0;
		/*
		flags = ifp->int_flags & IFF_INTERFACE ? MSG_DONTROUTE : 0;
		if(ifp->int_flags & IFF_MULTICAST )
			memcpy (&dst.sin6_addr, &ifp->int6_multi_addr,
			sizeof(struct in6_addr));		
		else
			memcpy (&dst.sin6_addr, &ifp->int6_site_addr,
			sizeof(struct in6_addr));*/		      
		
		memcpy (&dst.sin6_addr, &ifp->int6_multi_addr, sizeof(struct in6_addr));      
		dst.sin6_family = AF_INET6;
		dst.sin6_port = htons(OLSRPORT);
		dstptr = (struct sockaddr*)&dst;
		
		(*f)(dstptr, hmsg, tmsg, mmsg, nmsg, flags, ifp);

		//dumppacket_toscreen(dstptr, packet, outputsize, &now);
	}
	
}

/********************************* added by Y.Ge ****************/
void tointerface(void (*f)(struct sockaddr *, struct hello_message *, struct tc_message *, struct mid_message *, struct hna_message *,int, struct interface *), struct hello_message *hmsg, struct tc_message *tmsg, struct mid_message *mmsg, struct hna_message *nmsg,struct interface *skipif, int send_if_index)
{
	struct interface *ifp;
	/* CRC v4->v6 del
	struct sockaddr *dst;*/
	
	//CRC v4->v6 add	
	struct sockaddr *dstptr;
	struct sockaddr_in6 dst;
	
	int flags;
	
	//CRC v4->v6 add
	for (ifp = ifnet; ifp; ifp = ifp->int_next) 
	{
		if (ifp->my_index == send_if_index)
		{
			// printf ("...now look at inteface: %s\n",ifp->int_name);
		//	printf ("   this interface address is: %s\n", 
		//	convert_address_to_string(&ifp->int6_site_addr));//
			
			if (ifp->int_flags & IFF_PASSIVE || ifp == skipif)
			{
				continue;
			}
					
			flags=0;

			memcpy (&dst.sin6_addr, &ifp->int6_multi_addr, sizeof(struct in6_addr));      
			dst.sin6_family = AF_INET6;
			dst.sin6_port = htons(OLSRPORT);
			dstptr = (struct sockaddr*)&dst;
			
			msg->olsr_seqnum = htons(pkt_seqnum[ifp->my_index]++); 	//added by Y.Ge
		
			(*f)(dstptr, hmsg, tmsg, mmsg, nmsg, flags, ifp);

			//dumppacket_toscreen(dstptr, packet, outputsize, &now);
		}
	}
	
}
/************************ end of revision *********************/

/*
 * Output a preformed packet.
 * The two parameters message and messag are not used.
 */

void sndmsg(struct sockaddr *dst, struct hello_message *message, 
            struct tc_message *messag, struct mid_message *mmsg, struct hna_message *messg, int flags, struct interface *ifp)
{
	/*TRACE_OUTPUT(ifp, dst, outputsize);*/
	/*(*afswitch[dst->sa_family].af_output)(s6, flags,
		dst, outputsize);*/

	//inet6_output(s6, flags,	dst, outputsize);  //commented by Y.Ge
	inet6_output(s6[ifp->my_index], flags, dst, outputsize); //added by Y.Ge
}


/*
 * Supply dst with the contents of the routing tables or neighbor table.
 * If this won't fit in one packet, chop it up into several.
 */
 
void supply(struct sockaddr *dst, struct hello_message *h_packet, struct rt_message *r_packet, int type, int if_index) //"if_index" is added by Y.Ge
{
	struct interface *ifp = 0;
	
	if ( outputsize )
		toall(sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL );

	if (type == HELLO_REQUEST)
		hello_build(dst, h_packet, (struct tc_message *)NULL, (struct mid_message *)NULL, (struct hna_message*)NULL, 0,(struct interface *)NULL);

	else
	{
		if (type == ROUTE_REQUEST)
			rt_build(dst, r_packet, 0, (struct interface *)NULL);
	}
	if ( outputsize )
	{
		TRACE_OUTPUT(ifp, dst, outputsize);
		(*afswitch[dst->sa_family].af_output)(s6[if_index], 0, dst, outputsize); //add "if_index" by Y.Ge
	}

}


/*
 * Generate HELLO packet with the contents of the parameter "message".
 * If this won't fit in one packet, chop it up into several.
 * Send the packet if the size of the data contained in the output buffer
 * reach MAXPACKETSIZE. Can generate an empty HELLO packet if the 
 * neighbor table is empty. The parameter messag is not used.
 */
 
void hello_build(struct sockaddr *dst, struct hello_message *message, struct tc_message *messag,struct mid_message *mmessage, struct hna_message *messg, int flags, struct interface *ifp)
{

	/*void (*output)(int,int,struct sockaddr *,int) = afswitch[dst->sa_family].af_output;*/
	/*int remainsize = outputsize - (sizeof(msg->olsr_packlen) + sizeof(msg->olsr_res));*/ /* commented by Y.Ge */
	int remainsize = outputsize - comm_head_size;
	struct hello_neighbor *nb, *prev_nb;

	struct olsrmsg *m = outputsize ? (struct olsrmsg *)((char *)msg->olsr_msg + remainsize) : msg->olsr_msg;
	struct hellomsg *h = m->olsr_hello;
	struct hellinfo *hinfo = h->hell_info;

	//CRC v4->v6 ok
	struct olsr_ip_addr *haddr = hinfo->neigh_addr;
	int    size, i, sametype, npackets = 0, found =0;
	int lastpacket = 0; /* number of neighbors with the same greater link status in the last packet */

	if (!message)
		return;

	if (debug_level >3)
	{
		printf ("\n\n-------------- in hello build -----------------\n");
		fprintf (y_file,"\n\n-------------- in hello build -----------------\n");
		printf ("originator address: %s\n", convert_address_to_string(&message->originator_addr));
		fprintf (y_file, "originator address: %s\n", convert_address_to_string(&message->originator_addr));
		printf ("sending address: %s\n", convert_address_to_string(&addr_manet[ifp->my_index]));
		fprintf (y_file,"sending address: %s\n", convert_address_to_string(&addr_manet[ifp->my_index]));
	}

	if (outputsize > MAXPACKETSIZE - (int)sizeof (struct olsrmsg)) 
	{
		msg->olsr_packlen = htons(outputsize);
		//memset(msg->olsr_res, 0, sizeof(msg->olsr_res));  //commented by Y.Ge
		//msg->olsr_seqnum = htons(pkt_seqnum[ifp->my_index]++); 	//added by Y.Ge

		//TRACE_OUTPUT(ifp, dst, outputsize);
		//(*output)(s6, flags, dst, outputsize);
		//inet6_output(s6, flags, dst, outputsize); //commented by Y.Ge
		tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index); //added by Y.Ge
	}

	if (debug_level > 3)
	{
		printf ("     INCLUDING MESSAGE NEIGHBORS!     \n");
		fprintf (y_file,"     INCLUDING MESSAGE NEIGHBORS!     \n");
	}

	for (i = 0; i < HELLOSTATUS_MAX; i++) //revised by Y.Ge. Originally, i starts with 1
	{
		/* printf ("===== i= %d\n", i); */

		lastpacket = sametype = 0;
		outputsize = (char *)hinfo - packet;
		
		/*printf ("outputsize: %d\n, maxpacketsize: %d\n", outputsize, MAXPACKETSIZE);
		fprintf (y_file,"outputsize: %d\n, maxpacketsize: %d\n", outputsize, MAXPACKETSIZE);*/
		
		if (outputsize > MAXPACKETSIZE - (int)sizeof (struct hellinfo)) 
		{
			msg->olsr_packlen = htons(outputsize);
			//msg->olsr_seqnum = htons(pkt_seqnum[ifp->my_index]++); 	//added by Y.Ge
			//memset(msg->olsr_res, 0, sizeof(msg->olsr_res));  //commented by Y.Ge
			m->olsr_msgtype = HELLO_PACKET;
			m->olsr_vtime = olsr_compute_vtime (neighbor_hold_time); //added by Y.Ge
			//m->olsr_broad = MSG_NOBROADCAST;  //commented by Y.Ge
			//m->olsr_nextmsg = (char *)hinfo - (char *)m;  //commented by Y.Ge
			//m->olsr_nextmsg = htons(m->olsr_nextmsg);	//commented by Y.Ge
			m->olsr_msgsize = (char *)hinfo - (char *)m;  //addeded by Y.Ge
			m->olsr_msgsize = htons(m->olsr_msgsize);  //addeded by Y.Ge
			m->olsr_ttl = 1; //added by Y.Ge
			m->olsr_hopcount = 0; //added by Y.Ge
			
			/* deal with packe sequence number of the chunked HELLO */
			m->olsr_msgseq = htons(message->packet_seq_number + hello_chunk[ifp->my_index]);  //added by Y.Ge
			hello_chunk[ifp->my_index]++; //added by Y.Ge

			//CRC HNA add
			//---------------------------------------------------------------
			//memcpy(&(m->olsr_source),&addr_ip,sizeof(struct olsr_ip_addr));
			//memcpy(&(m->olsr_source),&(message->source_addr),sizeof(struct olsr_ip_addr));  //commented by Y.Ge
			memcpy(&(m->olsr_source), &(message->originator_addr), sizeof(struct olsr_ip_addr)); //added by Y.Ge
			//---------------------------------------------------------------
			
			//hinfo->next_linktype = (char *)haddr - (char *)hinfo;//commented by Y.Ge
			//hinfo->next_linktype = ntohs(hinfo->next_linktype);  //commented by Y.Ge

			hinfo->link_msg_size = (char *)haddr - (char *)hinfo;  //added by Y.Ge
			hinfo->link_msg_size = ntohs (hinfo->link_msg_size);  //added by Y.Ge

			//h->hell_seqnum = htons(hello_seqnum++);  //commented by Y.Ge
			//h->hell_mprseqnum = htons(message->mpr_seq_number);  //commented by Y.Ge

			memset(h->hell_resvd, 0, sizeof(h->hell_resvd));  //added by Y.Ge
			h->hell_htime = olsr_compute_vtime (hello_int); //added by Y.Ge
			h->hell_willingness = message->willingness; //added by Y.Ge
			
			if (debug_level > 3)
			{
				printf ("++++++++++ HELLO packet chunked!\n");
				//printf (" Message sequence: %d\n", ntohs(msg->olsr_seqnum));
				printf (" HELLO sequence: %d\n", ntohs(m->olsr_msgseq));
				printf (" HELLO packet size: %d\n", (char *)haddr - (char *)m);
				printf (" HELLO link message size: %d\n", (char *)haddr - (char *)hinfo);
				printf (" HELLO willingniss: %d\n", h->hell_willingness); 
					
				fprintf (y_file,"++++++++++ HELLO packet chunked!\n");
				//fprintf (y_file," Message sequence: %d\n", ntohs(msg->olsr_seqnum));
				fprintf (y_file," HELLO sequence: %d\n",ntohs(m->olsr_msgseq));
				fprintf (y_file," HELLO packet size: %d\n",  (char *)haddr - (char *)m);
				fprintf (y_file," HELLO link message size: %d\n", (char *)haddr - (char *)hinfo);
				fprintf (y_file," HELLO willingniss: %d\n", h->hell_willingness); 
			}

			//TRACE_OUTPUT(ifp, dst, outputsize);
			//(*output)(s6, flags, dst, outputsize);
			//inet6_output(s6, flags, dst, outputsize); //commented by Y.Ge
			tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index); //added by Y.Ge

			m = msg->olsr_msg;
			h = m->olsr_hello;
			hinfo = h->hell_info;
			haddr = hinfo->neigh_addr;
			npackets++;	
		}          
 	  
		for (nb = message->neighbors; nb != NULL; nb = nb->next) 
		{
			
			outputsize = (char *)haddr - packet;
			if (outputsize > MAXPACKETSIZE - (int)sizeof (haddr))
			{
				msg->olsr_packlen = htons(outputsize);
				//memset(msg->olsr_res, 0, sizeof(msg->olsr_res));  //commented by Y.Ge
				//msg->olsr_seqnum = htons(pkt_seqnum[ifp->my_index]++); 	//added by Y.Ge
				m->olsr_msgtype = HELLO_PACKET;
				m->olsr_vtime = olsr_compute_vtime (neighbor_hold_time); //added by Y.Ge
				//m->olsr_broad = MSG_NOBROADCAST;  //commented by Y.Ge
				//m->olsr_nextmsg = (char *)haddr - (char *)m;   //commented by Y.Ge
				//m->olsr_nextmsg = htons(m->olsr_nextmsg);      //commented by Y.Ge
				m->olsr_msgsize = (char *)haddr - (char *)m;     //added by Y.Ge
				m->olsr_msgsize = htons(m->olsr_msgsize);     //added by Y.Ge
				m->olsr_ttl = 1; //added by Y.Ge
				m->olsr_hopcount = 0; //added by Y.Ge
				
				/* deal with packe sequence number of the chunked HELLO */
				m->olsr_msgseq = htons(message->packet_seq_number + hello_chunk[ifp->my_index]);  //added by Y.Ge
				hello_chunk[ifp->my_index]++; //added by Y.Ge
			
				//CRC hna added-----------------------------------------------
				//memcpy(&(m->olsr_source),&addr_ip,sizeof(struct olsr_ip_addr));
				//memcpy(&(m->olsr_source),&(message->source_addr),sizeof(struct olsr_ip_addr)); //commented by Y.Ge
				memcpy(&(m->olsr_source), &(message->originator_addr), sizeof(struct olsr_ip_addr)); //added by Y.Ge
				//------------------------------------------------------------
				//hinfo->next_linktype = (char *)haddr - (char *)hinfo;  //commented by Y.Ge
				//hinfo->next_linktype = ntohs(hinfo->next_linktype);  //commented by Y.Ge
				//h->hell_seqnum = htons(hello_seqnum++);   //commented by Y.Ge
				//h->hell_mprseqnum = htons(message->mpr_seq_number);  //commented by Y.Ge

				hinfo->link_msg_size = (char *)haddr - (char *)hinfo;  //added by Y.Ge
				hinfo->link_msg_size = ntohs (hinfo->link_msg_size);  //added by Y.Ge

				memset(h->hell_resvd, 0, sizeof(h->hell_resvd));  //added by Y.Ge
				h->hell_htime = olsr_compute_vtime (hello_int); //added by Y.Ge
				h->hell_willingness = message->willingness; //added by Y.Ge
				
				if (debug_level > 3)
				{
					printf ("++++++++++ HELLO packet chunked!\n");
					//printf (" Message sequence: %d\n", ntohs(msg->olsr_seqnum));
					printf (" HELLO sequence: %d\n", ntohs(m->olsr_msgseq));
					printf (" HELLO packet size: %d\n", (char *)haddr - (char *)m);
					printf (" HELLO link message size: %d\n", (char *)haddr - (char *)hinfo);
					printf (" HELLO willingniss: %d\n", h->hell_willingness); 
					
					fprintf (y_file,"++++++++++ HELLO packet chunked!\n");
					//fprintf (y_file," Message sequence: %d\n", ntohs(msg->olsr_seqnum));
					fprintf (y_file," HELLO sequence: %d\n",ntohs(m->olsr_msgseq));
					fprintf (y_file," HELLO packet size: %d\n",  (char *)haddr - (char *)m);
					fprintf (y_file," HELLO link message size: %d\n", (char *)haddr - (char *)hinfo);
					fprintf (y_file," HELLO willingniss: %d\n", h->hell_willingness); 
				}

				//TRACE_OUTPUT(ifp, dst, outputsize);
				//(*output)(s6, flags, dst, outputsize);
				//inet6_output(s6, flags, dst, outputsize); //commented by Y.Ge
				tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index); //added by Y.Ge
            
				m = msg->olsr_msg;
				h = m->olsr_hello;
				hinfo = h->hell_info;
				haddr = hinfo->neigh_addr;
				//hinfo->link_type = i;  //commented by Y.Ge
				hinfo->link_code = i;	//added by Y.Ge
				lastpacket = sametype;
				npackets++;
			}

			//if (nb->status == i) //commented by Y.Ge
			if ((nb->neigh_status*4 + nb->link_status) == i) //added by Y.Ge, link_code = the combination of neigghbor status and lind status
			{
				if (debug_level > 3)
				{
					printf (" now consider neighbor %s with neighbor status: %d, link  status: %d\n", convert_address_to_string (&nb->address), nb->neigh_status, nb->link_status);
					printf ("neighbor's link code is: %d\n", i);
				
					fprintf (y_file," now consider neighbor %s with neighbor status: %d, link  status: %d\n", convert_address_to_string (&nb->address), nb->neigh_status, nb->link_status);
					fprintf (y_file,"neighbor's link code is: %d\n", i);
				}
				
				sametype++;
				if (sametype == 1)
				{
					//hinfo->link_type = i;	//commented by Y.Ge
					hinfo->link_code = i;	//added by Y.Ge
					memset(hinfo->res, 0, sizeof(hinfo->res));
				}

				//CRC v4->v6 del *haddr = nb->address;
				//CRC v4->v6
				memcpy(haddr,&(nb->address),sizeof(struct olsr_ip_addr));
				found = 1;

				haddr++;
			}
		}
       
		if (sametype && sametype > lastpacket)
		{
			//hinfo->next_linktype = (char *)haddr - (char *)hinfo;  //commented by Y.Ge
			//hinfo->next_linktype = ntohs(hinfo->next_linktype);  //commented by Y.Ge

			hinfo->link_msg_size = (char *)haddr - (char *)hinfo;  //added by Y.Ge
			hinfo->link_msg_size = ntohs (hinfo->link_msg_size);  //added by Y.Ge

			hinfo = (struct hellinfo *)((char *)haddr);
			haddr = hinfo->neigh_addr;
		}

	} /* for i*/

	//h->hell_seqnum = htons(hello_seqnum++);   //commented by Y.Ge
	//h->hell_mprseqnum = htons(message->mpr_seq_number);  //commented by Y.Ge

	memset(h->hell_resvd, 0, sizeof(h->hell_resvd));  //added by Y.Ge
	h->hell_htime = olsr_compute_vtime (hello_int); //added by Y.Ge
	h->hell_willingness = message->willingness; //added by Y.Ge
	
	m->olsr_msgtype = HELLO_PACKET;
	m->olsr_vtime = olsr_compute_vtime (neighbor_hold_time); //added by Y.Ge
	//m->olsr_broad = MSG_NOBROADCAST;  //commented by Y.Ge
	//m->olsr_nextmsg = (char *)hinfo - (char *)m;   //commented by Y.Ge
	//m->olsr_nextmsg = htons(m->olsr_nextmsg);      //commented by Y.Ge
	m->olsr_msgsize = (char *)hinfo - (char *)m;    //added by Y.Ge
	m->olsr_msgsize = htons(m->olsr_msgsize);    //added by Y.Ge
	m->olsr_ttl = 1; //added by Y.Ge
	m->olsr_hopcount = 0; //added by Y.Ge
	
	m->olsr_msgseq = htons(message->packet_seq_number + hello_chunk[ifp->my_index]);  //added by Y.Ge
	hello_chunk[ifp->my_index] ++;
	
	//CRC HNA add
	//---------------------------------------------------------------------------
	//memcpy(&(m->olsr_source),&addr_ip,sizeof(struct olsr_ip_addr));
	//memcpy(&(m->olsr_source),&(message->source_addr),sizeof(struct olsr_ip_addr)); //commented by Y.Ge
	memcpy(&(m->olsr_source), &(message->originator_addr), sizeof(struct olsr_ip_addr)); //added by Y.Ge
	//---------------------------------------------------------------------------
	
	outputsize = (char *)hinfo - packet;
	msg->olsr_packlen = htons(outputsize);
	//memset(msg->olsr_res, 0, sizeof(msg->olsr_res)); //commented by Y.Ge
	//msg->olsr_seqnum = htons(pkt_seqnum[ifp->my_index]++); 	//added by Y.Ge
	
	if (debug_level > 3)
	{
		printf ("++++++++++ HELLO packet built finished!\n");
		//printf (" Message sequence: %d\n", ntohs(msg->olsr_seqnum));
		printf (" HELLO sequence: %d\n", ntohs(m->olsr_msgseq));
		printf (" HELLO packet size: %d\n", (char *)haddr - (char *)m);
		printf (" HELLO link message size: %d\n", (char *)haddr - (char *)hinfo);
		printf (" HELLO willingniss: %d\n", h->hell_willingness); 
					
		fprintf (y_file,"++++++++++ HELLO packet built finished!\n");
		//fprintf (y_file," Message sequence: %d\n", ntohs(msg->olsr_seqnum));
		fprintf (y_file," HELLO sequence: %d\n", ntohs(m->olsr_msgseq));
		fprintf (y_file," HELLO packet size: %d\n",  (char *)haddr - (char *)m);
		fprintf (y_file," HELLO link message size: %d\n", (char *)haddr - (char *)hinfo);
		fprintf (y_file," HELLO willingniss: %d\n", h->hell_willingness); 
	}

	if (outputsize > MAXPACKETSIZE - (int)sizeof (struct olsrmsg)) 
	{
		//TRACE_OUTPUT(ifp, dst, outputsize);
		//(*output)(s6, flags, dst, outputsize);
		//inet6_output(s6, flags, dst, outputsize); //commented by Y.Ge
		tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index); //added by Y.Ge
	}

	/*
	 * Delete the list of neighbor messages.
	 */

	nb = message->neighbors;

	while (nb)
	{
		prev_nb = nb;
		nb = nb->next;
		free(prev_nb);
	}
	         
}

/*************************** added by Y.Ge ****************/
/*
 * if the node uses QoS functions,
 * after build HELLO, the node builds QOSHELLO
 * to inform its own busy time
 */
void qoshello_build(struct sockaddr *dst, int flags, struct interface *ifp)
{
	int                	remainsize = outputsize - comm_head_size;
	struct olsrmsg  	*m = outputsize ? (struct olsrmsg *)((char *)msg->olsr_msg + remainsize) : msg->olsr_msg;
	struct qoshellomsg	*qosh =m->olsr_qos_hello;
	olsr_u32_t        	*hello_busy_time = qosh->hell_busy;
	struct interface    	*interf;
	struct sockaddr   	address;
	
	if (debug_level >3)
	{
		printf ("\n\n-------------- in qos hello build -----------------\n");
		fprintf (y_file,"\n\n-------------- in qos hello build -----------------\n");
		printf ("originator address: %s\n", convert_address_to_string(&my_main_addr));
		fprintf (y_file, "originator address: %s\n", convert_address_to_string(&my_main_addr));
		printf ("sending address: %s\n", convert_address_to_string(&addr_manet[ifp->my_index]));
		fprintf (y_file,"sending address: %s\n", convert_address_to_string(&addr_manet[ifp->my_index]));
	}

	if (outputsize > MAXPACKETSIZE - (int)sizeof (struct olsrmsg)) 
	{
		msg->olsr_packlen = htons(outputsize);
		//memset(msg->olsr_res, 0, sizeof(msg->olsr_res));  //commented by Y.Ge
		//msg->olsr_seqnum = htons(pkt_seqnum[ifp->my_index]++); 	//added by Y.Ge

		//TRACE_OUTPUT(ifp, dst, outputsize);
		//(*output)(s6, flags, dst, outputsize);
		//inet6_output(s6, flags, dst, outputsize); //commented by Y.Ge
		tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index); //added by Y.Ge
	}
	
	//CRC v4->v6 add
	memset(&address, 0, sizeof(address));
	//memcpy(&(((struct sockaddr_in6 *)&address)->sin6_addr.s6_addr), &(message->originator),sizeof(struct in6_addr)); //commented by Y.Ge
	memcpy(&(((struct sockaddr_in6 *)&address)->sin6_addr.s6_addr), &addr_manet[ifp->my_index],sizeof(struct in6_addr)); //added by Y.Ge
	address.sa_family = AF_INET6;
	interf = if_ifwithaddr6(&address);
	
	if (debug_level > 3)
	{
		printf("hellobusytime: %d\n", my_busy_time);
		//printf ("outputsize: %d\n", outputsize);
	
		fprintf(y_file, "hellobusytime: %d\n", my_busy_time);
		//fprintf (y_file, "outputsize: %d\n", outputsize);
	}
	
	/* include the busy time */
	memcpy(hello_busy_time, &my_busy_time, sizeof(olsr_u32_t));
	hello_busy_time ++;
	
	/* build message headers */
	outputsize = (char *)hello_busy_time - packet;
	msg->olsr_packlen = htons(outputsize);
	//msg->olsr_seqnum = htons(pkt_seqnum[ifp->my_index]++);
	m->olsr_msgtype = QOSHELLO_PACKET;
	m->olsr_vtime = olsr_compute_vtime (neighbor_hold_time);
	m->olsr_msgsize = (char *)hello_busy_time - (char *)m;
	m->olsr_msgsize = htons(m->olsr_msgsize);
	m->olsr_ttl = 1;
	m->olsr_hopcount = 0;
	m->olsr_msgseq = htons(msg_seqnum.hello_seqnum + hello_chunk[ifp->my_index] - 1); //the qos message sequence number = the last hello sequence number
	memcpy(&(m->olsr_source), &my_main_addr, sizeof(struct olsr_ip_addr));
	
	if (debug_level > 3)
	{
		printf ("++++++++++ QoS HELLO packet built finished!\n");
		//printf (" Message sequence: %d\n", ntohs(msg->olsr_seqnum));
		printf (" QoS Hello sequence: %d\n", ntohs(m->olsr_msgseq)); 
		printf (" QoS HELLO packet size: %d\n", (char *)hello_busy_time - (char *)m);
				
		fprintf (y_file,"++++++++++ QoS HELLO packet built finished!\n");
		//fprintf (y_file," Message sequence: %d\n", ntohs(msg->olsr_seqnum));
		fprintf (y_file," QoS Hello sequence: %d\n", ntohs(m->olsr_msgseq));
		fprintf (y_file," QoS HELLO packet size: %d\n", (char *)hello_busy_time - (char *)m);
	}
	
	if (outputsize > MAXPACKETSIZE - (int)sizeof (struct olsrmsg)) 
	{
		//TRACE_OUTPUT(ifp, dst, outputsize);
		//(*output)(s6, flags, dst, outputsize);
		//inet6_output(s6, flags, dst, outputsize); //commented by Y.Ge
		tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index); //added by Y.Ge
	}

}

void qostc_build(struct sockaddr *dst, struct qos_tc_message *qosmessage, struct tc_message *message, int flags, struct interface *ifp, int relay)
{
	int remainsize = outputsize - comm_head_size;
	int outputsize_tmp = 0;
	int current_qostc_position;
	
	char packet_tmp[MAXPACKETSIZE+1];
	struct	olsr *msg_tmp = (struct olsr *)packet_tmp;
	struct olsrmsg *m_tmp = (struct olsrmsg *) msg_tmp->olsr_msg;
	struct tcmsg *tc = m_tmp->olsr_tc;	
	struct tc_mpr_addr *mprs, *prev_mprs;
	struct olsr_ip_addr *mprsaddr = tc->tc_mprsaddr;
	
	struct olsrmsg *m = outputsize ? (struct olsrmsg *)((char *)msg->olsr_msg + remainsize) : msg->olsr_msg;
	struct qostcmsg *qostc = m->olsr_qos_tc;
	struct tc_mpr_busy_time	*mprsbusy, *prev_mprsbusy;
	olsr_u32_t         *mprsbusytime = qostc->tc_mpr_busy;
	
	struct interface *interf;
	struct sockaddr address;
	int found = 0;
	int current_seq_number; 
	int i, j;

	if (debug_level > 3)
	{
		printf ("\n\n-------------- in qos tc build -----------------\n");
		fprintf (y_file,"\n\n-------------- in qos tc build -----------------\n");
		printf ("originator address: %s\n", convert_address_to_string(&message->originator));
		fprintf (y_file, "originator address: %s\n", convert_address_to_string(&message->originator));
		printf ("sending address: %s\n", convert_address_to_string(&addr_manet[ifp->my_index]));
		fprintf (y_file,"sending address: %s\n", convert_address_to_string(&addr_manet[ifp->my_index]));
	}
	
	/*printf ("mprsaddr: %d, packet_tmp: %d\n",(char *)mprsaddr, packet_tmp);
	printf("mprsbusytime: %d, packet: %d\n", (char *)mprsbusytime, packet);
	printf ("outputsize: %d\n", outputsize);
	
	printf("---address: %d\n", &address);
	printf ("--- address->sin6_addr.s6_addr: %d\n", &(((struct sockaddr_in6 *)&address)->sin6_addr.s6_addr));
	printf("1: mprsbusytime: %d, packet: %d\n", (char *)mprsbusytime, packet);
	memset(&address, 0, sizeof(address));
	printf("2: mprsbusytime: %d, packet: %d\n", (char *)mprsbusytime, packet);
	memcpy(&(((struct sockaddr_in6 *)&address)->sin6_addr.s6_addr), &(my_main_addr),sizeof(struct in6_addr)); //added by Y.Ge
	printf("3: mprsbusytime: %d, packet: %d\n", (char *)mprsbusytime, packet);
	address.sa_family = AF_INET6;
	printf("4: mprsbusytime: %d, packet: %d\n", (char *)mprsbusytime, packet);
	interf = if_ifwithaddr6(&address);
	printf("5: mprsbusytime: %d, packet: %d\n", (char *)mprsbusytime, packet);*/
	
	if (outputsize > MAXPACKETSIZE - (int)sizeof (struct olsrmsg)) 
	{
		msg->olsr_packlen = htons(outputsize);
		//msg->olsr_seqnum = htons(pkt_seqnum[ifp->my_index]++); 	//added by Y.Ge
 
		tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index); //added by Y.Ge
	}
            
	if (!message)
		return;
		
	if (!qosmessage)
		return;
		
	if(message->packet_seq_number != qosmessage->packet_seq_number)
	{
		printf("XXXXXXXXXXX something wrong with packet sequence number! XXXXXXXXX\n");
		printf("tc seq: %d, qostc seq: %d\n",message->packet_seq_number, qosmessage->packet_seq_number);
		exit(1); 
	}
		
	
	
	/*printf ("mprsaddr: %d, packet_tmp: %d\n",(char *)mprsaddr, packet_tmp);
	printf("mprsbusytime: %d, packet: %d\n", (char *)mprsbusytime, packet);
	printf ("outputsize: %d\n", outputsize);*/

	if (debug_level > 3)
	{
		printf ("include the mpr selector addresses and its busy time!\n");
		fprintf (y_file,"include the mpr selector addresses and its busy time!\n");
	}
	
	mprsbusy = qosmessage->multipoint_relay_selector_busy_time;
	
	for (mprs = message->multipoint_relay_selector_address; mprs != NULL; mprs = mprs->next) 
	{
		if (debug_level > 3)
		{
			printf ("+++ mpr selector: %s, busy time: %d\n", convert_address_to_string(&mprs->address), mprsbusy->busy_time);
			fprintf (y_file,"+++ mpr selector: %s, busy time: %d\n", convert_address_to_string(&mprs->address), mprsbusy->busy_time);
		}
	
		outputsize_tmp = (char *)mprsaddr - packet_tmp -4;		
		current_qostc_position = (char *)mprsbusytime - packet;		
		outputsize = current_qostc_position + outputsize_tmp;
		
		/*printf ("outputsize_tmp:%d, mprsaddr: %d, packet_tmp: %d\n", outputsize_tmp, (char *)mprsaddr, packet_tmp);
		printf("current_pos_position: %d, mprsbusytime: %d, packet: %d\n", current_qostc_position, (char *)mprsbusytime, packet);
		printf ("outputsize: %d\n", outputsize);*/
		
		/*fprintf (y_file, "outputsize_tmp: %d, current position: %d\n", outputsize_tmp, current_qostc_position);	
		fprintf (y_file,"outputsize: %d, maxpacketsize: %d, maxpacketsize - int: %d\n", outputsize, MAXPACKETSIZE, (MAXPACKETSIZE - (int)sizeof (mprsbusytime) - (int)sizeof(mprsaddr)));
		fprintf (y_file, "sizeof busy time: %d, sizeof addr: %d\n", (int)sizeof (mprsbusytime), (int)sizeof (mprsaddr));*/
		
		if (outputsize > (MAXPACKETSIZE - (int)sizeof (olsr_u32_t) - (int)sizeof (struct olsr_ip_addr)))
		{
			/* deal with the packet sequence number for the chunked packet */
			if (relay)
			{
				printf (" XXXXXXXX something wrong! relayed QoS TC and TC should not be chunked XXXXXX\n");
				exit (1);
			}
			current_seq_number = message->packet_seq_number + tc_chunk[ifp->my_index]; 
			tc_chunk[ifp->my_index] ++;
			
			/*m->olsr_msgseq = htons(message->packet_seq_number + tc_chunk[ifp->my_index]); 
			tc_chunk[ifp->my_index] ++;*/
			
			/* set the OLSR packet head */
			msg->olsr_packlen = htons(outputsize);
			//msg->olsr_seqnum = htons(pkt_seqnum[ifp->my_index]++);
			
			/* set the qos packet head */			
			m->olsr_msgtype = QOSTC_PACKET;
			m->olsr_vtime = message->v_time; 
			m->olsr_msgsize = (char *)mprsbusytime - (char *)m; 
			m->olsr_msgsize = htons(m->olsr_msgsize);   
			m->olsr_ttl = qosmessage->time_to_live;  
			m->olsr_hopcount = qosmessage->hop_count; 
			m->olsr_msgseq = htons(current_seq_number);			
			memcpy(&(m->olsr_source),&(qosmessage->originator),sizeof(struct olsr_ip_addr));
			qostc->qostc_mprseqnum = htons(qosmessage->mssn);
			memset(qostc->qostc_res, 0, sizeof(qostc->qostc_res));
			qostc->qostc_originator_busy = qosmessage->my_busy_time;
			
			/* set the TC packet head */ 
			m_tmp->olsr_msgtype = TC_PACKET;
			m_tmp->olsr_vtime = message->v_time;
			m_tmp->olsr_msgsize = (char *)mprsaddr - (char *)m_tmp;
			m_tmp->olsr_msgsize = htons(m_tmp->olsr_msgsize);
			m_tmp->olsr_ttl = message->time_to_live;
			m_tmp->olsr_hopcount = message->hop_count;
			m_tmp->olsr_msgseq = htons(current_seq_number);
			memcpy(&(m_tmp->olsr_source),&(message->originator),sizeof(struct olsr_ip_addr));
			tc->tc_mprseqnum = htons(message->mssn);
			memset(tc->tc_res, 0, sizeof(tc->tc_res));
			
			/* copy the TC packet from packet_tmp to packet */
			for (i=0; i<outputsize_tmp; i++)
			{
				packet[current_qostc_position+i]=packet_tmp[i+4];
			}
			
			if (debug_level > 3)
			{
				printf ("++++++++++ QoS TC packet chunked!\n");
				//printf (" Message sequence: %d\n", ntohs(msg->olsr_seqnum));
				printf (" TC sequence: %d\n", ntohs(m->olsr_msgseq));
				printf (" TC mssn: %d\n", message->mssn);
				printf (" Originator's busy time: %d\n", qostc->qostc_originator_busy);
				printf (" QoS TC packet size: %d\n", (char *)mprsbusytime - (char *)m);
				printf (" TC packet size: %d\n", (char *)mprsaddr - (char *)m_tmp);
				
				fprintf (y_file,"++++++++++ TC packet chunked!\n");
				//fprintf (y_file," Message sequence: %d\n", ntohs(msg->olsr_seqnum));
				fprintf (y_file," TC sequence: %d\n",ntohs(m->olsr_msgseq));
				fprintf (y_file," TC mssn: %d\n", message->mssn);
				fprintf (y_file," Originator's busy time: %d\n", qostc->qostc_originator_busy);
				fprintf (y_file," QoS TC packet size: %d\n", (char *)mprsbusytime - (char *)m);
				fprintf (y_file," TC packet size: %d\n", (char *)mprsaddr - (char *)m_tmp);
			}
			
			tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index);		
			m = msg->olsr_msg;
			m_tmp = msg_tmp->olsr_msg;
			qostc = m->olsr_qos_tc;
			tc = m_tmp->olsr_tc;
			mprsbusytime = qostc->tc_mpr_busy;
			mprsaddr = tc->tc_mprsaddr;
			found = 0;
		}
		
		found = 1;
		
		memcpy(mprsbusytime, &(mprsbusy->busy_time), sizeof(olsr_u32_t));
		mprsbusytime++;
		memcpy(mprsaddr,&(mprs->address),sizeof(struct olsr_ip_addr));
		mprsaddr++;
		
		mprsbusy = mprsbusy->next;
	}
	
	/*
	 * Y.Ge: TC (together with QoS TC) should be sent out in the following conditions
	 * a) mpr selector table is not NULL
	 * b) mpr selector table times out and become empty for the first time
	 * c) the TC is to be relayed
	 */
	if (found || (found ==0 && changes) || (found ==0 && relay))
	{
		/************** added by Y.Ge *************/
		if (debug_level > 3)
		{
			if (found ==0 && changes)
			{
				printf ("XXXXXXXXXXX MPR selector table becomes empty! Send out an empty QoS TC !\n");
				fprintf (y_file,"XXXXXXXXXXX MPR selector table becomes empty! Send out an empty QoS TC !\n");
			}
			
			if (found ==0 && relay)
			{
				printf ("XXXXXXXXXXX RELAY out an empty QoS TC !\n");
				fprintf (y_file,"XXXXXXXXXXX RELAY out an empty QoS TC !\n");
			}
		}
		
		outputsize_tmp = (char *)mprsaddr - (char *)m_tmp;;		
		current_qostc_position = (char *)mprsbusytime - packet;		
		outputsize = current_qostc_position + outputsize_tmp;
		
		/*printf ("outputsize_tmp:%d, mprsaddr: %d, packet_tmp: %d\n", outputsize_tmp, (char *)mprsaddr, packet_tmp);
		printf("current_pos_position: %d, mprsbusytime: %d, packet: %d\n", current_qostc_position, (char *)mprsbusytime, packet);
		printf ("outputsize: %d\n", outputsize);	
		printf ("outputsize_tmp: %d, current_qos_position: %d, outputsize: %d\n", outputsize_tmp, current_qostc_position, outputsize); */
		
		/* set the OLSR packet head */
		msg->olsr_packlen = htons(outputsize);
		//msg->olsr_seqnum = htons(pkt_seqnum[ifp->my_index]++);
		
		/* deal with packet sequence number */
		if(!relay)
		{	
			current_seq_number = message->packet_seq_number + tc_chunk[ifp->my_index]; 
			tc_chunk[ifp->my_index] ++;
		}
		
		/* set the qos packet head */			
		m->olsr_msgtype = QOSTC_PACKET;
		m->olsr_vtime = message->v_time; 
		m->olsr_msgsize = (char *)mprsbusytime - (char *)m; 
		m->olsr_msgsize = htons(m->olsr_msgsize);   
		m->olsr_ttl = qosmessage->time_to_live;  
		m->olsr_hopcount = qosmessage->hop_count; 
		if (relay)
			m->olsr_msgseq = htons(qosmessage->packet_seq_number);
		else
			m->olsr_msgseq = htons(current_seq_number);			
		memcpy(&(m->olsr_source),&(qosmessage->originator),sizeof(struct olsr_ip_addr));
		qostc->qostc_mprseqnum = htons(qosmessage->mssn);
		memset(qostc->qostc_res, 0, sizeof(qostc->qostc_res));
		qostc->qostc_originator_busy = qosmessage->my_busy_time;
		msg->olsr_packlen = htons(outputsize);
		//msg->olsr_seqnum = htons(pkt_seqnum[ifp->my_index]++);
		
		/* set the TC packet head */ 
		m_tmp->olsr_msgtype = TC_PACKET;
		m_tmp->olsr_vtime = message->v_time;
		m_tmp->olsr_msgsize = (char *)mprsaddr - (char *)m_tmp;
		m_tmp->olsr_msgsize = htons(m_tmp->olsr_msgsize);
		m_tmp->olsr_ttl = message->time_to_live;
		m_tmp->olsr_hopcount = message->hop_count;
		if (relay)
			m_tmp->olsr_msgseq = htons(message->packet_seq_number);
		else
			m_tmp->olsr_msgseq = htons(current_seq_number);
		memcpy(&(m_tmp->olsr_source),&(message->originator),sizeof(struct olsr_ip_addr));
		tc->tc_mprseqnum = htons(message->mssn);
		memset(tc->tc_res, 0, sizeof(tc->tc_res));
			
		/* copy the TC packet from packet_tmp to packet */
		for (i=0; i<outputsize_tmp; i++)
		{
			packet[current_qostc_position+i]=packet_tmp[i+4];
		}
		
		/*j=0;
		for (i=0; i<current_qostc_position; i++)
		{
			if (j==4)
			{
				printf("\n");
				j=0;
			}
			printf ("i=%d, packet: %d  ", i, packet[i]);
			j++;
		}
		printf("\n");
		
		j=0;
		for (i=0; i<outputsize_tmp; i++)
		{
			
			if (j==4)
			{
				printf("\n");
				j=0;
			}
			printf ("i=%d, packet_tmp: %d  ", i, packet_tmp[i]);
			j++;
			//packet[current_qostc_position + i] = packet_tmp[i];
		}
		printf("\n");*/
	}
	
	if (debug_level > 3)
	{
		printf ("++++++++++ QoS TC packet built finished!\n");
		//printf (" Message sequence: %d\n", ntohs(msg->olsr_seqnum));
		printf (" TC sequence: %d\n", ntohs(m->olsr_msgseq));
		printf (" TC mssn: %d\n", message->mssn);
		printf (" Originator's busy time: %d\n", qostc->qostc_originator_busy);
		printf (" QoS TC packet size: %d\n", (char *)mprsbusytime - (char *)m);
		printf (" TC packet size: %d\n", (char *)mprsaddr - (char *)m_tmp);
				
		fprintf (y_file,"++++++++++ QoS TC packet built finished!\n");
		//fprintf (y_file," Message sequence: %d\n", ntohs(msg->olsr_seqnum));
		fprintf (y_file," TC sequence: %d\n", ntohs(m->olsr_msgseq));
		fprintf (y_file," TC mssn: %d\n", message->mssn);
		fprintf (y_file," Originator's busy time: %d\n", qostc->qostc_originator_busy);
		fprintf (y_file," QoS TC packet size: %d\n", (char *)mprsbusytime - (char *)m);
		fprintf (y_file," TC packet size: %d\n", (char *)mprsaddr - (char *)m_tmp);
	}
	
	if (outputsize > MAXPACKETSIZE - (int)sizeof (struct olsrmsg)) 
	{
		tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index); //added by Y.Ge
	}

}
/********************************* end of revision *********************/

/*
 * Generate TC packet with the contents of the parameter "message".
 * If this won't fit in one packet, chop it up into several.
 * Send the packet if the size of the data contained in the output buffer
 * reach MAXPACKETSIZE. Don't generate an empty TC packet.
 * The parameter messag is not used.
 */
 
void tc_build(struct sockaddr *dst, struct hello_message *messag, struct tc_message *message, struct mid_message *mmessage, struct hna_message *messg, int flags, struct interface *ifp, int relay)           
{
	/*void (*output)(int,int,struct sockaddr *,int) = afswitch[dst->sa_family].af_output;*/

	/*int remainsize = outputsize - (sizeof(msg->olsr_packlen)  + sizeof(msg->olsr_res));*/ //commented by Y.Ge
	int remainsize = outputsize - comm_head_size;  //addedd by Y.Ge
	struct tc_mpr_addr *mprs, *prev_mprs;
	struct olsrmsg *m = outputsize ? (struct olsrmsg *)((char *)msg->olsr_msg + remainsize) : msg->olsr_msg;
	struct tcmsg *tc = m->olsr_tc;
	//CRC v4->v6 add
	struct olsr_ip_addr *mprsaddr = tc->tc_mprsaddr; 
	struct interface *interf;
	struct sockaddr address;
	int found = 0;

	if (debug_level > 3)
	{
		printf ("\n\n-------------- in tc build -----------------\n");
		fprintf (y_file,"\n\n-------------- in tc build -----------------\n");
		printf ("originator address: %s\n", convert_address_to_string(&message->originator));
		fprintf (y_file, "originator address: %s\n", convert_address_to_string(&message->originator));
		printf ("sending address: %s\n", convert_address_to_string(&addr_manet[ifp->my_index]));
		fprintf (y_file,"sending address: %s\n", convert_address_to_string(&addr_manet[ifp->my_index]));
	}
	
	if (outputsize > MAXPACKETSIZE - (int)sizeof (struct olsrmsg)) 
	{
		msg->olsr_packlen = htons(outputsize);
		//memset(msg->olsr_res, 0, sizeof(msg->olsr_res));  //commented by Y.Ge
		//msg->olsr_seqnum = htons(pkt_seqnum[ifp->my_index]++); 	//added by Y.Ge
 
		//TRACE_OUTPUT(ifp, dst, outputsize);
		//(*output)(s6, flags, dst, outputsize);
		//inet6_output(s6, flags, dst, outputsize); //commented by Y.Ge
		tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index); //added by Y.Ge
	}
            
	if (!message)
		return;
	
	/*CRC v4->v6 del
	memset(address.sa_data, 0, sizeof(address.sa_data));
	((struct sockaddr_in *)&address)->sin_addr.s_addr = message->originator;
	address.sa_family = AF_INET;
	interf = if_ifwithaddr(&address);*/

	//CRC v4->v6 add
	memset(&address, 0, sizeof(address));
	//memcpy(&(((struct sockaddr_in6 *)&address)->sin6_addr.s6_addr), &(message->originator),sizeof(struct in6_addr)); //commented by Y.Ge
	memcpy(&(((struct sockaddr_in6 *)&address)->sin6_addr.s6_addr), &addr_manet[ifp->my_index],sizeof(struct in6_addr)); //added by Y.Ge
	address.sa_family = AF_INET6;
	interf = if_ifwithaddr6(&address);

	if (debug_level > 3)
	{
		printf ("include the mpr selector addresses!\n");
		fprintf (y_file,"include the mpr selector addresses!\n");
	}
	
	for (mprs = message->multipoint_relay_selector_address; mprs != NULL; mprs = mprs->next) 
	{
		if (debug_level > 3)
		{
			printf ("+++ mpr selector: %s\n", convert_address_to_string(&mprs->address));
			fprintf (y_file,"+++ mpr selector: %s\n", convert_address_to_string(&mprs->address));
		}
		
		outputsize = (char *)mprsaddr - packet;
		
		/* printf ("+++++++ outputsize: %d\n", outputsize);
		fprintf (y_file,"+++++++ outputsize: %d, sizeof addr: %d, maxpacketsize: %d\n", outputsize,(int)sizeof (mprsaddr),MAXPACKETSIZE); */
		
		//the following IF is revised by Y.Ge. Original one: "if (outputsize > MAXPACKETSIZE - (int)sizeof (mprsaddr)))"
		if (outputsize > MAXPACKETSIZE - (int)sizeof (struct olsr_ip_addr))
		{
			msg->olsr_packlen = htons(outputsize);
			//memset(msg->olsr_res, 0, sizeof(msg->olsr_res));  //commented by Y.Ge
			//msg->olsr_seqnum = htons(pkt_seqnum[ifp->my_index]++); 	//added by Y.Ge
			m->olsr_msgtype = TC_PACKET;
			m->olsr_vtime = message->v_time; //added by Y.Ge
			//m->olsr_broad = MSG_BROADCAST;  //commented by Y.Ge
			//m->olsr_nextmsg = (char *)mprsaddr - (char *)m;   //commented by Y.Ge
			//m->olsr_nextmsg = htons(m->olsr_nextmsg);         //commented by Y.Ge
			m->olsr_msgsize = (char *)mprsaddr - (char *)m;    //added by Y.Ge
			m->olsr_msgsize = htons(m->olsr_msgsize);    //added by Y.Ge
			m->olsr_ttl = message->time_to_live;  //added by Y.Ge
			m->olsr_hopcount = message->hop_count;  //added by Y.Ge
			
			/************ added by Y.Ge *********/
			/* deal with the packet sequence number for the chunked packet */
			if (relay)
			{
				printf (" XXXXXXXX something wrong! relayed TC should not be chunked XXXXXX\n");
				exit (1);
			}
			m->olsr_msgseq = htons(message->packet_seq_number + tc_chunk[ifp->my_index]); 
			tc_chunk[ifp->my_index] ++;
			/************ end of revision ******/
			
			//CRC HNA add
			//------------------------------------------------------------------
			//memcpy(&(m->olsr_source),&addr_ip,sizeof(struct olsr_ip_addr));
			//memcpy(&(m->olsr_source),&(message->source_addr),sizeof(struct olsr_ip_addr)); //commented by Y.Ge
			memcpy(&(m->olsr_source),&(message->originator),sizeof(struct olsr_ip_addr)); //added by Y.Ge
			//-------------------------------------------------------------------
			/*if (interf)            
				//tc->tc_seqnum = htons(tc_seqnum++); 
			else
				tc->tc_seqnum = htons(message->packet_seq_number);*/ //commented by Y.Ge
			
			tc->tc_mprseqnum = htons(message->mssn);
			/*tc->tc_hopcnt = message->hop_count;*/ //commented by Y.Ge
			memset(tc->tc_res, 0, sizeof(tc->tc_res));
			//CRC v4->v6 del tc->tc_origaddr = message->originator;
			//CRC v4->v6 add
			/*memcpy(&(tc->tc_origaddr),&(message->originator),sizeof(struct olsr_ip_addr));*/ //commented by Y.Ge
			
			if (debug_level > 3)
			{
				printf ("++++++++++ TC packet chunked!\n");
				//printf (" Message sequence: %d\n", ntohs(msg->olsr_seqnum));
				printf (" TC sequence: %d\n", ntohs(m->olsr_msgseq));
				printf (" TC packet size: %d\n", (char *)mprsaddr - (char *)m);
				
				fprintf (y_file,"++++++++++ TC packet chunked!\n");
				//fprintf (y_file," Message sequence: %d\n", ntohs(msg->olsr_seqnum));
				fprintf (y_file," TC sequence: %d\n",ntohs(m->olsr_msgseq));
				fprintf (y_file," TC packet size: %d\n", (char *)mprsaddr - (char *)m);
			}

		  
			//TRACE_OUTPUT(ifp, dst, outputsize);
			//(*output)(s6, flags, dst, outputsize);
			//inet6_output(s6, flags, dst, outputsize); //commented by Y.Ge
			tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index); //added by Y.Ge 	
		
			m = msg->olsr_msg;
			tc = m->olsr_tc;
			mprsaddr = tc->tc_mprsaddr;
			found = 0;
		}
		found = 1;
 
		//CRC v4->v6 del *mprsaddr = mprs->address;
		//CRC v4->v6 add
		memcpy(mprsaddr,&(mprs->address),sizeof(struct olsr_ip_addr));
		mprsaddr++;
	}
	//if (found) //commented by Y.Ge
	
	/*
	 * Y.Ge: TC should be sent out in the following conditions
	 * a) mpr selector table is not NULL
	 * b) mpr selector table times out and become empty for the first time
	 * c) the TC is to be relayed
	 */
	if (found || (found ==0 && changes) || (found ==0 && relay))  //added by Y.Ge
	{
		/************** added by Y.Ge *************/
		if (debug_level > 3)
		{
			if (found ==0 && changes)
			{
				printf ("XXXXXXXXXXX MPR selector table becomes empty! Send out an empty TC !\n");
				fprintf (y_file,"XXXXXXXXXXX MPR selector table becomes empty! Send out an empty TC !\n");
			}
			
			if (found ==0 && relay)
			{
				printf ("XXXXXXXXXXX RELAY out an empty TC !\n");
				fprintf (y_file,"XXXXXXXXXXX RELAY out an empty TC !\n");
			}
		}
		outputsize = (char *)mprsaddr - packet;
		msg->olsr_packlen = htons(outputsize);
		//memset(msg->olsr_res, 0, sizeof(msg->olsr_res));  //commented by Y.Ge
		//msg->olsr_seqnum = htons(pkt_seqnum[ifp->my_index]++); 	//added by Y.Ge
		m->olsr_msgtype = TC_PACKET;
		m->olsr_vtime = message->v_time; //added by Y.Ge
		//m->olsr_broad = MSG_BROADCAST;  //commented by Y.Ge
		//m->olsr_nextmsg = (char *)mprsaddr - (char *)m;   //commented by Y.Ge
		//m->olsr_nextmsg = htons(m->olsr_nextmsg);         //commented by Y.Ge
		m->olsr_msgsize = (char *)mprsaddr - (char *)m;   //addeded by Y.Ge
		m->olsr_msgsize = htons(m->olsr_msgsize);         //addeded by Y.Ge
		m->olsr_ttl = message->time_to_live;  //added by Y.Ge
		m->olsr_hopcount = message->hop_count;  //added by Y.Ge
		
		/************ added by Y.Ge *********/
		if (relay)
			m->olsr_msgseq = htons(message->packet_seq_number);
		else
		{
			m->olsr_msgseq = htons(message->packet_seq_number + tc_chunk[ifp->my_index]);
			tc_chunk[ifp->my_index] ++;
		}
		/************ end of revision ******/
			
		//CRC HNA add
		//-----------------------------------------------------------
		//memcpy(&(m->olsr_source),&addr_ip,sizeof(struct olsr_ip_addr));
		//memcpy(&(m->olsr_source),&(message->source_addr),sizeof(struct olsr_ip_addr)); //commented by Y.Ge
		memcpy(&(m->olsr_source),&(message->originator),sizeof(struct olsr_ip_addr)); //added by Y.Ge
		//------------------------------------------------------------
		
		/*if (interf)            
			tc->tc_seqnum = htons(tc_seqnum++);
		else
			tc->tc_seqnum = htons(message->packet_seq_number);*/ //commented by Y.Ge

		tc->tc_mprseqnum = htons(message->mssn);
		/*tc->tc_hopcnt = message->hop_count;*/ //commented by Y.Ge
		memset(tc->tc_res, 0, sizeof(tc->tc_res));
		//CRC v4->v6 del tc->tc_origaddr = message->originator;
		//CRC v4->v6 add
		/*memcpy(&(tc->tc_origaddr),&(message->originator),sizeof(struct olsr_ip_addr));*/ //commented by Y.Ge
	}
	
	if (debug_level > 3)
	{
		printf ("++++++++++ TC packet built finished!\n");
		//printf (" Message sequence: %d\n", ntohs(msg->olsr_seqnum));
		printf (" TC sequence: %d\n", ntohs(m->olsr_msgseq));
		printf (" TC mssn: %d\n", message->mssn);
		printf (" TC packet size: %d\n", (char *)mprsaddr - (char *)m);
				
		fprintf (y_file,"++++++++++ TC packet built finished!\n");
		//fprintf (y_file," Message sequence: %d\n", ntohs(msg->olsr_seqnum));
		//fprintf (y_file," TC sequence: %d\n", message->packet_seq_number + tc_chunk[ifp->my_index] -1 );
		fprintf (y_file," TC sequence: %d\n", ntohs(m->olsr_msgseq));
		fprintf (y_file," TC mssn: %d\n", message->mssn);
		fprintf (y_file," TC packet size: %d\n", (char *)mprsaddr - (char *)m);
	}


	if (outputsize > MAXPACKETSIZE - (int)sizeof (struct olsrmsg)) 
	{
		//TRACE_OUTPUT(ifp, dst, outputsize);
		//(*output)(s6, flags, dst, outputsize);
		//inet6_output(s6, flags, dst, outputsize); //commented by Y.Ge
		tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index); //added by Y.Ge
	}

	/*
	 * commented by Y.Ge. 
	 * This is now done in timer(), as node with multiple interface should build same TC andd MID.
	 * it is not necessary to call this function in all interfaces
	 */ 
	  
	/*
 	 * Delete the list of mprs messages
 	 */
	/*mprs = message->multipoint_relay_selector_address;

	while (mprs)
	{
		prev_mprs = mprs;
		mprs = mprs->next;
		free(prev_mprs);
	}*/ 
}

//CRC HNA add
/*---------------------------------------------------------------------------*/
/*
 * Generate HNA packet with the contents of the parameter "message".
 * Send the packet if the size of the data contained in the output buffer
 * reach MAXPACKETSIZE. Don't generate an empty HNA packet.
 * The parameter messag, messg is not used.
 */
 
void hna_build(struct sockaddr *dst, struct hello_message *messag,struct tc_message *messg, struct mid_message *mmessage, struct hna_message *message, int flags, struct interface *ifp, int relay)           
{
	/*int remainsize = outputsize - (sizeof(msg->olsr_packlen) + sizeof(msg->olsr_res));*/  //commented by Y.Ge
	int remainsize = outputsize - comm_head_size;
	struct hna_net_addr *nodeptr;
	struct olsrmsg 	*m = outputsize ? (struct olsrmsg *)((char *)msg->olsr_msg + remainsize): msg->olsr_msg;
	struct hnamsg 		*hna = m->olsr_hna;
	struct hnaaddr 	*hnaaddrs = hna->hna_netaddr; 
	struct interface 	*interf;
	struct sockaddr 	address;
	int found = 0;
	int more = 0;
	
	if (outputsize > MAXPACKETSIZE - (int)sizeof (struct olsrmsg)) 
	{
		msg->olsr_packlen = htons(outputsize);
		//memset(msg->olsr_res,0,sizeof(msg->olsr_res)); //commented by Y.Ge
		//msg->olsr_seqnum = htons(pkt_seqnum[ifp->my_index]++); 	//added by Y.Ge
		//inet6_output(s6, flags, dst, outputsize);	//commented by Y.Ge
		tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index); //added by Y.Ge
	}
            
	if (!message)
		return;
	
	if (!message->hna_net_addrs)
		return;
	
	if (debug_level > 3)
	{	
		printf ("\n\n-------------- in hna build -----------------\n");
		fprintf (y_file,"\n\n-------------- in hna build -----------------\n");
		printf ("--- HNA sequence number: %d\n",message->packet_seq_number);
		fprintf (y_file,"--- HNA sequence number: %d\n",message->packet_seq_number);
		printf ("originator address: %s\n", convert_address_to_string(&message->originator));
		fprintf (y_file, "originator address: %s\n", convert_address_to_string(&message->originator));
		printf ("sending address: %s\n", convert_address_to_string(&addr_manet[ifp->my_index]));
		fprintf (y_file,"sending address: %s\n", convert_address_to_string(&addr_manet[ifp->my_index]));
	}

	memset(&address, 0, sizeof(address));
	/*memcpy(&(((struct sockaddr_in6 *)&address)->sin6_addr.s6_addr),&(message->originator),sizeof(struct in6_addr));*/ //commented by Y.Ge
	memcpy(&(((struct sockaddr_in6 *)&address)->sin6_addr.s6_addr),&addr_manet[ifp->my_index],sizeof(struct in6_addr)); //added by Y.Ge
	address.sa_family = AF_INET6;
	interf = if_ifwithaddr6(&address);
	
	for (nodeptr = message->hna_net_addrs; nodeptr!= NULL; nodeptr = nodeptr->next) 
	{
		/* if (more ==1 && wrap_around_compare(message->packet_seq_number,3))
			continue;		
		more =1; */
		 	
		outputsize = (char *)hnaaddrs - packet;
		//printf (" outputsize: %d\n", outputsize);
		
		//the following IF is revised by Y.Ge,original one: "if (outputsize > MAXPACKETSIZE - (int)sizeof (hnaaddrs))"
		if (outputsize > MAXPACKETSIZE - (int)sizeof (struct hnaaddr))
		{
			msg->olsr_packlen = htons(outputsize);
			//memset(msg->olsr_res, 0, sizeof(msg->olsr_res));  //commented by Y.Ge
			//msg->olsr_seqnum = htons(pkt_seqnum[ifp->my_index]++); 	//added by Y.Ge
			m->olsr_msgtype = HNA_PACKET;
			m->olsr_vtime = message->v_time; //added by Y.Ge
			//m->olsr_broad = MSG_BROADCAST;  //commented by Y.Ge
			//m->olsr_nextmsg = (char *)hnaaddrs - (char *)m;  //commented by Y.Ge
			//m->olsr_nextmsg = htons(m->olsr_nextmsg);   //commented by Y.Ge
			m->olsr_msgsize = (char *)hnaaddrs - (char *)m;  //commented by Y.Ge
			m->olsr_msgsize = htons(m->olsr_msgsize);   //commented by Y.Ge
			m->olsr_ttl = message->time_to_live;	//added by Y.Ge
			m->olsr_hopcount = message->hop_count;	//added by Y.Ge
			
			/*********** added by Y.Ge ***********/
			/* deal with sequence number of the chunked packet */
			if (relay)
			{
				printf (" XXXXXXXX something wrong! relayed HNA should not be chunked XXXXXX\n");
				exit (1);
			}
			m->olsr_msgseq = htons(message->packet_seq_number + hna_chunk[ifp->my_index]);
			hna_chunk[ifp->my_index] ++ ;
			/************** end of revision *********/
	    
			//CRC HNA add
			//-------------------------------------------------------------
			//memcpy(&(m->olsr_source),&addr_ip,sizeof(struct olsr_ip_addr));
			//memcpy(&(m->olsr_source),&(message->source_addr),sizeof(struct olsr_ip_addr)); //commented by Y.Ge
			memcpy(&(m->olsr_source),&(message->originator),sizeof(struct olsr_ip_addr)); //added by Y.Ge
			//------------------------------------------------------------
		  	
			/*if (interf)            
				hna->hna_seqnum = htons(hna_seqnum++);
			else
				hna->hna_seqnum = htons(message->packet_seq_number);*/ //commented by Y.Ge

			/*hna->hna_msgseqnum = htons(message->msg_seq_number);
			hna->hna_hopcnt = message->hop_count;
			memset(hna->hna_res, 0, sizeof(hna->hna_res));
			memcpy(&(hna->hna_origaddr),&(message->originator),sizeof(struct olsr_ip_addr));*/ //commented by Y.Ge
		  
		  	if (debug_level > 3)
			{
				printf ("++++++++++ HNA packet chunked!\n");
				//printf (" Message sequence: %d\n", ntohs(msg->olsr_seqnum));
				printf (" HNA sequence: %d\n", ntohs(m->olsr_msgseq));
				printf (" HNA packet size: %d\n", (char *)hnaaddrs - (char *)m);
				
				fprintf (y_file,"++++++++++ HNA packet chunked!\n");
				//fprintf (y_file," Message sequence: %d\n",ntohs(msg->olsr_seqnum));
				fprintf (y_file," HNA sequence: %d\n", ntohs(m->olsr_msgseq));
				fprintf (y_file," HNA packet size: %d\n", (char *)hnaaddrs - (char *)m);
			}

			//inet6_output(s6, flags, dst, outputsize); //commented by Y.Ge
			tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index); //added by Y.Ge
		
			m = msg->olsr_msg;
			hna = m->olsr_hna;
			hnaaddrs = hna->hna_netaddr;
			found = 0;
		}
		found = 1;
 
		memcpy(&(hnaaddrs->addr),&(nodeptr->net_addr),sizeof(struct olsr_ip_addr));
		hnaaddrs->prefix=htons(nodeptr->prefix);
		memset(hnaaddrs->res,0,sizeof(hnaaddrs->res));
		
		if (debug_level >3)
		{
			printf ("--- hna address: %s, prefix:%d\n", convert_address_to_string(&hnaaddrs->addr), hnaaddrs->prefix);
			fprintf (y_file,"--- hna address: %s, prefix:%d\n", convert_address_to_string(&hnaaddrs->addr), hnaaddrs->prefix);
		}
		
		hnaaddrs++;
	}

	if (found)
	{
		outputsize = (char *)hnaaddrs - packet;
		msg->olsr_packlen = htons(outputsize);
		//memset(msg->olsr_res, 0, sizeof(msg->olsr_res));  //commented by Y.Ge
		//msg->olsr_seqnum = htons(pkt_seqnum[ifp->my_index]++); 	//added by Y.Ge
		m->olsr_msgtype = HNA_PACKET;
		m->olsr_vtime = message->v_time; //added by Y.Ge
		//m->olsr_broad = MSG_BROADCAST;  //commented by Y.Ge
		//m->olsr_nextmsg = (char *)hnaaddrs - (char *)m;  //commented by Y.Ge
		//m->olsr_nextmsg = htons(m->olsr_nextmsg);   //commented by Y.Ge
		m->olsr_msgsize = (char *)hnaaddrs - (char *)m;  //added by Y.Ge
		m->olsr_msgsize = htons(m->olsr_msgsize);   //addeded by Y.Ge
		m->olsr_ttl = message->time_to_live;	//added by Y.Ge
		m->olsr_hopcount = message->hop_count;	//added by Y.Ge
		
		/************* added by Y.Ge *********/
		if (relay)
			m->olsr_msgseq = htons(message->packet_seq_number);
		else
		{	
			m->olsr_msgseq = htons(message->packet_seq_number + hna_chunk[ifp->my_index]);
			hna_chunk[ifp->my_index] ++;
		}
		/************* end of revision *******/
		
		/*printf ("now, the hna sequence is:%d\n", hna_chunk[ifp->my_index]);*/
	
		//CRC HNA add
		//---------------------------------------------------------------
		//memcpy(&(m->olsr_source),&addr_ip,sizeof(struct olsr_ip_addr));
		//memcpy(&(m->olsr_source),&(message->source_addr),sizeof(struct olsr_ip_addr)); //commented by Y.Ge
		memcpy(&(m->olsr_source),&(message->originator),sizeof(struct olsr_ip_addr)); //added by Y.Ge
		//---------------------------------------------------------------
		/*if (interf)            
			hna->hna_seqnum = htons(hna_seqnum++);
		else
			hna->hna_seqnum = htons(message->packet_seq_number);*/ //commented by Y.Ge
        
		/*hna->hna_msgseqnum = htons(message->msg_seq_number);
		hna->hna_hopcnt = message->hop_count;
		memset(hna->hna_res, 0, sizeof(hna->hna_res));
		//CRC v4->v6 del tc->tc_origaddr = message->originator;
		//CRC v4->v6 add
		memcpy(&(hna->hna_origaddr),&(message->originator),sizeof(struct olsr_ip_addr));*/ //commented by Y.Ge
	}
	
	if (debug_level > 3)
	{
		printf ("++++++++++ HNA packet built finished!\n");
		//printf (" Message sequence: %d\n", ntohs(msg->olsr_seqnum));
		printf (" HNA sequence: %d\n", ntohs(m->olsr_msgseq));
		printf (" HNA packet size: %d\n", (char *)hnaaddrs - (char *)m);
				
		fprintf (y_file,"++++++++++ HNA packet built finished!\n");
		//fprintf (y_file," Message sequence: %d\n", ntohs(msg->olsr_seqnum));
		fprintf (y_file," HNA sequence: %d\n", ntohs(m->olsr_msgseq));
		fprintf (y_file," HNA packet size: %d\n", (char *)hnaaddrs - (char *)m);
	}

	if (outputsize > MAXPACKETSIZE - (int)sizeof (struct olsrmsg)) 
	{
		//inet6_output(s6, flags, dst, outputsize); //commented by Y.Ge
		tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index); //added by Y.Ge
	}
}

/*---------------------------------------------------------------------------*/
/****************************** added by Y.Ge ***********************/
/*
 * Generate MID packet with the contents of the parameter "message".
 * Send the packet if the size of the data contained in the output buffer
 * reach MAXPACKETSIZE. Don't generate an empty MID packet.
 * The parameter messag, messg, mess is not used.
 */
 
void mid_build(struct sockaddr *dst, struct hello_message *messg,struct tc_message *messag, struct mid_message *message, struct hna_message *mess,int flags, struct interface *ifp, int relay)           
{
	int remainsize = outputsize - comm_head_size;
	struct mid_iface_addr	*addressptr;
	struct olsrmsg      	*m = outputsize ? (struct olsrmsg *)((char *)msg->olsr_msg + remainsize): msg->olsr_msg;
	struct midmsg 	     	*mid = m->olsr_mid;
	struct olsr_ip_addr 	*ifaceaddrs = mid->mid_ifaceaddr; 
	struct interface    	*interf;
	struct sockaddr     	address;
	int found = 0;

	if (debug_level > 3)
	{
		printf ("\n\n-------------- in mid build -----------------\n");
		fprintf (y_file,"\n\n-------------- in mid build -----------------\n");
		printf ("originator address: %s\n", convert_address_to_string(&message->originator_addr));
		fprintf (y_file, "originator address: %s\n", convert_address_to_string(&message->originator_addr));
		printf ("sending address: %s\n", convert_address_to_string(&addr_manet[ifp->my_index]));
		fprintf (y_file,"sending address: %s\n", convert_address_to_string(&addr_manet[ifp->my_index]));
	}
	
	if (outputsize > MAXPACKETSIZE - (int)sizeof (struct olsrmsg)) 
	{
		msg->olsr_packlen = htons(outputsize);
		//msg->olsr_seqnum = htons(pkt_seqnum[ifp->my_index]++); 
		tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index);
	}
            
	if (!message)
	{
		printf ("message = NULL!\n");
		return;
	}
	
	if (!message->mid_interface_address)
	{
		printf (" message->mid_interface_address = NULL!\n");
		return;
	}

	memset(&address, 0, sizeof(address));
	memcpy(&(((struct sockaddr_in6 *)&address)->sin6_addr.s6_addr),&addr_manet[ifp->my_index],sizeof(struct in6_addr));
	address.sa_family = AF_INET6;
	interf = if_ifwithaddr6(&address);

	if (debug_level > 3)
	{
		printf ("include the interface addresses!\n");
		fprintf (y_file,"include the interface addresses!\n");
	}
	
	for (addressptr = message->mid_interface_address; addressptr!= NULL; addressptr = addressptr->next) 
	{
		if (debug_level > 3)
		{
			printf (" ++ interface address: %s\n", convert_address_to_string (&addressptr->interface_address));
			fprintf (y_file," ++ interface address: %s\n", convert_address_to_string (&addressptr->interface_address));
		}
		
		outputsize = (char *)ifaceaddrs - packet;
		if (outputsize > MAXPACKETSIZE - (int)sizeof (struct olsr_ip_addr))
		{
			msg->olsr_packlen = htons(outputsize);
			//memset(msg->olsr_res, 0, sizeof(msg->olsr_res));  
			//msg->olsr_seqnum = htons(pkt_seqnum[ifp->my_index]++); 	
			m->olsr_msgtype = MID_PACKET;
			m->olsr_vtime = message->v_time; 
			//m->olsr_broad = MSG_BROADCAST;  
			//m->olsr_nextmsg = (char *)hnaaddrs - (char *)m;  
			//m->olsr_nextmsg = htons(m->olsr_nextmsg);  
			m->olsr_msgsize = (char *)ifaceaddrs - (char *)m; 
			m->olsr_msgsize = htons(m->olsr_msgsize); 
			m->olsr_ttl = message->time_to_live;
			m->olsr_hopcount = message->hop_count;
			
			/**************** added by Y.Ge ************/
			if (relay)
			{
				printf (" XXXXXXXX something wrong! relayed MID should not be chunked XXXXXX\n");
				exit (1);
			}
			m->olsr_msgseq = htons (message->packet_seq_number + mid_chunk[ifp->my_index]); 
			mid_chunk[ifp->my_index] ++;
			/*************** end of revision *********/
			
			//-------------------------------------------------------------
			//memcpy(&(m->olsr_source),&addr_ip,sizeof(struct olsr_ip_addr));
			//memcpy(&(m->olsr_source),&(message->source_addr),sizeof(struct olsr_ip_addr)); 
			memcpy (&(m->olsr_source), &(message->originator_addr), sizeof(struct olsr_ip_addr)); 
			//------------------------------------------------------------

			//hna->hna_msgseqnum = htons(message->msg_seq_number);
			//hna->hna_hopcnt = message->hop_count;
			//memset(hna->hna_res, 0, sizeof(hna->hna_res));
			//memcpy(&(hna->hna_origaddr),&(message->originator),sizeof(struct olsr_ip_addr));
			
			if (debug_level > 3)
			{
				printf ("++++++++++ MID packet chunked!\n");
				//printf (" Message sequence: %d\n", ntohs(msg->olsr_seqnum));
				printf (" MID sequence: %d\n", ntohs(m->olsr_msgseq));
				printf (" MID packet size: %d\n", (char *)ifaceaddrs - (char *)m);
				
				fprintf (y_file,"++++++++++ MID packet chunked!\n");
				//fprintf (y_file," Message sequence: %d\n", ntohs(msg->olsr_seqnum));
				fprintf (y_file," MID sequence: %d\n", ntohs(m->olsr_msgseq));
				fprintf (y_file," MID packet size: %d\n", (char *)ifaceaddrs - (char *)m);
			}
		  
			//inet6_output(s6, flags, dst, outputsize); //commented by Y.Ge
			tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index); //added by Y.Ge
		
			m = msg->olsr_msg;
			mid = m->olsr_mid;
			ifaceaddrs = mid->mid_ifaceaddr;
			found = 0;
		}
		found = 1;
 
		memcpy(ifaceaddrs,&(addressptr->interface_address),sizeof(struct olsr_ip_addr));
		
		ifaceaddrs++;
	}

	if (found)
	{
		outputsize = (char *)ifaceaddrs - packet;
		msg->olsr_packlen = htons(outputsize);
		//memset(msg->olsr_res, 0, sizeof(msg->olsr_res)); 
		//msg->olsr_seqnum = htons(pkt_seqnum[ifp->my_index]++);
		m->olsr_msgtype = MID_PACKET;
		m->olsr_vtime = message->v_time;
		//m->olsr_broad = MSG_BROADCAST;
		//m->olsr_nextmsg = (char *)hnaaddrs - (char *)m;
		//m->olsr_nextmsg = htons(m->olsr_nextmsg); 
		m->olsr_msgsize = (char *)ifaceaddrs - (char *)m; 
		m->olsr_msgsize = htons(m->olsr_msgsize);
		m->olsr_ttl = message->time_to_live;
		m->olsr_hopcount = message->hop_count;
		
		/********** added by Y.Ge *********/
		if (relay)
			m->olsr_msgseq = htons (message->packet_seq_number);
		else
		{	
			m->olsr_msgseq = htons (message->packet_seq_number + mid_chunk[ifp->my_index]);
			mid_chunk[ifp->my_index] ++;
		}
		/*************** end of revision ********/
	
		//CRC HNA add
		//---------------------------------------------------------------
		//memcpy(&(m->olsr_source),&addr_ip,sizeof(struct olsr_ip_addr));
		//memcpy(&(m->olsr_source),&(message->source_addr),sizeof(struct olsr_ip_addr));
		memcpy(&(m->olsr_source),&(message->originator_addr),sizeof(struct olsr_ip_addr));
		//---------------------------------------------------------------
		/*if (interf)            
			hna->hna_seqnum = htons(hna_seqnum++);
		else
			hna->hna_seqnum = htons(message->packet_seq_number);*/
        
		//hna->hna_msgseqnum = htons(message->msg_seq_number);
		//hna->hna_hopcnt = message->hop_count;
		//memset(hna->hna_res, 0, sizeof(hna->hna_res));
		//CRC v4->v6 del tc->tc_origaddr = message->originator;
		//CRC v4->v6 add
		//memcpy(&(hna->hna_origaddr),&(message->originator),sizeof(struct olsr_ip_addr));
	}
	
	if (debug_level > 3)
	{
		printf ("++++++++++ MID packet built finished!\n");
		//printf (" Message sequence: %d\n", ntohs(msg->olsr_seqnum));
		printf (" MID sequence: %d\n", ntohs(m->olsr_msgseq));
		printf (" MID packet size: %d\n", (char *)ifaceaddrs - (char *)m);
				
		fprintf (y_file,"++++++++++ MID packet built finished!\n");
		//fprintf (y_file," Message sequence: %d\n", ntohs(msg->olsr_seqnum));
		fprintf (y_file," MID sequence: %d\n", ntohs(m->olsr_msgseq));
		fprintf (y_file," MID packet size: %d\n", (char *)ifaceaddrs - (char *)m);
	}

	if (outputsize > MAXPACKETSIZE - (int)sizeof (struct olsrmsg)) 
	{
		//inet6_output(s6, flags, dst, outputsize); //commented by Y.Ge
		tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index); //added by Y.Ge
	}
}


/****************************** end of revision *********************/
/*---------------------------------------------------------------------------*/
/*
 * Generate a response packet to ROUTE REQUEST with the contents 
 * of the parameter "message". If this won't fit in one packet,
 * chop it up into several. Send the packet if the size of data 
 * contained in the output buffer reach MAXPACKETSIZE. 
 */
 
void rt_build(struct sockaddr *dst, struct rt_message *message, int flags, struct interface *ifp)           
{
	//CRC v4->v6 modify
	struct sockaddr_in6 *sin;
	u_short *p;

	/*void (*output)(int,int,struct sockaddr *,int) = afswitch[dst->sa_family].af_output;*/

	/*int remainsize = outputsize - (sizeof(msg->olsr_packlen)  + sizeof(msg->olsr_res));*/ //commented by Y.Ge
	int remainsize = outputsize - comm_head_size;
     
	struct olsrmsg *m = outputsize ? (struct olsrmsg *)((char *)msg->olsr_msg + remainsize): msg->olsr_msg;
	struct rtmsg *rt = m->olsr_rt;
	struct rt_message *hostrt, *prev_hostrt;

	if (!message)
		return;

	for (hostrt = message; hostrt != NULL; hostrt = hostrt->next) 
	{
		outputsize = (char *)rt - packet;
		if (outputsize > MAXPACKETSIZE - (int)sizeof (struct rtmsg)) 
		{
			msg->olsr_packlen = htons(outputsize);
			//memset(msg->olsr_res, 0, sizeof(msg->olsr_res));  //commented by Y.G
			//msg->olsr_seqnum = htons(pkt_seqnum[ifp->my_index]++); 	//added by Y.Ge
			m->olsr_msgtype = ROUTE_RESPONSE;
			//m->olsr_broad = MSG_NOBROADCAST;  //commented by Y.Ge
			//m->olsr_nextmsg = (char *)rt - (char *)m;  //commented by Y.Ge
			//m->olsr_nextmsg = htons(m->olsr_nextmsg);  //commented by Y.Ge
			m->olsr_msgsize = (char *)rt - (char *)m;    //added by Y.Ge
			m->olsr_msgsize = htons(m->olsr_msgsize);    //added by Y.Ge
			m->olsr_ttl = message->time_to_live;	//added by Y.Ge
			m->olsr_hopcount = message->hop_count;	//added by Y.Ge
			m->olsr_msgseq = htons (msg_seqnum.rt_seqnum ++); //added by Y.Ge

			//TRACE_OUTPUT(ifp, dst, outputsize);
			//(*output)(s6, flags, dst, outputsize);
			//inet6_output(s6, flags, dst, outputsize); //commented by Y.Ge
			tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index); //added by Y.Ge

			m = msg->olsr_msg;
			rt = m->olsr_rt;
		}

		rt->rt_hopcnt = hostrt->hop_count;
		memset(rt->rt_res, 0, sizeof(rt->rt_res));

		/* CRC v4->v6 del
		rt->rt_destaddr = hostrt->destination;
		rt->rt_nextaddr = hostrt->next_hop;
		rt++;*/
	       
		//CRC v4->v6 add
		memcpy(&(rt->rt_destaddr), &(hostrt->destination),sizeof(struct olsr_ip_addr));
		memcpy(&(rt->rt_nextaddr),&(hostrt->next_hop),sizeof(struct olsr_ip_addr));
		rt++;
	}

	outputsize = (char *)rt - packet;
	msg->olsr_packlen = htons(outputsize);
	//memset(msg->olsr_res, 0, sizeof(msg->olsr_res)); //commented by Y.Ge
	//msg->olsr_seqnum = htons(pkt_seqnum[ifp->my_index]++); 	//added by Y.Ge
	m->olsr_msgtype = ROUTE_RESPONSE;
	//m->olsr_broad = MSG_NOBROADCAST;  //commented by Y.Ge
	//m->olsr_nextmsg = (char *)rt - (char *)m;   //commented by Y.Ge
	//m->olsr_nextmsg = htons(m->olsr_nextmsg);   //commented by Y.Ge
	m->olsr_msgsize = (char *)rt - (char *)m;     //added by Y.Ge
	m->olsr_msgsize = htons(m->olsr_msgsize);     //added by Y.Ge
	m->olsr_ttl = message->time_to_live;	//added by Y.Ge
	m->olsr_hopcount = message->hop_count;	//added by Y.Ge
	m->olsr_msgseq = htons (msg_seqnum.rt_seqnum ++); //added by Y.Ge

	if (outputsize > MAXPACKETSIZE - (int)sizeof (struct olsrmsg)) 
	{
		//TRACE_OUTPUT(ifp, dst, outputsize);
		//(*output)(s6, flags, dst, outputsize);
		//inet6_output(s6, flags, dst, outputsize); //commented by Y.Ge
		tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index); //added by Y.Ge
	}

	/*
 	 * Delete the list of routing informations
 	 */

	hostrt = message;

	while (hostrt)
	{
		prev_hostrt = hostrt;
		hostrt = hostrt->next;
		free(prev_hostrt);
	}

}
