/*
 * Copyright (c) 2004 Ying Ge, Communication Research Center Canada.
 *
 * Copyright (c) 2002, 2003 Maoyu Wang, Communication Research Center Canada.
 *
 * By Ying Ge:
 * 1. Change the OLSR packet format and message processing procedure based 
 *    on the OLSR RFC.
 * 2. Add support of multiple interfaces to OLSR, including MID message 
 *    creating and processing procedure as specified in the OLSR RFC.
 * 3. Add QoS Support to OLSR
 *
 * By Maoyu Wang:
 * 1. Ported OLSR from IPv4 to IPv6.
 * 2. Added the Host and Network Association (HNA) functionality into OLSR.
 * 3. Added the default gateway functionality into OLSR by extending the HNA 
 *    message usage. The default gateway functionality supported the mobility
 *	  by cooperating with Mobile IPv6 for a mobile node as well as supported 
 * 	  Internet access for MANET nodes. 
 *
 * DISTRIBUTED WITH NO WARRANTY, EXPRESS OR IMPLIED.
 * See the GNU Library General Public License (file COPYING in the distribution)
 * for conditions of use and redistribution
 */
/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est-à-dire qu'il peut
 * être utilisé, copié, modifié, distribué à l'unique
 * condition que ce texte soit conservé afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant participé à l'élaboration de ce logiciel ne peut
 * être utilisé sans son accord préalable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilité d'aucune sorte.
 * Certaines parties de ce logiciel sont dérivées de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 */

/*
 * Routing table management daemon.
 */

char routing_table_rcsid[] = 
  "$Id: routing_table.c,v 1.2 2000/12/06 10:36:12 prima Exp $";


#include "defs.h"

struct rthash                     routingtable[HASHSIZE];

/*------------------------------------------------------------------------*/
struct rt_entry
*olsr_lookup_routing_table(struct olsr_ip_addr *dst)
{

	struct rt_entry *destination;
	struct rthash   *routing_hash;
	olsr_u32_t      hash;

	olsr_hashing(dst,&hash);
	routing_hash=&routingtable[hash & HASHMASK];
  
	for(destination=routing_hash->rt_forw;destination!=(struct rt_entry *)routing_hash;destination=destination->rt_forw)
	{
		if(destination->rt_hash!=hash)
		continue;

		if (0==memcmp(&destination->rt_dst,dst,sizeof(struct olsr_ip_addr)))
		return(destination);
	}
	return(NULL);  
}
/*------------------------------------------------------------------------*/
void
olsr_delete_routing_table(struct rt_entry* destination)
{

	if(destination->rt_ifp!=NULL)
		free(destination->rt_ifp);    
	olsr_remque((struct olsr_qelem *)destination);
	free((void *)destination);
}
/*------------------------------------------------------------------------*/
void
olsr_release_routing_table(struct rthash *table)
{ 
	struct rt_entry *destination;
	struct rt_entry *destination_tmp;
	struct rthash   *routing_hash;
	olsr_u8_t             index;
  
	for(index=0;index<HASHSIZE;index++)
	{
		routing_hash=&table[index];
		destination=routing_hash->rt_forw;
		while(destination!=(struct rt_entry *)routing_hash)
		{
			destination_tmp=destination;
			destination=destination->rt_forw;
			olsr_delete_routing_table(destination_tmp);
		}
	}
}
/*------------------------------------------------------------------------*/
struct rt_entry 
*olsr_insert_routing_table(struct olsr_ip_addr *dst)
{
	struct rt_entry *new_route_entry;
	struct rthash   *routing_hash;
	olsr_u32_t       hash;

	olsr_hashing(dst,&hash);
	routing_hash=&routingtable[hash & HASHMASK];

	new_route_entry=(struct rt_entry *)malloc(sizeof(struct rt_entry));
	new_route_entry->rt_hash=hash;

	memcpy(&new_route_entry->rt_dst,dst,sizeof(struct olsr_ip_addr));
  
  	/************* commented by Y.Ge  **********/
	/* 
	 * should set the rt_src based on the interface address
	 * do it in other functions - olsr_fill_routing_table_with_neighbors(), olsr_insert_routing_table_from_topo, etc
	 */ 
	//CRC HNA add
	//------------------------------------------------------------------------
	//memcpy(&new_route_entry->rt_src,&addr_ip,sizeof(struct olsr_ip_addr));
	//CRC: don't know if it should be addr_ip or addr_manet? 
	/*memcpy(&new_route_entry->rt_src,&addr_manet,sizeof(struct olsr_ip_addr));*/
	//------------------------------------------------------------------------
  	/************ end of revision *************/
  
	/*
	 No timeout is needed since the routing table is calculated for each network change
	*/

	new_route_entry->rt_metric=0;
	new_route_entry->rt_ifmetric=0;
	new_route_entry->rt_state=0; 
	new_route_entry->rt_flags=(RTF_UP|RTF_GATEWAY|RTF_HOST);
	new_route_entry->rt_ifp=NULL;
	//CRC HNA add
	/*--------------------------------------------------------------------------*/
	new_route_entry->rt_dst_prefix=128;
	new_route_entry->rt_src_prefix=128;
	new_route_entry->rt_router_prefix=128;
	/*--------------------------------------------------------------------------*/

	olsr_insque((struct olsr_qelem *)new_route_entry, (struct olsr_qelem *)routing_hash);
  
	return(new_route_entry);
	/*
	  We need this pointer to use it for the next routing calculation
	*/
}
/*------------------------------------------------------------------------*/
struct rt_entry 
*olsr_insert_routing_table_from_topo(struct rt_entry *r_last,struct topology_destination_entry *destination)
{
	struct rt_entry *new_route_entry=NULL;
	struct rthash   *routing_hash;
	olsr_u32_t      hash;

	/*
	the hash calculation is the same
	*/
	/*-----------------*/
	if (debug_level > 6)
		printf("Insertion from the topology/////////\n");
	/*-----------------*/
	new_route_entry=olsr_insert_routing_table(&destination->topology_destination_dst);
	new_route_entry->rt_flags=(RTF_UP|RTF_HOST|RTF_GATEWAY);
	memcpy(&new_route_entry->rt_router,&r_last->rt_router,sizeof(struct olsr_ip_addr));
	memcpy(&new_route_entry->rt_src, &r_last->rt_src, sizeof (struct olsr_ip_addr)); //added by Y.Ge
	new_route_entry->rt_metric=r_last->rt_metric+1;
	new_route_entry->rt_willingness = r_last->rt_willingness; //added by Y.Ge
	new_route_entry->rt_local_if = r_last->rt_local_if; //added by Y.Ge 

	return(new_route_entry);  
}
/*NRL code------------------------------------------------------------------
void
olsr_add_default_gateway()
{
  // This function adds a new entry for the default destination
  
  struct rt_entry *default_gateway_entry;
  struct rt_entry  *gateway_entry;
  struct olsr_ip_addr     inaddr_any = in6addr_any;
  
  //if(default_gateway!=0)
  if(memcpy(&default_gateway,0,sizeof(struct olsr_ip_addr)))
    
    if(NULL!=(gateway_entry=olsr_lookup_routing_table(&default_gateway)))
      {      
	default_gateway_entry=olsr_insert_routing_table(&inaddr_any); 
	default_gateway_entry->rt_flags=0;
	default_gateway_entry->rt_flags=(RTF_UP|RTF_GATEWAY);
	memcpy(&default_gateway_entry->rt_router,&gateway_entry->rt_router,sizeof(struct
olsr_ip_addr));
	default_gateway_entry->rt_metric=gateway_entry->rt_metric;
      }
}
-----------------------------------------------------------------------*/


/* CRC code------------------------------------------------------------*/
/*struct rt_entry
*olsr_add_default_gateway(struct olsr_ip_addr* gateway)
{
  //This function adds a new entry for the default destination
  
  struct rt_entry *default_gateway_entry;
  struct rt_entry  *gateway_entry;
  struct olsr_ip_addr     inaddr_any = in6addr_any;
  struct olsr_ip_addr	  inaddrany;

  memset(&inaddrany,0,sizeof(struct olsr_ip_addr));
  
  if(0!=memcmp(gateway,&inaddrany,sizeof(struct olsr_ip_addr)))
   {
     if(NULL!=(gateway_entry=olsr_lookup_routing_table(gateway)))
      {      
		default_gateway_entry=olsr_insert_routing_table(&inaddr_any); 
		memcpy(&(default_gateway_entry->rt_dst),&inaddr_any,sizeof(struct olsr_ip_addr));
		default_gateway_entry->rt_flags=0;
		default_gateway_entry->rt_flags=(RTF_UP|RTF_GATEWAY);
		memcpy(&default_gateway_entry->rt_router,&gateway_entry->rt_router,sizeof(struct
		olsr_ip_addr));
		default_gateway_entry->rt_metric=gateway_entry->rt_metric;
		return(default_gateway_entry);
      }
      else
      {
      	memset(&gwhna,0,sizeof(struct hna_message));
      	return(NULL);
      }
  }
  return(NULL);
}*/

/*void
olsr_del_default_gateway()
{
  //This function delete an entry for the default destination
  
  struct rt_entry *default_gateway_entry;
  struct olsr_ip_addr     inaddrany = in6addr_any;
  
  if(NULL!=(default_gateway_entry=olsr_lookup_routing_table(&inaddrany)))
  {      
  	default_gateway_entry->rt_flags=0;
  	default_gateway_entry->rt_flags=(RTF_UP|RTF_GATEWAY);
  	//memcpy(&default_gateway_entry->rt_router,&gateway_entry->rt_router,sizeof(struct
	//olsr_ip_addr));
  	//default_gateway_entry->rt_metric=gateway_entry->rt_metric;
  	olsr_ioctl_del_route(default_gateway_entry);
  	olsr_delete_routing_table(default_gateway_entry);
  }	
 }*/
/*-----------------------------------------------------------------------*/
/*
 * Y.Ge: based on the OLSR RFC, in the current implementation,
 * olsr_fill_routing_table_with_neighbors() not only deals with the 1-hop-neighbor,
 * but the 2-hop-neighbor as well
 */ 
struct destination_n
*olsr_fill_routing_table_with_neighbors()
{
	struct destination_n   *list_destination_n=NULL;
	struct destination_n   *list_destination_tmp=NULL;
	olsr_u8_t              index;
	struct neighbor_entry  *neighbor;
	struct neighborhash    *hash_neighbor;
	struct rt_entry        *new_route_entry=NULL;
	struct neighbor_entry  *neighbor_any, *next_hop_neighbor;	
	struct main_addr_neigh_entry	*main_addr_neigh; //added by Y.Ge
	struct main_addr_neighhash 	*main_addr_neighhash;  //added by Y.Ge
	struct neighbor_2_entry    *neighbor_2;  //added by Y.Ge
	struct neighbor2_hash      *hash_2_neighbor; //added by Y.Ge
	
	olsr_clear_main_addr_neigh_table_covered_in_routing(); //added by Y.Ge, no neighbor's main address has been included in the routing table 
	
	for(index=0;index<HASHSIZE;index++)
	{
		hash_neighbor=&neighbortable.neighborhash[index];
		for(neighbor=hash_neighbor->neighbor_forw;neighbor!=(struct neighbor_entry *)hash_neighbor;neighbor=neighbor->neighbor_forw)     
		{
			//if((neighbor->neighbor_status==NBS_SYM)||(neighbor->neighbor_status==NBS_MPR)) //commented by Y.Ge
			if(neighbor->neighbor_status==SYM_LINK) //added by Y.Ge
			{
				new_route_entry=olsr_insert_routing_table(&neighbor->neighbor_addr);
				memcpy(&new_route_entry->rt_src, &neighbor->my_iface_addr, sizeof (struct olsr_ip_addr)); //added by Y.Ge
				memcpy(&new_route_entry->rt_router,&neighbor->neighbor_addr,sizeof(struct olsr_ip_addr));
				new_route_entry->rt_flags=(RTF_UP|RTF_HOST);
				new_route_entry->rt_willingness = neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_willingness; //added by Y.Ge
				new_route_entry->rt_local_if = neighbor->my_iface_index; //added by Y.Ge
				
				/*The next router is the neighbor itself*/
				new_route_entry->rt_metric=1; 
				
				/************ added by Y.Ge *********/
				/*** check if the neighbor address is the main_addres ***/
				if (memcmp(&neighbor->neighbor_addr, &neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_addr, sizeof(struct olsr_ip_addr))==0)
				{
					neighbor->neighbor_main_info->main_neighbor->covered_in_routing = 1;	
				}
				/********** end of revision *********/
				
				/*fisrt hop. may be we should use another metric*/
				list_destination_tmp=(struct destination_n *)malloc(sizeof(struct destination_n));
				list_destination_tmp->destination=new_route_entry;
				list_destination_tmp->next=list_destination_n;
				list_destination_n=list_destination_tmp;
			}
		}
	}
	
	/******************* added by Y.Ge ****************/
	/*** add main address of the neighbor into routingg list ***/
	for(index=0;index<HASHSIZE;index++)
	{
		main_addr_neighhash=&mainaddrneightable[index];
		for(main_addr_neigh=main_addr_neighhash->mainaddrneigh_forw;main_addr_neigh!=(struct main_addr_neigh_entry *)main_addr_neighhash;main_addr_neigh=main_addr_neigh->main_addr_neigh_forw)
		{
			if ((main_addr_neigh->main_addr_neigh_status != NOT_NEIGH) && (main_addr_neigh->covered_in_routing != 1))
			{
				main_addr_neigh->covered_in_routing = 1;
				
				neighbor_any = olsr_find_sym_neighbor(main_addr_neigh);  //select a SYM neighbor that connect to the main address neigghbor
				
				if (neighbor_any == NULL)
				{
					printf ("XXXXXXXXX error in find neighbor_any!\n");
					exit (1);
				}
				
				new_route_entry=olsr_insert_routing_table(&main_addr_neigh->main_addr_neigh_addr);
				memcpy(&new_route_entry->rt_src, &neighbor_any->my_iface_addr, sizeof (struct olsr_ip_addr)); 
				memcpy(&new_route_entry->rt_router, &neighbor_any->neighbor_addr, sizeof (struct olsr_ip_addr));
				new_route_entry->rt_willingness = main_addr_neigh->main_addr_neigh_willingness;
				new_route_entry->rt_local_if = neighbor_any->my_iface_index; //added by Y.Ge
				
				/*The next router is the neighbor itself*/
				new_route_entry->rt_metric=2; 
				
				/*fisrt hop. may be we should use another metric*/
				list_destination_tmp=(struct destination_n *)malloc(sizeof(struct destination_n));
				list_destination_tmp->destination=new_route_entry;
				list_destination_tmp->next=list_destination_n;
				list_destination_n=list_destination_tmp;				
			} 
		}
	}
	
	/*** add the 2-hop-neighbors into routingg list ***/
	/* find the higest willingness neighbor as the route to the 2-hop-neighbor */
	for(index=0;index<HASHSIZE;index++)
	{
		hash_2_neighbor=&neighbor2table[index];
		for(neighbor_2=hash_2_neighbor->neighbor2_forw;neighbor_2!=(struct neighbor_2_entry *)hash_2_neighbor;neighbor_2=neighbor_2->neighbor_2_forw)
		{
			/* find the highest willingness neighbor that the 2-hop-neighbor is connected to */
			next_hop_neighbor = olsr_find_highest_willingness_neighbor(neighbor_2);
			
			if (next_hop_neighbor != NULL)
			{
				new_route_entry=olsr_insert_routing_table(&neighbor_2->neighbor_2_addr);
				memcpy(&new_route_entry->rt_src, &next_hop_neighbor->my_iface_addr, sizeof (struct olsr_ip_addr)); 
				memcpy(&new_route_entry->rt_router, &next_hop_neighbor->neighbor_addr, sizeof (struct olsr_ip_addr));
				new_route_entry->rt_willingness = next_hop_neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_willingness;
				new_route_entry->rt_local_if = next_hop_neighbor->my_iface_index; //added by Y.Ge
				
				/*The next router is the neighbor itself*/
				new_route_entry->rt_metric=2; 
				
				/*fisrt hop. may be we should use another metric*/
				list_destination_tmp=(struct destination_n *)malloc(sizeof(struct destination_n));
				list_destination_tmp->destination=new_route_entry;
				list_destination_tmp->next=list_destination_n;
				list_destination_n=list_destination_tmp;
			}			
		}
	}
	/******************* end of revision **************/
	return(list_destination_n);
}
/*------------------------------------------------------------------------*/
void 
olsr_calculate_routing_table()

{
	struct destination_n      	*list_destination_n=NULL;
	struct destination_n      	*list_destination_n_1=NULL;
	struct destination_n      	*destination_n_1=NULL;
	struct topology_last_entry 	*topo_last;
	struct destination_list   	*topo_dest;
	struct rt_entry           	*destination_entry;
	struct rt_entry            	*returned_destination=NULL, *new_route_entry=NULL; //added by Y.Ge
	struct main_addr_entry  	*main_address;	//added by Y.Ge
	struct main_addr_hash   	*main_address_hash; //added by Y.Ge
	struct interface_addr_list	*list_of_my_interface_address; //added by Y.Ge  
	int	index; //added by Y.Ge      

	/* For Debugging Messages: */
	char 				tmp_my_ip_string[INET6_ADDRSTRLEN];
	char 				str[INET6_ADDRSTRLEN];
	struct destination_n  		*tmp_list=NULL;
  
	olsr_routing_mirror();

	/*--------------------------*/
	if(debug_level > 6)
	{
		printf(">>>>Etat des tables apres copie dans la mirror\n");
		printf("La table de routage maintenant\n");
		olsr_print_routing_table(routingtable);
		printf("l'acienne table de routage est:\n");
		olsr_print_routing_table(mirror_table);
		printf("**************************************\n");
	}
	
	if (debug_level > 4)
	{
		printf ("\n\n+++++++++++++++++++ olsr calculate routing table! ++++++++++++++++++++\n");
		printf ("\nbefore calculation, the routing table is:\n");
		olsr_print_routing_table(mirror_table);
		
		fprintf (y_file,"\n\n+++++++++++++++++++ olsr calculate routing table! ++++++++++++++++++++\n");
		fprintf (y_file,"\nbefore calculation, the routing table is:\n");
		olsr_print_trace_routing_table(mirror_table);
	}
	/*--------------------------*/

	/*
	Delete old entries from the table
	*/
	/*
	 * Y.Ge: based on the OLSR RFC, in the current implementation,
	 * olsr_fill_routing_table_with_neighbors() not only deals with the 1-hop-neighbor,
	 * but the 2-hop-neighbor as well
	 */
	if(debug_level >4)
	{
		printf ("\n************* fill routinting table with neighbors\n");
		printf("the neighbor tables are:\n");
		olsr_print_neighbor_table();
		olsr_print_main_addr_neigh_table(); 
		olsr_print_two_hop_neighbor_table();
		
		fprintf (y_file,"\n************* fill routinting table with neighbors\n");
		fprintf(y_file,"the neighbor tables are:\n");
		olsr_print_trace_neighbor_table();
		olsr_print_trace_main_addr_neigh_table(); 
		olsr_print_trace_two_hop_neighbor_table();
	} 
	list_destination_n_1=olsr_fill_routing_table_with_neighbors();
	
	/*--------------------------*/
	if (debug_level > 4)
	{
		printf ("\n........after fill routing  table with neighbors, the routing table is:\n");
		olsr_print_routing_table(routingtable);
		
		fprintf (y_file,"\n........after fill routing  table with neighbors, the routing table is:\n");
		olsr_print_trace_routing_table(routingtable);
	}
	/*--------------------------*/
	
	list_destination_n=list_destination_n_1;
	if (debug_level > 6)
	{  
		//printf("\n\nRouting-Links List: %02d:%02d:%02d.%06lu\n",nowtm->tm_hour, nowtm->tm_min, nowtm->tm_sec, now.tv_usec);
		//CRC v4->v6 add
		//memcpy(tmp_my_ip_string, convert_address_to_string(&addr_ip),46);
		//memcpy(tmp_my_ip_string, convert_address_to_string(&addr_manet),46);//commented by Y.Ge 
		tmp_list = list_destination_n_1;
		while(tmp_list!=NULL)
		{
			//printf("%s -> %s\n", tmp_my_ip_string, convert_address_to_string(&tmp_list->destination->rt_dst));
			//printf("%s -> %s\n", tmp_my_ip_string, inet_ntop(AF_INET6,&(tmp_list->destination->rt_dst),str,sizeof(str))); //commented by Y.Ge
			//printf ("%s -> %s\n", inet_ntop(AF_INET6,&(tmp_list->destination->rt_src),str,sizeof(str)), inet_ntop(AF_INET6,&(tmp_list->destination->rt_dst),str,sizeof(str))); //added by Y.Ge
			tmp_list = tmp_list->next;
		}
	}
	if (debug_level >4)
	{
		printf ("\n**************** now add routing table from topology table!\n");
		printf ("topology table:\n");
		olsr_print_topology_table();
		
		fprintf (y_file,"\n**************** now add routing table from topology table!\n");
		fprintf (y_file,"topology table:\n");
		olsr_print_trace_topology_table();
	}

	while(list_destination_n_1!=NULL)
	{
      
		list_destination_n_1=NULL;
		while(list_destination_n!=NULL)
		{
			/*----------------------*/
			if (debug_level > 6)
			//printf("%s+++++++++++ last hop\n",convert_address_to_string(&list_destination_n->destination->rt_dst));
			printf("%s+++++++++++ last hop\n", inet_ntop(AF_INET6,&list_destination_n->destination->rt_dst,str,sizeof(str)));
			/*----------------------*/
	  
			if(NULL!=(topo_last=olsr_lookup_last_topology_table(&list_destination_n->destination->rt_dst)))
			{
				topo_dest=topo_last->topology_list_of_destinations;
				while(topo_dest!=NULL)
				{
					/*
					topo_dest is not in the routing table		   
					*/
					if(NULL==(returned_destination = olsr_lookup_routing_table(&topo_dest->destination_node->topology_destination_dst)))		    
					{
						/* PRINT OUT: Last Hop to Final Destination */
						/* The function convert_address_to_string has to be seperately */
						if (debug_level > 6) 
						{
							//printf("%s -> ", convert_address_to_string(&list_destination_n->destination->rt_dst));
							//printf("%s\n",convert_address_to_string(&topo_dest->destination_node->topology_destination_dst) );
							printf("%s -> ",inet_ntop(AF_INET6,&list_destination_n->destination->rt_dst,str,sizeof(str)));
							printf("%s\n",inet_ntop(AF_INET6,&topo_dest->destination_node->topology_destination_dst,str,sizeof(str)));
						}
						destination_n_1=(struct destination_n *)malloc(sizeof(struct destination_n));
						destination_n_1->destination=olsr_insert_routing_table_from_topo(list_destination_n->destination,topo_dest->destination_node);

						destination_n_1->next=list_destination_n_1;
						list_destination_n_1=destination_n_1;
					}
					/********************* added by Y.Ge ***************/
					/*
					 * check if we can optimize current route
					 * in terms of willingness, hop-count, etc
					 */
					else
					{
						if (returned_destination->rt_metric >= list_destination_n->destination->rt_metric + 1)
						{
							if (returned_destination->rt_metric > list_destination_n->destination->rt_metric + 1)
							{
								/* update the routing entry to reduce the number of hops */
								returned_destination->rt_flags = (RTF_UP|RTF_HOST|RTF_GATEWAY);
								memcpy(&returned_destination->rt_router, &list_destination_n->destination->rt_router, sizeof (struct olsr_ip_addr));
								memcpy(&returned_destination->rt_src, &list_destination_n->destination->rt_src,sizeof (struct olsr_ip_addr));
								returned_destination->rt_metric = list_destination_n->destination->rt_metric + 1;
								returned_destination->rt_willingness = list_destination_n->destination->rt_willingness;
								returned_destination->rt_local_if = list_destination_n->destination->rt_local_if;
								
								destination_n_1=(struct destination_n *)malloc(sizeof(struct destination_n));
								destination_n_1->destination = returned_destination;
							
								destination_n_1->next=list_destination_n_1;
								list_destination_n_1=destination_n_1;
							}
							else
							{
								/* Y.Ge : Do we need to do so??? */
								if (returned_destination->rt_willingness < list_destination_n->destination->rt_willingness)
								{
									/* update the routing entry to increase the willingness */
									returned_destination->rt_flags = (RTF_UP|RTF_HOST|RTF_GATEWAY);
									memcpy(&returned_destination->rt_router, &list_destination_n->destination->rt_router, sizeof (struct olsr_ip_addr));
									memcpy(&returned_destination->rt_src, &list_destination_n->destination->rt_src,sizeof (struct olsr_ip_addr));
									returned_destination->rt_metric = list_destination_n->destination->rt_metric + 1;
									returned_destination->rt_willingness = list_destination_n->destination->rt_willingness;
									returned_destination->rt_local_if = list_destination_n->destination->rt_local_if;
								
									destination_n_1=(struct destination_n *)malloc(sizeof(struct destination_n));
									destination_n_1->destination = returned_destination;
							
									destination_n_1->next=list_destination_n_1;
									list_destination_n_1=destination_n_1;									
								}
								
							}
						}
					}
					/******************** end of revision *************/
					topo_dest=topo_dest->next;
				}	
			}
	  
			destination_n_1=list_destination_n;
			list_destination_n=list_destination_n->next;
			free(destination_n_1);
		}
		list_destination_n=list_destination_n_1;
	}
      
	if (debug_level > 6)
	{
		printf("End of Routing-Links List.\n"); fflush(stdout);
	}
	
	/*--------------------------*/
	if (debug_level > 4)
	{
		printf ("\n..........after fill routing  table with topology table, the routing table is:\n");
		olsr_print_routing_table(routingtable);
		
		fprintf (y_file,"\n..........after fill routing  table with topology table, the routing table is:\n");
		olsr_print_trace_routing_table(routingtable);
	}
	
	/************************* added by Y.Ge  ************************/
	/* look at MID table */
	if(debug_level > 4)
	{
		printf("\n************ now adding routint table from MID table\n");
		olsr_print_main_address_table();
		
		fprintf(y_file,"\n************ now adding routint table from MID table\n");
		olsr_print_trace_main_address_table();
	}
	/*--------------------------*/
	
	for (index=0; index<HASHSIZE; index++)
	{
		main_address_hash = &nodemainaddresstable[index];
		
		for (main_address=main_address_hash->mainaddress_forw; main_address!=(struct main_addr_entry *)main_address_hash;main_address=main_address->main_addr_forw)
		{
			if(NULL!=(returned_destination = olsr_lookup_routing_table(&main_address->node_main_address)))
			{
				list_of_my_interface_address = main_address->list_of_interfaceaddress;
				while (list_of_my_interface_address!=NULL)
				{
					//printf ("%s, ", convert_address_to_string(&list_of_my_interface_address->interface_addresses->node_interface_address));
					
					if (NULL==olsr_lookup_routing_table(&list_of_my_interface_address->interface_addresses->node_interface_address))
					{
						new_route_entry=olsr_insert_routing_table(&list_of_my_interface_address->interface_addresses->node_interface_address);
						memcpy(&new_route_entry->rt_src, &returned_destination->rt_src, sizeof (struct olsr_ip_addr)); //added by Y.Ge
						memcpy(&new_route_entry->rt_router,&returned_destination->rt_router,sizeof(struct olsr_ip_addr));
						new_route_entry->rt_flags=(RTF_UP|RTF_HOST);
						new_route_entry->rt_willingness = returned_destination->rt_willingness; //added by Y.Ge
						new_route_entry->rt_local_if = returned_destination->rt_local_if; //added by Y.Ge
						
						/*The next router is the neighbor itself*/
						/* attention! different from the RFC */
						/* in RFC, it should be - new_route_entry->rt_metric=returned_destination->rt_metric */
						/* That is NOT corrrect ! */
						new_route_entry->rt_metric=returned_destination->rt_metric+1; 
					}
					
					list_of_my_interface_address=list_of_my_interface_address->next;
				}
				
			}
		}
	}
	/************************* end of revision ***********************/

	//CRC add for HNA
	olsr_add_routing_table_with_hna();
	
	if(debug_level > 4)
	{
		printf ("\n********************** after calculate routing table, the routing table is: ********************\n");
		olsr_print_routing_table(routingtable);
		
		fprintf (y_file,"\n********************** after calculate routing table, the routing table is: ********************\n");
		olsr_print_trace_routing_table(routingtable);
	}
	/*-----------------*/
  
	olsr_update_kernel_routes();
  
	/*-----------------*/
	if(debug_level > 6)
	{
		olsr_print_routing_table(routingtable);
		printf("l'acienne table de routage est:\n");
		olsr_print_routing_table(mirror_table);
	}

	olsr_release_routing_table(mirror_table);	
}
/*------------------------------------------------------------------------*/
/*void
olsr_print_trace_routing_table(struct rthash *table)
{

	struct rt_entry     *destination;
	struct rthash       *routing_hash;
	olsr_u8_t           index;

	fprintf(f_r_id,"Current time is %s\n",ctime((time_t *)&now.tv_sec));
	fprintf(f_r_id,"ROUTING TABLE\n");
	fprintf(f_r_id,"DESTINATION       NEXT HOP\n");
	for(index=0;index<HASHSIZE;index++)
	{
		routing_hash=&table[index];
		for(destination=routing_hash->rt_forw;destination!=(struct rt_entry *)routing_hash;destination=destination->rt_forw)
		{
			fprintf(f_r_id,"%s            ",convert_address_to_string(&destination->rt_dst));
			fprintf(f_r_id,"%s           \n ",convert_address_to_string(&destination->rt_router));
		}
	}
	fflush(f_r_id);
}*/ //commented by Y.Ge
/*------------------------------------------------------------------------*/
/***************** revised by Y.Ge ******************/
/* qos OLSR */

/* Y.Ge: in qos_olsr_fill_routing_table_with_neighbors,
 * different from  RFC, we don't look at two hop neighbors. 
 * the two hop neighbors will be added into routing table
 * as an oridinary node in topology table */
struct destination_n
*qos_olsr_fill_routing_table_with_neighbors()
{
	struct destination_n   *list_destination_n=NULL;
	struct destination_n   *list_destination_tmp=NULL;
	olsr_u8_t              index;
	struct neighbor_entry  *neighbor;
	struct neighborhash    *hash_neighbor;
	struct rt_entry        *new_route_entry=NULL;
	struct neighbor_entry  *neighbor_any, *next_hop_neighbor;	
	struct main_addr_neigh_entry	*main_addr_neigh; //added by Y.Ge
	struct main_addr_neighhash 	*main_addr_neighhash;  //added by Y.Ge
	struct neighbor_2_entry    *neighbor_2;  //added by Y.Ge
	struct neighbor2_hash      *hash_2_neighbor; //added by Y.Ge
	
	olsr_clear_main_addr_neigh_table_covered_in_routing(); //added by Y.Ge, no neighbor's main address has been included in the routing table 
	
	for(index=0;index<HASHSIZE;index++)
	{
		hash_neighbor=&neighbortable.neighborhash[index];
		for(neighbor=hash_neighbor->neighbor_forw;neighbor!=(struct neighbor_entry *)hash_neighbor;neighbor=neighbor->neighbor_forw)     
		{
			//if((neighbor->neighbor_status==NBS_SYM)||(neighbor->neighbor_status==NBS_MPR)) //commented by Y.Ge
			if(neighbor->neighbor_status==SYM_LINK) //added by Y.Ge
			{
				new_route_entry=olsr_insert_routing_table(&neighbor->neighbor_addr);
				memcpy(&new_route_entry->rt_src, &neighbor->my_iface_addr, sizeof (struct olsr_ip_addr)); //added by Y.Ge
				memcpy(&new_route_entry->rt_router,&neighbor->neighbor_addr,sizeof(struct olsr_ip_addr));
				new_route_entry->rt_flags=(RTF_UP|RTF_HOST);
				new_route_entry->rt_willingness = neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_willingness; //added by Y.Ge
				new_route_entry->rt_local_if = neighbor->my_iface_index; //added by Y.Ge
				
				/* qos OLSR */
				new_route_entry->rt_busy = my_busy_time > neighbor->neighbor_busy_time? my_busy_time : neighbor->neighbor_busy_time;
				
				/*The next router is the neighbor itself*/
				new_route_entry->rt_metric=1; 
				
				/************ added by Y.Ge *********/
				/*** check if the neighbor address is the main_addres ***/
				if (memcmp(&neighbor->neighbor_addr, &neighbor->neighbor_main_info->main_neighbor->main_addr_neigh_addr, sizeof(struct olsr_ip_addr))==0)
				{
					neighbor->neighbor_main_info->main_neighbor->covered_in_routing = 1;	
				}
				/********** end of revision *********/
				
				/*fisrt hop. may be we should use another metric*/
				list_destination_tmp=(struct destination_n *)malloc(sizeof(struct destination_n));
				list_destination_tmp->destination=new_route_entry;
				list_destination_tmp->next=list_destination_n;
				list_destination_n=list_destination_tmp;
			}
		}
	}
	
	/******************* added by Y.Ge ****************/
	/*** add main address of the neighbor into routing list ***/
	for(index=0;index<HASHSIZE;index++)
	{
		main_addr_neighhash=&mainaddrneightable[index];
		for(main_addr_neigh=main_addr_neighhash->mainaddrneigh_forw;main_addr_neigh!=(struct main_addr_neigh_entry *)main_addr_neighhash;main_addr_neigh=main_addr_neigh->main_addr_neigh_forw)
		{
			if ((main_addr_neigh->main_addr_neigh_status != NOT_NEIGH) && (main_addr_neigh->covered_in_routing != 1))
			{
				main_addr_neigh->covered_in_routing = 1;
				
				neighbor_any = olsr_find_highest_bw_neighbor(main_addr_neigh); //select a SYM neighbor that connect to the main address neigghbor
				
				if (neighbor_any == NULL)
				{
					printf ("XXXXXXXXX error in find QOS neighbor_any!\n");
					exit (1);
				}
				
				new_route_entry=olsr_insert_routing_table(&main_addr_neigh->main_addr_neigh_addr);
				memcpy(&new_route_entry->rt_src, &neighbor_any->my_iface_addr, sizeof (struct olsr_ip_addr)); 
				memcpy(&new_route_entry->rt_router, &neighbor_any->neighbor_addr, sizeof (struct olsr_ip_addr));
				new_route_entry->rt_willingness = main_addr_neigh->main_addr_neigh_willingness;
				new_route_entry->rt_local_if = neighbor_any->my_iface_index; //added by Y.Ge
				
				/*The next router is the neighbor itself*/
				new_route_entry->rt_metric=2; 
				
				/* qos information */
				new_route_entry->rt_busy = my_busy_time > neighbor_any->neighbor_busy_time? my_busy_time : neighbor_any->neighbor_busy_time;
				
				/*fisrt hop. may be we should use another metric*/
				list_destination_tmp=(struct destination_n *)malloc(sizeof(struct destination_n));
				list_destination_tmp->destination=new_route_entry;
				list_destination_tmp->next=list_destination_n;
				list_destination_n=list_destination_tmp;				
			} 
		}
	}
	
	return(list_destination_n);
}
/*------------------------------------------------------------------------*/
void
olsr_calculate_routing_table_with_qos()
{
	struct destination_n               *list_destination_n=NULL;	//(S_new)
	struct destination_n               *list_destination_n_1=NULL;  //(S_prev)
	struct destination_n               *destination_n_1=NULL;

	struct topology_last_entry       	*topo_last, *topo_current_dst;
	struct topology_destination_entry	*topo_destination;
	struct  destination_list        	*topo_dest;
	struct  last_list	         	*topo_last_list;	
	struct  rt_entry			*new_rt_entry;
	olsr_u32_t		         	current_bw, current_busy_time;
	
   	struct destination_n               	*tmp_list=NULL;

	double bw_factor, old_bw_factor;
	
	int index;
	struct rt_entry            	*returned_destination=NULL, *new_route_entry=NULL;
	struct main_addr_entry  	*main_address;
	struct main_addr_hash   	*main_address_hash; 
	struct interface_addr_list	*list_of_my_interface_address;   
	
	if (debug_level>4)
	{
		printf ("\n\n============= in olsr_calculate_routing_table with QOS ================\n");
		printf ("before calculation, the routing table is: \n");
		qos_olsr_print_routing_table(routingtable);
		
		fprintf (y_file,"\n\n============= in olsr_calculate_routing_table with QOS ================\n");
		fprintf (y_file,"before calculation, the routing table is: \n");
		qos_olsr_print_trace_routing_table(routingtable);
	}

	olsr_routing_mirror();
	
	if(debug_level >4)
	{
		printf ("\n************* fill routinting table with neighbors\n");
		printf("the neighbor tables are:\n");
		olsr_print_neighbor_table();
		olsr_print_main_addr_neigh_table(); 
		
		fprintf (y_file,"\n************* fill routinting table with neighbors\n");
		fprintf(y_file,"the neighbor tables are:\n");
		olsr_print_trace_neighbor_table();
		olsr_print_trace_main_addr_neigh_table(); 
	} 
	
	list_destination_n_1=qos_olsr_fill_routing_table_with_neighbors();
	
	/*--------------------------*/
	if (debug_level > 4)
	{
		printf ("\n........after fill routing  table with neighbors, the routing table is:\n");
		qos_olsr_print_routing_table(routingtable);
		
		fprintf (y_file,"\n........after fill routing  table with neighbors, the routing table is:\n");
		qos_olsr_print_trace_routing_table(routingtable);
	}
	/*--------------------------*/
	
	list_destination_n=list_destination_n_1;
	
	while(list_destination_n_1!=NULL)
	{
		if (debug_level > 5)
		{
			printf ("############### now, S_prev is: ############\n");
			fprintf (y_file,"############### now, S_prev is: ############\n");
			
			tmp_list = list_destination_n_1;
			while(tmp_list!=NULL)
			{
				printf("%s, ", convert_address_to_string(&tmp_list->destination->rt_entry_infos.rtu_dst));
				fprintf(y_file,"%s, ", convert_address_to_string(&tmp_list->destination->rt_entry_infos.rtu_dst));
				
				tmp_list = tmp_list->next;
			}
			printf ("\n");
		}

		list_destination_n_1=NULL;
		while(list_destination_n!=NULL)
		{
			if (debug_level > 5)
			{
			    printf("...........................now consider %s\n",convert_address_to_string(&list_destination_n->destination->rt_entry_infos.rtu_dst));
			    fprintf(y_file,"...........................now consider %s\n",convert_address_to_string(&list_destination_n->destination->rt_entry_infos.rtu_dst));
			}

			if(NULL!=(topo_last=olsr_lookup_last_topology_table(&list_destination_n->destination->rt_entry_infos.rtu_dst)))
			{
				topo_dest=topo_last->topology_list_of_destinations;
				while(topo_dest!=NULL)
				{
					if (debug_level > 5)
					{
						printf("%s(%d) -> ", convert_address_to_string(&list_destination_n->destination->rt_entry_infos.rtu_dst), list_destination_n->destination->rt_busy);
						printf("%s(%d) \n",convert_address_to_string(&topo_dest->destination_node->topology_destination_dst), topo_dest->destination_node->topology_destination_dst_busy);
						
						fprintf(y_file,"%s(%d) -> ", convert_address_to_string(&list_destination_n->destination->rt_entry_infos.rtu_dst), list_destination_n->destination->rt_busy);
						fprintf(y_file,"%s(%d) \n",convert_address_to_string(&topo_dest->destination_node->topology_destination_dst), topo_dest->destination_node->topology_destination_dst_busy);
					 }

					new_rt_entry = olsr_lookup_routing_table(&topo_dest->destination_node->topology_destination_dst);
					
					
					current_bw = list_destination_n->destination->rt_busy > topo_dest->destination_node->topology_destination_dst_busy? list_destination_n->destination->rt_busy : topo_dest->destination_node->topology_destination_dst_busy;
					
					if (debug_level > 5)
					{
						printf ("new_bw is:%d\n", current_bw);
						fprintf (y_file,"new_bw is:%d\n", current_bw);
					}

					
				 	if(new_rt_entry == NULL)
					{
						if (debug_level > 5)
						{
							printf ("*** add new entry to the table!\n");
							fprintf (y_file,"*** add new entry to the table!\n");
						}
						destination_n_1=(struct destination_n *)malloc(sizeof(struct destination_n));
						destination_n_1->destination=olsr_insert_routing_table_from_topo(list_destination_n->destination,topo_dest->destination_node);
						destination_n_1->destination->rt_busy = current_bw;
						
						destination_n_1->next=list_destination_n_1;
						list_destination_n_1=destination_n_1;
						
						if (debug_level > 5)
						{
							printf ("after add new entries, the routing table is:\n");
							qos_olsr_print_routing_table (routingtable);
							
							fprintf (y_file,"after add new entries, the routing table is:\n");
							qos_olsr_print_trace_routing_table (routingtable);
						}
					}
					else
					{
						if (new_rt_entry->rt_busy > current_bw && new_rt_entry->rt_entry_infos.rtu_metric!=1)
						{
						
							if (debug_level > 5)
							{
								printf ("*** WE MAY NEED to update old entry to the table!\n");
								fprintf (y_file,"*** WE MAY NEED to update old entry to the table!\n");
							}
							
							bw_factor = (double)(100-current_bw)/(sqrt(list_destination_n->destination->rt_entry_infos.rtu_metric+1));
							
							old_bw_factor = (double)(100-new_rt_entry->rt_busy)/(sqrt(new_rt_entry->rt_entry_infos.rtu_metric));
							
							
							if (debug_level > 5)
							{
								printf ("************************ bw_factor is: %f, old_bw_factor is: %f\n", bw_factor, old_bw_factor);
								fprintf (y_file,"************************ bw_factor is: %f, old_bw_factor is: %f\n", bw_factor, old_bw_factor);
							}
							
							if (bw_factor > old_bw_factor)
							{
								memcpy(&new_rt_entry->rt_entry_infos.rtu_dst,&topo_dest->destination_node->topology_destination_dst,sizeof(struct olsr_ip_addr));
								new_rt_entry->rt_entry_infos.rtu_flags=(RTF_UP|RTF_HOST|RTF_GATEWAY);
								memcpy(&new_rt_entry->rt_entry_infos.rtu_router,&list_destination_n->destination->rt_entry_infos.rtu_router,sizeof(struct olsr_ip_addr));
								new_rt_entry->rt_entry_infos.rtu_metric=list_destination_n->destination->rt_entry_infos.rtu_metric+1; 
								new_rt_entry->rt_entry_infos.rtu_busy = current_bw;
								new_rt_entry->rt_local_if = list_destination_n->destination->rt_entry_infos.rtu_local_if;
							
								destination_n_1=(struct destination_n *)malloc(sizeof(struct destination_n));
								destination_n_1->destination = new_rt_entry;
								destination_n_1->next=list_destination_n_1;
								list_destination_n_1=destination_n_1;
							
								if (debug_level > 5)
								{
									printf ("after update old entries, the routing table is:\n");
									qos_olsr_print_routing_table (routingtable);
									
									fprintf (y_file,"after update old entries, the routing table is:\n");
									qos_olsr_print_trace_routing_table (routingtable);
								}
							}
						 }
					}
					
					topo_dest=topo_dest->next;
				 }
			}
					
			destination_n_1=list_destination_n;
			list_destination_n=list_destination_n->next;
			free(destination_n_1);
		}
		list_destination_n=list_destination_n_1;
	}
	
	/* look at MID table */
	if(debug_level > 4)
	{
		printf("\n************ now adding routint table from MID table\n");
		olsr_print_main_address_table();
		
		fprintf(y_file,"\n************ now adding routint table from MID table\n");
		olsr_print_trace_main_address_table();
	}
	/*--------------------------*/
	
	for (index=0; index<HASHSIZE; index++)
	{
		main_address_hash = &nodemainaddresstable[index];
		
		for (main_address=main_address_hash->mainaddress_forw; main_address!=(struct main_addr_entry *)main_address_hash;main_address=main_address->main_addr_forw)
		{
			if(NULL!=(returned_destination = olsr_lookup_routing_table(&main_address->node_main_address)))
			{
				list_of_my_interface_address = main_address->list_of_interfaceaddress;
				while (list_of_my_interface_address!=NULL)
				{
					//printf ("%s, ", convert_address_to_string(&list_of_my_interface_address->interface_addresses->node_interface_address));
					
					if (NULL==olsr_lookup_routing_table(&list_of_my_interface_address->interface_addresses->node_interface_address))
					{
						new_route_entry=olsr_insert_routing_table(&list_of_my_interface_address->interface_addresses->node_interface_address);
						memcpy(&new_route_entry->rt_src, &returned_destination->rt_src, sizeof (struct olsr_ip_addr)); //added by Y.Ge
						memcpy(&new_route_entry->rt_router,&returned_destination->rt_router,sizeof(struct olsr_ip_addr));
						new_route_entry->rt_flags=(RTF_UP|RTF_HOST);
						new_route_entry->rt_willingness = returned_destination->rt_willingness; //added by Y.Ge
						new_route_entry->rt_local_if = returned_destination->rt_local_if; //added by Y.Ge
						
						/*The next router is the neighbor itself*/
						/* attention! different from the RFC */
						/* in RFC, it should be - new_route_entry->rt_metric=returned_destination->rt_metric */
						/* That is NOT corrrect ! */
						new_route_entry->rt_metric=returned_destination->rt_metric+1; 
						
						/* qos OLSR */
						new_route_entry->rt_busy = returned_destination->rt_busy;
					}
					
					list_of_my_interface_address=list_of_my_interface_address->next;
				}
				
			}
		}
	}

	//CRC add for HNA
	olsr_add_routing_table_with_hna();
	
	if(debug_level > 4)
	{
		printf ("\n********************** after calculate routing table, the routing table is: ********************\n");
		qos_olsr_print_routing_table(routingtable);
		
		fprintf (y_file,"\n********************** after calculate routing table, the routing table is: ********************\n");
		qos_olsr_print_trace_routing_table(routingtable);
	}
	/*-----------------*/
  
	olsr_update_kernel_routes();
	olsr_release_routing_table(mirror_table);
}
/*------------------------------------------------------------------------*/
void
olsr_print_trace_routing_table(struct rthash *table)
{
	struct rt_entry     *destination;
	struct rthash       *routing_hash;
	olsr_u8_t           index;
	char str[INET6_ADDRSTRLEN];

	fprintf(y_file,"\n\nROUTING TABLE\n");
	fprintf(y_file,"DESTINATION \tSOURCE \tNEXT HOP \tINTERFACE \tMETRIC\n");
	for(index=0;index<HASHSIZE;index++)
	{
		routing_hash=&table[index];
		for(destination=routing_hash->rt_forw;destination!=(struct rt_entry *)routing_hash;destination=destination->rt_forw)
		{
			//printf("%s            ",convert_address_to_string(&destination->rt_dst));
			//printf("%s           \n ",convert_address_to_string(&destination->rt_router));
			fprintf(y_file,"%s ",inet_ntop(AF_INET6,&destination->rt_dst,str,sizeof(str)));
			fprintf(y_file,"\t%s ",inet_ntop(AF_INET6,&destination->rt_src,str,sizeof(str)));
			fprintf(y_file,"\t%s ",inet_ntop(AF_INET6,&destination->rt_router,str,sizeof(str)));
			fprintf(y_file,"\t%d ", destination->rt_local_if);
			fprintf(y_file,"\t%d\n",destination->rt_metric);	
		}
	}
}
/*------------------------------------------------------------------------*/
void
qos_olsr_print_routing_table(struct rthash *table)
{
	struct rt_entry     *destination;
	struct rthash       *routing_hash;
	olsr_u8_t           index;
	char str[INET6_ADDRSTRLEN];

	printf("\n\nROUTING TABLE\n");
	printf("DESTINATION \tSOURCE \tNEXT HOP \tINTERFACE \tMETRIC \tBUSYTIME\n");
	for(index=0;index<HASHSIZE;index++)
	{
		routing_hash=&table[index];
		for(destination=routing_hash->rt_forw;destination!=(struct rt_entry *)routing_hash;destination=destination->rt_forw)
		{
			//printf("%s            ",convert_address_to_string(&destination->rt_dst));
			//printf("%s           \n ",convert_address_to_string(&destination->rt_router));
			printf("%s ",inet_ntop(AF_INET6,&destination->rt_dst,str,sizeof(str)));
			printf("\t%s ",inet_ntop(AF_INET6,&destination->rt_src,str,sizeof(str)));
			printf("\t%s ",inet_ntop(AF_INET6,&destination->rt_router,str,sizeof(str)));
			printf("\t%d ", destination->rt_local_if);
			printf("\t%d",destination->rt_metric);
			printf("\t%d\n", destination->rt_busy);	
		}
	}
}

void
qos_olsr_print_trace_routing_table(struct rthash *table)
{
	struct rt_entry     *destination;
	struct rthash       *routing_hash;
	olsr_u8_t           index;
	char str[INET6_ADDRSTRLEN];

	fprintf(y_file,"\n\nROUTING TABLE\n");
	fprintf(y_file,"DESTINATION \tSOURCE \tNEXT HOP \tINTERFACE \tMETRIC \tBUSYTIME\n");
	for(index=0;index<HASHSIZE;index++)
	{
		routing_hash=&table[index];
		for(destination=routing_hash->rt_forw;destination!=(struct rt_entry *)routing_hash;destination=destination->rt_forw)
		{
			//printf("%s            ",convert_address_to_string(&destination->rt_dst));
			//printf("%s           \n ",convert_address_to_string(&destination->rt_router));
			fprintf(y_file,"%s ",inet_ntop(AF_INET6,&destination->rt_dst,str,sizeof(str)));
			fprintf(y_file,"\t%s ",inet_ntop(AF_INET6,&destination->rt_src,str,sizeof(str)));
			fprintf(y_file,"\t%s ",inet_ntop(AF_INET6,&destination->rt_router,str,sizeof(str)));
			fprintf(y_file,"\t%d ", destination->rt_local_if);
			fprintf(y_file,"\t%d",destination->rt_metric);
			fprintf(y_file,"\t%d\n", destination->rt_busy);	
		}
	}
}
/****************** end of revision ****************/
/*------------------------------------------------------------------------*/
void
olsr_print_routing_table(struct rthash *table)
{
	struct rt_entry     *destination;
	struct rthash       *routing_hash;
	olsr_u8_t           index;
	char str[INET6_ADDRSTRLEN];

	printf("\n\nROUTING TABLE\n");
	printf("DESTINATION \tSOURCE \tNEXT HOP \tINTERFACE \tMETRIC\n");
	for(index=0;index<HASHSIZE;index++)
	{
		routing_hash=&table[index];
		for(destination=routing_hash->rt_forw;destination!=(struct rt_entry *)routing_hash;destination=destination->rt_forw)
		{
			//printf("%s            ",convert_address_to_string(&destination->rt_dst));
			//printf("%s           \n ",convert_address_to_string(&destination->rt_router));
			printf("%s ",inet_ntop(AF_INET6,&destination->rt_dst,str,sizeof(str)));
			printf("\t%s ",inet_ntop(AF_INET6,&destination->rt_src,str,sizeof(str)));
			printf("\t%s ",inet_ntop(AF_INET6,&destination->rt_router,str,sizeof(str)));
			printf("\t%d ", destination->rt_local_if);
			printf("\t%d\n",destination->rt_metric);	
		}
	}
}
/*------------------------------------------------------------------------*/


//CRC add for HNA
/*------------------------------------------------------------------------*/
void olsr_copy_routing_table(struct rthash *table)
{
	/*copy a routing table from previous the mirror table*/
	olsr_u32_t           	index;
	struct rthash		   	*table_hash;
	struct rt_entry	   	*destination;
	struct rt_entry		*start_entry=NULL, *new_entry=NULL;

	for(index=0;index<HASHSIZE;index++)
	{
		if(table[index].rt_forw==(struct rt_entry *)&table[index])
		{
			routingtable[index].rt_forw=(struct rt_entry *)&routingtable[index];
			routingtable[index].rt_back=(struct rt_entry *)&routingtable[index];
		}
		else
		{	
			table_hash=&table[index];
			for(destination=table_hash->rt_forw;destination!=(struct rt_entry *)table_hash;destination=destination->rt_forw)
			{
				if(destination==table_hash->rt_forw)
				{
					start_entry=(struct rt_entry *)malloc(sizeof(struct rt_entry));
					memcpy(start_entry,destination,sizeof(struct rt_entry));
					new_entry=start_entry;
				}
				else
				{
					new_entry->rt_forw=(struct rt_entry *)malloc(sizeof(struct rt_entry));
					memcpy(new_entry->rt_forw,destination,sizeof(struct rt_entry));
					new_entry->rt_forw->rt_back=new_entry; 
					new_entry=new_entry->rt_forw;
				}     		
			} 
			routingtable[index].rt_forw=start_entry;
			routingtable[index].rt_forw->rt_back=(struct rt_entry *)&routingtable[index];
			routingtable[index].rt_back=new_entry;
			routingtable[index].rt_back->rt_forw=(struct rt_entry *)&routingtable[index];
		}       
	}
}
     
void     
olsr_add_routing_table_with_hna()
{
	olsr_u32_t			index;
	olsr_u16_t			shortest_metric;
	struct hna_subnet_hash		*subnet_hash_hna;
	struct hna_subnet_entry		*subnet_entry;
	struct host_list		*list_of_hosts;
	struct rt_entry			*host_rt_entry;
	struct rt_entry			*shortest_rt_entry;
	struct rt_entry			*net_rt_entry;
	
	if(debug_level > 4)
	{
		printf("\n************ now adding routint table from HNA table\n");
		olsr_print_hna_host_table();
		olsr_print_hna_subnet_table();
		
		fprintf(y_file,"\n************ now adding routint table from HNA table\n");
		olsr_print_trace_hna_host_table();
		olsr_print_trace_hna_subnet_table();
	}
	
	for(index=0;index<HASHSIZE;index++)
	{
		subnet_hash_hna=&hnasubnettable[index];
		for(subnet_entry=subnet_hash_hna->hna_subnet_forw;subnet_entry!=(struct hna_subnet_entry*)subnet_hash_hna;subnet_entry=subnet_entry->hna_subnet_forw)
		{
			shortest_metric=10000;
			shortest_rt_entry=NULL;
			for(list_of_hosts=subnet_entry->hna_list_of_host;list_of_hosts!=NULL;list_of_hosts=list_of_hosts->next)
			{				
				if(NULL!=(host_rt_entry=olsr_lookup_routing_table(&(list_of_hosts->host->hna_host_addr))))
					if(shortest_metric>host_rt_entry->rt_metric)
					{
						shortest_metric=host_rt_entry->rt_metric; //"+1" is added by Y.Ge
						shortest_rt_entry=host_rt_entry;					
					}
			}
			
			if(NULL!=shortest_rt_entry)
			{
				/*add a new net entry into routing table*/
				net_rt_entry=olsr_insert_routing_table(&(subnet_entry->hna_subnet_addr));
				net_rt_entry->rt_dst_prefix=subnet_entry->hna_subnet_prefix;
				memcpy(&(net_rt_entry->rt_router),&(shortest_rt_entry->rt_router),sizeof(struct olsr_ip_addr));
				net_rt_entry->rt_router_prefix=shortest_rt_entry->rt_router_prefix;
				net_rt_entry->rt_metric=shortest_rt_entry->rt_metric+1; //"+1" is added by Y.Ge
				net_rt_entry->rt_flags=0;
				net_rt_entry->rt_flags=(RTF_UP|RTF_GATEWAY);
				net_rt_entry->rt_local_if = shortest_rt_entry->rt_local_if; //added by Y.Ge
				memcpy(&(net_rt_entry->rt_src), &(shortest_rt_entry->rt_src), sizeof(struct olsr_ip_addr)); //added by Y.Ge
			}
		}
	}
}

void
olsr_modify_routing_table_from_hna()
{
	olsr_routing_mirror();
	olsr_copy_routing_table(mirror_table);
      	
	if(debug_level > 4)
	{
		printf("\n************ now modify routing table from HNA table\n");
		olsr_print_hna_host_table();
		olsr_print_hna_subnet_table();
		
		fprintf(y_file,"\n************ now modify routing table from HNA table\n");
		olsr_print_trace_hna_host_table();
		olsr_print_trace_hna_subnet_table();
	}
      	
	olsr_add_routing_table_with_hna();
	olsr_update_kernel_routes();
	olsr_release_routing_table(mirror_table);
}

/*-------------------------------------------------------------------------------------------------------------------------*/
