/*
 * Copyright (c) 2004 Ying Ge, Communication Research Center Canada.
 *
 * Copyright (c) 2002, 2003 Maoyu Wang, Communication Research Center Canada.
 *
 * By Ying Ge:
 * 1. Change the OLSR packet format and message processing procedure based 
 *    on the OLSR RFC.
 * 2. Add support of multiple interfaces to OLSR, including MID message 
 *    creating and processing procedure as specified in the OLSR RFC.
 * 3. Add QoS Support to OLSR
 *
 * By Maoyu Wang:
 * 1. Ported OLSR from IPv4 to IPv6.
 * 2. Added the Host and Network Association (HNA) functionality into OLSR.
 * 3. Added the default gateway functionality into OLSR by extending the HNA 
 *    message usage. The default gateway functionality supported the mobility
 *	  by cooperating with Mobile IPv6 for a mobile node as well as supported 
 * 	  Internet access for MANET nodes. 
 *
 * DISTRIBUTED WITH NO WARRANTY, EXPRESS OR IMPLIED.
 * See the GNU Library General Public License (file COPYING in the distribution)
 * for conditions of use and redistribution
 */
/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est-à-dire qu'il peut
 * être utilisé, copié, modifié, distribué à l'unique
 * condition que ce texte soit conservé afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant participé à l'élaboration de ce logiciel ne peut
 * être utilisé sans son accord préalable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilité d'aucune sorte.
 * Certaines parties de ce logiciel sont dérivées de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 */

char timer_rcsid[] = 
  "$Id: timer.c,v 1.2 2000/12/06 10:36:12 prima Exp $";

/*
 * Routing Table Management Daemon
 */

#include "defs.h"

float sincelast_tc = 0.0;      /* Time last since we send last TC packet */
float sincelast_hello = 0.0;   /* Time last since we send last HELLO packetc */
float sincelast_mid = 0.0;	/* added by Y.Ge, Time last since we send last MID packet */
float sincelast_hello_poll = 0.0; /* Need to keep average hellos consistent */
float tc_jit_result = 0.0, hello_jit_result = 0.0;
float mid_jit_result = 0.0; //added by Y.Ge
float hna_jit_result = 0.0; //added by Y.Ge
float tc_relay_jitter = 0.0; //added by Y.Ge
float mid_relay_jitter = 0.0; //added by Y.Ge
float hna_relay_jitter = 0.0;	//added by Y.Ge
float tc_relay_interval = 0.0; //added by Y.Ge
float mid_relay_interval = 0.0; //added by Y.Ge
float hna_relay_interval = 0.0;	//added by Y.Ge

/* qos OLSR */
int current_busy_time = 10; //added by Y.Ge

/* for QoS debug */
float since_last_busy_update = 0.0;
float busy_update_interval = 5.0;

//CRC HNA add
/*-----------------------------------------------------------------*/
float sincelast_hna = 0.0;   /* TIme last since we send last HNA packet*/
/*-----------------------------------------------------------------*/

/******************** added by Y.Ge ***************/
/* reset the chunk count array */
void reset_chunk_count(int *array)
{
	int i;
	
	for (i=0; i<num_of_interface; i++)
	{
		array[i]=0;
	}
}

int check_chunk_count (int *array)
{
	int i;
	
	for (i=1; i<num_of_interface; i++)
	{
		if (array[i] != array[0])
		{
			return 0; // something is wrong, the chunked number should be same
		}
	}
	
	return 1;	
}

/* qos OLSR */
int read_busy_time()
{
	FILE *proc_fp = NULL;
	char text_to_print[150];
	
	int	rcvd_idle;
	int	rcvd_tx_busy;
	int	rcvd_rx_busy;
	
	if (proc_fp == NULL)	
		proc_fp = fopen("/proc/net/av_bw_algo_ath0", "r");
		
	if (proc_fp < 0)
	{
		printf ("XXXXXXX Error in opening /proc XXXXXXXXX\n");
		exit(1);
	}
	
	/* read sequence number */
	if (NULL==fgets(text_to_print, 20, proc_fp)) 
	{
		printf ("XXXXXXX Error in reading sequence number XXXXXXXXX\n");
		exit(1);	

	} 
	else
		text_to_print[strlen(text_to_print)-1] = (char)NULL;
		
	/* read idle time */
	if (NULL==fgets(text_to_print, 20, proc_fp))
	{
		printf ("XXXXXXX Error in reading idle time XXXXXXXXX\n");
		exit(1);
	}
	else
	{
		sscanf(text_to_print, "%d", &rcvd_idle);
		
		if (debug_level > 6)
		{
			printf ("current idle time: %d\n", rcvd_idle);
			fprintf (y_file, "current idle time: %d\n", rcvd_idle);
		}
		/* strcpy(text_to_print+strlen(text_to_print)-1, "%"); */
	}
		
	/* read Busy tx */
	if (NULL==fgets(text_to_print, 20, proc_fp))
	{
		printf ("XXXXXXX Error in reading busy tx XXXXXXXXX\n");
		exit(1);
	}
	else
	{
		sscanf(text_to_print, "%d", &rcvd_tx_busy);
		
		if (debug_level > 6)
		{
			printf ("current idle time: %d\n", rcvd_tx_busy);
			fprintf (y_file, "current idle time: %d\n", rcvd_tx_busy);
		}
		/* strcpy(text_to_print+strlen(text_to_print)-1, "%"); */
	}
		
	/* read Busy rx */
	if (NULL==fgets(text_to_print, 20, proc_fp))
	{
		printf ("XXXXXXX Error in reading busy rx XXXXXXXXX\n");
		exit(1);
	}
	else
	{
		sscanf(text_to_print, "%d", &rcvd_rx_busy);
		
		if (debug_level > 6)
		{
			printf ("current idle time: %d\n", rcvd_rx_busy);
			fprintf (y_file, "current idle time: %d\n", rcvd_rx_busy);
		}
		/* strcpy(text_to_print+strlen(text_to_print)-1, "%"); */
	}
	
	if (proc_fp != NULL) 
	{
		fclose(proc_fp); 
		proc_fp = NULL;
	}
	
	return (rcvd_tx_busy + rcvd_rx_busy);
		
}
/******************* end of revision *************/
/*
 * Timer routine. Performs neighbor and topology information supply
 * duties and manages timers on all tables' entries.
 */

void timer(int signum)
{
	struct itimerval itval;
	struct hello_message hellopacket;
	struct tc_message tcpacket;
	struct mid_message midpacket; //added by Y.Ge
	struct qos_tc_message	qostcpacket; //added by Y.Ge
	struct interface *ifp; //added by Y.Ge
	struct mid_iface_addr	*mid_addresses, *prev_mid_addresses; //added by Y.Ge
	struct tc_mpr_addr *mprs, *prev_mprs; //added by Y.Ge
	int    mid_built_finished = 0, tc_built_finished = 0, hna_built_finished = 0; //added by Y.Ge, only need to build MID and TC once FOR all interfaces
	int    hna_sent = 0;//added by Y.Ge
	
	int timetobroadcast_hello, timetobroadcast_tc, timetoreset_hello, timetobroadcast_hna, timetobroadcast_mid;
	int timetorelay_tc=0, timetorelay_mid=0, timetorelay_hna=0; //added by Y.Ge
	int send_tc_seperately=0, send_mid_seperately=0; //added by Y.Ge
	
	tcpacket.multipoint_relay_selector_address = (struct tc_mpr_addr *)NULL;

	hellopacket.neighbors = (struct hello_neighbor *)NULL;

	midpacket.mid_interface_address = (struct mid_iface_addr *)NULL; //added by Y.Ge
	

	(void)signum;

	(void) gettimeofday(&now, (struct timezone *)NULL);
	nowtm = gmtime(&now.tv_sec);
	
	sincelast_hello += polling_int;
	sincelast_hello_poll += polling_int;
	sincelast_tc += polling_int;
	sincelast_hna += polling_int;
	sincelast_mid += polling_int; //added by Y.Ge

	timetoreset_hello = supplier && sincelast_hello_poll >= hello_int;
	timetobroadcast_hello = supplier && sincelast_hello >= hello_int_wjit;
	timetobroadcast_tc = changes ? supplier && sincelast_tc >= min_tc_int : supplier && sincelast_tc >= tc_int;
	//CRC HNA modify
	timetobroadcast_hna = supplier && sincelast_hna >= HNA_INTERVAL;
	timetobroadcast_mid = supplier && sincelast_mid >= mid_int; //added by Y.Ge
	
	/*timetobroadcast_mid = timetobroadcast_hello;
	timetobroadcast_hna = timetobroadcast_hello; 
	timetobroadcast_tc = timetobroadcast_hello; */
	
	
	/* qos OLSR debug */
	since_last_busy_update += polling_int;
	
	
	/*if (since_last_busy_update >= busy_update_interval)
	{
		//randondom_busy +=2;
		current_busy_time += 2; 
		since_last_busy_update = 0;
		printf ("\n\n------------------ updating busy time! new busy time is: %d ----------\n", current_busy_time);
	}*/
		
	/******************** added by Y.Ge ****************/
	/************ busy time updating *******/
	/* qos OLSR */
	if (use_qos)
	{
		current_busy_time = read_busy_time();
		
		if (abs(current_busy_time - my_busy_time) > (my_busy_time * bw_update_threshold))
		{
		
			if (debug_level > 3)
			{
				printf ("\n\n------------------ updating busy time! old busy time is: %d, new busy time is: %d ----------\n", my_busy_time, current_busy_time);
				fprintf (y_file, "\n\n------------------ updating busy time! old busy time is: %d, new busy time is: %d ----------\n", my_busy_time, current_busy_time);
			}
			
			my_busy_time = current_busy_time;
		}		
	}
		
	if (tc_to_relay)
	{
		if (debug_level > 3)
		{
			printf ("----- should relay tc!\n");
			printf ("--- interval:%f, jitter:%f\n", tc_relay_interval, tc_relay_jitter);
		
			fprintf (y_file,"----- should relay tc!\n");
			fprintf (y_file,"--- interval:%f, jitter:%f\n", tc_relay_interval, tc_relay_jitter);
		}
		
		tc_relay_interval += polling_int;
		timetorelay_tc = tc_relay_interval >= tc_relay_jitter;
		if (tc_relay_size ==0)
		{
			printf ("XXXXXXXXX something wrong with the tc relay size! xxxxxxxxx\n");
			exit(1);
		}
	}
	if (mid_to_relay)
	{
		if (debug_level > 3)
		{
			printf ("----- should relay mid!\n");
			printf ("--- interval:%f, jitter:%f\n", mid_relay_interval, mid_relay_jitter);
		
			fprintf (y_file, "----- should relay mid!\n");
			fprintf (y_file, "--- interval:%f, jitter:%f\n", mid_relay_interval, mid_relay_jitter);
		}
		
		mid_relay_interval += polling_int;
		timetorelay_mid = mid_relay_interval > mid_relay_jitter;
		if (mid_relay_size ==0)
		{
			printf ("XXXXXXXXX something wrong with the mid relay size! xxxxxxxxx\n");
			exit(1);
		}
	}
	
	if (hna_to_relay)
	{
		if (debug_level > 3)
		{
			printf ("----- should relay hna!\n");
			printf ("--- interval:%f, jitter:%f\n", hna_relay_interval, hna_relay_jitter);
		
			fprintf (y_file, "----- should relay hna!\n");
			fprintf (y_file, "--- interval:%f, jitter:%f\n", hna_relay_interval, hna_relay_jitter);
		}
		
		hna_relay_interval += polling_int;
		timetorelay_hna = hna_relay_interval > hna_relay_jitter;
		if (hna_relay_size ==0)
		{
			printf ("XXXXXXXXX something wrong with the mid relay size! xxxxxxxxx\n");
			exit(1);
		}
	}
	
	/* decide if the  relayed TC/MID/HNA should be sent out seperately */
	if (timetorelay_tc)
	{
		/* the two or three messages could be relayed in one packets */
		/* but the message should NOT be trucked */
		/* as it will cause the confusion of packet sequence number */ 
		/* so check the message size to make sure the messages are not chunked */
		if (timetorelay_mid && timetorelay_hna)
		{
			if (debug_level > 3)
			{
				printf ("********* there are relayed TC, MID, HNA together ************\n");
				printf ("tc relay size: %d, mid relay size: %d, hna relay size: %d\n", tc_relay_size, mid_relay_size, hna_relay_size);
				
				fprintf (y_file,"********* there are relayed TC, MID, HNA together ************\n");
				fprintf (y_file,"tc relay size: %d, mid relay size: %d, hna relay size: %d\n", tc_relay_size, mid_relay_size, hna_relay_size);
			}
			
			if (tc_relay_size + mid_relay_size < MAXPACKETSIZE)
			{
				if (tc_relay_size + hna_relay_size + mid_relay_size >  MAXPACKETSIZE)
				{
					send_mid_seperately = 1;
					
					if (debug_level > 3)
					{
						printf ("----- send TC and MID together!\n");
						fprintf (y_file,"----- send TC and MID together!\n");
					}
				}
			}
			else
			{
				send_tc_seperately = 1;
				
				if (debug_level > 3)
				{
					printf ("----- send TC out first!\n");
					fprintf (y_file,"----- send TC out first!\n");
				}
				
				if (hna_relay_size + mid_relay_size >  MAXPACKETSIZE)
				{
					send_mid_seperately = 1;
					
					if (debug_level > 3)
					{
						printf ("----- send MID out after TC!\n");
						fprintf (y_file,"----- send MID out after TC!\n");
					}
				}
			}
		}
		else 
		{
			if (debug_level > 3 && (timetorelay_mid || timetorelay_hna))
			{
				printf ("********* there are relayed TC || (MID || HNA) together ************\n");
				printf ("tc relay size: %d, mid relay size: %d, hna relay size: %d\n", tc_relay_size, mid_relay_size, hna_relay_size);
				
				fprintf (y_file,"********* there are relayed TC || (MID || HNA) together ************\n");
				fprintf (y_file,"tc relay size: %d, mid relay size: %d, hna relay size: %d\n", tc_relay_size, mid_relay_size, hna_relay_size);
			}
			
			if ((timetorelay_mid && (tc_relay_size + mid_relay_size >  MAXPACKETSIZE)) || (timetorelay_hna && (tc_relay_size + hna_relay_size >  MAXPACKETSIZE)))
			{
				send_tc_seperately = 1;
				
				if (debug_level > 3)
				{
					printf ("^^^^^^^^^^^^^^ can not send TC and MID/HNA together ^^^^^^^^^^^^^^^^\n");
					fprintf (y_file,"^^^^^^^^^^^^^^ can not send TC and MID/HNA together ^^^^^^^^^^^^^^^^\n");
				}
			}
		}		
	}
	else
	{
		if (timetorelay_mid && timetorelay_hna)
		{
			if (debug_level > 3)
			{
				printf ("********* there are relayed MID, HNA together ************\n");
				printf ("mid relay size: %d, hna relay size: %d\n", mid_relay_size, hna_relay_size);
				
				fprintf (y_file,"********* there are relayed MID, HNA together ************\n");
				fprintf (y_file,"mid relay size: %d, hna relay size: %d\n", mid_relay_size, hna_relay_size);
			}
			
			if (mid_relay_size + hna_relay_size>MAXPACKETSIZE)
			{
				send_mid_seperately = 1;
				
				if (debug_level > 3)
				{
					printf ("^^^^^^^^^^^^^^ can not send MID  and HNA together ^^^^^^^^^^^^^^^^\n");
					fprintf (y_file,"^^^^^^^^^^^^^^ can not send MID  and HNA together ^^^^^^^^^^^^^^^^\n");
				}	
			}
		}
	}
	/******************* end of revision ***************/

	/* Advance time on differents tables entry */

	olsr_delete_time_out_entries();

	//timetobroadcast_hello = 1;
	//timetobroadcast_mid = 0;

	if ((timetoreset_hello)&&(hello_start==1))
	{
		sincelast_hello_poll = 0;
		hello_jit_result = (float) random()/RAND_MAX - 0.5;
		hello_jit_result *= hello_jitter;
		hello_int_wjit = hello_int + hello_jit_result;
		
		if (debug_level == -1)
			printf("hello int with jitter value = %f\n",hello_int_wjit);  
	}

	/*************** following "for" loop is added by Y.Ge to add interface index info *********/
	for (ifp = ifnet; ifp; ifp = ifp->int_next) 
	{

		/******************** added by Y.Ge ****************/
		/*** first of all, check if there is packets to be relayed ***/
		if(timetorelay_tc)
		{
			printf ("XXXXXXXXXXXXXXXXXXXX time  (%02d:%02d:%02d.%06lu) TC sent from interface %d!XXXXXXXXXXXXXXXXXX\n", nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec, ifp->my_index);
			fprintf (y_file, "XXXXXXXXXXXXXXXXXXXX time  (%02d:%02d:%02d.%06lu) TC sent from interface %d!XXXXXXXXXXXXXXXXXX\n", nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec, ifp->my_index);
			
			/* qos OLSR */
			if (qostcpacket_relay.multipoint_relay_selector_busy_time !=NULL)
				qostc_build((struct sockaddr *)NULL, &qostcpacket_relay, &tcpacket_relay,0, ifp, 1);
			else
				tc_build((struct sockaddr *)NULL,(struct hello_message *)NULL, &tcpacket_relay, (struct mid_message *)NULL, (struct hna_message *)NULL, 0, ifp, 1);
			tc_to_relay = 0;
			qostc_to_relay = 0;
			tc_relay_interval = 0;
			
			if (send_tc_seperately)
			{
				/******** should send tc out first *********/
				if (debug_level > 3)
				{
					printf ("=========== the relayed TC is sent out! ===========\n");
					fprintf (y_file,"=========== the relayed TC is sent out! ===========\n");
				}
				
				tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index);
			}	
		}
		
		if (timetorelay_mid)
		{
			printf ("XXXXXXXXXXXXXXXXXXXX time  (%02d:%02d:%02d.%06lu) MID sent from interface %d!XXXXXXXXXXXXXXXXXX\n", nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec, ifp->my_index);
			fprintf (y_file, "XXXXXXXXXXXXXXXXXXXX time  (%02d:%02d:%02d.%06lu) MID sent from interface %d!XXXXXXXXXXXXXXXXXX\n", nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec, ifp->my_index);
			
			mid_build((struct sockaddr *)NULL,(struct hello_message *)NULL, (struct tc_message *)NULL, &midpacket_relay, (struct hna_message *)NULL, 0, ifp, 1);
			mid_to_relay = 0;
			mid_relay_interval = 0;	
			
			if (send_mid_seperately)
			{
				/******** should send mid out first *********/
				if (debug_level > 3)
				{
					printf ("=========== the relayed MID is sent out! ===========\n");
					fprintf (y_file,"=========== the relayed MID is sent out! ===========\n");
				}
				
				tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index);
			}	
			
		}
		if (timetorelay_hna)
		{
			printf ("XXXXXXXXXXXXXXXXXXXX time  (%02d:%02d:%02d.%06lu) HNA sent from interface %d!XXXXXXXXXXXXXXXXXX\n", nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec, ifp->my_index);
			fprintf (y_file, "XXXXXXXXXXXXXXXXXXXX time  (%02d:%02d:%02d.%06lu) HNA sent from interface %d!XXXXXXXXXXXXXXXXXX\n", nowtm->tm_hour, nowtm->tm_min,nowtm->tm_sec,now.tv_usec, ifp->my_index);
			
			hna_build((struct sockaddr *)NULL,(struct hello_message *)NULL, (struct tc_message *)NULL,(struct mid_message *)NULL, &hnapacket_relay,  0, ifp, 1);
			hna_to_relay = 0;
			hna_relay_interval = 0;				
		}
		/******************* end of revision ***************/
//		printf("timer.c: outside the if: index: %d timetobroadcast_hello %d hello_start //%d\n",ifp->my_index,timetobroadcast_hello,hello_start);
		if ((timetobroadcast_hello)&&(hello_start==1)) 
		{
			olsr_build_hello_packet(&hellopacket, ifp->my_index);
			hello_build((struct sockaddr *)NULL, &hellopacket, (struct tc_message *)NULL,(struct mid_message *)NULL,(struct hna_message *)NULL, 0, ifp);
			
			/******** added by Y.Ge *********/
			/* qos OLSR */
			if (use_qos)
			{
				qoshello_build((struct sockaddr *)NULL, 0, ifp);
			}
			/******** end of revision *******/
			sincelast_hello = 0;
			if(debug_level > 6) 
			{
				olsr_print_neighbor_table();
				if(debug_level > 7) 
					olsr_print_topology_table();
			}
			
			
		}
        
		if (timetobroadcast_tc && mprstable.mssn!=0) //"mprstable.mssn!=0" added by Y.Ge, mpr selector table has been changed
		{
			/************ added by Y.Ge **************/
			if (tc_built_finished == 0)
			{
				if (use_qos)
				{
					olsr_build_qos_tc_packet(&qostcpacket, ifp->my_index);
				}
				olsr_build_tc_packet(&tcpacket, ifp->my_index);
				tc_built_finished = 1;
			}
			
			/*toall(tc_build, (struct hello_message *)NULL, &tcpacket,(struct hna_message *)NULL, (struct interface *)NULL);*/ //commented by Y.Ge
			
			if (use_qos)
				qostc_build ((struct sockaddr *)NULL, &qostcpacket, &tcpacket, 0, ifp, 0);
			else 
				tc_build ((struct sockaddr  *)NULL,(struct hello_message *)NULL, &tcpacket,(struct mid_message *)NULL,(struct hna_message *)NULL, 0, ifp, 0);
			/*********** end of revision ***********/
						
			sincelast_tc = 0;
			tc_jit_result = (float) random()/RAND_MAX - 0.5;
			tc_jit_result *= tc_jitter;
			if (debug_level > 6)
				printf("tc jitter value = %f\n",tc_jit_result);
			sincelast_tc += tc_jit_result;
			//changes = 0;	//commented by Y.Ge, reset changes later
		}

		//CRC HNA modify
		if (timetobroadcast_hna && hnainfo.hna_net_addrs != NULL) // revised by Y.Ge
		{
			/*toall(hna_build, (struct hello_message *)NULL,(struct tc_message *)NULL,(struct mid_message *)NULL, &hnainfo,(struct interface *)NULL);
			sincelast_hna = 0;*/ //commented by  Y.Ge
			
			/************** added by Y.Ge **********/
			hna_build ((struct sockaddr *)NULL, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL, &hnainfo, 0, ifp, 0);
			sincelast_hna = 0;
			hna_jit_result = (float) random()/RAND_MAX -0.5;
			hna_jit_result *= hna_jitter;
			sincelast_hna += hna_jit_result;
			hna_sent = 1;
			/************** end of revision *********/
		}

		/*************** added by Y.Ge ***************/
		if (timetobroadcast_mid && num_of_interface>1)
		{
			if (mid_built_finished == 0) //only need to build MID once
			{
				olsr_build_mid_packet (&midpacket, ifp->my_index);
				mid_built_finished = 1;
			}
			mid_build((struct sockaddr *)NULL, (struct hello_message *)NULL,(struct tc_message *)NULL, &midpacket, (struct hna_message *)NULL, 0, ifp, 0);
			sincelast_mid = 0;
			mid_jit_result = (float) random()/RAND_MAX - 0.5;
			mid_jit_result *= mid_jitter;
			sincelast_mid += mid_jit_result;
		}
		/****************** end of revision **************/
		if (outputsize)
		{
			//toall(sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL ); //commented by Y.Ge
			tointerface (sndmsg, (struct hello_message *)NULL, (struct tc_message *)NULL, (struct mid_message *)NULL,(struct hna_message *)NULL, (struct interface *)NULL, ifp->my_index); //added by Y.Ge
			//printf ("\n\nXXXXXXXXXXXXXXXXXXXXXXXXXX sending message finished!XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\n");
			//getchar();
		}	
			
	}
	/************** added by Y.Ge ************/
	/*
	 * reset the changes
	 * which is used in timetobroadcast_tc
	 */
	if (timetobroadcast_tc)
	{
		if (changes)
			changes = 0;
	}
	
	/* 
	 * update message seq number
	 * delete the addresses in various messages  
	 */
	if (midpacket.mid_interface_address != NULL)
	{
		if (check_chunk_count(mid_chunk) ==0)
		{
			printf ("XXXXXXXXXXX something wrong with mid truck! XXXXXXXXXXXXXXX\n");
			exit (1);
		}
		
		msg_seqnum.mid_seqnum += mid_chunk[0];
		reset_chunk_count (mid_chunk);
		
		mid_addresses = midpacket.mid_interface_address;
		while (mid_addresses)
		{
			prev_mid_addresses = mid_addresses;
			mid_addresses = mid_addresses->next;
			free(prev_mid_addresses);
		}
	}
	if (tcpacket.multipoint_relay_selector_address != NULL)
	{
		if (check_chunk_count(tc_chunk) ==0)
		{
			printf ("XXXXXXXXXXX something wrong with tc truck! XXXXXXXXXXXXXXX\n");
			exit (1);
		}
		
		msg_seqnum.tc_seqnum += tc_chunk[0];
		
		/*printf (" ++++++++ now tc seq is: %d\n", msg_seqnum.tc_seqnum);
		fprintf (y_file," ++++++++ now tc seq is: %d\n", msg_seqnum.tc_seqnum);*/
		
		reset_chunk_count (tc_chunk);

		mprs = tcpacket.multipoint_relay_selector_address;
		while (mprs)
		{
			prev_mprs = mprs;
			mprs = mprs->next;
			free(prev_mprs);
		}
	}
	if (hna_sent && hnainfo.hna_net_addrs != NULL)
	{
		if (check_chunk_count(hna_chunk) ==0)
		{
			printf ("XXXXXXXXXXX something wrong with hna truck! XXXXXXXXXXXXXXX\n");
			exit (1);
		}
		/*printf ("&&&&&&&&&&&&&& hnachunk is: %d\n", hna_chunk[0]);
		printf (" originally hna sequence is: %d\n", msg_seqnum.hna_seqnum);*/
		
		msg_seqnum.hna_seqnum += hna_chunk[0];
		hnainfo.packet_seq_number = msg_seqnum.hna_seqnum;
		/*printf (" now hna sequence is: %d\n", msg_seqnum.hna_seqnum);*/
		reset_chunk_count (hna_chunk);
	}
	
	/*
	 * free the memroy of the relayed TC and MID packets
	 */
	if (tc_to_relay == 0 && tcpacket_relay.multipoint_relay_selector_address!=NULL)
	{
		olsr_destroy_tc_message(&tcpacket_relay);
		tcpacket_relay.multipoint_relay_selector_address = NULL;
		tc_relay_size = 0;
	}
	if (qostc_to_relay == 0 && qostcpacket_relay.multipoint_relay_selector_busy_time!=NULL)
	{
		olsr_destroy_qos_tc_message(&qostcpacket_relay);
		qostcpacket_relay.multipoint_relay_selector_busy_time = NULL;
		qostc_relay_size = 0;
	}
	if (mid_to_relay ==0 && midpacket_relay.mid_interface_address!=NULL)
	{
		//printf ("XXXXXXXXXX should destroy the midpacket_relay! XXXXXXXXXXX\n");
		olsr_destroy_mid_message(&midpacket_relay);
		midpacket_relay.mid_interface_address=NULL;
		mid_relay_size = 0;
	}
	if (hna_to_relay == 0 && hnapacket_relay.hna_net_addrs != NULL)
	{
		olsr_destroy_hna_message(&hnapacket_relay);
		hnapacket_relay.hna_net_addrs = NULL;
		hna_relay_size = 0;
	}
	/*************** end of revision ***************/	
}































