/*
 * Copyright (c) 2004 Ying Ge, Communication Research Center Canada.
 *
 * Copyright (c) 2002, 2003 Maoyu Wang, Communication Research Center Canada.
 *
 * By Ying Ge:
 * 1. Change the OLSR packet format and message processing procedure based 
 *    on the OLSR RFC.
 * 2. Add support of multiple interfaces to OLSR, including MID message 
 *    creating and processing procedure as specified in the OLSR RFC.
 * 3. Add QoS Support to OLSR
 *
 * By Maoyu Wang:
 * 1. Ported OLSR from IPv4 to IPv6.
 * 2. Added the Host and Network Association (HNA) functionality into OLSR.
 * 3. Added the default gateway functionality into OLSR by extending the HNA 
 *    message usage. The default gateway functionality supported the mobility
 *	  by cooperating with Mobile IPv6 for a mobile node as well as supported 
 * 	  Internet access for MANET nodes. 
 *
 * DISTRIBUTED WITH NO WARRANTY, EXPRESS OR IMPLIED.
 * See the GNU Library General Public License (file COPYING in the distribution)
 * for conditions of use and redistribution
 */
/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est--dire qu'il peut
 * tre utilis, copi, modifi, distribu  l'unique
 * condition que ce texte soit conserv afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant particip  l'laboration de ce logiciel ne peut
 * tre utilis sans son accord pralable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilit d'aucune sorte.
 * Certaines parties de ce logiciel sont drives de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 */

/*
 * Routing table management daemon.
 */

char topology_table_rcsid[] = 
  "$Id: topology_table.c,v 1.2 2000/12/06 10:36:12 prima Exp $";


#include "defs.h"


struct topology_destination_hash  topologytable[HASHSIZE];
struct topology_last_hash         topologylasttable[HASHSIZE];

/*-----------------------------------------------------------------------*/
void
olsr_delete_list_of_last(struct topology_last_entry *last_entry,struct olsr_ip_addr *dst)
{
  
	struct destination_list            *dest_list;
	struct destination_list            *dest_list_tmp;
  
	dest_list=last_entry->topology_list_of_destinations;
	last_entry->topology_list_of_destinations=NULL;

	while(dest_list!=NULL)
	{ 
		if(0==memcmp(&dest_list->destination_node->topology_destination_dst,dst,sizeof(struct olsr_ip_addr)))    
		{
			dest_list_tmp=dest_list;
			dest_list=dest_list->next;
			free(dest_list_tmp);
		}
		else
		{
			dest_list_tmp=dest_list;
			dest_list=dest_list->next;
			dest_list_tmp->next=last_entry->topology_list_of_destinations;
			last_entry->topology_list_of_destinations=dest_list_tmp;
		}
	}
}
/*------------------------------------------------------------------------*/
void
olsr_delete_dest_topolgy_table(struct topology_destination_entry *dest_entry)
{
	struct last_list *list_of_last;
	struct topology_last_entry *last_entry;

	list_of_last=dest_entry->topology_destination_list_of_last;

	while(list_of_last!=NULL)
	{
		last_entry=list_of_last->last_neighbor;
		olsr_delete_list_of_last(last_entry,&dest_entry->topology_destination_dst);              
		dest_entry->topology_destination_list_of_last=list_of_last->next;
		free(list_of_last);  
		list_of_last=dest_entry->topology_destination_list_of_last;
       }
	olsr_remque((struct olsr_qelem *)dest_entry);
	free((void *)dest_entry);
}
/*------------------------------------------------------------------------*/
void
olsr_delete_list_of_dest(struct topology_destination_entry *dest_entry,struct olsr_ip_addr *dst)
{
	struct last_list  *list_of_last;
	struct last_list  *list_of_last_tmp;

	list_of_last=dest_entry->topology_destination_list_of_last;
	dest_entry->topology_destination_list_of_last=NULL;
	while(list_of_last!=NULL)
	{
		if(0==memcmp(&list_of_last->last_neighbor->topology_last,dst,sizeof(struct olsr_ip_addr)))
		{
			list_of_last_tmp=list_of_last;
			list_of_last=list_of_last->next;
			free(list_of_last_tmp);
		}
		else
		{
			list_of_last_tmp=list_of_last;
			list_of_last=list_of_last->next;
			list_of_last_tmp->next=dest_entry->topology_destination_list_of_last;
			dest_entry->topology_destination_list_of_last=list_of_last_tmp;
		}
	}

	if(dest_entry->topology_destination_list_of_last==NULL)
	{
		olsr_remque((struct olsr_qelem *)dest_entry);
		free((void *)dest_entry);
	}
}
/*------------------------------------------------------------------------*/
void
olsr_delete_last_topolgy_table(struct topology_last_entry *last_entry)
{
	struct destination_list    *list_of_dest;
	struct topology_destination_entry *dest_entry;

	list_of_dest=last_entry->topology_list_of_destinations;

	while(list_of_dest!=NULL)
	{
		dest_entry=list_of_dest->destination_node;
		olsr_delete_list_of_dest(dest_entry,&last_entry->topology_last);
		last_entry->topology_list_of_destinations=list_of_dest->next;      
		free(list_of_dest);  
		list_of_dest=last_entry->topology_list_of_destinations;
       }

	olsr_remque((struct olsr_qelem *)last_entry);
	free((void *)last_entry);
}
/*------------------------------------------------------------------------*/
void
olsr_insert_last_topology_table(struct topology_last_entry *last_entry,struct tc_message *message)
{
	olsr_u32_t                  hash;
	struct topology_last_hash   *top_last_hash;
  
	olsr_hashing(&message->originator,&hash);
	top_last_hash=&topologylasttable[hash & HASHMASK];

	memcpy(&last_entry->topology_last,&message->originator,sizeof(struct olsr_ip_addr));
	last_entry->topology_list_of_destinations=NULL;
	last_entry->topologylast_hash=hash;
	last_entry->topology_seq=message->mssn;

	timeradd(&now,&hold_time_topology,&last_entry->topology_timer);

	olsr_insque((struct olsr_qelem *)last_entry, (struct olsr_qelem *)top_last_hash);
}
/*------------------------------------------------------------------------*/
void
olsr_insert_dest_topology_table(struct topology_destination_entry *dest_entry,struct tc_mpr_addr *mpr)
{
	olsr_u32_t                         hash;
	struct topology_destination_hash   *top_dest_hash;

	olsr_hashing(&mpr->address,&hash);

	top_dest_hash=&topologytable[hash & HASHMASK];
                 
	memcpy(&dest_entry->topology_destination_dst,&mpr->address,sizeof(struct olsr_ip_addr));

	dest_entry->topology_destination_list_of_last=NULL;
	dest_entry->topologydestination_hash=hash;

	olsr_insque((struct olsr_qelem *)dest_entry, (struct olsr_qelem *)top_dest_hash);
}
/*------------------------------------------------------------------------*/
struct topology_last_entry  
*olsr_lookup_last_topology_table(struct olsr_ip_addr *last)
{
	struct topology_last_entry  *top_last;
	struct topology_last_hash   *top_last_hash;
	olsr_u32_t                  hash;

	olsr_hashing(last,&hash);

	top_last_hash=&topologylasttable[hash & HASHMASK];

	for(top_last=top_last_hash->topology_last_forw;top_last!=(struct topology_last_entry *)top_last_hash;top_last=top_last->topology_last_forw)
	{
		if(top_last->topologylast_hash!=hash)
			continue;

		if (0==memcmp(&top_last->topology_last,last,sizeof(struct olsr_ip_addr)))
			return(top_last);
      
	}
	return (NULL);
}
/*------------------------------------------------------------------------*/
struct topology_destination_entry
*olsr_lookup_dest_topology_table(struct olsr_ip_addr *dest)
{

	struct topology_destination_entry    *top_dest;
	struct topology_destination_hash     *top_dest_hash;
	olsr_u32_t                           hash;

	/*-------------------*/
	if (debug_level > 6)
	{
		printf("olsr_lookup_dest_topology_table\n");
	}
	/*-------------------*/
  
	olsr_hashing(dest,&hash);

	top_dest_hash=&topologytable[hash & HASHMASK];

	for(top_dest=top_dest_hash->topology_destination_forw;top_dest!=(struct topology_destination_entry *)top_dest_hash;top_dest=top_dest->topology_destination_forw)
	{
		if(top_dest->topologydestination_hash!=hash)
			continue;
 
		if (0==memcmp(&top_dest->topology_destination_dst,dest,sizeof(struct olsr_ip_addr)))
			return(top_dest);
	}
	return (NULL);
}
/*------------------------------------------------------------------------*/
struct  destination_list 
*olsr_in_list_dest_topo(struct topology_last_entry *last_entry,struct olsr_ip_addr *address)
{
	struct  destination_list *list_of_dest;

	list_of_dest=last_entry->topology_list_of_destinations;

	/*----------------*/
	if (debug_level > 6)
	{
		printf("olsr_in_list_dest_topo\n");
	}
	/*----------------*/
 
	while(list_of_dest!=NULL)
	{
		if(0==(memcmp(&list_of_dest->destination_node->topology_destination_dst,address,sizeof(struct olsr_ip_addr))))
			return(list_of_dest);
		list_of_dest=list_of_dest->next;
	}
	return(NULL);
}
/*------------------------------------------------------------------------*/
void
olsr_update_last_table(struct topology_last_entry *last_entry,struct tc_message *message)
{
	/*
	 This function updates the time out for the existant pointed by t_last and 
	 creates new entries for the new destinations    
	*/
  
	struct tc_mpr_addr                *mpr;
	struct destination_list           *topo_dest;
	struct topology_destination_entry *destination_entry;
	struct last_list                  *topo_last;

	mpr=message->multipoint_relay_selector_address;

	timeradd(&now,&hold_time_topology,&last_entry->topology_timer);
	/*-----------------*/
	if(debug_level > 6)
	{
		printf("UPDATING THIS TOP_LAST!!!!!!!!! le timeout est %d et le now is %d \n",ctime((time_t *)&last_entry->topology_timer.tv_sec),ctime((time_t *)&now.tv_sec));
		printf("The list of MPRs is\n:");
		while(mpr!=NULL)
		{
			printf("This mpr is %s\n",convert_address_to_string(&mpr->address));
			mpr=mpr->next;
		}
		mpr=message->multipoint_relay_selector_address;
	}
	/*-----------------*/

	while(mpr!=NULL)
	{
		//CRC modified
		//if(0!=memcmp(&mpr->address,&addr_ip,sizeof(struct olsr_ip_addr))) 
		//if(0!=memcmp(&mpr->address,&addr_manet,sizeof(struct olsr_ip_addr))) //commented by Y.Ge
		if(0!=memcmp(&mpr->address, &my_main_addr, sizeof(struct olsr_ip_addr)))//added by Y.Ge 
		{	
			if(NULL==(topo_dest=olsr_in_list_dest_topo(last_entry,&mpr->address)))
			{
				/*----------------*/
				if (debug_level > 6)
				{
					printf("Ce DEST n'est pas dans la liste liee a ce LAST\n");
				}
				/*----------------*/

				changes_topology=UP;
				if(NULL==(destination_entry=olsr_lookup_dest_topology_table(&mpr->address)))
				{	
					destination_entry=(struct topology_destination_entry * )malloc(sizeof(struct topology_destination_entry));
					olsr_insert_dest_topology_table(destination_entry,mpr);
					
					/* added by Y.Ge */
					/* qos OLSR, we don't received a QoS TC, so suppose all the dest busy time is 0 */
					if (use_qos)
						destination_entry->topology_destination_dst_busy= default_busy_time; //topo_dest's busy time
					/** end of revision **/
				}
	    
				/* creating the structure for the linkage*/
				topo_dest=(struct  destination_list *)malloc(sizeof(struct  destination_list));
				topo_dest->next=last_entry->topology_list_of_destinations;
				last_entry->topology_list_of_destinations=topo_dest;

				topo_last=(struct last_list *)malloc(sizeof(struct last_list));
				topo_last->next=destination_entry->topology_destination_list_of_last;
				destination_entry->topology_destination_list_of_last=topo_last;
	    
				/*linking the last and the dest*/
				topo_last->last_neighbor=last_entry;
				topo_dest->destination_node=destination_entry;
			}	
      
			/*-----------------*/
			if (debug_level > 6)
			{
				printf("This mpr is %s\n",convert_address_to_string(&mpr->address));
				if(mpr->next==NULL)
					printf("the next is NULL\n");
				else
					printf("the next is not NULL\n");
			}
			/*-----------------*/
		}
		mpr=mpr->next;
	}
}
/*------------------------------------------------------------------------*/
void
olsr_release_topology_table()
{
	olsr_u8_t                            index;

	struct topology_destination_entry    *top_dest;
	struct topology_destination_entry    *top_dest_tmp;
	struct topology_destination_hash     *top_dest_hash;

	struct topology_last_entry           *top_last;
	struct topology_last_entry           *top_last_tmp;
	struct topology_last_hash            *top_last_hash;

	for(index=0;index<HASHSIZE;index++)
	{
		top_dest_hash=&topologytable[index];
		top_dest=top_dest_hash->topology_destination_forw;
		while(top_dest!=(struct topology_destination_entry *)top_dest_hash)
		{
			top_dest_tmp=top_dest;
			top_dest=top_dest->topology_destination_forw;	
			olsr_delete_dest_topolgy_table(top_dest_tmp);
		}
	}

	for(index=0;index<HASHSIZE;index++)
	{
		top_last_hash=&topologylasttable[index];
		top_last=top_last_hash->topology_last_forw;
		while(top_last!=(struct topology_last_entry *)top_last_hash)
		{
			top_last_tmp=top_last;
			top_last=top_last->topology_last_forw;
			olsr_delete_last_topolgy_table(top_last_tmp);
		}
	}
}
/*------------------------------------------------------------------------*/
void
olsr_time_out_topology_table()
{
	/* 
	 We only check the last table because we didn't include a time out in the dest entries
	*/
	olsr_u8_t                            index;
	struct topology_last_entry           *top_last;
	struct topology_last_entry           *top_last_tmp;
	struct topology_last_hash            *top_last_hash;
	int	 topology_time_out = 0;
  
	for(index=0;index<HASHSIZE;index++)
	{
		top_last_hash=&topologylasttable[index];
		top_last=top_last_hash->topology_last_forw;
		while(top_last!=(struct topology_last_entry *)top_last_hash)
		{
			if(olsr_timed_out(&top_last->topology_timer))
			{
				topology_time_out = 1;
				
				/*-----------------*/
				if(debug_level > 6)
				{
					printf("TIMEOUT POUR CE TOP_LAST!!!!!!!!! le timeout est %d  %d et le now is %d %d \n",top_last->topology_timer.tv_sec,top_last->topology_timer.tv_usec,now.tv_sec,now.tv_usec);
				}
				
				if(debug_level > 1)
				{
					printf("\n\n............t_last %s timed out! t_last timer: %d  %d, now: %d %d.............. \n",convert_address_to_string(&top_last->topology_last),top_last->topology_timer.tv_sec,top_last->topology_timer.tv_usec,now.tv_sec,now.tv_usec);
					fprintf(y_file,"\n\n............t_last %s timed out! t_last timer: %d  %d, now: %d %d \n",convert_address_to_string(&top_last->topology_last),top_last->topology_timer.tv_sec,top_last->topology_timer.tv_usec,now.tv_sec,now.tv_usec);
				}
				/*----------------*/
				
				top_last_tmp=top_last;
				top_last=top_last->topology_last_forw;
				olsr_delete_last_topolgy_table(top_last_tmp);
	      
				changes_topology=UP;
			}
			else
				top_last=top_last->topology_last_forw;
		}
	}
	
	if (topology_time_out)
	{
		if (debug_level > 1)
		{
			printf (" ************ after topology table timed out, the topology table is:\n");
			olsr_print_topology_table();
			
			fprintf (y_file," ************ after topology table timed out, the topology table is:\n");
			olsr_print_trace_topology_table();
		}
		
		topology_time_out = 0;
	}
}
/*------------------------------------------------------------------------*/
void
olsr_print_topology_table()
{
	olsr_u8_t                            index;
	struct topology_last_entry           *top_last;
	struct topology_last_hash            *top_last_hash;
	struct destination_list              *top_dest_list;
	char str[46];

	printf("\n\nTOPOLOGY TABLE: %02d:%02d:%02d.%06lu\n",nowtm->tm_hour, nowtm->tm_min, nowtm->tm_sec, now.tv_usec);
	for(index=0;index<HASHSIZE;index++)
	{
		top_last_hash=&topologylasttable[index];
		for(top_last=top_last_hash->topology_last_forw;top_last!=(struct topology_last_entry *)top_last_hash;top_last=top_last->topology_last_forw)
		{
			if(use_qos)
				printf("%s,(%d);[last_hop_of; ",inet_ntop(AF_INET6,&top_last->topology_last,str,46), top_last->topology_last_busy);	
			else
				printf("%s;[last_hop_of; ",inet_ntop(AF_INET6,&top_last->topology_last,str,46));
	  
			top_dest_list=top_last->topology_list_of_destinations;
	  
			while(top_dest_list!=NULL)
			{
				if(use_qos)
					printf("%s, (%d);",inet_ntop(AF_INET6,(&top_dest_list->destination_node->topology_destination_dst),str,46), top_dest_list->destination_node->topology_destination_dst_busy);
				else				
					printf("%s;",inet_ntop(AF_INET6,(&top_dest_list->destination_node->topology_destination_dst),str,46));
				top_dest_list=top_dest_list->next;
			}
			printf("]\n");
		}
	}
}
/*------------------------------------------------------------------------*/
/************************* added by Y.Ge *****************/
void
olsr_print_trace_topology_table()
{
	olsr_u8_t                            index;
	struct topology_last_entry           *top_last;
	struct topology_last_hash            *top_last_hash;
	struct destination_list              *top_dest_list;
	char str[46];

	fprintf(y_file,"\n\nTOPOLOGY TABLE: %02d:%02d:%02d.%06lu\n", nowtm->tm_hour, nowtm->tm_min, nowtm->tm_sec, now.tv_usec);
  
	for(index=0;index<HASHSIZE;index++)
	{
		top_last_hash=&topologylasttable[index];
		for(top_last=top_last_hash->topology_last_forw;top_last!=(struct topology_last_entry *)top_last_hash;top_last=top_last->topology_last_forw)
		{
			if (use_qos)
				fprintf(y_file,"%s,(%d);[last_hop_of; ",inet_ntop(AF_INET6,&top_last->topology_last,str,46), top_last->topology_last_busy);
			else
				fprintf(y_file,"%s;[last_hop_of; ",inet_ntop(AF_INET6,&top_last->topology_last,str,46));
	  
			top_dest_list=top_last->topology_list_of_destinations;
	  
			while(top_dest_list!=NULL)
			{
				if (use_qos)
					fprintf(y_file,"%s, (%d);",inet_ntop(AF_INET6,(&top_dest_list->destination_node->topology_destination_dst),str,46), top_dest_list->destination_node->topology_destination_dst_busy);
				else
					fprintf(y_file,"%s;",inet_ntop(AF_INET6,(&top_dest_list->destination_node->topology_destination_dst),str,46));
				top_dest_list=top_dest_list->next;
			}
			fprintf(y_file,"]\n");
		}
	}
}
/*------------------------------------------------------------------------*/
/* qos OLSR */
void
olsr_update_last_table_with_qos(struct topology_last_entry *last_entry,struct tc_message *message, struct qos_tc_message *qosmessage)
{
	/*
	 This function updates the topology last table with QoS information
	*/
  
	struct tc_mpr_addr                *mpr;
	struct tc_mpr_busy_time           *mprbusy;
	struct destination_list           *topo_dest;
	struct topology_destination_entry *destination_entry;
	struct last_list                  *topo_last;

	mpr=message->multipoint_relay_selector_address;
	mprbusy=qosmessage->multipoint_relay_selector_busy_time;

	/* update last_entry */
	if (last_entry->topology_last_busy != qosmessage->my_busy_time)
	{
		changes_topology = UP;
		last_entry->topology_last_busy = qosmessage->my_busy_time;
	}
	timeradd(&now,&hold_time_topology,&last_entry->topology_timer);
	/*-----------------*/
	if(debug_level > 6)
	{
		printf("UPDATING THIS TOP_LAST!!!!!!!!! le timeout est %d et le now is %d \n",ctime((time_t *)&last_entry->topology_timer.tv_sec),ctime((time_t *)&now.tv_sec));
		printf("The list of MPRs is\n:");
		while(mpr!=NULL)
		{
			printf("This mpr is %s\n",convert_address_to_string(&mpr->address));
			mpr=mpr->next;
		}
		mpr=message->multipoint_relay_selector_address;
	}
	/*-----------------*/

	while(mpr!=NULL)
	{
		if (debug_level > 5)
		{
			fprintf (y_file,"++++++++ the mpr is: %s, mpr busy time is: %d\n", convert_address_to_string(&mpr->address), mprbusy->busy_time);
			printf ("++++++++ the mpr is: %s, mpr busy time is: %d\n", convert_address_to_string(&mpr->address), mprbusy->busy_time);
		}
		if(0!=memcmp(&mpr->address, &my_main_addr, sizeof(struct olsr_ip_addr))) 
		{	
			if(NULL==(topo_dest=olsr_in_list_dest_topo(last_entry,&mpr->address)))
			{
				/*----------------*/
				if (debug_level > 6)
				{
					printf("Ce DEST n'est pas dans la liste liee a ce LAST\n");
				}
				
				if (debug_level > 5)
				{
					printf ("insert new t_dest!\n");
					fprintf (y_file,"insert new t_dest!\n");
				}
				/*----------------*/

				changes_topology=UP;
				if(NULL==(destination_entry=olsr_lookup_dest_topology_table(&mpr->address)))
				{	
					destination_entry=(struct topology_destination_entry * )malloc(sizeof(struct topology_destination_entry));
					olsr_insert_dest_topology_table(destination_entry,mpr);
				}
				
				destination_entry->topology_destination_dst_busy=mprbusy->busy_time; //topo_dest's busy time
	    
				/* creating the structure for the linkage*/
				topo_dest=(struct  destination_list *)malloc(sizeof(struct  destination_list));
				topo_dest->next=last_entry->topology_list_of_destinations;
				last_entry->topology_list_of_destinations=topo_dest;

				topo_last=(struct last_list *)malloc(sizeof(struct last_list));
				topo_last->next=destination_entry->topology_destination_list_of_last;
				destination_entry->topology_destination_list_of_last=topo_last;
	    
				/*linking the last and the dest*/
				topo_last->last_neighbor=last_entry;
				topo_dest->destination_node=destination_entry;
			}
			else
			{
				if (debug_level > 5)
				{
					printf ("update t_dest!\n");
					fprintf (y_file,"update t_dest!\n");
				}
				/* update topology destination busy time */
				if(topo_dest->destination_node->topology_destination_dst_busy!=mprbusy->busy_time)
				{
					changes_topology=UP;
					topo_dest->destination_node->topology_destination_dst_busy=mprbusy->busy_time;
				}
			}	
      
			/*-----------------*/
			if (debug_level > 6)
			{
				printf("This mpr is %s\n",convert_address_to_string(&mpr->address));
				if(mpr->next==NULL)
					printf("the next is NULL\n");
				else
					printf("the next is not NULL\n");
			}
			/*-----------------*/
		}
		mpr=mpr->next;
		mprbusy=mprbusy->next;
	}
}
/*------------------------------------------------------------------------*/
/********************* end of revision **********************/





