/*
 * Copyright (c) 2004 Ying Ge, Communication Research Center Canada.
 *
 * Copyright (c) 2002, 2003 Maoyu Wang, Communication Research Center Canada.
 *
 * By Ying Ge:
 * 1. Change the OLSR packet format and message processing procedure based 
 *    on the OLSR RFC.
 * 2. Add support of multiple interfaces to OLSR, including MID message 
 *    creating and processing procedure as specified in the OLSR RFC.
 * 3. Add QoS Support to OLSR
 *
 * By Maoyu Wang:
 * 1. Ported OLSR from IPv4 to IPv6.
 * 2. Added the Host and Network Association (HNA) functionality into OLSR.
 * 3. Added the default gateway functionality into OLSR by extending the HNA 
 *    message usage. The default gateway functionality supported the mobility
 *	  by cooperating with Mobile IPv6 for a mobile node as well as supported 
 * 	  Internet access for MANET nodes. 
 *
 * DISTRIBUTED WITH NO WARRANTY, EXPRESS OR IMPLIED.
 * See the GNU Library General Public License (file COPYING in the distribution)
 * for conditions of use and redistribution
 */
/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est--dire qu'il peut
 * tre utilis, copi, modifi, distribu  l'unique
 * condition que ce texte soit conserv afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant particip  l'laboration de ce logiciel ne peut
 * tre utilis sans son accord pralable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilit d'aucune sorte.
 * Certaines parties de ce logiciel sont drives de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 * 
 *	$Id: topology_table.h,v 1.2 2000/12/06 10:36:12 prima Exp $
 */

/*
 * Routing table management daemon.
 *
 */


/*
 * Topology table structure.
 */


struct last_list{
        struct last_list  *next;
        struct topology_last_entry *last_neighbor;       
};

struct topology_destination_infos{
        olsr_u32_t          	topologydestination_hash;
        struct  olsr_ip_addr	topology_dst;
        olsr_u32_t          	topology_dst_busy;	//added by Y.Ge, qos OLSR
        struct  last_list   	*list_of_last;
};

struct topology_destination_entry{
       	struct	topology_destination_entry *topology_destination_forw;
	struct	topology_destination_entry *topology_destination_back; 
        struct  topology_destination_infos  topology_destination_info;
};

#define topologydestination_hash         topology_destination_info.topologydestination_hash

#define topology_destination_dst         topology_destination_info.topology_dst

#define topology_destination_dst_busy	topology_destination_info.topology_dst_busy  //added by Y.Ge

#define topology_destination_list_of_last topology_destination_info.list_of_last

struct destination_list{
        struct destination_list  *next;
        struct topology_destination_entry *destination_node;
};

struct topology_last_infos {
	olsr_u32_t              topologylast_hash;
        struct olsr_ip_addr     topology_last;
	olsr_u16_t              topology_seq;
	olsr_u32_t         	topology_last_busy; //added by Y.Ge, qos OLSR
	struct timeval	        topology_timer;		        
        struct destination_list *list_of_destinations;
};


struct topology_last_entry {
	struct	topology_last_entry *topology_last_forw;
	struct	topology_last_entry *topology_last_back;
        struct  topology_last_infos  topology_last_infos;
};


#define	topologylast_hash   	topology_last_infos.topologylast_hash
#define	topology_last	  	topology_last_infos.topology_last      /* node selected by topo_dst as a MultiPoint Relay */
#define	topology_seq       	topology_last_infos.topology_seq       /* sequence number with which topo_last announced this information */
#define topology_last_busy	topology_last_infos.topology_last_busy /* added by Y.Ge, qos OLSR */	
#define	topology_timer    	topology_last_infos.topology_timer      /* for invalidation */
#define topology_list_of_destinations topology_last_infos.list_of_destinations 

struct topology_last_hash {
		struct	topology_last_entry *topology_last_forw;
		struct	topology_last_entry *topology_last_back;
};



struct topology_destination_hash{
        struct	topology_destination_entry *topology_destination_forw;
		struct	topology_destination_entry *topology_destination_back;
};

void
olsr_delete_list_of_last(struct topology_last_entry *last_entry,struct olsr_ip_addr *dst);

void
olsr_delete_dest_topolgy_table(struct topology_destination_entry *dest_entry);

void
olsr_delete_list_of_dest(struct topology_destination_entry *dest_entry,struct olsr_ip_addr *dst);

void
olsr_delete_last_topolgy_table(struct topology_last_entry *last_entry);

void
olsr_insert_last_topology_table(struct topology_last_entry *last_entry,struct tc_message *message);

void
olsr_insert_dest_topology_table(struct topology_destination_entry *dest_entry,struct tc_mpr_addr *mpr);

struct topology_last_entry  
*olsr_lookup_last_topology_table(struct olsr_ip_addr *last);

struct topology_destination_entry
*olsr_lookup_dest_topology_table(struct olsr_ip_addr *dest);

struct  destination_list 
*olsr_in_list_dest_topo(struct topology_last_entry *last_entry,struct olsr_ip_addr *addr);

void
olsr_update_last_table(struct topology_last_entry *last_entry,struct tc_message *message);

void
olsr_update_last_table_with_qos(struct topology_last_entry *last_entry,struct tc_message *message, struct qos_tc_message *qosmessage); //added by Y.Ge, qos OLSR

void
olsr_release_topology_table();

void
olsr_time_out_topology_table();

void
olsr_print_topology_table();

void
olsr_print_trace_topology_table(); //added by Y.Ge
