/*
 * Copyright (c) 2004 Ying Ge, Communication Research Center Canada.
 *
 * Copyright (c) 2002, 2003 Maoyu Wang, Communication Research Center Canada.
 *
 * By Ying Ge:
 * 1. Change the OLSR packet format and message processing procedure based 
 *    on the OLSR RFC.
 * 2. Add support of multiple interfaces to OLSR, including MID message 
 *    creating and processing procedure as specified in the OLSR RFC.
 * 3. Add QoS Support to OLSR
 *
 * By Maoyu Wang:
 * 1. Ported OLSR from IPv4 to IPv6.
 * 2. Added the Host and Network Association (HNA) functionality into OLSR.
 * 3. Added the default gateway functionality into OLSR by extending the HNA 
 *    message usage. The default gateway functionality supported the mobility
 *	  by cooperating with Mobile IPv6 for a mobile node as well as supported 
 * 	  Internet access for MANET nodes. 
 *
 * DISTRIBUTED WITH NO WARRANTY, EXPRESS OR IMPLIED.
 * See the GNU Library General Public License (file COPYING in the distribution)
 * for conditions of use and redistribution
 */
/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est--dire qu'il peut
 * tre utilis, copi, modifi, distribu  l'unique
 * condition que ce texte soit conserv afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant particip  l'laboration de ce logiciel ne peut
 * tre utilis sans son accord pralable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilit d'aucune sorte.
 * Certaines parties de ce logiciel sont drives de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 */

/*
 * Routing table management daemon.
 */

char two_hop_neighbor_table_rcsid[] = 
  "$Id: two_hop_neighbor_table.c,v 1.2 2000/12/06 10:36:12 prima Exp $";


#include "defs.h"

struct neighbor2_hash             neighbor2table[HASHSIZE];

/*------------------------------------------------------------------------*/
void
olsr_delete_neighbor_pointer(struct neighbor_2_entry *two_hop_entry,struct olsr_ip_addr *address)
{
  /*
    This procedure deletes the pointer to this addr from the list contained
    in the two hop neighbor entry;
  */  
 struct	neighbor_list_entry *one_hop_list;
 struct	neighbor_list_entry *one_hop_list_tmp;


 one_hop_list= two_hop_entry->neighbor_2_nblist;
 
 two_hop_entry->neighbor_2_nblist=NULL;

 while(one_hop_list)
   {
     if(0==memcmp(&one_hop_list->neighbor->neighbor_addr,address,sizeof(struct olsr_ip_addr)))
       {
	 one_hop_list_tmp=one_hop_list;
	 one_hop_list=one_hop_list->neighbor_next;
	 free(one_hop_list_tmp);
       }
     else
       {
	 one_hop_list_tmp=one_hop_list;
	 one_hop_list=one_hop_list->neighbor_next;
	 one_hop_list_tmp->neighbor_next=two_hop_entry->neighbor_2_nblist;
	 two_hop_entry->neighbor_2_nblist=one_hop_list_tmp;
       }          
   }
}

/*------------------------------------------------------------------------*/
void
olsr_delete_two_hop_neighbor_table(struct neighbor_2_entry *two_hop_neighbor)

{
  /*
    Delete a two hop neighbor entry results the deletion of its 1 hop neighbors list!!!
    We don't try to remove the one hop neighbor even if it has no more 2 hop neighbors!!
   */
  struct neighbor_list_entry *one_hop_list; 
  struct neighbor_entry      *one_hop_entry;

  one_hop_list=two_hop_neighbor->neighbor_2_nblist;

  while(one_hop_list!=NULL)
    {
      one_hop_entry=one_hop_list->neighbor;
      
      olsr_delete_neighbor_2_pointer(one_hop_entry,&two_hop_neighbor->neighbor_2_addr);

      
      two_hop_neighbor->neighbor_2_nblist=one_hop_list->neighbor_next;
      free(one_hop_list);
      one_hop_list=two_hop_neighbor->neighbor_2_nblist;

    }
   olsr_remque((struct olsr_qelem *)two_hop_neighbor);
   free((void *)two_hop_neighbor);

}

/*------------------------------------------------------------------------*/
void
olsr_insert_two_hop_neighbor_table(struct neighbor_2_entry *two_hop_neighbor)
{

  olsr_u32_t               hash; 
  struct neighbor2_hash    *hash_two_neighbor;

  olsr_hashing(&two_hop_neighbor->neighbor_2_addr,&hash);
  
  two_hop_neighbor->neighbor_2_hash=hash;
  
  hash_two_neighbor=&neighbor2table[hash & HASHMASK];
  
  olsr_insque((struct olsr_qelem *)two_hop_neighbor, (struct olsr_qelem *)hash_two_neighbor);
  
}
/*------------------------------------------------------------------------*/
struct neighbor_2_entry 
*olsr_lookup_two_hop_neighbor_table(struct olsr_ip_addr *dest)
{

  struct neighbor_2_entry  *neighbor_2;
  struct neighbor2_hash    *hash_2_neighbor;
  olsr_u32_t               hash;

  olsr_hashing(dest,&hash);
  hash_2_neighbor=&neighbor2table[hash & HASHMASK];
  
  for(neighbor_2=hash_2_neighbor->neighbor2_forw;neighbor_2!=(struct neighbor_2_entry *)hash_2_neighbor;neighbor_2=neighbor_2->neighbor_2_forw)
    {
      if(neighbor_2->neighbor_2_hash!=hash)
	continue;

      if (0==memcmp(&neighbor_2->neighbor_2_addr,dest,sizeof(struct olsr_ip_addr)))
	return(neighbor_2);
      
    }
  return (NULL);  
}

/*------------------------------------------------------------------------*/
void
olsr_release_two_hop_neighbor_table()
{
  olsr_u8_t                index;
  struct neighbor_2_entry  *neighbor_2;
  struct neighbor_2_entry  *neighbor_2_tmp;
  struct neighbor2_hash    *hash_2_neighbor;

  for(index=0;index<HASHSIZE;index++)
    {
      hash_2_neighbor=&neighbor2table[index];
      neighbor_2=hash_2_neighbor->neighbor2_forw;
      while(neighbor_2!=(struct neighbor_2_entry *)hash_2_neighbor)
	{
	  neighbor_2_tmp=neighbor_2;
	  neighbor_2=neighbor_2->neighbor_2_forw;
	  olsr_delete_two_hop_neighbor_table(neighbor_2_tmp);
	}
    }
}

/*------------------------------------------------------------------------*/
void
olsr_print_two_hop_neighbor_table()
{
  olsr_u8_t                  index;
  struct neighbor_2_entry    *neighbor_2;
  struct neighbor2_hash      *hash_2_neighbor;

  struct neighbor_list_entry *list_1=NULL;
  char str[46];

  printf("\n\nTHE TWO HOP NEIGHBORS\n");

  for(index=0;index<HASHSIZE;index++)
    {
      hash_2_neighbor=&neighbor2table[index];
      for(neighbor_2=hash_2_neighbor->neighbor2_forw;neighbor_2!=(struct neighbor_2_entry *)hash_2_neighbor;neighbor_2=neighbor_2->neighbor_2_forw)
	{

	  printf("ADDRESS TWO HOP NEIGHBORS\n");
	  //printf("%s \n",convert_address_to_string(&neighbor_2->neighbor_2_addr)); 
	  printf("%s \n",inet_ntop(AF_INET6,(&neighbor_2->neighbor_2_addr),str,46)); 
	  printf("---------------------\n");
	  list_1=neighbor_2->neighbor_2_nblist;
	  
	  printf("POINTED BY\n");
	  while(list_1!=NULL)
	    {	      
	      //printf("%s \n",convert_address_to_string(&list_1->neighbor->neighbor_addr));
	      printf("%s \n",inet_ntop(AF_INET6,&(list_1->neighbor->neighbor_addr),str,46));
	      list_1=list_1->neighbor_next;
	    }
	  printf("---------------------\n");
	}
    }
}
/*------------------------------------------------------------------------*/
/***************** added by Y.Ge ****************/
void
olsr_print_trace_two_hop_neighbor_table()
{
  olsr_u8_t                  index;
  struct neighbor_2_entry    *neighbor_2;
  struct neighbor2_hash      *hash_2_neighbor;

  struct neighbor_list_entry *list_1=NULL;
  char str[46];

  fprintf(y_file,"\n\nTHE TWO HOP NEIGHBORS\n");

  for(index=0;index<HASHSIZE;index++)
    {
      hash_2_neighbor=&neighbor2table[index];
      for(neighbor_2=hash_2_neighbor->neighbor2_forw;neighbor_2!=(struct neighbor_2_entry *)hash_2_neighbor;neighbor_2=neighbor_2->neighbor_2_forw)
	{

	  fprintf(y_file,"ADDRESS TWO HOP NEIGHBORS\n");
	  //printf("%s \n",convert_address_to_string(&neighbor_2->neighbor_2_addr)); 
	  fprintf(y_file,"%s \n",inet_ntop(AF_INET6,(&neighbor_2->neighbor_2_addr),str,46)); 
	  printf("---------------------\n");
	  list_1=neighbor_2->neighbor_2_nblist;
	  
	  fprintf(y_file,"POINTED BY\n");
	  while(list_1!=NULL)
	    {	      
	      //printf("%s \n",convert_address_to_string(&list_1->neighbor->neighbor_addr));
	      fprintf(y_file,"%s \n",inet_ntop(AF_INET6,&(list_1->neighbor->neighbor_addr),str,46));
	      list_1=list_1->neighbor_next;
	    }
	  fprintf(y_file,"---------------------\n");
	}
    }
}
/*------------------------------------------------------------------------*/
struct neighbor_entry
*olsr_find_highest_willingness_neighbor(struct neighbor_2_entry *neighbor_2)
{
	 struct neighbor_list_entry	*list_1=NULL;
	 struct neighbor_entry      	*highest_will_neighbor = NULL;  
	 struct main_addr_neigh_entry	*main_neighbor; 
	 int    willingness = 0;
	 
	 list_1=neighbor_2->neighbor_2_nblist;
	 
	 while (list_1!=NULL)
	 {
	 	main_neighbor = list_1->neighbor->neighbor_main_info->main_neighbor;
	 	if (main_neighbor->main_addr_neigh_willingness > willingness && list_1->neighbor->neighbor_status == SYM_LINK)
		{
			willingness = main_neighbor->main_addr_neigh_willingness;
			highest_will_neighbor = list_1->neighbor;
		}
		list_1 = list_1->neighbor_next;
	}
	return (highest_will_neighbor);
}
