/*
 * Department of Systems and Computer Engineering
 * Carleton University, CANADA
 * Copyright (c) 2004 Liang Qin, Department of Systems and Computer Engineering,
 * Carleton University
 * 1. Porting CRC's OLSR with QoS support to XORP 1.0, this release enable OLSR
 *    adds and deletes route entries to RIB in XORP, instead of to Linux kernel. 
 *
*/
#include "config.h"
#include "libxorp/xorp.h"

#define DEBUG_LOGGING

#include "libxorp/debug.h"
#include "libxorp/status_codes.h"
#include "libxipc/xrl_router.hh"

#include "xrl_target6.hh"

int XrlOlsrTarget::status_code = PROC_READY;

XrlOlsrTarget::XrlOlsrTarget(XrlRouter* rtr): XrlOlsrTargetBase(rtr){ }

XrlOlsrTarget::~XrlOlsrTarget()
{
}


XrlCmdError
XrlOlsrTarget::common_0_1_get_target_name(string& n)
{
    n = "OLSR";
    printf("get the target name\n");
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::common_0_1_get_version(string& v)
{
    v = string(version());
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::common_0_1_get_status(uint32_t& status,
				     string&   reason)
{
    status = status_code;
    reason = "none";
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::common_0_1_shutdown()
{
    exit(0);
    //return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::olsr_3_0_set_hello_interval(const int32_t& new_int){
	helloInterval=new_int;
	printf("set hello interval\n");
	return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::olsr_3_0_get_hello_interval(int32_t& old_int)
{
    old_int=helloInterval;
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::olsr_3_0_set_tc_interval(const int32_t& new_int)
{
    tcInterval=new_int;
    printf("set TC interval\n");
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::olsr_3_0_get_tc_interval(int32_t& old_int)
{
    old_int=tcInterval;
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::olsr_3_0_set_mini_tc_interval(const int32_t& new_int)
{
    minitcInterval=new_int;
    printf("set mini tc interval\n");
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::olsr_3_0_get_mini_tc_interval(int32_t&	old_int){
    old_int=minitcInterval;
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::olsr_3_0_set_neighbor_timeout_mult(const int32_t& new_int)
{
    neighbormult=new_int;
    printf("set the nighbor timeout mult\n");
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::olsr_3_0_get_neighbor_timeout_mult(int32_t&	old_int){
    old_int=neighbormult;
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::olsr_3_0_set_topo_timeout_mult(const int32_t& new_int)
{
    topomult=new_int;
    printf("set the topo timeout mult\n");
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::olsr_3_0_get_topo_timeout_mult(int32_t&	old_int){
    old_int=topomult;
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::olsr_3_0_set_tos(const int32_t& new_int)
{
    tos=new_int;
    printf("set the tos\n");
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::olsr_3_0_get_tos(int32_t&	old_int){
    old_int=tos;
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::olsr_3_0_get_tc_jitter(int32_t&	old_int){
    old_int=tcJitter;
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::olsr_3_0_set_tc_jitter(const int32_t& new_int)
{
    tcJitter=new_int;
    printf("set the TC jitter\n");
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::olsr_3_0_get_hello_jitter(int32_t&	old_int){
    old_int=helloJitter;
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::olsr_3_0_set_hello_jitter(const int32_t& new_int)
{
    helloJitter=new_int;
    printf("set the Hello jitter\n");
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::olsr_3_0_get_willingness(int32_t&	old_int){
    old_int=nodeWillingness;
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::olsr_3_0_set_willingness(const int32_t& new_int)
{
    nodeWillingness=new_int;
    printf("set the node willingness\n");
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::olsr_3_0_get_bandwidth_update_threshold(int32_t&	old_int){
    old_int=bdwidthUpdate;
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::olsr_3_0_set_bandwidth_update_threshold(const int32_t& new_int)
{
    bdwidthUpdate=new_int;
    printf("set the node willingness\n");
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::olsr_3_0_get_debug_level(int32_t&	old_int){
    old_int=debugLevel;
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::olsr_3_0_set_debug_level(const int32_t& new_int)
{
    debugLevel=new_int;
    printf("set the debug level\n");
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::olsr_3_0_get_interface_number(int32_t&	old_int){
    old_int=interfaceNo;
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::olsr_3_0_set_interface_number(const int32_t& new_int)
{
    interfaceNo=new_int;
    printf("set the the number of interfaces in the node\n");
    return XrlCmdError::OKAY();
}


XrlCmdError
XrlOlsrTarget::olsr_3_0_get_hyst_enabled(int32_t&	old_int){
    old_int=hystused;
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::olsr_3_0_set_hyst_enabled(const int32_t& new_int)
{
    hystused=new_int;
    printf("set the hysteresis\n");
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::olsr_3_0_get_qos_enabled(int32_t&	old_int){
    old_int=qosused;
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::olsr_3_0_set_qos_enabled(const int32_t& new_int)
{
    qosused=new_int;
    printf("set the QoS\n");
    return XrlCmdError::OKAY();
}

XrlCmdError 
XrlOlsrTarget::olsr_3_0_set_interface_name(const string& ifname){
	wifName=ifname;
        printf("set interface name\n");
	return XrlCmdError::OKAY();
}

XrlCmdError 
XrlOlsrTarget::olsr_3_0_get_interface_name(string& ifname){
	ifname=wifName;
        printf("get interface name %s\n", ifname.c_str());
	return XrlCmdError::OKAY();
}

XrlCmdError 
XrlOlsrTarget::olsr_3_0_set_host_role(const string& role){
	hostRole=role;
        printf("set host role\n");
	return XrlCmdError::OKAY();
}

XrlCmdError 
XrlOlsrTarget::olsr_3_0_get_host_role(string& role){
	role=hostRole;
        printf("get host role %s\n", role.c_str());
	return XrlCmdError::OKAY();
}

XrlCmdError 
XrlOlsrTarget::olsr_3_0_set_multicast_addr(const string& maddr){
	mulAddr=maddr;
        printf("set multicast address\n");
	return XrlCmdError::OKAY();
}

XrlCmdError 
XrlOlsrTarget::olsr_3_0_get_multicast_addr(string& maddr){
	maddr=mulAddr;
        printf("get multicast address %s\n", maddr.c_str());
	return XrlCmdError::OKAY();
}

XrlCmdError
XrlOlsrTarget::olsr_3_0_activate_olsr(){
    status_code = PROC_READY;
    printf(" OLSR target is enabled\n");
    return XrlCmdError::OKAY();
}
