/*
 * Department of Systems and Computer Engineering
 * Carleton University, CANADA
 * Copyright (c) 2004 Liang Qin, Department of Systems and Computer Engineering,
 * Carleton University
 * 1. Porting CRC's OLSR with QoS support to XORP 1.0, this release enable OLSR
 *    adds and deletes route entries to RIB in XORP, instead of to Linux kernel. 
 *
*/

// $XORP: xorp/olsr/xrl_target6.hh,v 1.0 2004/01/28 23:39:59 L. Qin Exp $

#ifndef __OLSR_XRL_TARGET6_HH__
#define __OLSR_XRL_TARGET6_HH__

#include "libxorp/status_codes.h"
#include "xrl/targets/olsr_base.hh"

class XrlRouter;

class XrlOlsrTarget : public XrlOlsrTargetBase {
public:
    XrlOlsrTarget(XrlRouter* rtr);
    ~XrlOlsrTarget();

    XrlCmdError common_0_1_get_target_name(string& name);
    XrlCmdError common_0_1_get_version(string& version);
    XrlCmdError common_0_1_get_status(uint32_t& status, string& reason);
    XrlCmdError common_0_1_shutdown();

    XrlCmdError
    olsr_3_0_set_hello_interval(const int32_t&	new_int);

    XrlCmdError
    olsr_3_0_get_hello_interval(int32_t& old_int);

    XrlCmdError
    olsr_3_0_set_tc_interval(const int32_t& new_int);

    XrlCmdError 
    olsr_3_0_get_tc_interval(int32_t&	old_int);

    XrlCmdError 
    olsr_3_0_set_mini_tc_interval(const int32_t& new_int);

    XrlCmdError 
    olsr_3_0_get_mini_tc_interval(int32_t&	old_int);

    XrlCmdError 
    olsr_3_0_set_neighbor_timeout_mult(const int32_t& new_int);

    XrlCmdError 
    olsr_3_0_get_neighbor_timeout_mult(int32_t& old_int);

    XrlCmdError 
    olsr_3_0_set_topo_timeout_mult(const int32_t& new_int);

    XrlCmdError 
    olsr_3_0_get_topo_timeout_mult(int32_t& old_int);

    XrlCmdError 
    olsr_3_0_set_tos(const int32_t& new_int);

    XrlCmdError 
    olsr_3_0_get_tos(int32_t& old_int);

    XrlCmdError 
    olsr_3_0_set_tc_jitter(const int32_t& new_int);

    XrlCmdError 
    olsr_3_0_get_tc_jitter(int32_t& old_int);

    XrlCmdError 
    olsr_3_0_set_hello_jitter(const int32_t& new_int);

    XrlCmdError 
    olsr_3_0_get_hello_jitter(int32_t& old_int);

    XrlCmdError 
    olsr_3_0_set_willingness(const int32_t& new_int);

    XrlCmdError 
    olsr_3_0_get_willingness(int32_t& old_int);

    XrlCmdError 
    olsr_3_0_set_bandwidth_update_threshold(const int32_t& new_int);

    XrlCmdError 
    olsr_3_0_get_bandwidth_update_threshold(int32_t& old_int);

    XrlCmdError 
    olsr_3_0_set_debug_level(const int32_t& new_int);

    XrlCmdError 
    olsr_3_0_get_debug_level(int32_t& old_int);

    XrlCmdError 
    olsr_3_0_set_interface_number(const int32_t& new_int);

    XrlCmdError 
    olsr_3_0_get_interface_number(int32_t& old_int);

    XrlCmdError 
    olsr_3_0_set_hyst_enabled(const int32_t& new_int);

    XrlCmdError 
    olsr_3_0_get_hyst_enabled(int32_t& old_int);

    XrlCmdError 
    olsr_3_0_set_qos_enabled(const int32_t& new_int);

    XrlCmdError 
    olsr_3_0_get_qos_enabled(int32_t& old_int);

    XrlCmdError 
    olsr_3_0_set_interface_name(const string&	ifname);

    XrlCmdError 
    olsr_3_0_get_interface_name(string&	ifname);

    XrlCmdError 
    olsr_3_0_set_host_role(const string&	role);

    XrlCmdError 
    olsr_3_0_get_host_role(string&	role);

    XrlCmdError 
    olsr_3_0_set_multicast_addr(const string&	maddr);

    XrlCmdError 
    olsr_3_0_get_multicast_addr(string&	maddr);

    XrlCmdError 
    olsr_3_0_activate_olsr();

protected:
	static int status_code;
	int helloInterval;
	int tcInterval;
	int minitcInterval;
	int neighbormult;
	int topomult;
	int tos;
        int tcJitter;
        int helloJitter;
	int hystused;
        int qosused;
	int nodeWillingness;
        int bdwidthUpdate;
	int debugLevel;
	int interfaceNo;
	string hostRole;
	string wifName;
        string mulAddr;

};

#endif // __OLSR_XRL_TARGET6_HH__
